Set Implicit Arguments.
Require Import MyTactics.
Require Import Heaps.
Require Import AdoptionResources.
Require Import HeapResources.

(* Agreement between physical heaps and adoption resources. *)

(* ---------------------------------------------------------------------------- *)

(* The relation [adoptee_agrees] requires that the access right be equal to
   [AdopteeOwned _], which means that, someone, somewhere, owns this memory
   location as an adoptee. *)

(* More precisely, the access right must be [AdopteeOwned p], where [p] is
   the adopter pointer currently stored in the block. The fields of the
   block are ignored; see [BlockAgreement]. *)

Definition adoptee_agrees (b : block) (i : adoptee_status) :=
  let (p, _) := b in
  AdopteeOwned p = i.

Hint Unfold adoptee_agrees : agrees.

(* ---------------------------------------------------------------------------- *)

(* The relation [adopter_agrees] requires that the access right be equal to
   [AdopterOwned _], which means that, someone, somewhere, owns this memory
   location as an adopter. *)

Definition adopter_agrees (b : block) (i : adopter_status) :=
  AdopterOwned = i.

Hint Unfold adopter_agrees : agrees.

(* ---------------------------------------------------------------------------- *)

(* These relations are combined. *)

Inductive adoption_agrees : block -> adoption_status -> Prop :=
| AdoptionAgrees:
    forall b iee ier,
    adoptee_agrees b iee ->
    adopter_agrees b ier ->
    adoption_agrees b (iee, ier).

Hint Constructors adoption_agrees : agrees.

Ltac adoption_agrees_inversion :=
  match goal with h: adoption_agrees _ _ |- _ => inversion h; clear h; subst end;
  unfold adoptee_agrees, adopter_agrees in *;
  destruct_blocks;
  unpack; injections.

Inductive adoptee_exclusive_agrees : heap_access_right -> adoption_status -> Prop :=
| AdopteeExclusiveAgrees:
    forall v p ier,
    adoptee_exclusive_agrees (mv2dx Exclusive v) (AdopteeOwned p, ier).

Hint Constructors adoptee_exclusive_agrees : agrees.

Notation raw_adoption_resource_agrees h R :=
  (aMap_agrees adoption_agrees h R).

(* ---------------------------------------------------------------------------- *)

(* This relation is lifted to the level of (round) allocation maps. *)

Inductive adoption_resource_agrees : heap -> adoption_resource -> Prop :=
| AdoptionResourceAgrees:
    forall h R pf,
    aMap_agrees adoption_agrees h R ->
    adoption_resource_agrees h (exist _ R pf).

Hint Constructors adoption_resource_agrees : agrees.

Ltac adoption_resource_agrees_inversion :=
  match goal with
  h: adoption_resource_agrees _ _ |- _ =>
    inversion h; clear h; subst
  end.

Inductive adoptee_exclusive_resource_agrees : heap_resource -> adoption_resource -> Prop :=
| AdopteeExclusiveResourceAgrees:
    forall hR aR pf,
    aMap_agrees adoptee_exclusive_agrees hR aR ->
    adoptee_exclusive_resource_agrees hR (exist _ aR pf).

(* ---------------------------------------------------------------------------- *)

(* If [R] describes some machine state, then [R] is consistent. *)

Lemma adoptee_agrees_consistent:
  forall b i,
  adoptee_agrees b i ->
  consistent i.
Proof.
  unfold adoptee_agrees. intros. destruct_blocks. subst. eauto with consistent.
Qed.

Lemma adopter_agrees_consistent:
  forall b i,
  adopter_agrees b i ->
  consistent i.
Proof.
  unfold adopter_agrees. intros. destruct_blocks. subst. eauto with consistent.
Qed.

Lemma adoption_agrees_consistent:
  forall b i,
  adoption_agrees b i ->
  consistent i.
Proof.
  dependent_destruction. msa_pair.
  eauto using adoptee_agrees_consistent, adopter_agrees_consistent.
Qed.

Lemma adoption_resource_agrees_consistent:
  forall h R,
  adoption_resource_agrees h R ->
  consistent R.
Proof.
  dependent_destruction. msa_sig.
  eauto using adoption_agrees_consistent with consistent.
Qed.

Hint Resolve adoption_resource_agrees_consistent : consistent.

(* ---------------------------------------------------------------------------- *)

(* The empty heap agrees with the empty adoption resource. *)

Lemma adoption_agrees_void:
  adoption_resource_agrees empty_heap adoption_void.
Proof.
  unfold empty_heap. eauto with agrees.
Qed.

Hint Resolve adoption_agrees_void : agrees.

(* ---------------------------------------------------------------------------- *)

(* Agreement between the physical heap and the adoption resource is not
   affected by updates to the physical heap, provided the adopter pointer
   is not modified. *)

Lemma adoption_resource_agrees_update:
  forall h l p v1 v2 R,
  adoption_resource_agrees h R ->
  maps h l (Block p v1) ->
  adoption_resource_agrees (update h l (Block p v2)) R.
Proof.
  dependent_destruction. intros.
  forwards: @agrees_maps_left_to_right; eauto. unpack.
  adoption_agrees_inversion.
  econstructor.
  equates 1.
    { eauto 7 with agrees. }
    { symmetry. eauto using @update_identity. }
Qed.

Hint Resolve adoption_resource_agrees_update : agrees.

(* ---------------------------------------------------------------------------- *)

(* Agreement is preserved by an update to both the physical heap and the
   adoption resource. *)

Lemma adoption_resource_agrees_pointer_update:
  forall h R l p1 p2 v,
  raw_adoption_resource_agrees h R ->
  maps h l (Block p1 v) ->
  raw_adoption_resource_agrees (update  h l (Block p2 v))
                               (set_adopter_edge R l p2).
Proof.
  intros.
  forwards: @agrees_maps_left_to_right; eauto. unpack.
  adoption_agrees_inversion.
  equates 1.
    eauto with agrees.
    unfold set_adopter_edge. eauto using @maps_update_updatef, def_set_owned.
Qed.

Hint Resolve adoption_resource_agrees_pointer_update : agrees.

(* ---------------------------------------------------------------------------- *)

(* If the address [l] is currently unadopted according to the resource [R], and
   if the heap [h] and resource [R] agree, then the address [l] exists in the
   heap and has a [Null] adopter pointer. *)

Lemma adoption_resource_agrees_unadopted:
  forall h R,
  raw_adoption_resource_agrees h R ->
  forall l,
  unadopted R l ->
  exists v,
  maps h l (Block Null v).
Proof.
  unfold unadopted, mapsp, at_left. intros. unpack. destruct_adoption_status. subst.
  forwards: @agrees_maps_right_to_left; eauto. unpack.
  adoption_agrees_inversion. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the address [l] is adoptable according to the resource [R], and if the
   heap [h] and resource [R] agree, then the address [l] exists in the
   heap. *)

Lemma adoption_resource_agrees_adoptable:
  forall h R,
  raw_adoption_resource_agrees h R ->
  forall l,
  adoptable R l ->
  exists p v,
  maps h l (Block p v).
Proof.
  unfold adoptable, mapsp, at_left. intros. unpack. destruct_adoption_status. subst.
  forwards: @agrees_maps_right_to_left; eauto. unpack.
  adoption_agrees_inversion. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the address [x] currently has a [NonNull y] adopter pointer in the heap,
   and if the heap [h] and resource [R] agree, then the same pointer exists in
   the resource, i.e., [x] is adopted by [y]. *)

Lemma adoption_resource_agrees_adopted:
  forall h R x y v,
  raw_adoption_resource_agrees h R ->
  maps h x (Block (NonNull y) v) ->
  adopted R x y.
Proof.
  unfold adopted, mapsp, at_left. intros.
  forwards: @agrees_maps_left_to_right; eauto. unpack.
  destruct_adoption_status. adoption_agrees_inversion.
  xplit. eauto. simpl. eauto.
Qed.

