Set Implicit Arguments.
Require Import List.
Require Import Permutation.
Require Import MyTactics.
Require Export MSA.
Require Export MSASig.
Require Export AllocationMaps.
Require Export MSAAllocationMaps.
Require Export MSARoundAllocationMaps.
Require Export MSAProduct.
Require Export MSABit.
Require Import MSAExplode.
Require Import Heaps.
Require Import GetSet.

(* An adoption resource is another kind of instrumented heap. It records
   information about each memory location in its role as an adoptee and in its
   role as an adopter. *)

(* ---------------------------------------------------------------------------- *)

(* Every memory block contains a (possibly-null) pointer to its adopter. The
   adoptee status represents the ownership of (and information about the value
   of) this pointer. It is a nothing-or-exclusive access right: either one
   owns and knows nothing, or one owns the adopter field and knows its current
   value. *)

Notation adoptee_status :=
  (bit pointer).

Notation AdopteeNotOwned :=
  (BitZero _ : adoptee_status).

Notation AdopteeOwned p :=
  (BitOne p).

(* ---------------------------------------------------------------------------- *)

(* Every memory location serves as a role as an adopter, regardless of what is
   stored there. The adopter status represents the ownership of this location
   role in its role as an adopter. It is a nothing-or-exclusive access
   right. *)

Inductive adopter_brand : Type :=
  AdopterBrand.

Notation adopter_status :=
  (bit adopter_brand).

Notation AdopterNotOwned :=
  (BitZero _ : adopter_status).

Notation AdopterOwned :=
  (BitOne AdopterBrand).

(* ---------------------------------------------------------------------------- *)

(* We are interested in allocation maps that map each memory location to a
   pair of adoptee and adopter statuses. *)

Notation adoption_status :=
  (adoptee_status * adopter_status)%type.

Notation raw_adoption_resource :=
  (aMap adoption_status).

Ltac destruct_adoption_status :=
  repeat match goal with i : adoption_status |- _ => destruct i end.

(* ---------------------------------------------------------------------------- *)

(* We now wish to restrict our attention to (instrumented) heap fragments that
   are closed under adopter pointers, i.e., whenever one owns an adoptee, one
   owns its adopter too. We define this notion by instantiating the theory in
   [MSARoundAllocationMaps]. *)

(* ---------------------------------------------------------------------------- *)

(* The predicate [adoptee_pointsto] is used to instantiate the abstract
   predicate [pointsto] that appears in the definition of round allocation
   maps. It is true of [i] and [x] when the adoptee status [i] describes
   a non-null adopter pointer whose value is the memory location [x]. *)

Local Ltac adoptee :=
  decompose_pairs; msa_pair; unpack;
  repeat match goal with i : adoptee_status |- _ => destruct i end;
  simpl in *; injections;
  try solve [ tauto | congruence | equiv_dec_by_cases ].

Obligation Tactic := adoptee.

Definition adoptee_pointsto (i : adoption_status) x :=
  let (i, _) := i in
  match i with
  | AdopteeOwned (NonNull y) =>
      y = x
  | _ =>
      False
  end.

Program Instance pointsto_adoption_status : @Pointsto adoption_status _ := {
  pointsto := adoptee_pointsto
}.

Program Instance pointsto_more_adoption_status : @PointstoMore adoption_status _ _.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adopter_present] is used to instantiate the abstract
   predicate [present] that appears in the definition of round allocation
   maps. It is true of the exclusive permission [AdopterOwned]. This expresses
   the fact that, when an adopter edge points to some object [x], then we must
   own [x] as an adopter. This predicate is true also of [BitError], for a
   technical reason -- the proof obligation [present_star_construction] would
   otherwise fail. This is not a problem, as we are ultimately interested in
   consistent resources only. *)

Local Ltac adopter :=
  decompose_pairs; msa_pair; unpack;
  repeat match goal with i : adopter_status |- _ => destruct i end;
  repeat match goal with a : adopter_brand |- _ => destruct a end;
  simpl in *; try solve [ tauto | congruence ].

Obligation Tactic := adopter.

Definition adopter_present (i : adoption_status) :=
  let (_, i) := i in
  match i with BitZero (* AdopterNotOwned *) => False | _ => True end.

Program Instance present_adoption_status : @Present adoption_status _ := {
  present := adopter_present
}.

Program Instance present_more_adoption_status : @PresentMore adoption_status _ _.

(* ---------------------------------------------------------------------------- *)

(* We write just [round] for this particular instance of [round]. (Yes, this
   is a rather violent abuse of notation.) *)

Notation round :=
  (round (A := adoption_status)).

Hint Resolve @round_core @round_star : round.

(* The type [sig round] is the type of round allocation maps, i.e., the
   allocation maps of type [aMap adoption_status] that satisfy the predicate
   [round]. We refer to such a map as an `adoption resource'. *)

Notation adoption_resource :=
  (sig round).

Lemma round_adoption_void:
  round (empty_map _).
Proof.
  eapply round_empty.
Qed.

Notation adoption_void :=
  (exist _ _ round_adoption_void).

(* ---------------------------------------------------------------------------- *)

(* Adoption resources form a monotonic separation algebra. (Sanity check.) *)

Goal MSA adoption_resource.
Proof.
  eauto with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adoptable R l] means that, according to the adoption resource
   [R], the memory location [l] is adoptable, i.e., there is an adopter field at
   this address. *)

(* For the moment, every memory block contains an adopter field. Thus, every
   valid memory location has type [TyAdoptable]. *)

Definition adoptable (R : raw_adoption_resource) l :=
  mapsp R l (fun _ => True).

(* ---------------------------------------------------------------------------- *)

(* The predicate [adoptable] is affine. *)

Lemma adoptable_affine:
  forall R1 R2 R l,
  adoptable R1 l ->
  star R1 R2 = R ->
  consistent R ->
  adoptable R l.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [unadopted R l] means that, according to the adoption resource
   [R], the memory location [l] is (adoptable and) unadopted, i.e., there is an
   adopter field at this address, and it is currently [Null]. *)

Definition unadopted (R : raw_adoption_resource) l :=
  mapsp R l (at_left (eq (AdopteeOwned Null))).

(* ---------------------------------------------------------------------------- *)

(* The predicate [unadopted _ l] is affine. *)

Lemma unadopted_affine:
  forall R1 R2 R l,
  unadopted R1 l ->
  star R1 R2 = R ->
  consistent R ->
  unadopted R l.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adopts R l] means that, according to the adoption resource
   [R], the memory location [l] is an adopter, i.e., [R] owns this location
   in its role as an adopter. *)

Definition adopts (R : raw_adoption_resource) l :=
  mapsp R l (at_right (eq AdopterOwned)).

(* A consequence. *)

Lemma adopts_domain:
  forall R l,
  adopts R l ->
  domain R l.
Proof.
  unfold domain, adopts, mapsp, at_right. intros. unpack. destruct_adoption_status. subst.
  xplit. eauto. simpl. eauto.
Qed.

(* The reverse implication, [domain] implies [adopts], is true only in the
   presence of a consistency hypothesis. *)

(* ---------------------------------------------------------------------------- *)

(* The predicate [adopts _ l] is affine. *)

Lemma adopts_affine:
  forall R1 R2 R l,
  adopts R1 l ->
  star R1 R2 = R ->
  consistent R ->
  adopts R l.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* The predicate [adopts _ l] is stable. *)

Lemma adopts_stable:
  forall R1 R2 l,
  adopts R1 l ->
  rely R1 R2 ->
  adopts R2 l.
Proof.
  intros. eapply mapsp_stablep; eauto with stable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adopted R x y] means that, according to the adoption resource
   [R], the memory location [x] is adopted by the memory location [y], i.e.,
   [x.adopter] is [y]. *)

Definition adopted (R : raw_adoption_resource) x y :=
  mapsp R x (at_left (eq (AdopteeOwned (NonNull y)))).

(* An equivalent formulation. *)

Lemma def_adopted:
  forall R x y,
  adopted R x y <->
  exists i,
  maps R x (AdopteeOwned (NonNull y), i).
Proof.
  unfold adopted, mapsp, at_left, at_right.
  split; intros; unpack; decompose_pairs; subst.
  eauto.
  eexists; split; [ eauto | instantiate; simpl; eauto ].
Qed.

(* Another equivalent formulation. *)

Lemma adopted_edge:
  forall R x y,
  adopted R x y <->
  edge R x y.
Proof.
  unfold adopted, at_left, edge.
  split; intros; eapply mapsp_covariant; eauto;
  intros; destruct_adoption_status; subst; simpl in *.
  (* One direction. *)
  reflexivity.
  (* Reverse direction. *)
  match goal with |- _ = ?i => destruct i end;
  destruct_pointers; subst; tauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adopted _ x y] is affine. *)

Lemma adopted_affine:
  forall R1 R2 R x y,
  adopted R1 x y ->
  star R1 R2 = R ->
  consistent R ->
  adopted R x y.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* The predicate [adopted _ x y] is stable. *)

Lemma adopted_stable:
  forall R1 R2 x y,
  adopted R1 x y ->
  rely R1 R2 ->
  adopted R2 x y.
Proof.
  intros. eapply mapsp_stablep; eauto with stable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The following lemma is a reformulation of [edge_narrow_left]. If [star R1 R2]
   has an edge of [x] to [y], and if [y] as an adopter is owned by [R1], then
   -- under a roundness hypothesis about [R2] -- this edge of [x] to [y] lies in
   [R1]. *)

Lemma adopted_narrow_left:
  forall R1 R2 x y,
  adopted (star R1 R2) x y ->
  adopts R1 y ->
  round R2 ->
  consistent (star R1 R2) ->
  adopted R1 x y.
Proof.
  intros. rewrite adopted_edge in *.
  eauto using edge_narrow_left, adopts_domain with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Unadopted implies adoptable. *)

Lemma unadopted_implies_adoptable:
  forall R l,
  unadopted R l ->
  adoptable R l.
Proof.
  unfold unadopted, adoptable. eauto using mapsp_covariant.
Qed.

(* Adopted implies adoptable. *)

Lemma adopted_adoptable:
  forall R x y,
  adopted R x y ->
  adoptable R x.
Proof.
  unfold adopted, adoptable. eauto using mapsp_covariant.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [set_owned] replaces the adoptee status with [AdopteeOwned p]. This can
   be used to delete an adopter edge (when [p] is [Null]) or to install one
   (when [p] is [NonNull _]). *)

Definition set_owned (p: pointer) : adoption_status -> adoption_status :=
  set (AdopteeOwned p).

(* An equivalent formulation. *)

Lemma def_set_owned:
  forall p iee ier,
  set_owned p (iee, ier) = (AdopteeOwned p, ier).
Proof.
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If we have [AdopteeNotOwned] one side, then [set_owned] commutes
   with (the other side of) [star]. *)

Lemma star_set_owned_left:
  forall i1 i2 p,
  at_left (eq AdopteeNotOwned) i2 ->
  star (set_owned p i1) i2 = set_owned p (star i1 i2).
Proof.
  intros. destruct_adoption_status. simpl in *. subst.
  msa_pair.
  do 2 rewrite def_set_owned. (* [simpl] is too strong. *)
  msa_pair. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* We lift [set_owned] up to the level of allocation maps via [updatef]. We
   obtain two functions, [adopt] and [abandon], which respectively install
   and remove an adopter edge. *)

Definition set_adopter_edge (R: raw_adoption_resource) x p :=
  updatef R x (set_owned p).

(* Adoption installs an adopter pointer from [x] to [y]. *)

Notation adopt R x y :=
  (set_adopter_edge R x (NonNull y)).

(* Abandon removes an adopter pointer from [x] to [y]. *)

Notation abandon R x :=
  (set_adopter_edge R x Null).

(* ---------------------------------------------------------------------------- *)

(* Abandon undoes adoption. *)

Lemma abandon_adopt:
  forall R x y,
  unadopted R x ->
  abandon (adopt R x y) x = R.
Proof.
  unfold unadopted, set_adopter_edge. intros.
  erewrite compose_updatef_updatef with (g := set_owned Null).
    2: intros; destruct_adoption_status; eauto.
  eapply updatef_identity; eauto.
  intros. destruct_adoption_status. unfold at_left in *. subst. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Adoption establishes adoption, i.e., [adopt] establishes [adopted]. *)

Lemma adopted_adopt:
  forall R x y,
  adoptable R x ->
  adopted (adopt R x y) x y.
Proof.
  intros.
  eapply mapsp_updatef_there; [ | eauto | eauto ].
  simpl. eauto.
Qed.

(* [abandon] establishes [unadopted]. *)

Lemma unadopted_abandon:
  forall R x,
  adoptable R x ->
  unadopted (abandon R x) x.
Proof.
  intros.
  eapply mapsp_updatef_there; [ | eauto | eauto ].
  simpl. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Two successive abandon operations commute. This implies that the order of the
   adoptees [xs] in the predicate [all_adoptees R xs y] does not matter. See the
   lemma [the_last_will_be_the_first]. *)

Lemma abandon_abandon:
  forall R x1 x2,
  abandon (abandon R x1) x2 = abandon (abandon R x2) x1.
Proof.
  unfold set_adopter_edge. eauto using updatef_updatef.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [adopt] and [abandon] both preserve consistency. *)

Lemma consistent_set_adopter_edge:
  forall R x p,
  consistent R ->
  consistent (set_adopter_edge R x p).
Proof.
  intros. eapply consistent_updatef; eauto.
  unfold set_owned. intros. destruct_adoption_status. simpl. msa_pair. unpack.
  eauto with consistent.
Qed.

Hint Resolve consistent_set_adopter_edge : consistent.

(* ---------------------------------------------------------------------------- *)

(* [abandon] preserves roundness. *)

Lemma round_abandon:
  forall R x,
  round R ->
  round (abandon R x).
Proof.
  intros. eapply round_updatef; eauto.
    (* present *)
    unfold set_owned. intros. destruct_adoption_status. simpl. tauto.
    (* pointsto *)
    intros. false.
Qed.

(* [adopt] also preserves roundness, provided one owns the target of the new
   edge as an adopter. *)

Lemma round_adopt:
  forall R x y,
  adopts R y ->
  round R ->
  round (adopt R x y).
Proof.
  introv hadopts ?. eapply round_updatef; eauto.
    (* present *)
    unfold set_owned. intros. destruct_adoption_status. simpl. tauto.
    (* pointsto *)
    simpl. intros. subst. eauto using adopts_domain.
Qed.

Hint Resolve round_abandon round_adopt : round.

(* ---------------------------------------------------------------------------- *)

(* If an adoptee is owned on one side of a split, then it is not owned on the
   other side. This auxiliary fact is used in the following two lemmas. *)

Lemma star_adoptee_owned:
  forall (i1 i2: adoption_status) p,
  consistent (star i1 i2) ->
  at_left (eq (AdopteeOwned p)) i1 ->
  at_left (eq AdopteeNotOwned) i2.
Proof.
  intros. destruct_adoption_status. msa_pair. unpack.
  simpl at_left in *. subst.
  forwards: @use_solitary; eauto with solitary.
Qed.

(* [abandon] commutes with [star]. More precisely, if [x] is adopted in [R1],
   then deleting this adopter pointer commutes with conjoining [R2]. *)

Lemma star_abandon_left:
  forall R1 R2 x y,
  consistent (star R1 R2) ->
  adopted R1 x y ->
  star (abandon R1 x) R2 = abandon (star R1 R2) x.
Proof.
  unfold set_adopter_edge.
  eauto using star_updatef_left, star_set_owned_left, star_adoptee_owned.
Qed.

(* Similarly, [adopts] commutes with [star]. More precisely, if [x] is
   unadopted in [R1], then installing an adopter pointer commutes with
   conjoining [R2]. *)

Lemma star_adopt_left:
  forall R1 R2 x y,
  consistent (star R1 R2) ->
  unadopted R1 x ->
  star (adopt R1 x y) R2 = adopt (star R1 R2) x y.
Proof.
  unfold set_adopter_edge.
  eauto using star_updatef_left, star_set_owned_left, star_adoptee_owned.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [rely] is preserved by [abandon] at an existing address. *)

Lemma rely_abandon:
  forall R1 R2 x y,
  rely R1 R2 ->
  adopted R1 x y ->
  rely (abandon R1 x) (abandon R2 x).
Proof.
  intros. rewrite def_adopted in *. unpack.
  eapply rely_updatef_updatef; eauto.
  intros. destruct_adoption_status. msa_pair. unpack. simpl in *. subst.
  eauto using @rely_reflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [adopted] is preserved by [abandon] at a distinct address. *)

Lemma adopted_abandon:
  forall R x y,
  adopted R x y ->
  forall z,
  z <> x ->
  adopted (abandon R z) x y.
Proof.
  intros. rewrite def_adopted in *. (* hypothesis and goal! *) unpack.
  forwards: @maps_updatef_elsewhere x z.
    eauto.
    omega.
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Conversely, if, after abandoning [x1], we find that [x2] is still adopted by
   [y], then it must the case that [x1] and [x2] are distinct and that [x2] was
   adopted by [y] prior to abandoning [x1]. *)

Lemma adopted_abandon_reverse_1:
  forall R x1 x2 y,
  adopted (abandon R x1) x2 y ->
  adopted R x2 y.
Proof.
  intros. rewrite def_adopted in *. unpack.
  forwards f: @maps_updatef_inversion. eauto.
  destruct f; unfold mapsp in *; unpack; subst; [ eauto | false ].
Qed.

Lemma adopted_abandon_reverse_2:
  forall R x1 x2 y,
  adopted (abandon R x1) x2 y ->
  x2 <> x1.
Proof.
  intros. rewrite def_adopted in *. unpack.
  forwards f: @maps_updatef_inversion. eauto.
  destruct f; unfold mapsp in *; unpack; subst; [ eauto | false ].
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [all_adoptees R xs y] means that [xs] is a list of all of
   [y]'s adoptees according to [R]. It also means that [R] owns [y] as an
   adopter and owns each of the [xs] as an adoptee. *)

Inductive all_adoptees : raw_adoption_resource -> list nat -> nat -> Prop :=
| AllAdopteesBase:
    forall R y,
    adopts R y ->
    no_antecedents R y ->
    all_adoptees R nil y
| AllAdopteesStep:
    forall R x xs y,
    adopted R x y ->
    (* Deleting the edge out of [x] ensures that [x] cannot appear
       twice in the list. *)
    all_adoptees (abandon R x) xs y ->
    all_adoptees R (x :: xs) y.

Hint Constructors all_adoptees : all_adoptees.

(* ---------------------------------------------------------------------------- *)

(* [all_adoptees] is affine. *)

Lemma all_adoptees_affine:
  forall R1 xs y,
  all_adoptees R1 xs y ->
  forall R2 R,
  star R1 R2 = R ->
  consistent R ->
  round R2 ->
  all_adoptees R xs y.
Proof.
  induction 1; intros; subst.
  { eauto using adopts_affine, @no_antecedents_star with typeclass_instances all_adoptees. }
  { econstructor.
    { eauto using adopted_affine. }
    { erewrite <- star_abandon_left by eauto.
      eapply IHall_adoptees; eauto.
      erewrite star_abandon_left by eauto.
      eauto with consistent. }}
Qed.

(* [all_adoptees] is stable. *)

Lemma all_adoptees_stable:
  forall R1 xs y,
  all_adoptees R1 xs y ->
  forall R2,
  rely R1 R2 ->
  all_adoptees R2 xs y.
Proof.
  induction 1; eauto using adopts_stable, no_antecedents_rely,
    adopted_stable, rely_abandon with all_adoptees.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [abandon R x] owns [y] as an adopter, then so does [R]. This is in fact
   an equivalence, but one half of it suffices for our purposes. The intuitive
   reason why this is true is that [abandon] affects only the adoptee status,
   while [adopts] cares only about the adopter status. *)

Lemma adopts_abandon:
  forall R x y,
  adopts (abandon R x) y ->
  adopts R y.
Proof.
  unfold adopts, set_adopter_edge. intros.
  eapply mapsp_updatef_inversion; [ | eauto ].
  unfold at_right. intros. destruct_adoption_status. assumption.
Qed.

(* As a consequence, [all_adoptees R xs y] implies [adopts R y]. *)

Lemma all_adoptees_adopts:
  forall R xs y,
  all_adoptees R xs y ->
  adopts R y.
Proof.
  induction 1; eauto using adopts_abandon.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The act of adopting a (previsouly unadopted) address [x] extends the list of
   all adoptees with [x]. *)

Lemma all_adoptees_cons:
  forall R1 R2 R x xs y,
  unadopted R1 x ->
  round R1 ->
  all_adoptees R2 xs y ->
  star R1 R2 = R ->
  consistent R ->
  all_adoptees (adopt R x y) (cons x xs) y.
Proof.
  intros; subst.
  eapply AllAdopteesStep.
    { eauto using adopted_adopt, unadopted_implies_adoptable, unadopted_affine. }
    { rewrite abandon_adopt by eauto using unadopted_affine.
      eauto using all_adoptees_affine, @star_commutative. }
Qed.

(* ---------------------------------------------------------------------------- *)

(* A key property of the predicate [all_adoptees] is that the list [ys]
   contains all of the current adoptees of [x]. That is, if [y.adopter]
   is [x], then [y] must be a member of the list [ys]. *)

Lemma no_lost_sheep:
  forall R xs y,
  all_adoptees R xs y ->
  forall x,
  adopted R x y ->
  In x xs.
Proof.
  induction 1; intros; simpl.
  (* Base case. *)
  (* [y] has no antecedents, yet [x] is adopted by [y]. Contradiction. *)
  { unfold no_antecedents, into, not in *.
    rewrite def_adopted in *. unpack.
    match goal with h: forall x: nat, _ |- _ => eapply h; eauto end.
    simpl. eauto. }
  (* Step case. *)
  (* If the two adoptees are equal, the result is immediate. Otherwise,
     apply the induction hypothesis. *)
  match goal with |- ?x1 = ?x2 \/ _ =>
    destruct (eq_nat_dec x1 x2)
  end;
  eauto using adopted_abandon.
Qed.

(* ---------------------------------------------------------------------------- *)

(* As part of the [take] operation, once we have proven that some object [x]
   is a member of the list [xs], we wish to move [y] to the front of the list,
   so as to be able to then easily take it out of the list. *)

Lemma the_last_will_be_the_first:
  forall R xs y,
  all_adoptees R xs y ->
  forall x,
  In x xs ->
  exists xs',
  Permutation xs (x :: xs') /\
  adopted R x y /\
  all_adoptees (abandon R x) xs' y.
Proof.
  induction 1; simpl; introv h; [ tauto | destruct h ].
  (* If we have reached [x], the result is immediate. *)
  subst. eauto.
  (* If we are looking at some other element [z], then we must apply
     the induction hypothesis to bring [x] in front of the list tail,
     then permute [x] and [z]. *)
  forwards: IHall_adoptees; [ eauto | unpack; subst ].
  (* rewrite abandon_abandon in *. *)
  xplit; eauto using perm_trans, perm_swap.
  eauto using adopted_abandon_reverse_1.
  econstructor.
  eapply adopted_abandon. eauto. eauto using adopted_abandon_reverse_2.
  eauto using adopted_abandon, adopted_abandon_reverse_2.
  rewrite abandon_abandon. eauto.
Qed.

(* The last conjunct in the conclusion of the above lemma implies that
   [x] does not appear in the list of the remaining adoptees. *)

Lemma all_adoptees_abandon_implies_not_in:
  forall R x xs y,
  all_adoptees (abandon R x) xs y ->
  ~ In x xs.
Proof.
  intros. intro.
  forwards: the_last_will_be_the_first; eauto. unpack.
  forwards: adopted_abandon_reverse_2; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [x] is not in the list of adoptees [xs], then losing the ownership of [x]
   does not invalidate [all_adoptees _ xs y]. This is used in the proof of type
   preservation for [TTakeCommitted], while arguing that we recover the permission
   [x @ adoptable]. *)

Lemma all_adoptees_lose:
  forall R xs y,
  all_adoptees R xs y ->
  forall x,
  ~ In x xs ->
  all_adoptees (updatef R x (set AdopteeNotOwned)) xs y.
Proof.
  (* Boring proof. Just proving that the update commutes with everything. *)
  induction 1; simpl; intros.
  (* Base case. *)
  econstructor.
    { eapply mapsp_updatef_anywhere; eauto.
      intros. destruct_adoption_status. eauto. }
    { unfold no_antecedents in *. unpack. split.
      { eapply domain_updatef; eauto.
        intros. destruct_adoption_status. eauto. }
      { eauto using into_updatef. }}
  (* Step case. *)
  econstructor.
    { eapply mapsp_updatef_elsewhere; eauto. }
    { equates 3; [ | eapply updatef_updatef_elsewhere ]; eauto. }
Qed.

