Set Implicit Arguments.
Require Import MyTactics.
Require Import Syntax.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import Resources.
Require Import BlockAgreement.
Require Import AdoptionAgreement.
Require Import LockAgreement.
Require Import Typing.
Require Import JBasic.
Require Import JAffinity.
Require Import JStability.
Require Import Entails.
Require Import HiddenInvariants.

(* ---------------------------------------------------------------------------- *)

(* The conjunction of [heap_agrees] and [lock_heap_agrees]. *)

Inductive agrees : state -> resource -> Prop :=
| Agrees:
    (* We have a concrete heap [h] and a concrete lock heap [lckh]. *)
    forall h lckh,
    (* We have a logical heap [Rh], a logical adoption heap [Ra], and
       a logical lock heap [Rlckh]. *)
    forall Rh Ra Rlckh,
    (* We require agreement of the heap components, *)
    heap_agrees h Rh ->
    (* agreement of the (same!) heap with the adoption heap, *)
    adoption_resource_agrees h Ra ->
    (* (* the adoptable resources are exclusive *) *)
    (* adoptee_exclusive_resource_agrees Rh Ra -> *)
    (* and agreement of the lock heap components. *)
    lock_heap_agrees lckh Rlckh ->
    (* And we are happy. *)
    agrees (h, lckh) (Rh, Ra, Rlckh).

Hint Constructors agrees : agrees.

Ltac agrees_inversion :=
  match goal with h: agrees _ _ |- _ => inversion h; clear h; try subst end.

(* ---------------------------------------------------------------------------- *)

(* Agreement implies consistency. *)

Lemma agrees_consistent_unified:
  forall s R,
  agrees s R ->
  consistent R.
Proof.
  intros. agrees_inversion. explode; eauto with consistent.
Qed.

Ltac agrees_consistent_unified :=
  forwards: agrees_consistent_unified; [ eauto | ].

(* ---------------------------------------------------------------------------- *)

(* The relation [jsr s R] means that the machine state [s] and the resource
   [R] agree. It is analogous to the view projection function in the Views
   framework. *)

(* It is somewhat striking that the definition of the jugement [jsr] depends
   on the typing judgement [jpi]. It is not just a simple notion of agreement
   between concrete heaps and instrumented heaps; it is more deeply recursive.
   In the abstract presentation of Core Mezzo, we are tempted to present [jsr]
   as a parameter of the system; but in fact, we see that it cannot be defined
   a priori. *)

Inductive jsr : state -> resource -> Prop :=
| Jsr:
    forall s R Ravl Rhdd,
    (* We require agreement between the machine state and the resource [R]. *)
    agrees s R ->
    (* Now, we wish to split [R] between a part that is available to the
       program, [Ravl], and a part that is protected by the available locks,
       hence hidden from the program, [Rhdd]. *)
    star Ravl Rhdd = R ->
    (* How do we split? The (lock heap component of) [R] tells us which
       locks are free. It also tells us what their invariants are. We take
       the conjunction of the invariants of these locks, and require
       [Rhdd] to justify this permission. *)
    jpi Rhdd zilch (hidden_invariants (get R)) ->
    (* In the conclusion of the judgement, we expose only the
       resource that is available to the program. *)
    jsr s Ravl.

Hint Constructors jsr : jsr.

(* TEMPORARY *)
Ltac jsr_inversion_light :=
  match goal with h: jsr _ _ |- _ => inversion h; clear h; subst end;
  agrees_consistent_unified.

Ltac jsr_inversion :=
  jsr_inversion_light; agrees_inversion.

(* ---------------------------------------------------------------------------- *)

(* If [R] describes some machine state, then [R] is consistent. *)

Lemma jsr_consistent:
  forall s R,
  jsr s R ->
  consistent R.
Proof.
  intros. jsr_inversion. explode_resources. eauto 6 with consistent_split.
Qed.

(* This tactic applies the above lemma to any [jsr] hypotheses found in the
   context. *)

Ltac jsr_consistent :=
  repeat match goal with
  | h: jsr _ _ |- _ =>
      generalize (jsr_consistent h); intro; revert h
  end;
  intros.

(* ---------------------------------------------------------------------------- *)

(* The initial machine state agrees with the resource [void]. *)

Lemma jsr_initial:
  jsr initial void.
Proof.
  econstructor.
    unfold initial. eauto with agrees.
    eapply @core_right_unit.
    (* We must argue that [lock_void] has no hidden invariants. *)
    simpl. eauto with jpi.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Reasoning about changes in the machine state and in the resource. *)

(* [accord agreement changeA changeB] means that a change of the left-hand
   side (the machine state) along [changeA] and of the right-hand side (the
   resource) along [changeB] preserves [agreement]. *)

Definition accord A B
  (agreement : A -> B -> Prop)
  (changeA : A -> A -> Prop)
  (changeB : B -> B -> Prop)
:=
  forall a1 b1 a2 b2,
  agreement a1 b1 ->
  changeA a1 a2 ->
  changeB b1 b2 ->
  agreement a2 b2.

(* [preserves_lock_resource evolution] means just that: an evolution along
   [evolution] does not affect the lock resource component. *)

Definition preserves_lock_resource (evolution : resource -> resource -> Prop) :=
  forall R1 R2,
  evolution R1 R2 ->
  (get R1 = get R2 :> lock_resource).

Ltac prove_preserves_lock_resource :=
  unfold preserves_lock_resource;
  intros; unpack; subst; explode_resources_locks; eauto.

(* ---------------------------------------------------------------------------- *)

(* If an [evolution] of the resource is admissible and does not affect the lock
   resource component, then it suffices to prove that [agrees] is preserved,
   and there follows that [jsr] is preserved too. The intuitive idea is, in
   that case, the hidden resource behaves just as if it was framed out. *)

Lemma lift_accord:
  forall evolution,
  admissible evolution ->
  preserves_lock_resource evolution ->
  forall change,
  accord agrees change evolution ->
  accord jsr change evolution.
Proof.
  unfold accord. introv hadm hacc. intros. jsr_inversion_light.
  forwards: hadm; eauto. unpack.
  forwards: hacc; eauto.
  econstructor; eauto.
  (* There remains to check that the invariants of the available locks are
     still satisfied. This is the case because the set of available locks
     has not changed and the hidden resource has evolved along [rely]. *)
  eapply j_rely.
    equates 1. eauto. explode_resources_locks. congruence.
    eauto with consistent_split.
    eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [s] and [R] agree, then their heap components have a common limit. *)

Lemma jsr_limit_limit_heap:
  forall s R,
  jsr s R ->
  forall h : heap,
  forall Rh : heap_resource,
  h = get s ->
  Rh = get R ->
  limit h = limit Rh.
Proof.
  intros. jsr_inversion. explode_resources_blocks.
  forwards f: (@agrees_limit block); [ eassumption | ].
  rewrite f. clear f.
  eauto with limit_limit consistent_split.
Qed.

(* The same statement holds about the adoption resource. *)

Lemma jsr_limit_limit_adoption:
  forall s R,
  jsr s R ->
  forall h : heap,
  forall Ra : raw_adoption_resource,
  h = get s ->
  Ra = get R ->
  limit h = limit Ra.
Proof.
  intros. jsr_inversion. explode_resources_adoption_strong.
  adoption_resource_agrees_inversion.
  forwards f: @agrees_limit; [ eassumption | ].
  rewrite f. clear f.
  eauto with limit_limit consistent_split.
Qed.

(* The same statement holds about the lock heap components. *)

Lemma jsr_limit_limit_lock:
  forall s R,
  jsr s R ->
  forall lckh : lock_heap,
  forall Rlckh : lock_resource,
  lckh = get s ->
  Rlckh = get R ->
  limit lckh = limit Rlckh.
Proof.
  intros. jsr_inversion. explode_resources_locks.
  forwards f: (@agrees_limit lock_status); [ eassumption | ].
  rewrite f. clear f.
  eauto with limit_limit consistent_split.
Qed.

Hint Resolve jsr_limit_limit_heap jsr_limit_limit_adoption
jsr_limit_limit_lock : limit_limit.

(* ---------------------------------------------------------------------------- *)

(* If a memory location is allocated in the heap and in the instrumented heap,
   and if the machine state and resource are in agreement, then the values
   found in the heap and in the instrumented heap at this location must be the
   same. *)

Lemma ag_maps_maps:
  forall s1 R1 R1a R1b l m p v1 v2,
  agrees s1 R1 ->
  maps (get s1) l (Block p v1) ->
  maps (get R1a) l (mv2dx m v2) ->
  star R1a R1b = R1 -> (* TEMPORARY keep this built-in maps_maximal? *)
  v1 = v2.
Proof.
  intros. agrees_consistent_unified. agrees_inversion. explode_resources_blocks.
  forwards: @agrees_maps_maps. 2: eauto. eauto. maps_maximal.
  block_agrees_inversion. destruct_modes; dx_agrees_inversion; eauto.
Qed.

Lemma jsr_maps_maps:
  forall s1 R1 R1a R1b l m p v1 v2,
  jsr s1 R1 ->
  maps (get s1) l (Block p v1) ->
  maps (get R1a) l (mv2dx m v2) ->
  star R1a R1b = R1 ->
  v1 = v2.
Proof.
  intros. jsr_inversion_light. eauto using ag_maps_maps, @star_associative.
Qed.

Ltac jsr_maps_maps :=
  match goal with
    | h: maps _ ?l (mv2dx _ _) |- _ =>
        forwards: >> jsr_maps_maps l
    | h: maps _ ?l (DX_EXCLUSIVE _) |- _ =>
        forwards: >> jsr_maps_maps l Exclusive
    | h: maps _ ?l (DX_DUPLICABLE _) |- _ =>
        forwards: >> jsr_maps_maps l Duplicable
  end;
  [ eassumption | eauto | eauto | aac_guess | ];
  instantiate; subst.

(* ---------------------------------------------------------------------------- *)

(* If a memory location is allocated in the current resource, and if the heap
   and resource are in agreement, then this memory location is also allocated
   in the heap, and the values found in the heap and in the resource at this
   location must be the same. *)

Lemma ag_maps_right_to_left: (* TEMPORARY bad name *)
  forall s R l m v,
  agrees s R ->
  maps (get R) l (mv2dx m v) ->
  exists p,
  maps (get s) l (Block p v).
Proof.
  intros. agrees_inversion. explode_resources_blocks.
  forwards: @agrees_maps_right_to_left. 2: eauto. eauto. unpack.
  destruct_modes; destruct_blocks; block_agrees_inversion; eauto.
Qed.

Lemma jsr_maps_right_to_left:
  forall s R l m v,
  jsr s R ->
  maps (get R) l (mv2dx m v) ->
  exists p,
  maps (get s) l (Block p v).
Proof.
  intros. jsr_inversion_light. eauto using ag_maps_right_to_left with explode_resources maps_maximal.
Qed.

Ltac jsr_maps_right_to_left :=
  match goal with
  | h: maps _ ?l (mv2dx ?m ?v) |- _ =>
      forwards: >> jsr_maps_right_to_left l m v
  | h: maps _ ?l (DX_EXCLUSIVE _) |- _ =>
      forwards: >> jsr_maps_right_to_left l Exclusive
  | h: maps _ ?l (DX_DUPLICABLE _) |- _ =>
      forwards: >> jsr_maps_right_to_left l Duplicable
  end;
  [ eauto | explode_resources_blocks; maps_maximal | unpack ].

(* ---------------------------------------------------------------------------- *)

(* If [s] and [R] agree, then allocating a new (locked) lock (at both levels)
   preserves agreement. *)

Lemma jsr_allocate_lock:
  forall s R s' R' P,
  jsr s R ->
  s' = set (allocate (get s) LOCKED) s ->
  R' = set (allocate (get R) (LockOne P)) R ->
  jsr s' R'.
Proof.
  intros. jsr_inversion. explode_resources_locks.
  econstructor.
    { eauto with agrees. }
    { explode. eapply star_allocate_core_right_unit. }
    (* We must check that the hidden resource is not affected by this
       allocation. This works because this allocation is permitted by
       the relation [rely] and because creating a new lock (which is
       initially locked) does not enlarge the hidden invariants. *)
    eapply hidden_invariants_allocate.
    eapply j_rely.
      eassumption.
      explode; eauto with agrees consistent consistent_split.
      explode; eauto using @rely_allocate_core, @rely_reflexive with explode consistent.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [s] and [R] agree, and if the lock [lck] is acquired, yielding a new
   machine state [s'], then [s'] agrees with the conjunction of [RP] -- a
   new resource which justifies the lock invariant [P] -- and [R'] -- the
   updated version of [R]. *)

Lemma jsr_acquire:
  forall s s' lckh lckh' lck R P,
  (* If [s] and [R] agree, *)
  jsr s R ->
  (* if the old and new lock heaps are as described by [RedAcquire], *)
  lckh = get s ->
  maps lckh lck UNLOCKED ->
  lckh' = update lckh lck LOCKED ->
  s' = set lckh' s ->
  (* if the lock invariant is [P], *)
  invariant (get R) lck P ->
  (* then the lock was initially unowned, *)
  maps (get R) lck (LockZero P) /\
  (* and [s'] agrees with the conjunction of [RP] and [R'], where [RP] justifies
     [P] and [R'] is [R] updated with the ownership of the lock. *)
  exists RP,
  jpi RP zilch P /\
  jsr s' (star
            RP
            (set (update (get R) lck (LockOne P)) R)).
Proof.
  intros. invariant_inversion. jsr_inversion. explode_resources_locks.
  (* If [lck] exists and is unlocked in the lock heap, then it exists and is
     unowned in the (global) lock resource. *)
  forwards: (@agrees_maps_left_to_right lock_status); [ eauto | eauto | unpack ].
  lock_agrees_inversion.
  (* The hidden resource satisfies the hidden invariants. We can extract
     out of it a sub-resource that satisfies [P]. *)
  fire_entailment entails_hidden_acquire.
  jpi_inversion. explode_resources_locks.
  (* The lock is unowned in the hidden resource and in the visible resource. *)
  maps_star_inversion_aatomic.
  (* Furthermore, there is just one lock invariant [P]. *)
  maps_injective. injections.
  (* Let us now prove the goal. *)
  xplit; eauto.
  econstructor; [ eauto with agrees | | explode_resources_locks; eauto ].
    (* Sub-goal 2. *)
    explode; try aac_reflexivity.
    rewrite <- @star_update_core_left. aac_reflexivity. assumption.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The converse lemma. If [s] and [R] and the lock is taken/owned in both; if
   (more precisely) [R] can be split as [star R1 R2] where [R2] owns the lock
   and [R1] justifies the lock invariant [P]; then the updated machine state [s']
   and the updated resource [R2] (alone) agree. *)

Lemma jsr_release:
  forall s R R1 R2 s' R' lck P,
  jsr s R ->
  maps (get s) lck LOCKED ->
  maps (get R2) lck (LockOne P) ->
  jpi R1 zilch P ->
  star R1 R2 = R ->
  s' = set (update (get s) lck UNLOCKED) s ->
  R' = set (update (get R2) lck (LockZero P)) R2 ->
  jsr s' R'.
Proof.
  introv ? ? hm2 ? ? ? ?. subst. jsr_inversion. explode_resources_locks.
  (* The global resource [R] has full permission over this lock. *)
  hoist_maps_maximal.
  (* We must insert into the hidden resource a sub-resource that
     satisfies [P]. This is done by exploiting this entailment. *)
  forwards entailment: entails_hidden_release; [ eauto | eauto | ].
  forwards: entailment.
    econstructor; eauto.
      equates 1. eauto. f_equal. aac_reflexivity.
  clear entailment.
  (* Split. *)
  (* This assertion helps solve the third subgoal. *)
  assert (j:
    forall R R1 R2,
    jpi R zilch (hidden_invariants R1) ->
    R1 = R2 ->
    jpi R zilch (hidden_invariants R2)
  ); [ intros; subst; assumption | ].
  econstructor; [ clear j | clear j | eauto with getpair ].
    (* Sub-goal 1. *)
    econstructor; eauto.
    eapply agrees_update; eauto with agrees.
    equates 1. eauto. aac_reflexivity.
    (* Sub-goal 3. *)
    explode; try aac_reflexivity.
    erewrite @star_update_solitary_left by eauto with solitary consistent_aac.
    f_equal. aac_reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the lock [lck] is well-typed, then it exists in the lock heap. *)

Lemma jsr_invariant:
  forall s R,
  jsr s R ->
  forall lck P,
  invariant (get R) lck P ->
  exists st,
  maps (get s) lck st.
Proof.
  intros. invariant_inversion. jsr_inversion. explode_resources_locks.
  (* Extends the [maps] assertion to the global resource. *)
  forwards: @maps_star; [ eauto | eauto | unpack ].
  forwards: @maps_star_construction; eauto.
  (* Because we have agreement, there follows that the lock exists
     in the lock heap. *)
  forwards: (@agrees_maps_right_to_left lock_status) lck; [ eauto | eauto | unpack ].
  eauto.
Qed.

Ltac jsr_invariant :=
  let st := fresh in
  forwards [ st ? ]: jsr_invariant; [
    eauto
  | eauto using invariant_affine with explode_resources aac_guess
  | destruct st ].

(* ---------------------------------------------------------------------------- *)

(* If the lock [lck] has type [TyLocked], then this lock exists in the
   lock heap and is really [LOCKED]. *)

Lemma jsr_taken:
  forall s R,
  jsr s R ->
  forall lck,
  taken (get R) lck ->
  maps (get s) lck LOCKED.
Proof.
  intros. taken_inversion. jsr_inversion. explode_resources_locks.
  (* Extends the [maps] assertion to the global resource. *)
  hoist_maps_maximal.
  (* Because we have agreement, there follows that the lock exists
     in the lock heap. *)
  forwards: (@agrees_maps_right_to_left lock_status) lck; [ eauto | eauto | unpack ].
  lock_agrees_inversion.
  assumption.
Qed.

Ltac jsr_taken :=
  forwards: jsr_taken; [
    eauto
  | eauto using taken_affine with explode_resources aac_guess
  | ].

(* ---------------------------------------------------------------------------- *)

(* If the address [l] is currently unadopted according to the resource [R], and
   if the machine state [s] and resource [R] agree, then the address [l] exists
   in the heap and has a [Null] adopter pointer. *)

Lemma jsr_unadopted:
  forall s R,
  jsr s R ->
  forall l,
  unadopted (get R) l ->
  exists v,
  maps (get s) l (Block Null v).
Proof.
  intros. jsr_inversion.
  explode_resources_adoption_strong. adoption_resource_agrees_inversion.
  eauto using adoption_resource_agrees_unadopted, unadopted_affine.
Qed.

Ltac jsr_unadopted :=
  forwards: jsr_unadopted; [
    eauto
  | eauto using unadopted_affine with explode_resources aac_guess
  | unpack ].

(* ---------------------------------------------------------------------------- *)

(* If the address [l] is adoptable according to the resource [R], and if the
   machine state [s] and resource [R] agree, then the address [l] exists in
   the heap. *)

Lemma jsr_adoptable:
  forall s R,
  jsr s R ->
  forall l,
  adoptable (get R) l ->
  exists p v,
  maps (get s) l (Block p v).
Proof.
  intros. jsr_inversion.
  explode_resources_adoption_strong. adoption_resource_agrees_inversion.
  eauto using adoption_resource_agrees_adoptable, adoptable_affine.
Qed.

Ltac jsr_adoptable :=
  forwards: jsr_adoptable; [
    eauto
  | eauto using adoptable_affine with explode_resources aac_guess
  | unpack ].

(* ---------------------------------------------------------------------------- *)

(* If the address [x] currently has a [NonNull y] adopter pointer in the heap,
   and if the machine state [s] and resource [R] agree, then the same pointer
   exists in the resource, i.e., [x] is adopted by [y]. *)

(* We must argue that the visible resource [R] owns [y] as an adopter, i.e.,
   [adopts (get R) y]. Indeed, if [y] was owned by the hidden resource, then
   the edge of [x] to [y] would be owned by the hidden resource as well. *)

Lemma jsr_adopted:
  forall s R x y v,
  jsr s R ->
  maps (get s) x (Block (NonNull y) v) ->
  adopts (get R) y ->
  adopted (get R) x y.
Proof.
  intros. jsr_inversion.
  explode_resources_adoption_strong. adoption_resource_agrees_inversion.
  eauto using adoption_resource_agrees_adopted, adopted_narrow_left.
Qed.

