Set Implicit Arguments.
Require Import MyTactics.
Require Import GetSet.
Require Import Syntax.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.

(* ---------------------------------------------------------------------------- *)

(* States. *)

(* The nature of machine states should be thought of as open-ended, or partly
   unspecified. Some parts of the operational semantics or of the type system
   care about one component of the machine state (e.g., the heap), but there
   could be other components. Furthermore, many parts of the proof in fact do
   not care about the nature of machine states. *)

(* When we wish to access or update one component of a machine state, we do
   so via the [get] and [set] overloaded functions. Thus, we remain oblivious
   to whatever other components may be present in a machine state. *)

(* At present, a machine state is a pair of an ordinary heap and a lock heap. *)

Definition state : Type :=
  (heap * lock_heap)%type.

(* Program execution begins in a fixed initial state. *)

Definition initial : state :=
  (empty_heap, empty_lock_heap).

(* ---------------------------------------------------------------------------- *)

(* A machine state is well-formed when all of its components are. *)

Inductive wks : state -> Prop :=
| Wks:
    forall h lckh,
    (* The heap contains closed values. *)
    wkh h ->
    (* No requirement on the lock heap. *)
    wks (h, lckh).

Hint Constructors wks : wk.

(* The initial state is well-kinded. *)

Lemma wks_initial:
  wks initial.
Proof.
  econstructor. eauto using wkh_empty_heap.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Configurations. *)

(* A configuration is a pair of a machine state and a term. This term may have
   kind [TERM] and represent a single thread. It may also have kind [SOUP] and
   represent a thread soup. *)

Inductive configuration :=
| Cf: state -> term -> configuration.

(* ---------------------------------------------------------------------------- *)

(* Well-kindedness for configurations. *)

Inductive wkc : configuration -> kind -> Prop :=
| Wkc:
    forall s t kappa,
    wks s ->
    wk zilch t kappa ->
    wkc (Cf s t) kappa.

Hint Constructors wkc : wk.

(* ---------------------------------------------------------------------------- *)

(* Inversion tactics. *)

Ltac destruct_states :=
  repeat match goal with s: state |- _ => destruct s as [ ? ? ] end.

Ltac destruct_configurations :=
  match goal with c: configuration |- _ => destruct c end.

Ltac select_wks action :=
  match goal with
  | h: wks _ |- _ =>
      action h
  end.

Ltac wks_inversion :=
  gji select_wks.

Ltac select_wkc action :=
  match goal with
  | h: wkc _ _ |- _ =>
      action h
  end.

Ltac wkc_inversion :=
  gji select_wkc.

