Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import Agreement.
Require Import JConf.
Require Import SubsumptionSoundness.

(* Well-typed programs are data race free. *)

(* ---------------------------------------------------------------------------- *)

(* The predicate [access mode t l] means that the term [t] is just about to
   access the memory location [l]. The mode [am] indicates whether this is a
   read access or a write access. *)

Inductive access_mode :=
  | Read
  | Write
.

Inductive access: access_mode -> term -> nat -> Prop :=
 
  | AccRead:
      forall l,
      access Read (TRead (TLoc l)) l

  | AccWrite:
      forall l v,
      access Write (TWrite (TLoc l) v) l

  | AccApp:
      forall am v t l,
      access am t l ->
      access am (TApp v t) l

  | AccThread:
      forall am t l,
      access am t l ->
      access am (TThread t) l

  | AccParLeft:
      forall am t1 t2 l,
      access am t1 l ->
      access am (TPar t1 t2) l

  | AccParRight:
      forall am t1 t2 l,
      access am t2 l ->
      access am (TPar t1 t2) l
.

(* ---------------------------------------------------------------------------- *)

(* A sanity check -- just to make sure that we haven't forgotten some
   evaluation contexts. *)

Goal
  forall am t l,
  access am t l ->
  forall E,
  eval_context E ->
  access am (fill E t) l.
Proof.
  intros; destruct_eval_context; eauto using AccApp.
Qed.

(* ---------------------------------------------------------------------------- *)

Local Ltac select_access action :=
  match goal with h: access _ _ _ |- _ => action h end.
Local Ltac access_inversion :=
  gji select_access.

(* ---------------------------------------------------------------------------- *)

(* A program exhibits a data race if two distinct threads access the same
   location [l] and at least one of these accesses is a write. *)

Inductive data_race : term -> Prop :=

  | DataRace:
      forall am1 am2 t1 t2 l,
      access am1 t1 l ->
      access am2 t2 l ->
      (am1 = Write \/ am2 = Write) ->
      data_race (TPar t1 t2)

  | DRLeft:
      forall t1 t2,
      data_race t1 ->
      data_race (TPar t1 t2)

  | DRRight:
      forall t1 t2,
      data_race t2 ->
      data_race (TPar t1 t2)

.

Inductive data_race_free : configuration -> Prop :=
  | DataRaceFree:
      forall s t,
      ~ data_race t ->
      data_race_free (Cf s t).

(* ---------------------------------------------------------------------------- *)

(* A value cannot access a memory location. *)

Lemma access_value:
  forall am t l,
  access am t l ->
  wk zilch t VALUE ->
  False.
Proof.
  induction 1; intros; try destruct_eval_context; wk_inversion.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the term [t] is well-typed under [R] and is about to access the memory
   location [l], then [R] contains an access right [m] for this location;
   furthermore, if this is a write access, then [m] is [Exclusive]. *)

(* This lemma is analogous, in its statement and in its proof, to the progress
   lemma for terms. *)

Lemma typed_access_TERM:
  forall n R1 K P t T,
  jt n R1 K P t T -> (* [t] is well-typed *)
  wk zilch t TERM -> (* [t] is closed *)
  forall R2 R,
  jpi R2 K P ->
  star R1 R2 = R ->
  consistent R ->
  forall am l,
  access am t l ->
  exists m v,
  (am = Write -> m = Exclusive) /\
  maps (get R) l (mv2dx m v).

(* A tactic to help apply the induction hypothesis. *)
Local Ltac typed_access_TERM ih :=
  forwards: ih; [
    eauto
  | eauto
  | eauto
  | eauto with jpi
  | eauto
  | eauto with consistent_aac
  | eauto
  | unpack ].

Proof.
  intro n. induction_wf ih: lt_wf n.
  introv h; dependent destruction h; intros; try subst;
  (* The term [t] cannot be a value. This discards some cases. *)
  try solve [ false; eauto using access_value, wk_empty_reverse ];
  wk_inversion;
  access_inversion;
  decomposition.

  (* JForallIntro *)
  closed. solve [ eauto ].

  (* Case: JExistsElim *)
  substitution. solve [ eauto ].

  (* JFrame *)
  typed_access_TERM ih.
  solve [ eauto 6 with explode_resources maps_maximal ].

  (* JCut *)
  typed_access_TERM ih.
  solve [ eauto with explode_resources maps_maximal ].

  (* JSubLeft *)
  solve [ eauto using jps_sound with consistent_aac ].

  (* JSubRight *)
  solve [ eauto ].

  (* JApplication *)
  typed_access_TERM ih.
  solve [ eauto 6 with explode_resources maps_maximal ].

  (* JRead *)
  solve [ eauto with explode_resources maps_maximal congruence ].
  
  (* JWrite *)
  solve [ eauto with explode_resources maps_maximal ].

Qed.

(* ---------------------------------------------------------------------------- *)

(* This property is extended to soups. *)

Lemma typed_access_SOUP:
  forall R t,
  js R t ->
  forall am l,
  access am t l ->
  consistent R ->
  exists m v,
  (am = Write -> m = Exclusive) /\  
  maps (get R) l (mv2dx m v).
Proof.
  induction 1; inversion 1; intros; subst.
  (* JsThread *)
  solve [ eauto using typed_access_TERM, @core_right_unit with wk jpi ].
  (* JsPar/left *)
  forwards: IHjs1; [ eauto | eauto with consistent_split | unpack ].
  solve [ eauto with explode_resources maps_maximal ].
  (* JsPar/right. *)
  forwards: IHjs2; [ eauto | eauto with consistent_split | unpack ].
  solve [ eauto with explode_resources maps_maximal ].
Qed.

(* ---------------------------------------------------------------------------- *)

(* Well-typed programs are data-race-free. Preliminary statement. *)

Lemma data_race_freedom_preliminary:
  forall R t,
  js R t ->
  data_race t ->
  consistent R ->
  False.
Proof.
  induction 1; inversion 1; intros; subst; eauto with consistent_split.
  (* Case: DataRace *)
  (* We have two terms, each of which accesses memory location [l]. So, each
     of them has a permission for this location. *)
  do 2 match goal with h: js ?R ?t |- _ =>
    forwards: typed_access_SOUP; [
      eexact h
    | eauto
    | eauto with consistent_split
    | clear h ]
  end.
  unpack.
  (* Contradiction. *)
  explode_resources_blocks.
  forwards: @maps_star_solitary_left; [ eauto | eauto with solitary | eauto | ].
  maps_injective.
  (* Two symmetric cases, depending on which side performs a write. *)
  match goal with h: _ \/ _ |- _ => destruct h end;
  (* Deduce that the access right is exclusive in one of [R1] and [R2]. *)
  cleanup; subst;
  (* Further case analysis of the mode. *)
  destruct_modes;
  (* Conclude. *)
  simpl in *; discriminate.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Well-typed programs are data-race-free. *)

Lemma data_race_freedom:
  forall c,
  jconf c ->
  data_race_free c.
Proof.
  inversion 1; jsr_consistent; econstructor; eauto using data_race_freedom_preliminary.
Qed.

