Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.

(* ---------------------------------------------------------------------------- *)

(* Subtyping is transitive. *)

Lemma sub_transitive:
  forall n1 n2 K T1 T2 T3,
  sub n1 K T1 T2 ->
  sub n2 K T2 T3 ->
  sub (S (mymax n1 n2)) K T1 T3.
Proof.
  econstructor. sub_inversion.
  equates 4.
    eauto 4 using JpsTransitive, (get1 jps_index_monotonic) with mymax.
    mymax.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A version of [JpsReflexive] that does not require immediate syntactic
   equality. *)

Lemma JpsReflexiveEq:
  forall n K P Q,
  P = Q ->
  jps n K P Q.
Proof.
  intros. subst. eapply JpsReflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The permission [x @ (=y | y @ T)] can be transformed to [x @ T]. *)

Lemma JpsSingletonElimination:
  forall K x y T,
  wk K T TYPE ->
  jps 1 K
    (TyAnchored x (TyMixStar (TySingleton y) (TyAnchored y T)))
    (TyAnchored x T).
Proof.
  intros.
  jps_left JpsMixStarElim.
  eapply JpsEqualsForEqualsReverseDrop with (P := (TyAnchored (TVar 0) (shift 0 T)));
    eauto using wk_weakening with wk;
    simpl_subst_goal; eauto with f_equal subst_lift.
Qed.

(* Hence, the type [=v | v @ T] is a subtype of [T]. *)

Lemma SubSingletonElimination:
  forall K v T,
  wk K T TYPE ->
  sub 2 K
    (TyMixStar (TySingleton v) (TyAnchored v T))
    T.
Proof.
  (* Unfold the definition of [sub] in terms of [jps]. *)
  econstructor.
  simpl_lift_goal.
  (* The result is then immediate. *)
  eauto using JpsSingletonElimination, wk_weakening.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Some of the rules that define [jps] can be re-formulated in terms of [sub].
   Here is an example. *)

Lemma SubCoArrow:
  forall m n K T1 T2 U1 U2,
  sub m K U1 T1 ->
  sub m K T2 U2 ->
  wk K U1 TYPE ->
  sub (S n) K
    (TyArrow T1 T2)
    (TyArrow U1 U2).
Proof.
  intros.
  eapply Sub.
  simpl_lift_goal.
  eapply JpsCoArrow.
    eapply j_weakening; eauto.
    eapply j_weakening; eauto.
    eauto using wk_weakening.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The type constructor [TyAnchored] is covariant. This is trivial: it is
   a direct consequence of the manner in which [sub] is defined in terms
   of [jps]. *)

Lemma JpsCoAnchored:
  forall n K T U,
  sub (S n) K T U ->
  forall v,
  wk K v VALUE ->
  jps n K
    (TyAnchored v T)
    (TyAnchored v U).
Proof.
  intros. sub_inversion.
  eapply j_substitution; eauto;
    simpl_subst_goal; eauto with f_equal subst_lift substitutable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The constructor [TyMixStar], which represents a conjunction of a type and a
   permission, is covariant. *)

Lemma SubCoMixStar:
  forall n K T U P Q,
  sub n K T U ->
  jps n K P Q ->
  sub (S (S (S (S n)))) K (TyMixStar T P) (TyMixStar U Q).
Proof.
  econstructor. simpl_lift_goal.
  (* The goal is to transform [x @ (T | P)] into [x @ (U | Q)]. *)
  jps_left JpsMixStarElim.
  jps_left JpsCoStar.
    sub_inversion. eapply jps_index_monotonic; eauto with omega.
    eapply j_weakening; eauto.
  eapply JpsMixStarIntro.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Universal types can be eliminated. *)

Lemma SubForallElim:
  forall K U kappa T,
  wk K U kappa ->
  sub 3 K (TyForall kappa T) (subst U 0 T).
Proof.
  econstructor.
  simpl_lift_goal.
  jps_left JpsForallType2Perm.
    simpl_lift_goal. reflexivity.
  jps_left JpsForallElimPerm.
    simpl_subst_goal. reflexivity.
    eauto using wk_weakening.
  simpl.
  rewrite lift_subst_2 by omega.
  replace (1 + 0) with 1 by omega.
  eapply JpsReflexive.
Qed.

(* Universal types are covariant. *)

Lemma SubCoForall:
  forall n K kappa T U,
  sub n (insert 0 kappa K) T U ->
  sub (S (S (S n))) K (TyForall kappa T) (TyForall kappa U).
Proof.
  intros. sub_inversion. econstructor. simpl_lift_goal. rewrite plus_0_r.
  jps_left JpsForallType2Perm; [ simpl_lift_goal; rewrite plus_0_r; reflexivity | ].
  jps_right JpsForallPerm2Type; [ | simpl_lift_goal; rewrite plus_0_r; reflexivity].
  eapply JpsCoForall.
  (* We are almost there, but we must exchange the indices 0 and 1. Tricky.
     I am doing this as an exercise, although probably making this subtyping
     axiom primitive would be much simpler than attempting to derive it in
     this way. Could the DeBruijn library be extended to facilitate this kind
     of exchange? *)
  (* Introduce a fresh name at index 2. *)
  forwards h: (get1 j_weakening); eauto.
  instantiate (1 := 1 + 1) in h.
  simpl_lift_all.
  (* Replace 0 with the new name at 2, i.e. technically, substitute TVar 1 for 0. *)
  forwards i: (get1 j_substitution).
    eauto.
    rewrite insert_insert. f_equal. omega.
    eauto.
    eauto.
    instantiate (1 := TVar 1). eapply WkTVar. eauto with lookup_insert.
    eauto with substitutable.
  simpl_subst_all.
  (* Now, check the goal. *)
  equates 1. equates 2. equates 3. eassumption.
  eauto with insert_insert.
  rewrite <- lift_lift by omega. rewrite subst_lift. reflexivity.
  rewrite <- lift_lift by omega. rewrite subst_lift. reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Existential types can be introduced. *)

Lemma SubExistsIntro:
  forall K U kappa T,
  wk K U kappa ->
  wk (insert 0 kappa K) T TYPE ->
  substitutable kappa ->
  sub 3 K (subst U 0 T) (TyExists kappa T).
Proof.
  econstructor.
  simpl_lift_goal.
  jps_right JpsExistsPerm2Type; [ | reflexivity ].
  simpl_lift_goal.
  jps_right JpsExistsIntro; eauto using wk_weakening with wk insert_insert.
  simpl_subst_goal. rewrite lift_subst_2 by omega. simpl. replace (1 + 0) with 1 by omega.
  eapply JpsReflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [TyBottom] is the least type. *)

Lemma SubBottom:
  forall K U,
  wk K U TYPE ->
  sub 3 K TyBottom U.
Proof.
  intros. equates 1. eauto using SubForallElim. simpl_subst_goal. reflexivity.
Qed.

(* [TyTop] is the greatest type. *)

Lemma SubTop:
  forall K U,
  wk K U TYPE ->
  sub 4 K U TyTop.
Proof.
  (* [TyTop] is defined as [{y : VALUE} =y]. The idea of the proof is
     as follows. As soon as we have [x @ U], for an arbitrary U, we
     have [x @ =x], by [JpsEqualityReflexive], hence we also have
     [x @ {y : VALUE} =y], by [JpsExistsIntro]. In other words, the
     fact that [x] is a well-typed value plays no role; all we need
     to know is that [x] is well-kinded, and off we go. *)
  econstructor. simpl_lift_goal.
  jps_right JpsExistsPerm2Type; [ | reflexivity ]. simpl_lift_goal.
  jps_right JpsExistsIntro.
    jps_left JpsEmptyTop. eapply JpsEqualityReflexive. eauto using WkTVar, lookup_insert_bingo.
    simpl_subst_goal. eauto.
    eauto with wk.
    eauto with wk.
    eauto with substitutable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* It is always permitted to introduce a name for a value: the type [T] is a
   subtype of [{x : VALUE} (=x | x @ T)]. *)

Lemma SubNameIntro:
  forall K T,
  wk K T TYPE ->
  sub 6 K
    T
    (TyExists VALUE
      (TyMixStar
        (TySingleton (TVar 0))
        (TyAnchored (TVar 0) (shift 0 T)))).
Proof.
  econstructor. simpl_lift_goal.
  (* x @ T <= empty * x @ T *)
  jps_left JpsEmptyAppearsLeft.
  (*   ... <= x = x * x @ T *)
  jps_left JpsCoStar.
    eapply JpsEqualityReflexive. eapply WkTVar with (x := 0). eauto with wk.
    eapply JpsReflexive.
  (*   ... <= x @ (=x | x @ T) *)
  jps_left JpsMixStarIntro.
  (*   ... <= x @ {x: VALUE} (=x | x @ T) *)
  jps_right JpsExistsPerm2Type; [ | reflexivity ]. simpl_lift_goal.
  eapply JpsExistsIntro.
    simpl_subst_goal. do 3 f_equal. replace (TVar 0) with (var 0); eauto using pun_2.
    eauto with wk.
    eauto 10 using wk_weakening with wk insert_insert.
    eauto with substitutable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* As a sanity check, we verify that if [dup T] holds, then the permission
   [TyDuplicable T] is provable via permission subsumption. *)

Lemma dup_complete:
  forall K T kappa,
  wk K T kappa ->
  logic kappa ->
  dup T ->
  exists n,
  jps n K TyEmpty (TyDuplicable T).
Local Ltac dup_complete_ih :=
  match goal with
  | ih: logic VALUE -> _ |- _ =>
      clear ih
  | ih: logic _ -> dup _ -> _ |- _ =>
      forwards: ih; [ eauto | eauto | clear ih; unpack ]
  end.
Proof.
  assert (forall i j, mymax i j >= i). intros. mymax.
  assert (forall i j, mymax i j >= j). intros. mymax.
  (* Because [dup] is co-inductive, the proof must be by induction on the
     syntax of the type [T]. *)
  induction 1; intros; subst; dup_inversion;
  try solve [
    exists 0; eauto with j
  | false
  ];
  repeat dup_complete_ih; eexists.
  (* TyForall *)
  jps_left JpsVacuousForallIntro. eauto. simpl_lift_goal; eauto.
  jps_left JpsCoForall. eauto. eapply JpsDupForall.
  (* TyExists *)
  jps_left JpsVacuousForallIntro. eauto. simpl_lift_goal; eauto.
  jps_left JpsCoForall. eauto. eapply JpsDupExists.
  (* TyMixStar *)
  jps_left JpsEmptyAppearsLeft.
  jps_right JpsDupMixStar; eauto with wk.
  eauto using JpsCoStar, (get1 jps_index_monotonic).
  (* TyAnchored *)
  eauto with j.
  (* TyStar *)
  jps_left JpsEmptyAppearsLeft.
  jps_right JpsDupStar; eauto with wk.
  eauto using JpsCoStar, (get1 jps_index_monotonic).
  (* RowCons *)
  jps_left JpsEmptyAppearsLeft.
  jps_right JpsDupCons; eauto with wk.
  eauto using JpsCoStar, (get1 jps_index_monotonic).
  (* TyRef *)
  jps_right JpsDupRef; eauto with wk.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Thus, if [dup P] holds, then [P] can be converted to [P * P]. This was
   a primitive subtyping rule before we introduced [TyDuplicable]. *)

Lemma dup_complete_corollary:
  forall K P,
  wk K P PERM ->
  dup P ->
  exists n,
  jps n K P (TyStar P P).
Proof.
  intros.
  forwards: dup_complete; eauto with logic. unpack.
  eexists.
  jps_left JpsEmptyAppearsLeft.
  jps_left JpsCoStar.
    eassumption.
    eapply JpsReflexive.
  eapply JpsDuplicate.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Other possible definitions of the top type. *)

(* The usual definition of the top type in System F is [{a : TYPE} a]. *)

Notation TyTopStandard :=
  (TyExists TYPE (TVar 0)).

Lemma wk_TyTopStandard:
  forall K,
  wk K TyTopStandard TYPE.
Proof.
  eauto with wk logic substitutable.
Qed.

(* [TyTopStandard] is also a top element, hence is interconvertible with
   [TyTop]. *)

Lemma SubTopStandard:
  forall K U,
  wk K U TYPE ->
  sub 3 K U TyTopStandard.
Proof.
  intros. equates 2.
    eauto 8 using SubExistsIntro with wk logic substitutable.
    simpl_subst_goal. eauto.
Qed.

(* Another top type in Mezzo is [{a : TYPE} (a | duplicable a)]. *)

Notation TyTopDup :=
  (TyExists TYPE (TyMixStar (TVar 0) (TyDuplicable (TVar 0)))).

Lemma wk_TyTopDup:
  forall K,
  wk K TyTopDup TYPE.
Proof.
  eauto 10 with wk logic substitutable lookup_insert.
Qed.

(* [TyTopDup] is also a top element, hence is interconvertible with
   [TyTop] and [TyTopStandard]. *)

Lemma SubTopDup:
  forall K U,
  wk K U TYPE ->
  exists n,
  sub n K U TyTopDup.
Proof.
  intro K.
  (* They key is to show that [TyTop] is a subtype of [TyTopDup].  This is
     true because [TyTop] is duplicable, hence [TyTop] can be converted to
     [TyMixStar TyTop (TyDuplicable TyTop)], and then it suffices to use an
     existential introduction rule. *)
  (* We begin with this intermediate lemma. *)
  assert (exists n, sub n K TyTop (TyMixStar TyTop (TyDuplicable TyTop))).
    forwards: dup_complete; [ eapply wk_TyTop | eauto with logic | eapply  dup_TyTop | unpack ].
    eexists. econstructor. simpl_lift_goal.
    jps_left JpsEmptyAppearsRight.
    jps_right JpsMixStarIntro.
    eapply JpsCoStar; [ eapply JpsReflexive | eassumption ].
  unpack. intros. eexists.
  (* Now. We know that [U] is a subtype of [TyTop]. *)
  eapply sub_transitive; [ eapply SubTop; eauto | ].
  (* So there remains to show that [TyTop] is a subtype of [TyTopDup]. *)
  eapply sub_transitive; [ eassumption | ].
  (* The existential introduction step. *)
  equates 2.
    eapply SubExistsIntro.
      eapply wk_TyTop. (* Force the witness. *)
      eauto 8 with wk logic substitutable lookup_insert.
      eauto with logic substitutable.
  simpl_subst_goal. eauto. (* Phew. *)
Qed.

