Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import DerivedSubtyping.

(* The following rules are meant to be used for type-checking source programs.
   They are not used in the type soundness proof. For this reason, instead of
   trying to allow an arbitrary resource [R], we specialize the rules and work
   with [void]. *)

(* ---------------------------------------------------------------------------- *)

(* Lemmas that help adjust the resource that appears in a typing judgement. *)

Lemma JSameResource:
  forall f n R1 R2 K P t T,
  j f n R1 K P t T ->
  R2 = R1 ->
  j f n R2 K P t T.
Proof.
  intros. subst. eauto.
Qed.

Lemma JpiSameResource:
  forall R1 R2 K P,
  jpi R1 K P ->
  R2 = R1 ->
  jpi R2 K P.
Proof.
  intros. subst. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derivation for a judgement of the form [jt _ _ _ TyEmpty t T] often begins
   with an application of [JCut]. In order to facilitate its use, we prove a
   derived rule. *)

Lemma JtTopCut:
  forall n R1 R2 R K P t T,
  jt n R2 K P t T ->
  jpi R1 K P ->
  star R1 R2 = R ->
  jt (S (S n)) R K TyEmpty t T.
Proof.
  eauto using JCut, (@JSubLeft 0), JpsDisappearsRight.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A variant of [JFrame] that frames everything. *)

Lemma JFrameAll:
  forall n K Q v T,
  jt n void K TyEmpty v T ->
  jt (S (S n)) void K Q v (TyMixStar T Q).
Proof.
  eauto using (@JSubLeft 0), JpsEmptyAppearsLeft, JFrame.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derived typing rule for values. *)

(* The native typing rule for values, [JSingleton], asserts that a value [v]
   has the singleton type [=v]. This judgement does not require any permission
   as part of its pre-condition. Here, we develop a variant of this rule,
   which states that, if we have the permission [v @ T], then the value [v]
   has type [T]. *)

Lemma JValue:
  forall K v T,
  wk K v VALUE ->
  wk K T TYPE ->
  jt 3 void K (TyAnchored v T) v T.
Proof.
  intros.
  (* We build the derivation from the bottom up. *)
  eapply JSubRight; [ | eapply SubSingletonElimination; eauto ].
  eapply JFrameAll.
  eapply JSingleton; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A simplified variant of [JFunction]. In the conclusion, the precondition is
   just [TyEmpty]. And in the premise, the function body [t] is type-checked
   under just [x @ T] instead of [P * x @ T]. Oh, and the conclusion is a [jpi]
   judgement. *)

Lemma JFunctionEmpty:
  forall n R K T t U,
  wk K T TYPE ->
  jt
    n
    (core R)
    (insert 0 VALUE K)
    (TyAnchored (TVar 0) (shift 0 T))
    t
    (shift 0 U)
  ->
  consistent R ->
  jpi R K (TyAnchored (TAbs t) (TyArrow T U)). (* jctop R K (TAbs t) (TyArrow T U) *)
Proof.
  intros.
  econstructor; [ eauto with wk | | | eapply core_left_unit ].
  (* Premise 1. Type-check the function body. *)
  eapply JFunction; [ eauto | ].
  solve [ eauto using (@JSubLeft 0), JpsDisappearsLeft, JSameResource, @core_idempotent ].
  (* Premise 2: Justify [TyDuplicable Empty] and [TyEmpty]. *)
  eauto using @core_left_unit with jpi wk logic dup.
Qed.

(* ---------------------------------------------------------------------------- *)

(* We may express the typing rule for functions in (almost) the form it
   assumed before we introduced [TyDuplicable]. That is, instead of requiring
   a permission [TyDuplicable P], we use a premise of the form [dup P]. *)

Lemma OldJFunction:
  forall f n R K P T t U,
  wk K P PERM ->
  wk K T TYPE ->
  dup P ->
  j
    Unrestricted
    n
    (core R)
    (insert 0 VALUE K)
    (intro P T)
    t
    (shift 0 U)
  ->
  j f (S (S n)) R K P (TAbs t) (TyArrow T U).
Proof.
  intros.
  eapply JCut.
    eapply JFunction; eauto.
    (* This works out of the box because [dup P] justifies the permission
       [TyDuplicable P]. *)
    eauto with jpi logic.
    eauto using @core_left_unit.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Derived typing rules for sequencing. *)

(* In this version, all of the permissions that are initially available,
   represented by [P], are sent into the left-hand side [t]. Thus, the
   only permission that is available to the right-hand side [u] is the
   permission [x @ T]. *)

(* K |- { P } t : T
   K; x: VALUE ||- { x @ T } u : U
   x # T, U
   -------------------------------------------
   K |- { P } let x = t in u : U            *)

Lemma JLet:
  forall m n K P T U t u,
  jt n void K P t T ->
  jt m void
    (insert 0 VALUE K)
    (TyAnchored (TVar 0) (shift 0 T))
    u
    (shift 0 U)
  ->
  wk K T TYPE ->
  jt (S (S n)) void K P (TLet t u) U.
Proof.
  intros.
  eapply JCut; [ | | eapply duplicable_void ].
  (* The application [TLet t u] is well-typed under
     the pre-condition [(TAbs u) @ (TyArrow T U)]. *)
  solve [ eapply JApplication; eauto with wk ].
  (* The permission [(TAbs u) @ (TyArrow T U)] is justified. *)
  eauto using JFunctionEmpty, JSameResource, core_void, consistent_void.
Qed.

(* In this version, we have built in an application of the frame rule. The
   permissions that are initially available are split as [P * Q]. The
   permission [P] is sent into the left-hand side [t], whereas [Q] is not made
   available to [t] and remains available for consumption by [u]. *)

(* K |- { P } t : T
   K; x: VALUE ||- { Q * x @ T } u : U
   x # Q, T, U
   -----------------------------------------------
   K |- { P * Q } let x = t in u : U            *)

Lemma JLetFrame:
  forall m n K P Q T U t u,
  jt n void K P t T ->
  jt m void
    (insert 0 VALUE K)
    (TyStar (shift 0 Q) (TyAnchored (TVar 0) (shift 0 T)))
    u
    (shift 0 U) 
  ->
  wk K T TYPE ->
  wk K Q PERM ->
  jt (S (S (S n))) void K (TyStar P Q) (TLet t u) U.
Proof.
  intros.
  eapply JCut; [ | | eapply duplicable_void ].
  (* [t] has type [T] under the pre-condition [P]. We frame [Q] around
     this judgement, so [t] has type [T | Q] under [P * Q]. *)
  eapply JApplication; [ eauto with wk | ].
  solve [ eapply JFrame; eauto ].
  (* The permission [(TAbs u) @ (TyArrow (T | Q) U)] is justified. *)
  eapply JFunctionEmpty; [ eauto with wk | simpl_lift_goal | eauto using consistent_void ].
  rewrite core_void.
  eapply (@JSubLeft 1); [ eassumption | simpl_lift_goal ].
  jps_left JpsMixStarElim.
  eauto with j.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derived typing rule for applications. *)

(* The rule states that if the function [t] has type [U -> T] under the
   precondition [P] and if the argument [u] has type [U] under the
   precondition [Q] then the application has type [T] under the combined
   precondition [P * Q]. *)

(* The application is not written under the form [t u]. Indeed, this form
   is not considered well-kinded unless [t] is a value. Furthermore, even
   if we were to require that [t] be a value, that would not be sufficient:
   we would have to also require a canonical derivation for [t] under an
   empty precondition, a very strong requirement. Thus, the present
   formulation seems more satisfactory. It reflects the fact that the
   type system requires A-normal form. *)

Lemma JNormalApp:
  forall m n K P Q t u T U,
  jt n void K P t (TyArrow U T) ->
  jt m void K Q u U ->
  wk K P PERM ->
  wk K Q PERM ->
  jt (S (S (S n))) void K (TyStar P Q) (TLet t (TApp (TVar 0) (shift 0 u))) T.
Proof.
  intros.
  eapply JLetFrame.
    eauto.
    eapply (@JSubLeft 0).
      eapply JApplication.
        eauto with wk.
        eapply j_weakening; eauto.
      simpl_lift_goal. eapply JpsStarCommutative.
    eauto with wk.
    eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A variant of [JForallIntro]. Instead of allowing an arbitrary [P] in the
   premise and introducing [forall x.P] in the conclusion, we require that
   [P] do not mention [x], and we use [P] in the premise and conclusion. *)

Lemma JForallIntroVariant:
  forall n R K P t kappa T,
  harmless t ->
  jt n R (insert 0 kappa K) (shift 0 P) (shift 0 t) T ->
  substitutable kappa ->
  jt (S (S n)) R K P t (TyForall kappa T).
Proof.
  intros.
  (* Apply JForallIntro. *)
  forwards: JForallIntro; eauto.
  (* We now have a vacuous quantification [forall x.P], where [x]
     does not occur in [P]. Use permission subsumption to argue
     that this is equivalent to [P]. *)
  eauto using (@JSubLeft 1), JpsVacuousForallIntro.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Any well-kinded value has type [TyTop]. *)

Lemma JTop:
  forall f R K v,
  wk K v VALUE ->
  j f 1 R K TyEmpty v TyTop.
Proof.
  intros.
  eapply JExistsIntro with (U := v); eauto with wk logic substitutable.
  simpl_subst_goal. eauto with j.
Qed.

(* Hence, any value has type [TyMixStar TyTop P] under the assumption [P]. *)

Lemma JTopMix:
  forall f R K P v, 
  wk K v VALUE ->
  j f 3 R K P v (TyMixStar TyTop P).
Proof.
  intros.
  eapply JCut.
    eapply JFrame. eauto using JTop.
    eauto with j.
    eauto using @core_right_unit.
Qed.

(* This can be formulated also as follows. *)

Lemma JtTopMix:
  forall R1 R2 R K P v,
  wk K v VALUE ->
  jpi R2 K P ->
  star R1 R2 = R ->
  consistent R ->
  jt 5 R K TyEmpty v (TyMixStar TyTop P).
Proof.
  intros; subst. eapply JtTopCut; eauto using JTopMix, @star_commutative.
Qed.

