Set Implicit Arguments.
Require Import List.
Require Import MyList.
Require Import MyTactics.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JAffinity.

(* ---------------------------------------------------------------------------- *)

(* A notion of permission entailment is useful when reasoning about the
   evolution of the hidden invariants. (There is no need to use [jps],
   which is incomplete and clumsy, for this purpose.) *)

Definition entails P Q :=
  forall R,
  jpi R zilch P ->
  jpi R zilch Q.

(* ---------------------------------------------------------------------------- *)

(* This tactic proves many leaf entailments. *)

Ltac entails :=
  intros;
  unfold entails in *;
  intros;
  eauto; (* first attempt *)
  jpi_inversion;
  try solve [
    (* second attempt *)
    eauto using @core_left_unit, @core_right_unit with jpi
  | (* third attempt, up to AC. *)
    (
      equates 3; [
        eauto using @core_left_unit, @core_right_unit with jpi
      | instantiate; aac_reflexivity
      ]
    )
  | (* last attempt, up to AC and affinity. *)
    (
      eapply jpi_drop_right; [
        eauto using @core_left_unit, @core_right_unit with jpi
      | instantiate; aac_guess
      ]
    )
  ]
.

(* ---------------------------------------------------------------------------- *)

(* A collection of basic properties of entailment. *)

Lemma entails_transitive:
  forall P1 P2 P3,
  entails P1 P2 ->
  entails P2 P3 ->
  entails P1 P3.
Proof.
  entails.
Qed.

Lemma entails_TyStar_left:
  forall P1 P2 Q,
  entails P1 P2 ->
  entails (TyStar P1 Q) (TyStar P2 Q).
Proof.
  entails.
Qed.

Lemma entails_TyStar_right:
  forall P1 P2 Q,
  entails P1 P2 ->
  entails (TyStar Q P1) (TyStar Q P2).
Proof.
  entails.
Qed.

Lemma entails_TyStar_left_iterated:
  forall Ps P Q,
  entails P Q ->
  entails
    (fold_left TyStar Ps P)
    (fold_left TyStar Ps Q).
Proof.
  induction Ps; simpl; eauto using entails_TyStar_left.
Qed.

Lemma entails_TyStar_right_iterated:
  forall Ps P Q,
  entails P Q ->
  entails
    (fold_right TyStar P Ps)
    (fold_right TyStar Q Ps).
Proof.
  induction Ps; simpl; eauto using entails_TyStar_right.
Qed.

Lemma entails_extract_zero:
  forall P Q Ps,
  entails
    (fold_left TyStar Ps (TyStar P Q))
    (TyStar Q (fold_left TyStar Ps P)).
Proof.
  intros P Q. apply rev_ind; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  intros.
  repeat rewrite fold_left_app. simpl.
  eapply entails_transitive.
    eapply entails_TyStar_left. eassumption.
    entails.
Qed.

Lemma entails_insert_zero:
  forall P Q Ps,
  entails
    (TyStar Q (fold_left TyStar Ps P))
    (fold_left TyStar Ps (TyStar P Q)).
Proof.
  intros P Q. apply rev_ind; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  intros.
  repeat rewrite fold_left_app. simpl.
  entails.
Qed.

Lemma entails_left_right:
  forall P Ps,
  entails
    (fold_left TyStar Ps P)
    (fold_right TyStar P Ps).
Proof.
  induction Ps; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  eapply entails_transitive.
    eapply entails_extract_zero.
    eapply entails_TyStar_right. eauto.
Qed.

Lemma entails_right_left:
  forall P Ps,
  entails
    (fold_right TyStar P Ps)
    (fold_left TyStar Ps P).
Proof.
  induction Ps; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  eapply entails_transitive.
    eapply entails_TyStar_right. eauto.
    eapply entails_insert_zero.
Qed.

Lemma entails_TyIteratedStar_extract:
  forall n Ps P,
  is_nth n Ps P ->
  entails
    (TyIteratedStar Ps)
    (TyStar P (TyIteratedStar (set_nth n Ps TyEmpty))).
Proof.
  induction 1; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  eapply entails_transitive.
    eapply entails_TyStar_right. eauto.
    entails.
Qed.

Lemma entails_TyIteratedStar_insert:
  forall n Ps P,
  is_nth n Ps TyEmpty ->
  entails
    (TyStar P (TyIteratedStar Ps))
    (TyIteratedStar (set_nth n Ps P)).
Proof.
  induction n; inversion 1; subst; simpl.
  (* Base case. *)
  entails.
  (* Step case. *)
  eapply entails_transitive; [
  | eapply entails_TyStar_right; eauto (* uses IHn *) ].
  entails.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This tactic fires a lemma whose conclusion is an entailment assertion.
   The [jpi] fact that serves to justify the left-hand side of the
   entailment must appear as a hypothesis, and is consumed. *)

Ltac fire_entailment lemma :=
  let e := fresh in
  (* Create the entailment statement by applying the lemma. *)
  (forwards e: lemma; eauto);
  (* Then, use the entailment statement by applying it to a [jpi] fact. *)
  match goal with h: jpi _ _ _ |- _ =>
    (forwards: e h; eauto);
    (* Clear the entailment statement, and clear the [jpi] hypothesis
       that was used too, so as to avoid confusion. *)
    clear h e
  end.

