Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WellKindedness.

(* ---------------------------------------------------------------------------- *)

(* Evaluation contexts. *)

(* Evaluation contexts are defined as special terms, in which the hole is
   represented by a variable of index 0. This is the only free variable in an
   evaluation context -- the rest is closed. *)

Notation hole :=
  (TVar 0).

(* Filling the hole of an evaluation context is substituting for 0. *)

Notation fill E t :=
  (subst t 0 E).

(* Here, we are only interested in evaluation contexts of kind [TERM] with a
   hole of kind [TERM]. In order to allow reduction under the constructs
   [TThread] and [TPar], we use ad hoc reduction rules; we do not make these
   constructs part of evaluation contexts. *)

(* ---------------------------------------------------------------------------- *)

(* We begin with a definition of shallow (one level deep) evaluation contexts.
   They are used directly to allow reduction under a context. They are used
   also in the definition of deep evaluation contexts, which are used as part
   of the reduction rule for [TSpawn]. *)

Inductive eval_context : term -> Prop :=

  (* The right-hand side of an application is an evaluation context. *)

  | EvalContextAppRight:
      forall v,
      wk zilch v VALUE ->
      eval_context (TApp v hole)

.

Hint Constructors eval_context : eval_context.

(* ---------------------------------------------------------------------------- *)

(* Deep evaluation contexts are lists of shallow evaluation contexts. *)

Inductive deep_eval_context : term -> Prop :=

  | DeepEvalContextHole:
      deep_eval_context hole

  | DeepEvalContextInd:
      forall E D D',
      eval_context E ->
      deep_eval_context D ->
      fill E D = D' ->
      deep_eval_context D'

.

Hint Constructors deep_eval_context : deep_eval_context.

(* ---------------------------------------------------------------------------- *)

(* Filling a term that does not contain a hole has no effect. *)

Lemma fill_zilch:
  forall v k t,
  wk zilch v k ->
  fill v t = v.
Proof.
  intros.
  eapply closed_subst_invariant. 
    eauto using wk_closed.
    eauto using length_empty.
Qed.

Ltac rewrite_fill_zilch :=
  repeat match goal with
  | hzilch: wk zilch ?v _ |- context[fill ?v ?t] =>
      rewrite (fill_zilch t hzilch)
end.

Ltac rewrite_fill_zilch_all :=
  rewrite_fill_zilch;
  repeat match goal with
  | hzilch: wk zilch ?v _, hf: context[fill ?v ?t] |- _ =>
      rewrite (fill_zilch t hzilch) in hf
end.

Ltac destruct_eval_context :=
  match goal with
  | h: eval_context _ |- _ =>
      destruct h; simpl_subst_all; rewrite_fill_zilch_all
  end.

(* ---------------------------------------------------------------------------- *)

(* Filling is associative. *)

Lemma fill_fill:
  forall D t E,
  eval_context E ->
  fill (fill E D) t = fill E (fill D t).
Proof.
  intros. destruct_eval_context. reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Filling a shallow evaluation context cannot yield a value. *)

Lemma eval_context_value:
  forall E K t,
  eval_context E ->
  wk K (fill E t) VALUE ->
  False.
Proof.
  intros. destruct_eval_context.
  match goal with h: wk _ (TApp _ _) VALUE |- _ => inversion h end.
Qed.

(* Filling a deep evaluation context cannot yield a value, unless
   the context is empty and the term in the hole is a value. *)

Lemma deep_eval_context_value:
  forall D K t,
  deep_eval_context D ->
  wk K (fill D t) VALUE ->
  wk K t VALUE.
Proof.
  induction 1; subst; simpl_subst_goal; intros.
  eauto.
  false. eapply eval_context_value; eauto.
  rewrite fill_fill in * by eauto.
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Filling an evaluation context replaces the hole with the term of interest. *)

Lemma fill_eval_context:
  forall v t,
  wk zilch v VALUE ->
  fill (TApp v hole) t = TApp v t.
Proof.
  intros. simpl_subst_goal. rewrite_fill_zilch. auto.
Qed.

(* Filling an evaluation context cannot produce [TFail]. *)

Lemma fill_fail:
  forall E t,
  eval_context E ->
  fill E t = TFail ->
  False.
Proof.
  intros. destruct_eval_context; discriminate.
Qed.

(* Filling an evaluation context cannot produce [TSpawn]. *)

Lemma fill_spawn:
  forall E t v1 v2,
  eval_context E ->
  fill E t = TSpawn v1 v2 ->
  False.
Proof.
  intros. destruct_eval_context; discriminate.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Deep evaluation contexts are well-kinded. *)

Lemma wk_deep_eval_context:
  forall D K,
  deep_eval_context D ->
  wk (insert 0 TERM K) D TERM.
Proof.
  induction 1; subst.
  eauto with wk.
  destruct_eval_context. eauto using wk_empty with wk.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Filling a deep evaluation context preserves well-kindedness. *)

Lemma wk_fill_deep_eval_context:
  forall D K t,
  deep_eval_context D ->
  wk K t TERM ->
  wk K (fill D t) TERM.
Proof.
  eauto using wk_substitution, wk_deep_eval_context.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This tactic solves a goal by contradiction, in the situation where there is
   an inconsistent hypothesis of the form [t1 = fill D t2]. For instance, [t1]
   could be [TAbs _], and [t2] could be [TSpawn _ _]. Regardless of [D], there
   is no way that this equation can hold. *)

Ltac inconsistent_fill :=
  match goal with
  | h1: deep_eval_context ?D,
    h2: _ = fill ?D _ |- _ =>
      inversion h1; subst; simpl_subst_all; [
        discriminate
      | destruct_eval_context; discriminate
      ]
  end.

