Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.

(* ------------------------------------------------------------------------- *)

(* These type classes describe the behavior of a pair of [get] and [set]
   functions that allow accessing and updating a component of type [A]
   in a larger structure of type [S]. *)

Class GetOp `{A : Type, St : Type} := {

  get : St -> A

}.

Class SetOp `{A : Type, St : Type} := {

  set : A -> St -> St

}.

Class GetSet `{GetOp A St, SetOp A St} := {

  get_set:
    forall a s,
    get (set a s) = a;

  set_get:
    forall s,
    set (get s) s = s;

  set_set:
    forall a1 a2 s,
    set a2 (set a1 s) = set a2 s

}.

(* ------------------------------------------------------------------------- *)

Section GetSetProperties.

Context `{GetSet A St}.

(* [set] is injective in its first argument. *)

Lemma set_injective:
  forall a1 a2 s1 s2,
  set a1 s1 = set a2 s2 ->
  a1 = a2.
Proof.
  intros.
  rewrite <- (get_set a1 s1).
  rewrite <- (get_set a2 s2).
  congruence.
Qed.

End GetSetProperties.

(* ------------------------------------------------------------------------- *)

(* Our goal is to allow any component of a tuple to be accessed via a [GetSet]
   instance. Of course, every component of the tuple must have a distinct type,
   otherwise which component is desired is ambiguous. *)

(* Ideally, we would give an identity instance that allows accessing a
   component of type [A] in a structure of type [A]; two instances that allow
   accessing the first and second component of a pair; and a transitivity
   instance that allows composing two arbitrary instances. However, that would
   be too liberal; it would lead the search for instances to diverge. We must
   formulate our instances in such a way that the search for instances looks
   for a smaller instance at each step. *)

(* ------------------------------------------------------------------------- *)

(* The identity instance. *)

Instance get_id (A : Type) : @GetOp A A := {
  get := fun a => a
}.

Instance set_id (A : Type) : @SetOp A A := {
  set := fun a _ => a
}.

Program Instance getset_id (A : Type) : (@GetSet A A (@get_id A) (@set_id A)).

(* ------------------------------------------------------------------------- *)

(* The combined first projection and transitivity instance. *)

(* This says that if one knows how to extract a [C] out of an [A], then, by
   composing with the first projection, one knows how to extract a [C] out of
   a pair of type [A * B]. Note that, when read from bottom to top, during
   proof search, this reduces the problem of finding an instance of [GetSet C
   (A * B)] to the smaller problem of finding an instance of [GetSet C A]. *)

Section Transitivity.

Variables A B C : Type.

Variable g : @GetOp C A.

Variable s : @SetOp C A.

Variable gs : @GetSet C A g s.

Global Instance get_fst : @GetOp C (A * B) := {
  get := (fun ab => get (fst ab))
}.

Global Instance set_fst : @SetOp C (A * B) := {
  set := (fun c ab => (set c (fst ab), snd ab))
}.

Global Program Instance getset_fst : @GetSet C (A * B) get_fst set_fst.

Next Obligation.
  rewrite get_set. reflexivity.
Qed.

Next Obligation.
  rewrite set_get. reflexivity.
Qed.

Next Obligation.
  rewrite set_set. reflexivity.
Qed.

(* ------------------------------------------------------------------------- *)

(* The combined second projection and transitivity instance. *)

Global Instance get_snd : @GetOp C (B * A) := {
  get := (fun ba => get (snd ba))
}.

Global Instance set_snd : @SetOp C (B * A) := {
  set := (fun c ba => (fst ba, set c (snd ba)))
}.

Global Program Instance getset_snd : @GetSet C (B * A) get_snd set_snd.

Next Obligation.
  rewrite get_set. reflexivity.
Qed.

Next Obligation.
  rewrite set_get. reflexivity.
Qed.

Next Obligation.
  rewrite set_set. reflexivity.
Qed.

End Transitivity.

(* ------------------------------------------------------------------------- *)

(* We can define an instance of [get] for the subset type [@sig A P]. We cannot
   define a corresponding [set] instance, because not every element of type [A]
   satisfies [P]. *)

Instance get_sig (A C : Type) (P : A -> Prop) (g : @GetOp C A)
: @GetOp C (@sig A P) := {
  get := fun a => get (proj1_sig a)
}.

(* ------------------------------------------------------------------------- *)

(* Examples. *)

Goal
  get 2 = 2.
Proof.
  reflexivity.
Qed.

Goal
  get (2, tt) = 2.
Proof.
  reflexivity.
Qed.

Goal
  get (2, tt) = tt.
Proof.
  reflexivity.
Qed.

Goal
  set 3 (2, tt) = (3, tt).
Proof.
  reflexivity.
Qed.

Goal
  get (2, True, tt) = 2.
Proof.
  reflexivity.
Qed.

Goal
  get (2, True, tt) = True.
Proof.
  reflexivity.
Qed.

Goal
  get (2, True, tt) = tt.
Proof.
  reflexivity.
Qed.

Goal
  exists c,
  set 3 (True, tt, 0) = (True, tt, c).
Proof.
  eexists. simpl. reflexivity.
Qed.

Goal
  set 3 (True, 0, tt) = (True, 3, tt).
Proof.
  reflexivity.
Qed.

Goal
  set 3 (0, True, tt) = (3, True, tt).
Proof.
  reflexivity.
Qed.

Goal
  get (0, (True, tt)) = True.
Proof.
  reflexivity.
Qed.

Goal
  get (0, (True, tt)) = tt.
Proof.
  reflexivity.
Qed.

Goal
  forall A : Type,
  forall P : A -> Prop,
  forall x : A,
  forall h : P x,
  get (exist P x h) = x.
Proof.
  reflexivity.
Qed.

Goal
  forall A B : Type,
  forall P : A * B -> Prop,
  forall x : A,
  forall y : B,
  forall h : P (x, y),
  get (exist _ _ h) = y. (* would work with [x], or [(x, y)], too! *)
Proof.
  reflexivity.
Qed.

(* ------------------------------------------------------------------------- *)

(* These lemmas help [eauto] instantiate a meta-variable with a pair. *)

Lemma fst_pair:
  forall A B (a : A) (b : B),
  get (a, b) = a.
Proof.
  reflexivity.
Qed.

Lemma snd_pair:
  forall A B (a : A) (b : B),
  get (a, b) = b.
Proof.
  reflexivity.
Qed.

Hint Resolve fst_pair snd_pair : getpair.

