Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MyList.
Require Import GetSet.
Require Import AllocationMaps.

(* ------------------------------------------------------------------------- *)

(* Allocation maps are looked up using [maps], which is a relation, not a
   total function. Thus, we cannot define a [GetOp] instance for them.
   However, one thing we can do is lifting [get] up to the level of allocation
   maps, as follows: if one can get an [A] out of a [B], then one can get an
   [aMap A] out of an [aMap B]. *)

(* This is convenient because it allows changing [get . maps] to [maps . get],
   and (as a result) changing [get . maps . get] to just [maps . get]. *)

Instance get_aMap (A B : Type) (g: @GetOp A B) : @GetOp (aMap A) (aMap B) := {
  get := transform get
}.

(* ------------------------------------------------------------------------- *)

(* The corresponding [SetOp] instance. This [set] operation works as intended
   if one works with defined maps and the allocation limits agree. *)

Instance set_aMap (A B : Type) (s: @SetOp A B) : @SetOp (aMap A) (aMap B) := {
  set := mzip set
}.

(* ------------------------------------------------------------------------- *)

(* Unfortunately, we cannot prove that the above instances are coherent by
   defining an instance of [GetSet], because two of the laws, namely [get_set]
   and [set_set], have side conditions. See the lemmas that involve [transform]
   and [mzip] in [AllocationMaps]. *)

(*

Program Instance getset_aMap
  (A B : Type)
  (g : @GetOp A B) (s : @SetOp A B) (H : @GetSet A B g s)
: (@GetSet (aMap A) (aMap B) (@get_aMap A B g) (@set_aMap A B s)).

*)

(* ------------------------------------------------------------------------- *)

(* The following properties concern allocation maps whose codomain is a
   product. Perhaps they could be stated in a more general way. *)

Require Import MSA.
Require Import MSAAllocationMaps.
Require Import MSAProduct.
Transparent msa_map. (* Just for us *)

Section MapProduct.

Context `{MSA A, MSA B}.

Variable R1 R2 : aMap (A * B).

Lemma get_star_fst:
  get (star R1 R2) = star (get R1) (get R2 : aMap A).
Proof.
  eapply transform_mzip. decompose_pairs. eauto.
Qed.

Lemma get_star_snd:
  get (star R1 R2) = star (get R1) (get R2 : aMap B).
Proof.
  eapply transform_mzip. decompose_pairs. eauto.
Qed.

Lemma get_core_fst:
  get (core R1) = core (get R1 : aMap A).
Proof.
  simpl.
  erewrite transform_transform with (h := fun ab => core (fst ab)).
  erewrite transform_transform; reflexivity.
  (* The essence of the proof is that [fst] and [core] commute. *)
  decompose_pairs. msa_pair. simpl. reflexivity.
Qed.

Lemma get_core_snd:
  get (core R1) = core (get R1 : aMap B).
Proof.
  simpl.
  erewrite transform_transform with (h := fun ab => core (snd ab)).
  erewrite transform_transform; reflexivity.
  decompose_pairs. msa_pair. simpl. reflexivity.
Qed.

Lemma consistent_get_fst:
  consistent (star R1 R2) ->
  consistent (get (star R1 R2) : aMap A).
Proof.
  simpl; intros.
  eapply everywhere_transform; [ eauto | ].
  decompose_pairs. msa_pair. unpack. eauto.
Qed.

Lemma consistent_get_snd:
  consistent (star R1 R2) ->
  consistent (get (star R1 R2) : aMap B).
Proof.
  simpl; intros.
  eapply everywhere_transform; [ eauto | ].
  decompose_pairs. msa_pair. unpack. eauto.
Qed.

Lemma rely_get_fst:
  rely R1 R2 ->
  rely (get R1) (get R2 : aMap A).
Proof.
  destruct R1; destruct R2; simpl; intros; try tauto.
  eapply growth_map_map; [ | intros; eapply Forall_map; [ eauto | ] | eauto ];
  (* Two subgoals, involving basic properties of the pair MSA. *)
  decompose_pairs; msa_pair; injections; simpl; tauto.
Qed.

Lemma rely_get_snd:
  rely R1 R2 ->
  rely (get R1) (get R2 : aMap B).
Proof.
  destruct R1; destruct R2; simpl; intros; try tauto.
  eapply growth_map_map; [ | intros; eapply Forall_map; [ eauto | ] | eauto ];
  (* Two subgoals, involving basic properties of the pair MSA. *)
  decompose_pairs; msa_pair; injections; simpl; tauto.
Qed.

End MapProduct.

