Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WellKindedness.
Require Import EvaluationContexts.
Require Import Heaps.
Require Import Configurations.
Require Import Reduction.

(* ------------------------------------------------------------------------- *)

(* The simplest form of the value restriction states that [JForallIntro] is
   restricted to values. We wish to be slightly more permissive. In
   particular, we wish to allow a memory allocation expression to receive a
   polymorphic type, otherwise we would not be able to construct a polymorphic
   empty list. Thus, we define a subcategory of harmless terms. This
   subcategory must be stable by substitution, preserved by reduction, and
   must have the property that a harmless term cannot allocate hidden state
   (i.e., cannot create a new lock). *)

Inductive harmless: term -> Prop :=

  (* An application of an unknown function is conservatively considered
     harmful. An application of a known function, [TApp (TAbs t) u], is
     in a fact a sequencing construct, [TLet u t], and is harmless if [u]
     adn [t] are harmless. *)

  | HlTLet:
      forall u t,
      harmless u ->
      harmless t ->
      harmless (TLet u t)

  (* [TSpawn v1 v2] is harmless if and only if [TApp v1 v2] is harmless.
     We write this in the following way, because it makes the proofs
     (below) easier. *)

  | HlTSpawn:
      forall t1 v2,
      harmless t1 ->
      harmless v2 ->
      harmless (TSpawn (TAbs t1) v2)

  (* The allocation of a new lock, [TNewLock], is ``harmful''. The lock
     invariant must be closed, hence [TNewLock] cannot be allowed under
     a capital-Lambda abstraction. *)

  (* Every other form is harmless. For greater modularity, we write
     this as follows. *)

  | HlTOther:
      forall t,
      (forall v u, t <> TApp v u) ->
      (forall v1 v2, t <> TSpawn v1 v2) ->
      (t <> TNewLock) ->
      harmless t

.

Hint Constructors harmless: harmless.

Hint Extern 1 => congruence : harmless.

Local Ltac harmless_inversion :=
  match goal with h: harmless _ |- _ =>
    let t := fresh in
    inversion h as [ | | t ]; destruct t;
    try solve [ discriminate | congruence ]
  end.

Local Ltac harmless_induction :=
  let t := fresh in
  induction 1 as [ | | t ]; [ | | destruct t ];
  intros;
  try solve [ congruence ].

(* ------------------------------------------------------------------------- *)

(* [TSpawn v1 v2] is harmless if and only if [TApp v1 v2] is harmless. *)

Lemma harmless_TSpawn_TApp:
  forall v1 v2,
  harmless (TSpawn v1 v2) ->
  harmless (TApp v1 v2).
Proof.
  dependent_destruction; eauto with harmless.
Qed.

Goal
  forall v1 v2,
  harmless (TApp v1 v2) ->
  harmless (TSpawn v1 v2).
Proof.
  dependent_destruction; eauto with harmless.
Qed.

(* ------------------------------------------------------------------------- *)

(* Values are harmless. *)

Lemma harmless_value:
  forall K t,
  wk K t VALUE ->
  harmless t.
Proof.
  dependent_induction; solve [ false | eauto with harmless ].
Qed.

(* ------------------------------------------------------------------------- *)

(* [TNewLock] is harmful. This is stated in a form that is convenient for
   use in the main subject reduction lemma. *)

Lemma harmful_TNewLock:
  forall K,
  (K <> zilch -> harmless TNewLock) ->
  K = zilch.
Proof.
  destruct K.
  eauto.
  introv h. forwards: h. (intro; discriminate). harmless_inversion.
Qed.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by weakening. *)

Lemma harmless_lift:
  forall t,
  harmless t ->
  forall k l,
  harmless (lift k l t).
Proof.
  harmless_induction; simpl_lift_all; eauto with harmless.
Qed.

(* Harmlessness is preserved by inverse weakening. *)

Lemma harmless_lift_inverse:
  forall t',
  harmless t' ->
  forall k l t,
  lift k l t = t' ->
  harmless t.
Proof.
  harmless_induction;
  repeat match goal with h: lift _ _ ?t = _ |- _ =>
    destruct t;
    try solve [ discriminate | eauto with harmless ];
    simpl_lift_in h;
    injections;
    try solve [ eauto with harmless ]
  end.
Qed.

Hint Resolve harmless_lift : j_weakening j_substitution.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by substitution of a harmless term for
   a variable. *)

Lemma harmless_subst_harmless:
  forall t,
  harmless t ->
  forall u k,
  harmless u ->
  harmless (subst u k t).
Proof.
  harmless_induction; simpl_subst_goal;
  eauto using harmless_lift with harmless.
  (* TVar *)
  unfold subst_idx. dblib_by_cases; simpl; eauto with harmless.
Qed.

(* Harmlessness is also preserved by substitution at a logical kind.
   Indeed, such a substitution, applied to a term, is the identity. *)

(* The two situations where harmlessness is preserved by substitution
   are summarized as follows. *)

Lemma harmless_subst:
  forall K' t2 kappa2,
  harmless t2 ->
  wk K' t2 kappa2 ->
  ~ logic kappa2 ->
  forall t1 x K kappa1,
  insert x kappa1 K = K' ->
  wk K t1 kappa1 ->
  substitutable kappa1 ->
  harmless (subst t1 x t2).
Proof.
  (* By unfolding [substitutable], we get two cases. *)
  introv ? ? ? ? [ ? | ? ].
  (* In this case, the result of [subst t1 x t2] is just [t2], up to
     an inverse shift, so the goal follows from the hypothesis. *)
  eauto using harmless_lift_inverse, wk_substitution_logic.
  (* In this case, [kappa1] is [VALUE], so we [t1] is harmless and
     we use the lemma [harmless_subst_harmless]. *)
  subst. eauto using harmless_subst_harmless, harmless_value.
Qed.

Hint Resolve harmless_subst : j_substitution.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by reduction. *)

Lemma harmless_red:
  forall t1 t2 h1 h2,
  red (Cf h1 t1) (Cf h2 t2) ->
  harmless t1 ->
  wkc (Cf h1 t1) TERM ->
  harmless t2.
Proof.
  introv hred hl ?. dependent induction hred;
  try destruct_eval_context;
  wkc_inversion; wk_inversion;
  try solve [
    dependent destruction hl;
    eauto using harmless_subst_harmless with harmless wk
  ].
  (* RedRead *)
  wks_inversion. wkh_inversion. eauto using harmless_value.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [D[t]] is harmless, where [D] is a deep evaluation context, then [t]
   is harmless. *)

Lemma harmless_TApp_inversion:
  forall v t,
  harmless (TApp v t) ->
  harmless t.
Proof.
  intros. harmless_inversion.
Qed.

Lemma harmless_fill_inverse:
  forall D t,
  deep_eval_context D ->
  harmless (fill D t) ->
  harmless t.
Proof.
  induction 1; intros; subst; simpl_subst_all.
  eauto.
  rewrite fill_fill in * by eauto. destruct_eval_context. eauto using harmless_TApp_inversion.
Qed.

(* If [D[t1]] is harmless and [t2] is harmless, then [D[t2]] is harmless. *)

Lemma harmless_replace:
  forall D t1 t2,
  deep_eval_context D ->
  harmless (fill D t1) ->
  harmless t2 ->
  harmless (fill D t2).
Proof.
  induction 1; intros; subst; simpl_subst_all.
  eauto.
  rewrite fill_fill in * by eauto. destruct_eval_context. harmless_inversion. eauto with harmless.
Qed.

