Set Implicit Arguments.
Require Import MyTactics.
Require Import AllocationMaps.
Require Import Syntax.
Require Import WellKindedness.

(* ---------------------------------------------------------------------------- *)

(* A pointer is either null or a memory location. *)

Inductive pointer :=
|    Null:        pointer
| NonNull: nat -> pointer.

Ltac destruct_pointers :=
  repeat match goal with p: pointer |- _ => destruct p end.

Lemma eq_pointer_dec:
  forall p1 p2 : pointer,
  { p1 = p2 } + { p1 <> p2 }.
Proof.
  repeat decide equality.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A memory block holds one pointer and one value. The pointer is the `adopter
   pointer'; for now, every block contains such a pointer. The value is the
   content of the block; for now, every block has exactly one field. *)

Inductive block :=
| Block: pointer -> term -> block.

Ltac destruct_blocks :=
  repeat match goal with b: block |- _ => destruct b end.

(* ---------------------------------------------------------------------------- *)

(* A (concrete) heap is an allocation map. It keeps track of which memory
   locations are currently allocated, and maps every allocated memory location
   to a value (for now). *)

Definition heap := aMap block.

(* ---------------------------------------------------------------------------- *)

(* The empty heap. *)

Definition empty_heap : heap :=
  (empty_map _).

(* ---------------------------------------------------------------------------- *)

(* A block is well-kinded if its fields hold closed values. *)

Inductive wkb : block -> Prop :=
| Wkb:
    forall p v,
    wk zilch v VALUE ->
    wkb (Block p v).

Hint Constructors wkb : wk.

(* ---------------------------------------------------------------------------- *)

(* A heap is well-kinded if it maps memory locations to well-kinded blocks. *)

Inductive wkh : heap -> Prop :=
| Wkh:
    forall h,
    everywhere wkb h ->
    wkh h.

Hint Constructors wkh : wk.

(* The empty heap is well-kinded. *)

Lemma wkh_empty_heap:
  wkh empty_heap.
Proof.
  econstructor. eauto using everywhere_empty.
Qed.

(* ---------------------------------------------------------------------------- *)

(* An inversion tactic. *)

Ltac select_wkh action :=
  match goal with
    (* Apply the lemma [everywhere_maps] if possible. *)
  | h1: everywhere _ ?h,
    h2: maps ?h _ _ |- _ =>
      forwards: everywhere_maps; [eexact h1 | eexact h2 | revert h2 ]
    (* Destruct the judgement. *)
  | h: wkh _ |- _ =>
      action h
  | h: wkb _ |- _ =>
      action h
  end.

Ltac wkh_inversion :=
  gji select_wkh; simpl in *.

