Set Implicit Arguments.
Require Import MyList.
Require Import MyTactics.
Require Import Syntax.
Require Import WellKindedness.
Require Import Sugar.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JStability.
Require Import Entails.
Transparent allocate update. (* Just for us. *)

(* ---------------------------------------------------------------------------- *)

(* When a lock is available, its invariant is effectively hidden. When the
   lock is free, the invariant is not hidden; it is in use by some thread. *)

Definition hidden_invariant (i : lock_info) :=
  match i with
  | LockZero P =>
      (* The lock is not taken. Its invariant is hidden. *)
      P
  | (_, _) =>
      (* The lock is taken. Its invariant is not hidden. *)
      TyEmpty
  end.

(* This function is in principle applied to the global resource, which
   is in agreement with the runtime lock heap. Hence, the ownership bit
   reflects the lock status. *)
    
Definition hidden_invariants (R : lock_resource) :=
  TyIteratedStar (
    match R with
    | MapUndefined =>
        nil
    | Map m =>
        map hidden_invariant m
    end
  ).

(* ---------------------------------------------------------------------------- *)

(* Allocating a new lock in the locked state does not cause the hidden
   invariants to grow. Indeed, the new lock contributes nothing to the
   conjunction. *)

Lemma hidden_invariants_allocate:
  forall R P,
  entails
    (hidden_invariants R)
    (hidden_invariants (allocate R (LockOne P))).
Proof.
  unfold hidden_invariants. intros.
  destruct R; simpl. entails.
  rewrite map_app. simpl.
  rewrite fold_right_app. simpl.
  eapply entails_TyStar_right_iterated. entails.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If a lock is unowned/free, then, by marking it owned/taken, we are able to
   extract the lock invariant [P] out of the hidden invariants. *)

Lemma entails_hidden_acquire:
  forall R R' lck P,
  maps R lck (LockZero P) ->
  R' = update R lck (LockOne P) ->
  entails
    (hidden_invariants R)
    (TyStar P (hidden_invariants R')).
Proof.
  introv hm. intros; subst.
  destruct R;
  dependent destruction hm;
  unfold hidden_invariants;
  simpl.
  rewrite map_set_nth. simpl.
  eapply entails_TyIteratedStar_extract.
  equates 1; eauto using is_nth_map.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If a lock is owned/taken, and if we wish to mark it unowned/free, then we
   must put the lock invariant [P] back into the hidden invariants. *)

Lemma entails_hidden_release:
  forall R R' lck P,
  maps R lck (LockOne P) ->
  R' = update R lck (LockZero P) ->
  entails
    (TyStar P (hidden_invariants R))
    (hidden_invariants R').
Proof.
  introv hm. intros; subst.
  destruct R;
  dependent destruction hm;
  unfold hidden_invariants;
  simpl.
  rewrite map_set_nth. simpl.
  eapply entails_TyIteratedStar_insert.
  equates 1; eauto using is_nth_map.
Qed.

