Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Heaps.
Require Import Resources.
Require Import Typing.
Require Import JBasic.

(* The typing judgements are affine with respect to [R]: the more resources
   are available, the better. *)

(* This property is needed in order to prove that [empty] is a unit for the
   conjunction of permissions: indeed, in order to convert [P * empty] to [P],
   we must abandon the resource that is owned by the [empty] conjunct. This
   resource is not necessarily trivial, because we have an affine
   interpretation of permissions. *)

(* It turns out that this property can be proved for [j] and for [jpi]
   independently, even though these judgements are mutually recursive.
   The trick is that the rules that mention the ``other'' judgement in
   a premise also mention ``this'' judgement in some other premise, so
   it is possible to always send the ``excess resource'' into ``this''
   side. *)

(* ---------------------------------------------------------------------------- *)

(* It is clear that [jpi] is affine. *)

Lemma jpi_drop_right:
  forall R1 K P,
  jpi R1 K P ->
  forall R2 R,
  star R1 R2 = R ->
  jpi R K P.
Proof.
  induction 1; intros; subst; eauto using @star_associative with jpi.
Qed.

Lemma jpi_drop_left:
  forall R2 K P,
  jpi R2 K P ->
  forall R1 R,
  star R1 R2 = R ->
  jpi (star R1 R2) K P.
Proof.
  intros; subst. eauto using jpi_drop_right, @star_commutative.
Qed.

Ltac jpi_drop_left :=
  eapply jpi_drop_left; [ | eauto ].

Hint Resolve jpi_drop_right jpi_drop_left : jpi_drop.

(* ---------------------------------------------------------------------------- *)

(* The predicate [invariant _ lck P] is affine. *)

Lemma invariant_affine:
  forall R1 R2 R lck P,
  invariant R1 lck P ->
  star R1 R2 = R ->
  consistent R ->
  invariant R lck P.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* The predicate [taken _ lck] is affine. *)

Lemma taken_affine:
  forall R1 R2 R lck,
  taken R1 lck ->
  star R1 R2 = R ->
  consistent R ->
  taken R lck.
Proof.
  intros. eapply mapsp_maximalp; eauto with maximal.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The typing judgement [j] is affine. *)

(* This lemma is named [j_drop_right] because, when read from bottom to top,
   it can be used to drop the right-hand side of a [star]. *)

Lemma j_drop_right:
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  forall R2 R,
  star R1 R2 = R ->
  consistent R ->
  j f n R K P t T.
Proof.
  induction 1; intros; subst;
  (* Clear the original type derivation, which can mislead [eauto]
     when attempting to construct a new type derivation. *)
  repeat match goal with h: j _ _ _ _ _ _ _ |- _ => clear h end;
  (* The lemma [star_associative] has the effect of sending the
     excess resource into the right-hand side of a split of the
     form [star R1 R2 = R]. *)
  (* The lemma [jpi_drop_right] is used in the case of [JAdopts].
     There, we could argue that the predicate [all_adoptees] is
     affine; yet, we can avoid this argument by sending the
     excess resource into the [jpi] premise. *)
    eauto using @star_core, @star_associative,
                invariant_affine, taken_affine,
                adoptable_affine, unadopted_affine, adopted_affine,
                @consistent_core, @consistent_12p3_23,
                jpi_drop_right
           with maps_maximal explode_resources j.
Qed.

(* The symmetric version can also be useful. *)

Lemma j_drop_left:
  forall f n R2 K P t T,
  j f n R2 K P t T ->
  forall R1 R,
  star R1 R2 = R ->
  consistent R ->
  j f n R K P t T.
Proof.
  introv ? ?; subst. rewrite star_commutative. eauto using @j_drop_right.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This variant states that if [t] is well-typed under [core R],
   then it is also well-typed under [R]. *)

Lemma j_core:
  forall f n R K P t T,
  j f n (core R) K P t T ->
  consistent R ->
  j f n R K P t T.
Proof.
  eauto using j_drop_right, @core_left_unit.
Qed.

