Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.

(* Basic properties of the typing judgements. *)

(* ---------------------------------------------------------------------------- *)

(* This derived typing rule is a special case of [JpiAtomic]. *)

Lemma JpiAtomicEmpty:
  forall n R K v T,
  wk K v VALUE ->
  jc n R K TyEmpty v T ->
  jpi R K (TyAnchored v T).
Proof.
  eauto using @core_right_unit with j.
Qed.

(* The following two derived typing rules are combinations of
   [JpiEmpty] and [JpiStar]. *)

Lemma JpiEmptyStar:
  forall R K P,
  jpi R K P ->
  jpi R K (TyStar TyEmpty P).
Proof.
  eauto using @core_left_unit with j.
Qed.

Lemma JpiStarEmpty:
  forall R K P,
  jpi R K P ->
  jpi R K (TyStar P TyEmpty).
Proof.
  eauto using @core_right_unit with j.
Qed.


Hint Resolve JpiEmptyStar JpiStarEmpty : jxtra.

(* ---------------------------------------------------------------------------- *)

(* The following tactics help a little bit when proving permission subsumption
   statements. [jps_left r] applies the permission subsumption rule [r] to the
   left-hand side of the goal. [jps_right r] applies [r] to the right-hand side
   of the goal. *)

Ltac jps_left r :=
  eapply JpsTransitive; [ eapply r | idtac ].

Ltac jps_right r :=
  eapply JpsTransitive; [ idtac | eapply r ].

(* ------------------------------------------------------------------------- *)

(* Destructing the permission interpretation judgement. *)

(* [jpi_inversion] destructs an arbitrary [jpi] judgement, as long as this
   produces 0 or 1 subgoals. See also the tactic [decompose] in
   [JDecomposition], which destructs [jctop] and [jpi] judgements in a
   sensible way. *)

Ltac select_jpi action :=
  match goal with h: jpi _ _ _ |- _ => action h end.
Ltac jpi_inversion :=
  gji select_jpi.

(* Destructing the subtyping judgement. *)

Ltac sub_inversion :=
  repeat match goal with h: sub _ _ _ _ |- _ => inversion h; subst; revert h end;
  intros.

(* ---------------------------------------------------------------------------- *)

(* Mutual induction schemes. *)

(* I thought that the need for these schemes was removed by [Lemma ... with],
   but the latter construct seems to be buggy -- I have seen Qed loop or take
   forever. Furthermore, proving a single lemma whose statement is a conjunction
   seems preferable to proving multiple lemmas, as it allows sharing the proofs. *)

Scheme j_mutind3   := Induction for j   Sort Prop
  with jbs_mutind3 := Induction for jbs Sort Prop
  with jpi_mutind3 := Induction for jpi Sort Prop.

Combined Scheme j_jbs_jpi_mutind from j_mutind3, jbs_mutind3, jpi_mutind3.

Scheme j_mutind2   := Induction for j   Sort Prop
  with jbs_mutind2 := Induction for jbs Sort Prop.

Combined Scheme j_jbs_mutind from j_mutind2, jbs_mutind2.

Scheme j_mutindT   := Induction for j   Sort Prop
  with jpi_mutindT := Induction for jpi Sort Prop.

Combined Scheme j_jpi_mutind from j_mutindT, jpi_mutindT.

Scheme jps_mutind3     := Induction for jps     Sort Prop
  with sub_mutind3     := Induction for sub     Sort Prop
  with row_sub_mutind3 := Induction for row_sub Sort Prop.

Combined Scheme jps_sub_mutind from jps_mutind3, sub_mutind3, row_sub_mutind3.

Scheme jps_mutind6     := Induction for jps     Sort Prop
  with sub_mutind6     := Induction for sub     Sort Prop
  with row_sub_mutind6 := Induction for row_sub Sort Prop
  with j_mutind6       := Induction for j       Sort Prop
  with jbs_mutind6     := Induction for jbs     Sort Prop
  with jpi_mutind6     := Induction for jpi     Sort Prop.

Combined Scheme the_great_mutind from jps_mutind6, sub_mutind6,
row_sub_mutind6, j_mutind6, jbs_mutind6, jpi_mutind6.

(* ---------------------------------------------------------------------------- *)

(* A canonical derivation can (obviously) serve as an unrestricted derivation. *)

Lemma jc_jt:
(
  forall f n R K P t T,
  j f n R K P t T ->
  jt n R K P t T
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  jbs Unrestricted n R K P tag Tss ts T
).
Proof.
  apply j_jbs_mutind; econstructor (solve [ eauto ]).
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [t] is well-typed, then [t] is a term, i.e. it has kind [TERM]. *)

Lemma j_term:
(
  forall f n R K P t T,
  j f n R K P t T ->
  wk K t TERM
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  wk K ts (ROW TERM)
).
Proof.
  apply j_jbs_mutind; intros; subst; wk_inversion; eauto using wk_empty with wk.
Qed.

Hint Resolve (get1 j_term) : wk.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements respect well-kindedness. *)

(* If [K |- P1 <= P2] holds and [P1] is a valid permission, then [P2] is a
   valid permission.

   If [K |- T1 <= T2] holds and [T1] is valid type, then [T2] is a valid type.

   If [R, K |- { P } v : T] holds and [P] is a valid permission, then [T] is a
   valid type.

   If [P |- vs @ Ts] holds and [P] is a valid permission, then [Ts] is a valid
   row of types.

   If [R, K |- P] holds, then [P] is a valid permission. *)

Lemma j_wk:
(
  forall n K P1 P2,
  jps n K P1 P2 ->
  wk K P1 PERM ->
  wk K P2 PERM
) /\ (
  forall n K T1 T2,
  sub n K T1 T2 ->
  wk K T1 TYPE ->
  wk K T2 TYPE
) /\ (
  forall n K Ts1 Ts2,
  row_sub n K Ts1 Ts2 ->
  wk K Ts1 (ROW TYPE) ->
  wk K Ts2 (ROW TYPE)
) /\ (
  forall f n R K P t T,
  j f n R K P t T ->
  wk K P PERM ->
  wk K T TYPE
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  True
) /\ (
  forall R K P,
  jpi R K P ->
  wk K P PERM
).
Proof.
  apply the_great_mutind; intros; subst;
  (* The pre-condition is well-kinded; what can we deduce from this? *)
  wk_inversion;
  (* This should allow us to use the induction hypotheses... *)
  try solve [
    tauto
  | repeat match goal with ih: wk _ _ _ -> wk _ _ _ |- _ =>
      specializes ih ___; [ eauto 8 using wk_weakening with wk |
      (* ... which in turn allow us to deduce more information. *)
      wk_inversion ]
    end;
    (* It is then easy to conclude! Although some goals require heavy strength. *)
    solve [
      eauto    using wk_weakening, wk_substitution, wk_empty with wk logic substitutable
    | eauto  8 using wk_weakening, wk_substitution, wk_empty with wk logic substitutable
    | eauto 14 using wk_weakening with wk logic substitutable
    ]
  ].
Qed.

Hint Resolve (get1 j_wk) (get2 j_wk) (get3 j_wk) (get4 j_wk) (last6 j_wk) : wk.

(* ---------------------------------------------------------------------------- *)

(* A sanity check: because the index carried by the subtyping judgements is
   supposed to be an over-approximation of their height, these judgements
   are preserved when this index grows. *)

Lemma jps_index_monotonic:
(
  forall n K P Q,
  jps n K P Q ->
  forall m,
  m >= n ->
  jps m K P Q
) /\ (
  forall n K T U,
  sub n K T U ->
  forall m,
  m >= n ->
  sub m K T U
) /\ (
  forall n K Ts Us,
  row_sub n K Ts Us ->
  forall m,
  m >= n ->
  row_sub m K Ts Us
).
Proof.
  eapply jps_sub_mutind; intros;
  try match goal with h: ?m >= S ?n |- _ =>
    destruct m; [ false; omega | ]
  end;
  eauto with j omega.
Qed.

Lemma j_index_monotonic:
(
  forall f n R K P t T,
  j f n R K P t T ->
  forall m,
  m >= n ->
  j f m R K P t T
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  forall m,
  m >= n ->
  jbs f m R K P tag Tss ts T
).
Proof.
  apply j_jbs_mutind; intros;
  try match goal with h: ?m >= S ?n |- _ =>
    destruct m; [ false; omega | ]
  end;
  eauto with j omega.
Qed.

