Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JTop.
Require Import SubsumptionSoundness.

(* ---------------------------------------------------------------------------- *)

(* A canonicalization lemma. If the value [v] is well-typed according to [jt],
   then it must be well-typed according to [jctop]. In other words, an
   unrestricted type derivation can be turned into a canonical type
   derivation. The permission subsumption steps are eliminated by appeal to
   the lemma [jps_sound]. *)

(* The proof is by well-founded induction over the height of the type
   derivation -- as opposed to by structural induction over the type
   derivation -- because [JExistsElim] requires that the substitution lemma be
   applied to the premise before the induction hypothesis for the premise is
   invoked. *)

Lemma canonicalization:
  forall n R1 K P1 v T,
  jt n R1 K P1 v T ->
  forall R2 R,
  jpi R2 K P1 ->
  star R1 R2 = R ->
  consistent R ->
  wk K v VALUE ->
  jctop R K v T.
Proof.
  intro n. induction_wf ih: lt_wf n.
  introv h; dependent destruction h; intros; subst;
  wk_inversion; jpi_inversion.

  (* JSingleton *)
  { eauto using JcTopSingleton. }

  (* JFunction *)
  { eauto 6 with j wk. }

  (* JForallIntro *)
  (* We are establishing canonicalization under an arbitrary environment [K],
     so the induction hypothesis is strong enough to take us under this
     capital Lambda. *)
  { eauto 6 using JcTopForallIntro, (last6 j_weakening), wk_weakening. }

  (* JExistsIntro *)
  { eauto using JcTopExistsIntro. }

  (* JExistsElim *)
  (* The term [t] is well-typed in the presence of an abstract name [x] and
     under a precondition [P] which may refer to [x]. Now, we know that [P]
     holds when [x] is replaced with [U]. Thus, we replace [x] with [U]
     in the type derivation for [t]. The term [t] and its type [T] actually
     do not refer to [x], so they are unaffected. *)
  substitution.
  (* Applying the induction hypothesis to this new type derivation for [t]
     yields the goal. *)
  { eauto. }

  (* JFrame *)
  (* Apply the induction hypothesis; conclude by using [JcTopFrame]. *)
  { eauto using JcTopFrame with consistent_aac aac. }

  (* JCut *)
  (* Apply the induction hypothesis, and we are done. *)
  { eauto using JpiStar with aac. }

  (* JSubLeft *)
  (* Apply [jps_sound], then apply the induction hypothesis, and we are done. *)
  { eauto using jps_sound with consistent_split. }

  (* JSubRight *)
  (* Apply the induction hypothesis, then apply [sub_sound], and we are done. *)
  { eauto using sub_sound. }

  (* JRef *)
  { eauto using (JRef 0) with wk jpi. }

  (* JLock *)
  { eauto using (JLock 0) with wk jpi. }

  (* JLocked *)
  { eauto using (JLocked 0) with wk jpi. }

  (* JAdoptable *)
  { eauto using (JAdoptable 0) with wk jpi. }

  (* JUnadopted *)
  { eauto using (JUnadopted 0) with wk jpi. }

  (* JAdopts *)
  { eauto using (JAdopts 0) with wk jpi. }

Qed.

(* The following corollary is more readable. It is unused at the moment. *)

Goal
  forall n R K v T,
  jt n R K TyEmpty v T ->
  consistent R ->
  wk K v VALUE ->
  jctop R K v T.
Proof.
  eauto using canonicalization, @core_right_unit with j.
Qed.

