Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JDuplication.
Require Import JTop.
Require Import Zap.

(* Decomposing a canonical typing judgement [jctop]. *)

(* ------------------------------------------------------------------------- *)

(* If [v1] has type [=v2], then [v1] and [v2] are equal. *)

Lemma decomposition_TySingleton:
  forall R K v1 v2,
  jctop R K v1 (TySingleton v2) ->
  v1 = v2.
Proof.
  jctop_induction; eauto 3 with j.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a function [\x.t] has type [T -> U] under the resource [R], then the
   function body [t] has type [U] under the resource [core R] and under the
   precondition [x @ T]. *)

Lemma decomposition_TyArrow:
  forall R K t T U,
  jctop R K (TAbs t) (TyArrow T U) ->
  consistent R ->
  exists n,
  jt n
    (core R)
    (insert 0 VALUE K)
    (TyAnchored (TVar 0) (shift 0 T))
    t
    (shift 0 U).
Proof.
  jctop_induction.

  (* JFunction *)
  clear IHh. jpi_inversion. eexists.

  (* Use affinity [j_drop_right] in order to drop the resource that was used
     to justify [TyDuplicable P]. It is of no use. *)
  eapply j_drop_right; [ | | eauto using @consistent_core ].

  (* Use [JCut] in order to hide the fact that the body of the function relies
     on the permission [P]. This suggests that, instead of letting [JFunction]
     capture a duplicable permission [P], we could have adopted a stricter
     version of [JFunction] that does not allow this, and we would have
     recovered the lenient version by combining [JFunction] and [JCut]. In the
     end, it does not make much difference. The present lemma effectively
     gives us inversion for the strict variant. *)
    eapply JCut.
      eassumption.
      eapply j_weakening; [ | eauto | eauto ].
        eapply j_duplicable; eauto with consistent_split.
      reflexivity.

  (* What remains here is the second premise of [j_drop_right]. *)
  rewrite @star_core by eauto with consistent_aac.
  rewrite @star_core.
    f_equal. aac_guess.
    eauto with consistent_aac.

  (* JCut *)
  (* Just apply the induction hypothesis. *)
  forwards: IHh; eauto with j wk consistent_aac. unpack.
  eauto using JSameResource with aac.

Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has type [forall a.T], then [v] has type [T] in an environment
   extended with [a]. *)

Lemma decomposition_TyForall:
  forall R K v kappa T,
  jctop R K v (TyForall kappa T) ->
  jctop R (insert 0 kappa K) (shift 0 v) T /\
  substitutable kappa.
Proof.
  jctop_induction; jpi_inversion.
  (* JForallIntro. *)
  solve [ eauto using wk_weakening with j ].
  (* JCut *)
  forwards: IHh; [ eauto with j | eauto | unpack ].
  eauto using JpiSameResource with aac.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has type [exists a.T], then [v] has type [[U/a]T] for some [U]. *)

Lemma decomposition_TyExists:
  forall R K v kappa T,
  jctop R K v (TyExists kappa T) ->
  exists U,
  jctop R K v (subst U 0 T) /\
  wk K U kappa /\
  wk (insert 0 kappa K) T TYPE /\
  substitutable kappa.
Proof.
  jctop_induction; jpi_inversion.
  (* JExistsIntro. *)
  solve [ eauto 6 with j ].
  (* JCut *)
  forwards: IHh; [ eauto with j | eauto | unpack ].
  eauto 6 using JpiSameResource with aac.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has type [T | Q] under [R], then under the same resource,
   separately, [v] has type [T] and the permission [Q] is justified. *)

Lemma decomposition_TyMixStar:
  forall R K v T Q,
  jctop R K v (TyMixStar T Q) ->
  exists R1 R2,
  jctop R1 K v T /\
  jpi R2 K Q /\
  R = star R1 R2.
Proof.
  jctop_induction; jpi_inversion.
  (* JFrame *)
  solve [ xplit; eauto with j aac ].
  (* JCut *)
  forwards: IHh; eauto with j. unpack.
  xplit; eauto.
  equates 2. eassumption. aac_reflexivity.
Qed.

(* ------------------------------------------------------------------------- *)

(* Decomposition for memory locations. *)

Lemma decomposition_TyRef:
  forall R K l m T,
  jctop R K (TLoc l) (TyRef m T) ->
  exists R1 R2 v,
  star R1 R2 = R /\
  wk zilch v VALUE /\
  jpi R1 K (TyAnchored v T) /\
  maps (get R2) l (mv2dx m v).
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk. unpack. subst.
  match goal with h: star ?R1 ?R2 = _ |- _ => exists R1 R2; rewrite h end.
  solve [ eauto with aac ].
  (* JRef *)
  xplit. 4: eauto. aac_guess. eauto. jpi_drop_left. eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* Decomposition for locks. *)

Lemma decomposition_TyLock:
  forall R K lck P,
  jctop R K (TLock lck) (TyLock P) ->
  consistent R ->
  wk zilch P PERM /\
  invariant (get R) lck P.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk consistent_aac. unpack.
  split. eauto. equates 3. eauto. f_equal. aac_reflexivity.
  (* JLock *)
  explode_resources_locks.
  eauto using invariant_affine.
Qed.

Lemma decomposition_TyLocked:
  forall R K lck,
  jctop R K (TLock lck) TyLocked ->
  consistent R ->
  taken (get R) lck.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk consistent_aac.
  equates 2. eauto. f_equal. aac_reflexivity.
  (* JLocked *)
  explode_resources_locks.
  eauto using taken_affine.
Qed.

(* ------------------------------------------------------------------------- *)

(* Decomposition for adopters/adoptees. *)

Lemma decomposition_TyAdoptable:
  forall R K l,
  jctop R K (TLoc l) TyAdoptable ->
  consistent R ->
  adoptable (get R) l.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk consistent_aac.
  equates 2. eauto. f_equal. aac_reflexivity.
  (* JAdoptable *)
  explode_resources.
  eauto using adoptable_affine.
Qed.

Lemma decomposition_TyUnadopted:
  forall R K l,
  jctop R K (TLoc l) TyUnadopted ->
  consistent R ->
  unadopted (get R) l.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk consistent_aac.
  equates 2. eauto. f_equal. aac_reflexivity.
  (* JUnadopted *)
  explode_resources.
  eauto using unadopted_affine.
Qed.
  
Lemma decomposition_TyAdopts:
  forall R K l U,
  jctop R K (TLoc l) (TyAdopts U) ->
  consistent R ->
  exists R1 R2 xs,
  wk K U TYPE /\
  all_adoptees (get R1) xs l /\
  jpi R2 K (zap xs U) /\
  star R1 R2 = R.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk consistent_aac. unpack.
  xplit; eauto. equates 1. eauto. aac_reflexivity.
  (* JAdopts *)
  eauto 8 using jpi_drop_right, @star_associative.
Qed.

(* ------------------------------------------------------------------------- *)

(* The tactic [decomposition] applies the decomposition lemmas as much
   as possible. *)

(* It also decomposes [jpi] hypotheses, when they involve existentials
   and stars, so as to get access to the [jctop] judgements that might
   be hidden in them. It kind-of-subsumes [gentle_jpi_inversion]. *)

Ltac select_decomposition action :=
  match goal with
  | h: jctop _ _ ?v1 (TySingleton ?v2) |- _ =>
      (* [h] implies that [v1] and [v2] coincide *)
      generalize (decomposition_TySingleton h);
      (* [h] also implies that this value is well-kinded *)
      generalize (jctop_value h);
      (* we must clear [h] in order to avoid looping *)
      clear h; intros ? ?; try subst v1
  | h: jctop _ _ (TAbs _) (TyArrow _ _) |- _ =>
      (* [h] implies that the type [TyArrow _ _] is well-kinded;
         let's not forget that. *)
      generalize (jctop_wk h); intro; wk_inversion;
      (* Now, decompose. *)
      forwards: decomposition_TyArrow; [
        eexact h
      | eauto with consistent_split
      | clear h; unpack
      ]
  | h: jctop _ _ _ (TyForall _ _) |- _ =>
      generalize (decomposition_TyForall h); clear h;
      intros; unpack
  | h: jctop _ _ _ (TyExists _ _) |- _ =>
      generalize (decomposition_TyExists h); clear h;
      intros [ ? [ h ? ]]; unpack;
      (* The new hypothesis [h] contains a substitution applied to a
         type. Try to simplify it. *)
      simpl_subst_in h;
      repeat rewrite subst_lift in h
  | h: jctop _ _ _ (TyMixStar _ _) |- _ =>
      generalize (decomposition_TyMixStar h); clear h;
      intro; unpack; try subst
  | h: jctop _ _ (TLoc _) (TyRef _ _) |- _ =>
      generalize (decomposition_TyRef h); clear h;
      intro; unpack; try subst
  | h: jctop _ _ (TLock _) (TyLock _) |- _ =>
      forwards: (decomposition_TyLock h); [
        solve [ eauto with consistent_split ]
      | clear h; unpack
      ]
  | h: jctop _ _ (TLock _) TyLocked |- _ =>
      forwards: (decomposition_TyLocked h); [
        solve [ eauto with consistent_split ]
      | clear h
      ]
  | h: jctop _ _ (TLoc _) TyAdoptable |- _ =>
      forwards: (decomposition_TyAdoptable h); [
        solve [ eauto with consistent_split ]
      | clear h
      ]
  | h: jctop _ _ (TLoc _) TyUnadopted |- _ =>
      forwards: (decomposition_TyUnadopted h); [
        solve [ eauto with consistent_split ]
      | clear h
      ]
  | h: jctop _ _ (TLoc _) (TyAdopts _) |- _ =>
      forwards: (decomposition_TyAdopts h); [
        solve [ eauto with consistent_split ]
      | clear h; unpack; try subst
      ]
  (* We do not destruct judgements of the form [jpi _ _ (TyAnchored _ _)],
     since these are [jctop] judgements. For the same reason, we avoid
     touching judgements of the form of [jpi _ _ (shift _ _)] or
     [jpi _ _ (subst _ _ _)], as there might a [TyAnchored] under the
     shift or substitution operator. The user will have to manually
     simplify such a judgement if necessary. *)
  | h: jpi _ _ (TyAnchored _ _) |- _ =>
      revert h
  | h: jpi _ _ (subst _ _ _) |- _ =>
      revert h
  | h: jpi _ _ (shift _ _) |- _ =>
      revert h
  (* Any other [jpi] judgement can be destructed. *)
  | h: jpi _ _ _ |- _ => action h
  end.

Ltac decomposition :=
  gji select_decomposition.

(* ------------------------------------------------------------------------- *)

(* The permission [TyFalse] is unsatisfiable. *)

Lemma jpi_TyFalse:
  forall R K,
  jpi R K TyFalse ->
  False.
Proof.
  intros. jpi_inversion. (* Now, that is quite amazing! *)
Qed.

(* ------------------------------------------------------------------------- *)

(* This little lemma offers a pleasant way of decomposing a conjunction that
   involves a duplicable permission. *)

Lemma decomposition_TyStar_TyDuplicable_left:
  forall R K P Q,
  jpi R K (TyStar (TyDuplicable P) Q) ->
  dup P /\ jpi R K Q.
Proof.
  intros. jpi_inversion. eauto using jpi_drop_left.
Qed.

(* This one is more general, but is unused at present. *)

Goal
  forall R K P Q,
  jpi R K (TyStar P Q) ->
  dup P ->
  jpi R K P /\ jpi R K Q.
Proof.
  intros. jpi_inversion. eauto using jpi_drop_left, jpi_drop_right.
Qed.

