Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import DerivedTyping.
Require Import JBasic.

(* ---------------------------------------------------------------------------- *)

(* The predicate [invariant _ lck P] is duplicable. *)

Lemma invariant_core:
  forall R lck P,
  invariant R lck P ->
  invariant (core R) lck P.
Proof.
  intros. eapply mapsp_duplicablep; eauto with duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adoptable _ l] is duplicable. *)

Lemma adoptable_core:
  forall R l,
  adoptable R l ->
  adoptable (core R) l.
Proof.
  intros. eapply mapsp_duplicablep; eauto with duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The judgement [pcore P Q] means that the permission [Q] is *some* duplicable
   syntactic fragment of the permission [P]. *)

(* This predicate does not play an important role. It only serves in the
   statement of the auxiliary lemma [j_duplicable_preliminary]. *)

Inductive pcore : term -> term -> Prop :=

(* If [P] is a conjunction, then each of the conjuncts can be treated
   independently. *)

| PCoreStar:
    forall P1 P2 Q1 Q2,
    pcore P1 Q1 ->
    pcore P2 Q2 ->
    pcore (TyStar P1 P2) (TyStar Q1 Q2)

(* One might think that we need the following rule: if [P] is an existential
   permission, then we take a duplicable fragment of its body, and preserve
   the existential quantifier. Actually, we are able to work without it. The
   need for it does not arise in the proof of the lemma
   [j_duplicable_preliminary]. *)

(* We do need such a rule for universal quantifiers. *)

| PCoreForall:
    forall kappa P Q,
    pcore P Q ->
    pcore (TyForall kappa P) (TyForall kappa Q)

(* Any permission [P] can be discarded, regardless of whether it is or is not
   duplicable. This rule makes the relation [pcore] non-deterministic. It is
   required because a variable of kind [PERM] that is not known to be
   duplicable may later be instantiated with a duplicable permission. *)

| PCoreEmpty:
    forall P,
    pcore P TyEmpty

(* If [P] is known to be duplicable, then it can be kept. *)

| PCoreDuplicable:
    forall P,
    dup P ->
    pcore P P

(* The typing rule does not have a premise of the form [dup P]; instead,
   the typing judgement in its conclusion has a precondition of the form
   [TyStar (TyDuplicable P) P]. For this reason, we must make a special
   case and declare that a permission of this form can be kept (i.e., is
   a duplicable fragment of itself). *)

(* Adding this constructor allows the case of [JFunction] to go through.
   It does complicate the lemma [pcore_star_inversion], by adding one
   disjunct to the conclusion of this lemma, but (perhaps miraculously)
   in the two places where this lemma is used, the proof still goes
   through. *)

| PCoreSpecial:
    forall P,
    pcore (TyStar (TyDuplicable P) P) (TyStar (TyDuplicable P) P)

.

Hint Constructors pcore : pcore.

(* The tactic [pcore_inversion] destructs a [pcore _ _] hypothesis,
   even if this introduces multiple sub-cases. *)

Local Ltac pcore_inversion :=
  match goal with
  | h: pcore _ _ |- _ =>
      inversion h; subst; clear h
  end.

(* ---------------------------------------------------------------------------- *)

(* The relation [pcore] is preserved by substitution. *)

Lemma pcore_subst:
  forall P Q,
  pcore P Q ->
  forall v x,
  pcore (subst v x P) (subst v x Q).
Proof.
  induction 1; intros; simpl_subst_goal;
  econstructor (solve [ eauto using dup_substitution ]).
Qed.

(* ---------------------------------------------------------------------------- *)

(* A duplicable fragment of the empty permission must be the empty permission. *)

Lemma pcore_empty_inversion:
  forall Q,
  pcore TyEmpty Q ->
  Q = TyEmpty.
Proof.
  inversion 1; subst; eauto.
Qed.

(* A duplicable fragment of a conjunction [P1 * P2] must be either: 1- empty;
   or 2- a conjunction of two fragments of [P1] and [P2]; or 3- it can be
   [P1 * P2] itself, if [P1 * P2] is of the form [TyDuplicable P2 * P2]. *)

Lemma pcore_star_inversion:
  forall P1 P2 Q,
  pcore (TyStar P1 P2) Q ->
  Q = TyEmpty \/
  (exists Q1 Q2, Q = TyStar Q1 Q2 /\ pcore P1 Q1 /\ pcore P2 Q2) \/
  (P1 = TyDuplicable P2 /\ Q = TyStar (TyDuplicable P2) P2).
Proof.
  inversion 1; subst; dup_inversion; eauto 10 with pcore.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [jpi R K P] holds and [P] is duplicable, then [jpi (core R) K P] holds
   as well. In other words, duplicable permissions are justified by duplicable
   resources. *)

(* The above statement is not directly amenable to an inductive proof. In the
   case [JpiAtomic], for instance, the hidden permission is not necessarily
   duplicable, so the induction hypothesis cannot be applied. We generalize
   this statement as follows: if [jpi R K P] holds, then for *every* duplicable
   fragment [Q] of [P], the judgement [jpi (core R) K Q] holds as well. This is
   the reason why we need to define the relation [pcore]. *)

(* In order to prove this property of [jpi], we must state a related property
   about the typing judgement [j]. Fortunately, this property needs only
   concern values (as opposed to terms) and needs only concern canonical
   typing derivations (as opposed to arbitrary derivations). We prove that, if
   a canonical typing derivation concludes that a value [v] has duplicable
   type [T] under [R] and [P], then the same is true under [core R] and under
   *some* duplicable fragment [Q] of [P]. *)

(* This proof has a rather syntactic and fragile feel to it, due to the use
   of [pcore]. I am not fully happy with this, but have been unable so far
   to find another approach. *)

Lemma j_duplicable_preliminary:
(
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  f = Canonical ->
  wk K P PERM ->
  wk K t VALUE ->
  consistent R1 ->
  dup T ->
  forall R2,
  core R1 = R2 ->
  exists m Q,
  jc m R2 K Q t T /\
  pcore P Q
) /\ (
  forall R1 K P,
  jpi R1 K P ->
  consistent R1 ->
  forall R2,
  core R1 = R2 ->
  forall Q,
  pcore P Q ->
  jpi R2 K Q
).
(* A tactic for recognizing and invoking the induction hypothesis. *)
Local Ltac j_duplication_ih :=
  match goal with
  ih: Canonical = Canonical -> _,
  hj: jc _ _ _ _ _ _  |- _ =>
    forwards: ih; eauto using wk_weakening, dup_substitution
                         with wk dup consistent_split;
    clear ih hj;
    unpack
  end.
Proof.
  apply j_jpi_mutind; intros; subst; try discriminate;
  dup_inversion; wk_inversion;
  (* Clear the induction hypothesis when it is clearly not applicable. *)
  try match goal with ih: Unrestricted = Canonical -> _ |- _ => clear ih end.

  (* JSingleton *)
  solve [ eauto using (JSingleton 0), PCoreEmpty ].

  (* JFunction *)
  (* The captured permission must be duplicable, so it is its own
     core. Furthermore, the function body is type-checked under
     [core R], which is the same as [core (core R)]. Thus, we are
     able to reconstruct an instance of [JFunction]. *)
  solve [ eauto 8 using JFunction, JSameResource, @core_idempotent, PCoreSpecial ].

  (* JForallIntro *)
  j_duplication_ih.
  solve [ eauto 8 using JForallIntro, PCoreForall ].

  (* JExistsIntro *)
  j_duplication_ih.
  solve [ eauto using JExistsIntro ].

  (* JFrame *)
  j_duplication_ih.
  (* The framed out permission must be duplicable, so it is a duplicable
     fragment of itself. We are able to reconstruct an instance of the
     frame rule. *)
  solve [ eauto 8 using JFrame, PCoreStar, PCoreDuplicable ].

  (* JCut *)
  j_duplication_ih.
  (* The induction hypothesis yields the existence of *some* duplicable
     fragment [Q] of the conjunction [P1 * P2]. By analysis of the definition
     of [pcore], [Q] must be either 1- [TyEmpty]; or 2- a conjunction of two
     fragments of [P1] and [P2]; or 3- [P1 * P2] itself, if [P1] is
     [TyDuplicable P2]. *)
  forwards f: pcore_star_inversion; [ eauto | ].
  destruct f as [ | [ | ]]; unpack; subst; jpi_inversion.
  (* Sub-case 1. The result follows directly. *)
  solve [ eauto 6 using JSameResource, @core_star_right, PCoreEmpty ].
  (* Sub-case 2. The result follows from an application of [JCut]. *)
  solve [ eauto 8 using JCut, @star_core with consistent_split ].
  (* Sub-case 3. The result again follows from [JCut]. The knowledge
     that [P2] is duplicable allows us to use [PCoreDuplicable] and
     keep all of [P2] to type-check the term. Miracle? *)
  solve [ eauto 8 using JCut, @star_core, PCoreDuplicable with jpi ].

  (* JRef *)
  solve [ eauto 12 using (JRef 0), @star_core, PCoreDuplicable, @maps_core, PCoreEmpty
                    with consistent_split dup explode_resources ].

  (* JLock *)
  solve [ eauto 7 using (JLock 0), invariant_core, PCoreEmpty with explode_resources ].

  (* JAdoptable *)
  solve [ eauto 7 using (JAdoptable 0), adoptable_core, PCoreEmpty with explode_resources ].

  (* JpiEmpty *)
  (* A fragment of the empty permission must be itself empty. *)
  forwards: pcore_empty_inversion; eauto. subst.
  solve [ eauto using JpiEmpty ].

  (* JpiStar *)
  (* Again, a fragment of [P1 * P2] can be of one three forms. *)
  forwards f: pcore_star_inversion; [ eauto | ].
  destruct f as [ | [ | ]]; unpack; subst; jpi_inversion.
  (* Sub-case 1. The result is immediate. *)
  solve [ eauto using JpiEmpty ].
  (* Sub-case 2. The result follows from the induction hypotheses. *)
  solve [ eauto 7 using JpiStar, @star_core with consistent_split ].
  (* Sub-case 3. Because [P2] is known to be duplicable, there is no
     difficulty in arguing that we can keep all of it. *)
  solve [ eauto 7 using PCoreDuplicable, @star_core with jpi logic consistent_split ].

  (* JpiForall *)
  (* A fragment of a universal permission must be either a universal
     permission or empty or a (duplicable) universal permission. In
     either case, the result is easily obtained. *)
  pcore_inversion; dup_inversion;
  solve [ eauto using PCoreDuplicable with j ].

  (* JpiExists *)
  (* A fragment of an existential permission must be either empty or
     a (duplicable) existential permission. In the former case, the
     result is immediate. In the latter case, it follows from the
     induction hypothesis.*)
  pcore_inversion; dup_inversion.
  solve [ eauto using JpiEmpty ].
  solve [ eauto using JpiExists, pcore_subst, PCoreDuplicable ].

  (* JpiAtomic *)
  (* A fragment of the permission [v @ T] must be either empty or
     [v @ T] itself, where [T] is duplicable. In the former case, the
     result is immediate. In the latter case, it follows from the
     induction hypothesis.*)
  pcore_inversion; dup_inversion.
  solve [ eauto using JpiEmpty ].
  j_duplication_ih.
  solve [ eauto using JpiAtomic, @star_core with consistent_split ].

  (* JpiDuplicable *)
  (* A fragment of the permission [TyDuplicable T] must be either
     empty or [TyDuplicable T] itself. In either case, the result
     is immediate. *)
  pcore_inversion; solve [ eauto using JpiEmpty, JpiDuplicable ].

Qed.

(* ---------------------------------------------------------------------------- *)

(* This is the sought-after corollary: duplicable permissions are justified by
   duplicable resources. *)

Lemma j_duplicable:
  forall R K P,
  jpi R K P ->
  consistent R ->
  dup P ->
  jpi (core R) K P.
Proof.
  intros. eapply j_duplicable_preliminary; eauto using PCoreDuplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This corollary (which is not explicitly used) states that if [P] is duplicable,
   then [P] entails [P * P]. This justifies the permission subsumption rule
   [JpsDuplicate]. *)

Goal
  forall R K P,
  jpi R K P ->
  consistent R ->
  dup P ->
  jpi R K (TyStar P P).
Proof.
  eauto 3 using j_duplicable with jpi core_unit.
Qed.

