Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import MyList.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JSubstitution.
Require Import DerivedTyping.
Require Import Zap.

(* A typing judgement for closed values. *)

(* ---------------------------------------------------------------------------- *)

(* The judgement [jctop R K v T] states that the value [v] has type [T] under
   resource [R] and under the environment [K]. A non-empty environment is
   required in order to deal with polymorphic values, i.e., values that are
   well-typed in the presence of certain type variables.*)

(* We don't need to invent new rules for this judgement: we posit that, by
   definition, this judgement holds if and only if the resource [R] justifies
   the permission [v @ T]. *)

Notation jctop R K v T :=
  (jpi R K (TyAnchored v T)).

(* This tactic can help build a [jctop] judgement. The parameter [rule] is
   the typing rule that should be applied. *)

Ltac jctop rule :=
  eapply JpiAtomicEmpty; [ eauto with wk | eapply (@rule 0) ].

(* ---------------------------------------------------------------------------- *)

(* In a judgement of the form [jctop R K v T], [v] must be a value, and [T]
   must be a type. *)

Lemma jctop_value:
  forall R K v T,
  jctop R K v T ->
  wk K v VALUE.
Proof.
  intros. jpi_inversion. assumption.
Qed.

Lemma jctop_wk:
  forall R K v T,
  jctop R K v T ->
  wk K T TYPE.
Proof.
  intros. jpi_inversion. eauto with wk.
Qed.

Hint Resolve jctop_value jctop_wk : wk.

(* ---------------------------------------------------------------------------- *)

(* The definition of [jctop] in terms of [jpi] makes it rather inconvenient
   to build [jctop] judgements. In order to facilitate this task, we prove
   a number of derived rules. *)

(* Singleton. *)

Lemma JcTopSingleton:
  forall R K v,
  wk K v VALUE ->
  jctop R K v (TySingleton v).
Proof.
  eauto using JpiAtomicEmpty, (JSingleton 0).
Qed.

Lemma JcTopSingletonEq:
  forall R K v1 v2,
  v1 = v2 ->
  wk K v1 VALUE ->
  jctop R K v1 (TySingleton v2).
Proof.
  intros; subst. eauto using JcTopSingleton.
Qed.

(* Frame. *)

(* R1 |- v : T
   R2 |- Q
   --------------------
   R1 * R2 |- v : T | Q *)

Lemma JcTopFrame:
  forall R1 R2 K R Q v T,
  jctop R1 K v T ->
  jpi R2 K Q ->
  star R1 R2 = R ->
  jctop R K v (TyMixStar T Q).
Proof.
  intros. jpi_inversion. econstructor; eauto with j aac.
Qed.

Lemma JcTopMix:
  forall R K P v,
  wk K v VALUE ->
  jpi R K P ->
  jctop R K v (TyMixStar TyTop P).
Proof.
  intros. econstructor. eauto.
  econstructor. eauto using JTop. eauto using @core_left_unit with jpi. eauto using @core_left_unit.
Qed.

Lemma JcTopForallIntro:
  forall R K v kappa T,
  jctop R (insert 0 kappa K) (shift 0 v) T ->
  substitutable kappa ->
  jctop R K v (TyForall kappa T).
Proof.
  intros. jpi_inversion.
  assert (wk K v VALUE).
    eauto using wk_weakening_inverse with wk.
  (* There is a subtle technical point here. The permission [P]
     which justifies that [shift 0 v] has type [T] may mention
     the variable [0]. Yet, in order to obtain the desired
     conclusion, we need a permission that does not mention this
     variable. The trick is to universally quantify [P]. Thus, we
     seem to need universal permissions -- and their interpretation
     rule, [JpiForall] -- in order to justify universal types. *)
  econstructor.
    eauto.
    eapply JForallIntro; eauto using harmless_value.
    eapply JpiForall; eauto.
    eauto.
Qed.

Lemma JcTopExistsIntro:
  forall R K v kappa U T,
  jctop R K v (subst U 0 T) ->
  wk K U kappa ->
  wk (insert 0 kappa K) T TYPE ->
  substitutable kappa ->
  jctop R K v (TyExists kappa T).
Proof.
  intros. jpi_inversion. eauto with j.
Qed.

Lemma JcTopApp:
  forall R1 R2 R K v1 v2 T U,
  jctop R1 K v1 (TyArrow T U) ->
  jctop R2 K v2 T ->
  wk K v1 VALUE ->
  wk K v2 VALUE ->
  star R1 R2 = R ->
  exists n,
  jt n R K TyEmpty (TApp v1 v2) U.
Proof.
  intros. jpi_inversion. eexists.
  eapply JtTopCut.
    eapply JApplication.
      eauto.
      eauto using (get1 jc_jt).
      eauto with jpi.
    aac_reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* An introduction rule for the intersection type constructor. If the value [v]
   separately admits each of the types [Ts], then [v] has type [intersection Ts]. *)

Lemma JcTopIntersection:
  forall R K v Ts,
  wk K v VALUE ->
  jpi R K (TyIteratedStar (map (TyAnchored v) Ts)) ->
  jctop R K v (intersection Ts).
Proof.
  intros.
  eapply JcTopExistsIntro; [ | eauto | | eauto with substitutable ].
  { unfold intersection_body. simpl_subst_goal.
    eapply JcTopFrame; [ eauto using JcTopSingleton | | eapply core_left_unit ].
      erewrite subst_TyIteratedStar. eauto.
      rewrite map_map. eapply map_ext; intro.
      simpl_subst_goal. rewrite subst_lift. reflexivity. }
  { eapply wk_intersection_body.
    eapply Forall_map_reverse.
      { eauto using wk_TyIteratedStar_inverse with wk. }
      { simpl. intros. wk_inversion. eauto. }}
Qed.

(* ---------------------------------------------------------------------------- *)

(* A well-typed value is a well-typed term. *)

Lemma jctop_jt:
  forall R K v T,
  jctop R K v T ->
  exists n,
  jt n R K TyEmpty v T.
Proof.
  intros. jpi_inversion. eexists.
  eapply JtTopCut; eauto using (get1 jc_jt), @star_commutative.
Qed.

(* ------------------------------------------------------------------------- *)

(* A [jctop] judgement can be substituted into a [jt] judgement. *)

Lemma jctop_substitution:
  forall n R1 R2 R K U t v T,
  jctop R1 K v U ->
  jt n R2 (insert 0 VALUE K) (TyAnchored (TVar 0) (shift 0 U)) t (shift 0 T) ->
  wk K v VALUE ->
  star R1 R2 = R ->
  jt (S (S n)) R K TyEmpty (subst v 0 t) T.
Proof.
  intros; subst.
  eapply JtTopCut.
    eapply j_substitution; eauto with subst_lift substitutable.
    simpl_subst_goal. rewrite subst_lift. eauto.
    eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* [jctop] is defined as an abbreviation for [jpi _ _ (TyAnchored _ _)],
   which itself is essentially a conjunction of three judgements, [jc],
   [jpi], and [wk]. We need high-level lemmas and tactics that work at
   the level of [jctop] without exposing the lower level. *)

(* ------------------------------------------------------------------------- *)

(* The following tactic helps prove a lemma whose statement begins with a
   [jctop] hypothesis. The hypothesis is destructed, which gives rise to [jc],
   [jpi], and [wk] hypotheses; then, an induction is performed on the [jc]
   hypothesis. *)

Ltac jctop_induction :=
  (* Introduce the [jctop] hypothesis and destruct it. *)
  introv h; dependent destruction h; subst;
  (* We now have three hypotheses. Re-introduce them in the goal, in an
     appropriate order, so that the [jc] hypothesis becomes the primary
     hypothesis and can be used in an induction. *)
  match goal with h: wk _ _ _ |- _ => revert h end;
  match goal with h: jpi ?R2 _ _ |- _ => generalize R2 h; clear R2 h end;
  match goal with h: jc _ _ _ _ _ _ |- _ => revert h end;
  (* Attack the induction. *)
  intro h; dependent induction h;
  intros; wk_inversion; subst.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps prove a lemma whose statement begins with a
   [jctop] hypothesis. The hypothesis is destructed, which gives rise to [jc],
   [jpi], and [wk] hypotheses; only the first and last one are kept; then, an
   induction is performed on the [jc] hypothesis. *)

Ltac jctop_simplified_induction :=
  (* Introduce the [jctop] hypothesis and destruct it. *)
  introv h; dependent destruction h; subst;
  (* We now have three hypotheses. Drop the [jpi] hypothesis. Re-introduce
     the other two in the goal, in an appropriate order, so that the [jc]
     hypothesis becomes the primary hypothesis and can be used in an induction. *)
  match goal with h: wk _ _ _ |- _ => revert h end;
  match goal with h: jpi ?R2 _ _ |- _ => clear R2 h end;
  match goal with h: jc _ _ _ _ _ _ |- _ => revert h end;
  (* Attack the induction. *)
  intro h; dependent induction h;
  intros; wk_inversion; subst.

