Set Implicit Arguments.
Require Import MyTactics.
Require Export MSA.
Require Export AllocationMaps.
Require Export MSAAllocationMaps.
Require Export MSAProduct.
Require Export MSADiscrete.
Require Export MSABit.
Require Import Syntax.

(* ---------------------------------------------------------------------------- *)

(* A lock resource, or instrumented lock heap, maps lock addresses to pairs of
   a syntactic permission (i.e., a lock invariant) and a (nothing-or-exclusive)
   access right, which tells whether the lock is taken by the current thread. *)

(* The mapping of lock addresses to lock invariants is analogous to the
   ``store typing'' in a proof of type soundness for ML with weak
   references. *)

(* The type [lock_brand] is isomorphic to [unit]. It helps distinguish this
   use of the type [bit] from other uses. *)

Inductive lock_brand : Type :=
  LockBrand.

Notation lock_info :=
  (discrete term * bit lock_brand)%type.

Notation LockZero P :=
  (DiscreteValue P, BitZero _).

Notation LockOne P :=
  (DiscreteValue P, BitOne LockBrand).

Notation lock_resource :=
  (aMap lock_info).

Notation lock_void :=
  (empty_map _).

(* ---------------------------------------------------------------------------- *)

(* Lock resources form a monotonic separation algebra. (Sanity check.) *)

Goal MSA lock_resource.
Proof.
  eauto with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [invariant R lck P] means that, according to the lock resource
   [R], the invariant of the lock [lck] is [P]. This provides a useful notation,
   which is used in the typing rule [JLock]. *)

Definition invariant (R : lock_resource) lck P :=
  (* We do not care whether the lock is free or taken. *)
  mapsp R lck (at_left (eq (DiscreteValue P))).

(* Construction. *)

Hint Unfold invariant : mapsp.

(* Inversion. *)

Ltac invariant_inversion :=
  repeat match goal with h: invariant _ _ _ |- _ =>
    unfold invariant, mapsp, at_left in h;
    destruct h as [ [ ? ? ] [ ? ? ]]; subst
  end.

(* An equivalent formulation. *)

Goal
  forall R lck P,
  invariant R lck P <->
  exists bit, maps R lck (DiscreteValue P, bit).
Proof.
  split; intros; unpack; invariant_inversion; eauto 6 with mapsp.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [taken R lck] means that, according to the lock resource [R],
   the lock is taken. *)

Definition taken (R : lock_resource) lck :=
  (* This time, we do not care about the lock invariant. *)
  mapsp R lck (at_right (eq (BitOne LockBrand))).

(* Construction. *)

Hint Unfold taken : mapsp.

(* Inversion. *)

Ltac taken_inversion :=
  repeat match goal with h: taken _ _ |- _ =>
    unfold taken, mapsp, at_right in h;
    destruct h as [ [ ? ? ] [ ? ? ]]; subst
  end.

(* An equivalent formulation. *)

Goal
  forall R lck,
  taken R lck <->
  exists d, maps R lck (d, BitOne LockBrand).
Proof.
  split; intros; unpack; taken_inversion; eauto 6 with mapsp.
Qed.

(* ---------------------------------------------------------------------------- *)

(* These predicates hold of a newly allocated lock, provided the lock resource
   is extended with appropriate information. This is stated by the following
   two lemmas. *)

Lemma new_lock_invariant:
  forall R1 R2 P b lck,
  R2 = allocate R1 (DiscreteValue P, b) ->
  consistent R1 ->
  lck = limit R1 ->
  invariant R2 lck P.
Proof.
  intros. subst. eauto 7 using @maps_allocate_there with mapsp.
Qed.

Lemma new_lock_taken:
  forall R1 R2 P lck,
  R2 = allocate R1 (LockOne P) ->
  consistent R1 ->
  lck = limit R1 ->
  taken R2 lck.
Proof.
  intros. subst. eauto 7 using @maps_allocate_there with mapsp.
Qed.

(* An ad hoc variant. *)

Lemma new_lock_invariant_core:
  forall R1 R2 P b lck,
  R2 = allocate (core R1) (DiscreteValue P, b) ->
  consistent R1 ->
  lck = limit R1 ->
  invariant R2 lck P.
Proof.
  intros. subst. eapply new_lock_invariant; eauto using @consistent_core, @limit_core.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The left-hand component of the pair [LockOne P] justifies [TyLock P], while
   the right-hand component justifies [TyLocked]. The following lemma
   facilitates this split. *)

Lemma invariant_taken_split:
  forall (R : lock_resource) d,
  star (allocate (core R) (d, BitZero _)) (allocate R (d, BitOne LockBrand)) =
  allocate R (d, BitOne LockBrand).
Proof.
  (* Note that the way in which [R] is split is not important. *)
  intros. rewrite @star_allocate. rewrite core_left_unit. f_equal.
  msa_pair. f_equal. eauto with duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* When both [invariant] and [taken] hold, we have the following situation. *)

Lemma invariant_taken:
  forall R1 R2 lck P,
  invariant R2 lck P ->
  taken R1 lck ->
  consistent (star R1 R2) ->
  maps (star R1 R2) lck (LockOne P).
Proof.
  intros. unpack. invariant_inversion. taken_inversion.
  forwards: @maps_star_construction; eauto.
  forwards: @maps_consistent; eauto.
  msa_pair. unpack.
  equates 1. eassumption. f_equal; eauto with use_maximal maximal.
Qed.

(* ---------------------------------------------------------------------------- *)

(* An update from [LockZero] to [LockOne] establishes [taken]. *)

Lemma update_taken:
  forall R R' lck P,
  maps R lck (LockZero P) ->
  R' = update R lck (LockOne P) ->
  taken R' lck.
Proof.
  intros. subst. eauto 7 using @maps_update_there with mapsp.
Qed.

