Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.

(* ------------------------------------------------------------------------- *)

(* The following axioms characterize a monotonic separation algebra. *)

Class MSA (resource : Type) := {

  (* Resources are combined using [star]. *)

  star: resource -> resource -> resource;

  (* [star] is commutative and associative. *)

  star_commutative:
    forall R1 R2,
    star R1 R2 = star R2 R1;

  star_associative:
    forall R1 R2 R3,
    star R1 (star R2 R3) = star (star R1 R2) R3;

  (* Because we require [star] to be a total function, we must allow for the
     possibility that some resources are inconsistent. An inconsistent
     resource is semantically equivalent to [false]. There may exist several
     syntactically distinct inconsistent resources. *)

  consistent: resource -> Prop;

  (* If a combination is consistent, then its parts are consistent as well. *)

  consistent_split_left:
    forall R1 R2,
    consistent (star R1 R2) ->
    consistent R1;

  (* Every resource [R] has a duplicable core. *)

  core: resource -> resource;

  (* The resource [core R] acts as a unit element for [R] with respect to
     [star]. That is, it is compatible with [R], and adds nothing to it. *)

  core_right_unit:
    forall R,
    star R (core R) = R;

  (* Two compatible resources have a common core. *)

  compatible_implies_common_core_left:
    forall R1 R2 R,
    star R1 R2 = R ->
    consistent R ->
    core R1 = core R;

  (* A duplicable resource is its own core. *)

  duplicable_implies_self_core:
    forall R,
    star R R = R ->
    R = core R;

  (* Every core is duplicable. *)

  cores_are_duplicable:
    forall R,
    star (core R) (core R) = core R;

  (* If a combination is consistent and duplicable, then its parts are
     duplicable as well. *)

  duplicable_split_left:
    forall R1 R2 R,
    star R1 R2 = R ->
    consistent R ->
    star R R = R ->
    star R1 R1 = R1;

  (* The information relation represents how the currently executing thread is
     allowed to modify some other (currently inactive) thread's resource.
     That is, the information relation describes the changes that can be
     imposed to one thread by some other thread of execution. *)

  rely: resource -> resource -> Prop;

  rely_reflexive:
    forall R,
    rely R R;

  (* Note that, although the relation [rely] is usually transitive, this property
     is not technically required anywhere, so we do not require it. *)

  (* [rely] preserves consistency. A consistent resource cannot be forced to
     evolve to an inconsistent one. *)

  (* Technically, one could work without this property, because the axiom
     [rely_star] splits such a move into two smaller moves whose destinations
     are also guaranteed to be consistent. In effect, if we did not require
     [rely] to preserve consistency, we would have to explicitly work within
     the intersection of [rely] and consistency. Life should be slightly
     simpler for our clients if we prove consistency up front. *)

  rely_consistent:
    forall R1 R2,
    consistent R1 ->
    rely R1 R2 ->
    consistent R2;

  (* The information relation is compatible with separating conjunction in
     the following sense: if we have a split and if the outside world requires
     that the sum grow, then each summand is able to grow to reflect this.
     Intuitively, the new duplicable information is simply sent down into
     both sides of the split. *)

  rely_star:
    forall R1 R2 R',
    rely (star R1 R2) R' ->
    consistent (star R1 R2) ->
    exists R'1 R'2,
    star R'1 R'2 = R' /\ rely R1 R'1 /\ rely R2 R'2;

  rely_core:
    forall R1 R2,
    rely R1 R2 ->
    rely (core R1) (core R2)

}.

(* ------------------------------------------------------------------------- *)

(* An element [R] is duplicable if and only if it combines with itself to
   produce itself. *)

Notation duplicable R :=
  (star R R = R).

Section Definitions.

Context `{MSA A}.

(* A property [P] of a resource is duplicable if and only if it is preserved
   by [core]. *)

Definition duplicablep (P : A -> Prop) :=
  forall R, P R -> P (core R).

(* A property [P] of a resource is maximal if and only if it is preserved
   by addition of resources. *)

Definition maximalp (P : A -> Prop) :=
  forall R1 R2 R, P R1 -> star R1 R2 = R -> consistent R -> P R.

(* An element [R1] is maximal if and only if every consistent combination of
   it with some resource [R2] yields [R1] again; i.e., equality with [R1] is
   maximal. *)

Definition maximal R1 :=
  maximalp (eq R1).

(* A property [P] of a resource is stable if and only if it is preserved by
   [rely]. *)

Definition stablep (P : A -> Prop) :=
  forall R1 R2, P R1 -> rely R1 R2 -> P R2.

(* A resource [R] is stable if and only if equality with [R] is stable. *)

Definition stable R :=
  stablep (eq R).

(* An element [R] is atomic if every split of it is [star (core R) R] or
   [star R (core R)]. *)

Definition atomic R :=
  forall R1 R2,
  star R1 R2 = R ->
  R1 = core R /\ R2 = R \/
  R1 = R /\ R2 = core R.

(* An element [R] is absolutely atomic if the only way of split it is
   into two copies of itself. This is equivalent to the conjunction of
   [duplicable] and [atomic]. *)

Definition aatomic R :=
  forall R1 R2,
  star R1 R2 = R ->
  R1 = R /\ R2 = R.

(* An element [R] is solitary if it is compatible only with its own core.
   This implies that it is maximal (see below). The reverse implication
   seems false. *)

Definition solitary R :=
  forall R',
  consistent (star R R') ->
  R' = core R.

(* An element [R] is narcissic if it is compatible only with itself.
   This is in fact equivalent to being solitary and duplicable. *)

Definition narcissic R :=
  forall R',
  consistent (star R R') ->
  R' = R.

(* An evolution relation is admissible if it is compatible with [star] and
   [rely] in the following sense. If we have a split between an active thread
   and a passive thread, and the active thread evolves along [evolution], then
   the change that is imposed on the passive thread is along [rely], and the
   group as a whole evolves along [evolution]. *)

Definition admissible (evolution : A -> A -> Prop) :=
  forall R1 R2,
  evolution R1 R2 ->
  forall R1f,
  consistent (star R1 R1f) ->
  exists R2f,
  evolution (star R1 R1f) (star R2 R2f) /\
  rely R1f R2f.

End Definitions.

(* ------------------------------------------------------------------------- *)

(* The following properties are valid in all monotonic separation algebras. *)

Section Properties.

Context `{MSA A}.

(* Some of the above definitions are usefully rephrased as lemmas. *)

Lemma use_maximal:
  forall a1 a2 a,
  star a1 a2 = a ->
  consistent a ->
  maximal a1 ->
  a1 = a.
Proof.
  eauto.
Qed.

Lemma use_maximal_symmetric:
  forall a1 a2 a,
  star a1 a2 = a ->
  consistent a ->
  maximal a2 ->
  a2 = a.
Proof.
  intros. subst. rewrite star_commutative.
  eapply use_maximal; eauto.
  equates 1; eauto using @star_commutative.
Qed.

Lemma use_atomic:
  forall a1 a2 a,
  star a1 a2 = a ->
  atomic a ->
  a1 = core a /\ a2 = a \/
  a1 = a /\ a2 = core a.
Proof.
  eauto.
Qed.

Lemma use_aatomic:
  forall a1 a2 a,
  star a1 a2 = a ->
  aatomic a ->
  a1 = a /\ a2 = a.
Proof.
  eauto.
Qed.

Lemma use_solitary:
  forall a1 a2,
  consistent (star a1 a2) ->
  solitary a1 ->
  a2 = core a1.
Proof.
  eauto.
Qed.

Lemma use_narcissic:
  forall a1 a2,
  consistent (star a1 a2) ->
  narcissic a1 ->
  a2 = a1.
Proof.
  eauto.
Qed.

(* The symmetric counterpart of the definition of [admissible]. *)

Lemma use_admissible_symmetric:
  forall evolution,
  admissible evolution ->
  forall R1 R2,
  evolution R1 R2 ->
  forall R1f,
  consistent (star R1f R1) ->
  exists R2f,
  evolution (star R1f R1) (star R2f R2) /\
  rely R1f R2f.
Proof.
  introv hadm. intros.
  forwards: hadm; [ eauto | rewrite star_commutative; eauto | unpack ].
  xplit.
    equates 1; [ equates 2 | ]; eauto using @star_commutative. 
    eauto.
Qed.

(* The symmetric counterpart of the axiom [star_associative]. *)

Lemma star_associative_symmetric:
  forall R1 R2 R3,
  star (star R1 R2) R3 = star R1 (star R2 R3).
Proof.
  symmetry. apply star_associative.
Qed.

(* The symmetric counterpart of the axiom [consistent_split_left]. *)

Lemma consistent_split_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  consistent R2.
Proof.
  intros. eapply consistent_split_left. rewrite star_commutative. eassumption.
Qed.

(* The symmetric counterpart of the axiom [core_right_unit]. *)

Lemma core_left_unit:
  forall R,
  star (core R) R = R.
Proof.
  intros. rewrite star_commutative. eapply core_right_unit.
Qed.

(* Variants of [core_left_unit] and [core_right_unit]. *)

Lemma core_right_unit_variant:
  forall R1 R2,
  core R1 = R2 ->
  star R1 R2 = R1.
Proof.
  intros; subst. eapply core_right_unit.
Qed.

Lemma core_left_unit_variant:
  forall R1 R2,
  core R1 = R2 ->
  star R2 R1 = R1.
Proof.
  intros; subst. eapply core_left_unit.
Qed.

(* The symmetric counterpart of the axiom [duplicable_split_left]. *)

Lemma duplicable_split_right:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  duplicable R ->
  duplicable R2.
Proof.
  intros. eapply duplicable_split_left.
    rewrite star_commutative. eassumption.
    eassumption.
    eassumption.
Qed.

(* The symmetric counterparts of the axiom [compatible_implies_common_core_left]. *)

Lemma compatible_implies_common_core_right:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  core R2 = core R.
Proof.
  intros. eapply compatible_implies_common_core_left.
    rewrite star_commutative. eassumption.
    eassumption.
Qed.

Lemma compatible_implies_common_core_symmetric:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  core R1 = core R2.
Proof.
  intros.
  transitivity (core R). 
  eauto using compatible_implies_common_core_left.
  symmetry. eauto using compatible_implies_common_core_right.
Qed.

Lemma compatible_implies_common_core_both:
  forall R1 R2,
  consistent (star R1 R2) ->
  core R1 = core (star R1 R2) /\ core R2 = core (star R1 R2).
Proof.
  eauto using compatible_implies_common_core_left, compatible_implies_common_core_right.
Qed.

(* The symmetric counterpart of the axiom [duplicable_implies_self_core]. *)

Lemma duplicable_implies_self_core_symmetric:
  forall R,
  star R R = R ->
  core R = R.
Proof.
  symmetry. eauto using @duplicable_implies_self_core.
Qed.

(* Other ways of proving two cores equal. *)

Lemma simplify_core_core_left_left:
  forall R1 R2 R,
  core R1 = core R ->
  consistent (star R1 R2) ->
  core (star R1 R2) = core R.
Proof.
  intros.
  erewrite <- @compatible_implies_common_core_left by eauto.
  assumption.
Qed.

Lemma simplify_core_core_left_right:
  forall R1 R2 R,
  core R2 = core R ->
  consistent (star R1 R2) ->
  core (star R1 R2) = core R.
Proof.
  intros.
  rewrite @star_commutative.
  eapply simplify_core_core_left_left; eauto.
  rewrite @star_commutative.
  eauto.
Qed.

(* There are several ways to define what it means for a resource [R] to be
   duplicable. One is to state that two copies of [R] amount to one, that is,
   [star R R = R] holds. One other is to state that [R] is equal to its
   core. The two definitions are equivalent. One implication is an axiom
   ([duplicable_implies_self_core]) while the converse implication is derived
   below ([core_implies_duplicable]). A third way would be to state that [R]
   is duplicable if it is a core, that is, if there exists some resource [R0]
   such that [R = core R0]. Again, this definition is equivalent to the first
   one. This is another axiom ([cores_are_duplicable]). *)

Lemma core_implies_duplicable:
  forall R,
  R = core R ->
  duplicable R.
Proof.
  introv h. rewrite h at 2. apply core_right_unit.
Qed.

(* Core is idempotent. *)

Lemma core_idempotent:
  forall R,
  consistent R ->
  core (core R) = core R.
Proof.
  eauto using compatible_implies_common_core_left, core_left_unit.
Qed.

Lemma core_idempotent_symmetric:
  forall R,
  consistent R ->
  core R = core (core R).
Proof.
  symmetry. eauto using core_idempotent.
Qed.

(* A consistent resource has a consistent core. *)

Lemma consistent_core:
  forall R,
  consistent R ->
  consistent (core R).
Proof.
  intros.
  eapply consistent_split_left.
  rewrite core_left_unit.
  assumption.
Qed.

(* The core of a consistent combination is the combination of the cores. *)

Lemma core_star:
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = star (core R1) (core R2).
Proof.
  intros.
  rewrite <- cores_are_duplicable.
  symmetry. f_equal.
  eauto using compatible_implies_common_core_left.
  eauto using compatible_implies_common_core_right.
Qed.

Lemma star_core: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core (star R1 R2).
Proof.
  intros. symmetry. eauto using core_star.
Qed.

Lemma star_core_right: (* another formulation *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core R2.
Proof.
  intros.
  rewrite star_core by eauto.
  symmetry. eapply compatible_implies_common_core_right; eauto.
Qed.

Lemma star_core_left: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core R1.
Proof.
  intros.
  rewrite star_core by eauto.
  symmetry. eapply compatible_implies_common_core_left; eauto.
Qed.

Lemma core_star_right: (* yet another formulation *)
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = core R2.
Proof.
  intros.
  rewrite core_star by eauto.
  eapply star_core_right. eauto.
Qed.

Lemma core_star_left: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = core R1.
Proof.
  intros.
  rewrite core_star by eauto.
  eapply star_core_left. eauto.
Qed.

Lemma star_core_3:
  forall R1 R2 R3,
  consistent (star R1 (star R2 R3)) ->
  star (core R1) (star (core R2) (core R3)) = core (star R1 (star R2 R3)).
Proof.
  intros.
  rewrite star_core by eauto using consistent_split_right.
  eapply star_core. eauto.
Qed.

(* If [R] is duplicable and atomic, then it is absolutely atomic,
   and conversely. *)

Lemma duplicable_atomic_implies_aatomic:
  forall R,
  duplicable R ->
  atomic R ->
  aatomic R.
Proof.
  introv hd ha. intro; intros.
  forwards [ [ hc ? ] | [ ? hc ] ]: ha; [ eauto | | ];
  rewrite <- duplicable_implies_self_core in hc;
  eauto.
Qed.

Lemma aatomic_implies_duplicable:
  forall R,
  aatomic R ->
  duplicable R.
Proof.
  introv ha.
  forwards [ _ f ]: ha; [ eapply core_right_unit | ].
  rewrite <- f. eapply cores_are_duplicable.
Qed.

Lemma aatomic_implies_atomic:
  forall R,
  aatomic R ->
  atomic R.
Proof.
  introv ha.
  forwards d: aatomic_implies_duplicable. eauto.
  intro; intros.
  rewrite <- (duplicable_implies_self_core d).
  eauto.
Qed.

(* A solitary element is maximal. *)

Lemma solitary_implies_maximal:
  forall a,
  solitary a ->
  maximal a.
Proof.
  unfold solitary, maximal, maximalp. introv h. intros; subst.
  forwards: h; [ eauto | ]. subst.
  eauto using @core_right_unit.
Qed.

(* A maximal, atomic element is solitary. *)

Lemma maximal_atomic_implies_solitary:
  forall a,
  maximal a ->
  atomic a ->
  solitary a.
Proof.
  unfold maximal, maximalp, atomic, solitary.
  introv hm ha. intros.
  (* Because [a] is maximal, [star a _] is [a]. *)
  forwards eq: hm; eauto.
  (* Because [a] is atomic, [_] is [core a]. *)
  forwards cases: ha. symmetry. eexact eq.
  destruct cases; unpack; congruence.
Qed.

(* For consistent elements, narcissic is equivalent to solitary and duplicable. *)

Lemma narcissic_implies_duplicable:
  forall a,
  narcissic a ->
  consistent a ->
  duplicable a.
Proof.
  unfold narcissic, solitary. introv h ?.
  forwards eq: h. rewrite core_right_unit. eassumption.
  rewrite <- eq.
  eapply cores_are_duplicable.
Qed.

Lemma narcissic_implies_solitary:
  forall a,
  narcissic a ->
  consistent a ->
  solitary a.
Proof.
  intros.
  forwards: narcissic_implies_duplicable. eauto. eauto.
  unfold solitary. intros.
  rewrite <- duplicable_implies_self_core by eauto.
  eauto.
Qed.

Lemma solitary_duplicable_implies_narcissic:
  forall a,
  solitary a ->
  duplicable a ->
  narcissic a.
Proof.
  unfold solitary, narcissic. introv hs hd. intros.
  rewrite duplicable_implies_self_core by eauto.
  eauto.
Qed.

(* Perhaps surprisingly (?), the only way of splitting a (consistent)
   narcissic element is into two copies of itself. In other words, a
   (consistent) narcissic element is absolutely atomic. *)

Lemma split_narcissic_left:
  forall a1 a2 a,
  narcissic a ->
  consistent a ->
  star a1 a2 = a ->
  a1 = a.
Proof.
  introv hn hc eq.
  (* [a] is duplicable. *)
  forwards: narcissic_implies_duplicable; eauto.
  (* Thus, [a1] and [a2] are duplicable. At the time of writing, this
     is the only place where the property [duplicable_split_left] is
     exploited. *)
  forwards d1: duplicable_split_left eq; eauto.
  forwards d2: duplicable_split_right eq; eauto.
  (* Thus, [a = a1 * a2 = a1 * a1 * a2 = a1 * a]. *)
  assert (star a a1 = a).
    rewrite <- eq.
    rewrite (star_commutative a1 a2).
    rewrite <- star_associative.
    rewrite d1.
    reflexivity.
  (* Because [a1] is narcissic, this implies [a1 = a]. *)
  eapply hn. equates 1; eauto.
Qed.

Lemma split_narcissic_right:
  forall a1 a2 a,
  narcissic a ->
  consistent a ->
  star a1 a2 = a ->
  a2 = a.
Proof.
  intros. eapply split_narcissic_left. eauto. eauto. rewrite star_commutative. eauto.
Qed.

Lemma narcissic_implies_aatomic:
  forall a,
  narcissic a ->
  consistent a ->
  aatomic a.
Proof.
  unfold aatomic. eauto using split_narcissic_left, split_narcissic_right.
Qed.

(* A trivial property is duplicable, maximal, stable. *)

Lemma duplicablep_trivial:
  duplicablep (fun _ => True).
Proof.
  unfold duplicablep. tauto.
Qed.

Lemma maximalp_trivial:
  maximalp (fun _ => True).
Proof.
  unfold maximalp. tauto.
Qed.

Lemma stablep_trivial:
  stablep (fun _ => True).
Proof.
  unfold stablep. tauto.
Qed.

(* The property of being equal to some duplicable element [a] is duplicable. *)

Lemma duplicablep_eq:
  forall a,
  duplicable a ->
  duplicablep (eq a).
Proof.
  unfold duplicablep. intros. subst. eauto using duplicable_implies_self_core.
Qed.

Lemma maximalp_eq:
  forall a,
  maximal a ->
  maximalp (eq a).
Proof.
  (* This lemma is really an unfolding hint. *)
  eauto.
Qed.

(* The identity is an admissible evolution. *)

Definition identity (a1 a2 : A) :=
  a1 = a2.

Lemma admissible_identity:
  admissible identity.
Proof.
  unfold admissible, identity. intros. subst. eauto using rely_reflexive.
Qed.

End Properties.

Hint Resolve @core_left_unit @core_right_unit : core_unit.

Hint Resolve @consistent_core : consistent.

Hint Resolve @consistent_split_left @consistent_split_right : consistent_split.

Hint Resolve @core_left_unit @core_right_unit : core_unit.

Hint Resolve @duplicable_atomic_implies_aatomic : aatomic.

Hint Resolve @maximalp_trivial @maximalp_eq : maximal.

Hint Resolve @use_maximal @use_maximal_symmetric : use_maximal.

Hint Resolve @stablep_trivial : stable.

Hint Resolve @cores_are_duplicable @duplicablep_trivial @duplicablep_eq : duplicable.

Hint Resolve @admissible_identity : admissible.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps exploit the property [rely_star]. It should work
   iteratively when there are nested stars. The database [rely_star_hints] is
   used to prove the consistency subgoal. *)

Hint Resolve @consistent_split_left @consistent_split_right : rely_star_hints.

Ltac rely_star :=
  repeat match goal with
  h: rely (star ?R1a ?R1b) ?R2 |- _ =>
    forwards: @rely_star; [
      eexact h
    | eauto with rely_star_hints
    | unpack; subst;
      (* Move [h] away, so we look for another instance of [rely_star]. *)
      revert h
    ]
  end;
  (* Re-introduce the hypotheses that were moved away. *)
  intros.

Ltac rely_star_clear :=
  rely_star;
  repeat match goal with h: rely (star _ _) (star _ _) |- _ => clear h end.

