Set Implicit Arguments.
Generalizable All Variables.
Require Import MyList.
Require Import MyTactics.
Require Import MSA.
Require Import AllocationMaps.
Transparent empty_map allocate limit update.

(* ------------------------------------------------------------------------- *)

(* Allocation maps over a monotonic separation algebra form a monotonic
   separation algebra. *)

(* The relation [rely] allows new memory addresses to be allocated. At existing
   addresses, the old and new values must be related by the relation [rely]
   (one level down). At new addresses, the new value can be arbitrarily
   chosen, provided it is consistent and duplicable. The consistency
   requirement allows us to establish the property [rely_consistent]. The
   duplicability requirement means that a passive thread cannot be ``given''
   non-duplicable access rights. It is used in the theory of round allocation
   maps. *)

Definition aMap_rely `{MSA A} (R1 R2 : aMap A) : Prop :=
  match R1, R2 return _ with
  | MapUndefined, MapUndefined =>
      True
  | Map m1, Map m2 =>
      growth rely (fun a => consistent a /\ duplicable a) m1 m2
  | _, _ =>
      False
  end.

(* ------------------------------------------------------------------------- *)

(* We now gather the above definitions in an instance declaration, and
   prepare to meet the proof obligations that ensue. *)

Program Instance msa_map `{MSA A} : MSA (aMap A) := {

  (* The allocation limit must be the same on both sides of [star]. This is
     required so that when one side decides to allocate a fresh address, the
     new address really is globally fresh. Thus, in the definition of [star],
     we produce an inconsistent result unless both sides agree on the
     allocation limit. When the allocation limits agree, the two maps are
     combined pointwise. *)
  star := mzip star;

  (* [consistent] is defined pointwise. *)
  consistent := everywhere consistent;

  (* [core] is defined pointwise. *)
  core := transform core;

  rely := aMap_rely
}.

(* ------------------------------------------------------------------------- *)

(* The tactic [msadam] -- for ``destruct allocation maps'' -- does the bulk of
   the preparatory work. *)

Local Ltac msadam :=
  unfold mzip, aMap_rely in *; dam.

(* ------------------------------------------------------------------------- *)

(* These ad hoc lemmas are used below. *)

Lemma Forall_map_core_1:
  forall `{MSA A} (xs : list A),
  Forall consistent xs ->
  Forall (fun x => consistent x /\ duplicable x) (map core xs).
Proof.
  intros. eapply Forall_map; eauto using @consistent_core, @cores_are_duplicable.
Qed.

Lemma Forall_map_core_2:
  forall `{MSA A} (xs : list A),
  Forall (fun x => consistent x /\ duplicable x) xs ->
  Forall (fun x => consistent x /\ duplicable x) (map core xs).
Proof.
  intros. eapply Forall_map.
    eauto.
    simpl. intuition eauto using @consistent_core, @cores_are_duplicable.
Qed.

(* ------------------------------------------------------------------------- *)

(* This lemma corresponds to the proof obligation for [rely_star]. I state it
   explicitly because I wish to control its statement precisely. (Earlier, I
   was able to prove it on the fly, without stating a separate lemma, using
   [dependent induction]; this worked in Coq 8.3 but for some reason does not
   produce the desired induction hypothesis in Coq 8.4.) *)

Lemma aMap_rely_star:
  forall `{MSA A} (R1 R2 : list A),
  growth rely (fun a => consistent a /\ duplicable a) R1 R2 ->
  forall R1a R1b,
  zip star R1a R1b = R1 ->
  length R1a = length R1b ->
  Forall consistent R1 ->
  exists R2a R2b,
  length R2a = length R2b /\
  zip star R2a R2b = R2 /\
  growth rely (fun a => consistent a /\ duplicable a) R1a R2a /\
  growth rely (fun a => consistent a /\ duplicable a) R1b R2b.
Proof.
  induction 1; intros; my_list_cleanup.
  (* Case: GrowthZero *)
  (* The list [ys] represents the newly allocated memory locations.
     We distribute them in an arbitrary manner into the two sides.
     The lemma [core_right_unit] offers one way of doing so. *)
  match goal with ys: list _ |- _ =>
    exists ys (map core ys)
  end.
  repeat split.
    (* Obligation 1. *)
    symmetry. eapply map_length.
    (* Obligation 2. *)
    eauto using zip_map_right, @core_right_unit.
    (* Obligation 3. *)
    econstructor; eauto.
    (* Obligation 4. *)
    econstructor; eauto using Forall_map_core_2.
  (* Case: GrowthSucc *)
  (* Apply [rely_star] (one level down) for the head of the list. *)
  rely_star.
  (* Apply the induction hypothesis for the tail. *)
  forwards [ R2a [ R2b ? ]]: IHgrowth; [ eauto | eauto | eauto | clear IHgrowth; unpack ].
  (* Exhibit our witnesses and conclude. *)
  match goal with h: rely (star _ _) (star ?a ?b) |- _ =>
    exists (a :: R2a) (b :: R2b)
  end.
  repeat split; simpl; solve [ congruence | econstructor; eauto ].
Qed.

(* ------------------------------------------------------------------------- *)

(* Here come the proof obligations for the above instance declaration. *)

Next Obligation. (* star_commutative *)
  msadam. eauto 2 using zip_commutative, @star_commutative.
Qed.

Next Obligation. (* star_associative *)
  msadam; try solve [ 
    (* Some cases are impossible. *)
    rewrite zip_length_right in * by eauto; false
  ].
  eauto 2 using zip_associative, @star_associative.
Qed.

Next Obligation. (* consistent_split_left *)
  msadam. induction_lists; econstructor; eauto using @consistent_split_left.
Qed.

Next Obligation. (* core_right_unit *)
  msadam; eauto 2 using zip_map_right, @core_right_unit;
  (* The remaining cases are impossible. *)
  rewrite map_length in *; false.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  msadam. induction_lists; eauto 2 using @compatible_implies_common_core_left.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  msadam. induction_lists; eauto 2 using @duplicable_implies_self_core.
Qed.

Next Obligation. (* cores_are_duplicable *)
  msadam. induction_lists; eauto 2 using @cores_are_duplicable.
Qed.

Next Obligation. (* duplicable_split_left *)
  msadam. induction_lists; eauto 2 using @duplicable_split_left.
Qed.

Next Obligation. (* rely_reflexive *)
  msadam; eauto using growth_reflexive, @rely_reflexive.
Qed.

Next Obligation. (* rely_consistent *)
  msadam. induction_growth; destruction_Forall; unpack; econstructor; eauto using @rely_consistent.
  (* A little bit of left-over manual work. *)
  eapply Forall_covariant. eauto. simpl. tauto.
Qed.

Next Obligation. (* rely_star *)
  msadam.
  forwards [ m2a [ m2b ? ]]: aMap_rely_star; [ eauto | eauto | eauto | eauto | unpack ].
  exists (Map m2a) (Map m2b).
  dblib_by_cases.
  repeat split; congruence.
Qed.

Next Obligation. (* rely_core *)
  msadam. induction_growth; econstructor; eauto using Forall_map_core_2, @rely_core.
Qed.

(* ------------------------------------------------------------------------- *)

(* MSA-related properties of the operations on allocation maps. *)

Section Operations.

Variable A : Type.

Context `{MSA A}.

(* ------------------------------------------------------------------------- *)

(* Properties of the empty allocation map. *)

Lemma consistent_empty_map:
  consistent (@empty_map A).
Proof.
  econstructor.
Qed.

Lemma duplicable_empty_map:
  duplicable (@empty_map A).
Proof.
  econstructor.
Qed.

(* ------------------------------------------------------------------------- *)

(* Properties of maps. *)

(* A consistent resource maps every address to a consistent value. *)

Lemma maps_consistent:
  forall R r a,
  maps R r a ->
  consistent R ->
  consistent a.
Proof.
  simpl. intros. eapply everywhere_maps; eauto.
Qed.

(* If [R] maps [r] to [a], then [core R] maps [r] to [core a]. *)

Lemma maps_core:
  forall R r a b,
  maps R r a ->
  core a = b ->
  maps (core R) r b.
Proof.
  simpl. eauto using maps_transform.
Qed.

(* If [core R] maps [r] to [b], then [R] maps [r] to some [a] whose core is
   [b]. *)

Lemma maps_core_reverse:
  forall R r b,
  maps (core R) r b ->
  exists a,
  maps R r a /\ core a = b.
Proof.
  simpl. eauto using maps_transform_reverse.
Qed.

(* If [R1] maps [l] to [a1], [R2] maps [l] to [a2], and if [R1] and [R2] are
   compatible, then [R1 * R2] maps [l] to [a1 * a2]. This is a direct
   consequence of the definition of [star]. *)

Lemma maps_star_construction:
  forall R1 R2 l a1 a2,
  consistent (star R1 R2) ->
  maps R1 l a1 ->
  maps R2 l a2 ->
  maps (star R1 R2) l (star a1 a2).
Proof.
  introv ?; do 2 dependent_destruction; msadam; econstructor. eauto using is_nth_zip.
Qed.

(* This lemma combines [maps_star] and [maps_star_construction], in a weaker variant
   that does not require the consistency of [R1 * R2]. It only requires [R1] and [R2]
   to agree on the allocation limit. *)

Lemma maps_star_extension:
  forall R1 l a1,
  maps R1 l a1 ->
  forall R2,
  limit R1 = limit R2 ->
  exists a2,
  maps (star R1 R2) l (star a1 a2).
Proof.
  intros.
  assert (l < limit R1). rewrite limit_maps. eauto.
  assert (exists a2, maps R2 l a2). rewrite <- limit_maps. omega.
  unpack.
  msadam. repeat match goal with h: maps _ _ _ |- _ => inversion h; clear h end; subst.
  eexists. econstructor. eauto using is_nth_zip.
Qed.

(* If [R1] owns [l1] and [R2] owns [l2], and the combination of [R1] and [R2]
   is consistent, then [l1] and [l2] must be distinct addresses. *)

Lemma maps_star_construction_contrapositive:
  forall R1 R2 l1 l2 a1 a2,
  maps R1 l1 a1 ->
  maps R2 l2 a2 ->
  consistent (star R1 R2) ->
  ~ consistent (star a1 a2) ->
  l1 <> l2.
Proof.
  intros; intro; subst.
  forwards: @maps_star_construction; eauto.
  forwards: @maps_consistent; eauto.
Qed.

(* A combination of [maps_star_construction] and [maps_consistent]. *)

Lemma maps_star_consistent:
  forall R1 R2 l a1 a2,
  consistent (star R1 R2) ->
  maps R1 l a1 ->
  maps R2 l a2 ->
  consistent (star a1 a2).
Proof.
  eauto using maps_star_construction, maps_consistent.
Qed.

(* If [R1] maps [l] to [a1], and if [R1] and [R2] are compatible,
   then [R2] maps [l] to [a2], which is compatible with [a1]. *)

Lemma maps_star:
  forall R1 R2 l a1,
  maps R1 l a1 ->
  consistent (star R1 R2) ->
  exists a2,
  maps R2 l a2 /\ consistent (star a1 a2).
Proof.
  dependent_destruction. msadam.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eexists. split.
    eauto with maps.
    eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

(* If [star R1 R2] maps [l] to [a], then [a] can be split into [a1],
   which is contributed by [R1], and [a2], which is contributed by
   [R2]. *)

Lemma maps_star_inversion:
  forall R1 R2 l a,
  maps (star R1 R2) l a ->
  exists a1 a2,
  maps R1 l a1 /\ maps R2 l a2 /\ star a1 a2 = a.
Proof.
  dependent_destruction; msadam.
  forwards: length_is_nth.
    erewrite <- zip_length_right by eauto. eauto using is_nth_length.
  forwards: length_is_nth.
    erewrite <- zip_length_left by eauto. eauto using is_nth_length.
  unpack.
  repeat econstructor; eauto using is_nth_injective, is_nth_zip. (* cool *)
Qed.

(* A useful special case, where the only way of splitting [a] is into
   two copies of [a]. *)

Lemma maps_star_inversion_aatomic:
  forall R1 R2 l a,
  maps (star R1 R2) l a ->
  aatomic a ->
  maps R1 l a /\ maps R2 l a.
Proof.
  intros.
  forwards: maps_star_inversion; [ eauto | unpack ].
  forwards: use_aatomic; [ eauto | eauto | unpack ].
  split; congruence.
Qed.

(* If the address [r] exists in [R1] and this resource evolves along [rely] to
   [R2], then [r] still exists and its value has evolved along [rely], one
   level down. *)

Lemma maps_rely:
  forall R1 R2 r a1,
  maps R1 r a1 ->
  rely R1 R2 ->
  exists a2,
  maps R2 r a2 /\ rely a1 a2.
Proof.
  dependent_destruction; msadam.
  forwards: growth_is_nth. eauto. eauto. unpack.
  eauto with maps.
Qed.

(* Conversely, if [R1] evolves along [rely] to [R2] and if [r] exists in [R2],
   then either it existed previously and its value has evolved along [rely],
   one level down, or [r] is a newly allocated address and its value is
   consistent and duplicable. *)

Lemma maps_rely_reverse:
  forall R1 R2 r a2,
  maps R2 r a2 ->
  rely R1 R2 ->
  (exists a1, maps R1 r a1 /\ rely a1 a2) \/
  (consistent a2 /\ duplicable a2).
Proof.
  dependent_destruction; msadam.
  forwards [ ? | ? ]: growth_is_nth_reverse; [ eauto | eauto | unpack | unpack ];
  eauto with maps.
Qed.

(* If [R1] maps [l] to a maximal element [a], then [star R1 R2] must
   map [l] to the same element, provided it is consistent. *)

(* The tactic [maps_maximal], defined further on, helps use these two lemmas. *)

Lemma maps_maximal_left:
  forall R1 R2 l a,
  maps R1 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  maps (star R1 R2) l a.
Proof.
  unfold maximal, maximalp. dependent_destruction; introv ? hmax; msadam. econstructor.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eapply is_nth_zip_eq; eauto.
  eapply hmax; eauto.
  eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

Lemma maps_maximal_right:
  forall R1 R2 l a,
  maps R2 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  maps (star R1 R2) l a.
Proof.
  intros.
  rewrite star_commutative.
  eapply maps_maximal_left; eauto.
  rewrite star_commutative.
  eauto.
Qed.

(* A special case of [maps_star_construction], where [a1] is solitary.
   In that case, [a2] must be [core a1], and [star a1 a2] is just [a1]. *)

Lemma maps_star_construction_solitary:
  forall R1 R2 l a1 a2,
  consistent (star R1 R2) ->
  maps R1 l a1 ->
  maps R2 l a2 ->
  solitary a1 ->
  maps (star R1 R2) l a1 /\ a2 = core a1.
Proof.
  intros.
  forwards h: @maps_star_construction; eauto.
  forwards: @maps_consistent; eauto.
  forwards eq: @use_solitary; eauto.
  rewrite eq in h.
  rewrite @core_right_unit in h.
  eauto.
Qed.

(* If one side of a split has a solitary element [a], then the other side must
   have ``nothing'', that is, [core a]. This is the definition of [solitary],
   lifted up to allocation maps. *)

Lemma maps_star_solitary_left:
  forall R1 R2 l a,
  maps R1 l a ->
  solitary a ->
  consistent (star R1 R2) ->
  maps R2 l (core a).
Proof.
  intros.
  forwards: maps_star; [ eauto | eauto | unpack ].
  equates 1; [ eauto | symmetry ].
  eauto. (* this exploits [solitary a] *)
Qed.

Lemma maps_star_solitary_right:
  forall R1 R2 l a,
  maps R2 l a ->
  solitary a ->
  consistent (star R1 R2) ->
  maps R1 l (core a).
Proof.
  intros. eapply maps_star_solitary_left.
    eauto.
    eauto.
    rewrite star_commutative. eauto.
Qed.

(* If [P] is a duplicable property, then the property of mapping [l] to some
   element that satisfies [P] is itself duplicable. *)

Lemma mapsp_duplicablep:
  forall l P,
  duplicablep P ->
  duplicablep (fun R => mapsp R l P).
Proof.
  unfold duplicablep, mapsp. intros. unpack. eauto using maps_core.
Qed.

(* If [P] is a maximal property, then the property of mapping [l] to some
   element that satisfies [P] is itself maximal. *)

Lemma mapsp_maximalp:
  forall l P,
  maximalp P ->
  maximalp (fun R => mapsp R l P).
Proof.
  unfold maximalp, mapsp. intros. unpack. subst.
  forwards: maps_star; eauto. unpack.
  eauto using maps_star_construction.
Qed.

(* If [R1] maps [l] to an element that satisfies some stable property [P], and
   if [R1] evolves to [R2] via [rely], then [R2] maps [l] to some element that
   satisfies [P]. In other words, the property of mapping [l] to some element
   that satisfies [P] is stable. *)

Lemma mapsp_stablep:
  forall l P,
  stablep P ->
  stablep (fun R => mapsp R l P).
Proof.
  unfold stablep, mapsp. intros. unpack.
  forwards: maps_rely; [ eauto | eauto | unpack ].
  eauto.
Qed.

(* If [R1] maps [l] to a stable element [a], and if [R1] evolves to [R2] via
   [rely], then [R2] maps [l] to [a]. *)

Lemma maps_stable:
  forall R1 R2 l a,
  maps R1 l a ->
  rely R1 R2 ->
  stable a ->
  maps R2 l a.
Proof.
  unfold stablep, stable. dependent_destruction; introv ? hstable; msadam. econstructor.
  forwards: growth_is_nth; [ eauto | eauto | unpack ].
  erewrite hstable; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* Properties of allocate. *)

(* Allocation preserves consistency. *)

Lemma consistent_allocate:
  forall R a,
  consistent R ->
  consistent a ->
  consistent (allocate R a).
Proof.
  simpl. eauto using everywhere_allocate.
Qed.

(* [star] commutes with [allocate]. Each side of the split receives a partial
   access right for the new address. *)

Lemma star_allocate:
  forall R1 R2 a1 a2,
  star (allocate R1 a1) (allocate R2 a2) = allocate (star R1 R2) (star a1 a2).
Proof.
  msadam; try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

(* A particular case, which shows that [allocate R a] can be decomposed as a
   [star]-combination of a suitably-extended [R] and a suitably-extended [a]. *)

Lemma star_allocate_core:
  forall R a,
  star (allocate R (core a)) (allocate (core R) a) = allocate R a.
Proof.
  intros.
  rewrite <- (core_right_unit R) at 3.
  rewrite <- (core_left_unit a) at 3.
  eapply star_allocate.
Qed.

(* A particular case, where the resource [a] is split into [a] and [core a].
   Roughly speaking, all of [a] is sent into one side. *)

Lemma star_allocate_core_left_unit:
  forall R1 R2 a,
  star (allocate R1 (core a)) (allocate R2 a) = allocate (star R1 R2) a.
Proof.
  intros.
  rewrite <- (core_left_unit a) at 3.
  eapply star_allocate.
Qed.

Lemma star_allocate_core_right_unit:
  forall R1 R2 a,
  star (allocate R1 a) (allocate R2 (core a)) = allocate (star R1 R2) a.
Proof.
  intros.
  rewrite <- (core_right_unit a) at 3.
  eapply star_allocate.
Qed.

(* The allocation of a new address, with a consistent and duplicable
   access right, is permitted by [rely]. *)

Lemma rely_allocate:
  forall R a,
  consistent a ->
  duplicable a ->
  rely R (allocate R a).
Proof.
  msadam; eauto using growth_app, rely_reflexive.
Qed.

Lemma rely_allocate_core:
  forall R a,
  consistent a ->
  rely R (allocate R (core a)).
Proof.
  intros. eapply rely_allocate; eauto using @consistent_core, @cores_are_duplicable.
Qed.

(* ------------------------------------------------------------------------- *)

(* Properties of limit. *)

Lemma limit_split_left:
  forall R1 R2,
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R1.
Proof.
  msadam. rewrite zip_length_left; eauto.
Qed.

Lemma limit_split_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R2.
Proof.
  msadam. rewrite zip_length_right; eauto.
Qed.

Lemma limit_core:
  forall R,
  limit (core R) = limit R.
Proof.
  eapply limit_transform.
Qed.

Lemma maps_allocate_there:
  forall R r a,
  consistent R ->
  r = limit R ->
  maps (allocate R a) r a.
Proof.
  msadam. eauto using is_nth_app_far with maps is_nth.
Qed.

Lemma maps_allocate_there_core:
  forall R r a,
  consistent R ->
  r = limit R ->
  maps (allocate (core R) a) r a.
Proof.
  intros. eapply maps_allocate_there.
    eauto using @consistent_core.
    rewrite limit_core. eauto.
Qed.

(* If, in the heap [allocate R a], [r] points to [b], then either this
   fact was true in [R] already, or [b] is [a]. *)

Lemma maps_allocate_inversion:
  forall R r (a b : A),
  maps (allocate R a) r b ->
  maps R r b \/ a = b.
Proof.
  destruct R; dependent_destruction.
  forwards [ | ]: is_nth_inversion; eauto with maps.
Qed.

Lemma limit_rely:
  forall R1 R2,
  rely R1 R2 ->
  limit R1 <= limit R2.
Proof.
  msadam; eauto using growth_length.
Qed.

Lemma maps_star_limit:
  forall R1 R2 l a,
  maps (star R1 R2) l a ->
  limit R1 = limit R2.
Proof.
  dependent_destruction; msadam.
Qed.

Lemma consistent_star_limit:
  forall R1 R2,
  consistent (star R1 R2) ->
  limit R1 = limit R2.
Proof.
  intros.
  forwards: limit_split_left. eauto.
  forwards: limit_split_right. eauto.
  congruence.
Qed.

(* ------------------------------------------------------------------------- *)

(* Properties of update. *)

(* Update preserves consistency. *)

Lemma consistent_update:
  forall R l a,
  consistent R ->
  consistent a ->
  consistent (update R l a).
Proof.
  simpl. eauto using everywhere_update.
Qed.

(* [star] commutes with [update]. Each side of the split receives a partial
   access right for the updated address. *)

Lemma star_update:
  forall R1 R2 s a1 a2 a,
  star a1 a2 = a ->
  star (update R1 s a1) (update R2 s a2) = update (star R1 R2) s a.
Proof.
  msadam; try solve [ do 2 rewrite length_set_nth in *; simpl in *; false; omega ].
  eapply zip_set_nth.
Qed.

(* As a particular case of [star_update], if the new value [a] can be
   split as [star a1 a2], where [R2] already maps [l] to [a2], then
   [R2] does not change. *)

Lemma star_update_identity_left:
  forall R1 R2 l a1 a2 a,
  maps R2 l a2 ->
  star a1 a2 = a ->
  star (update R1 l a1) R2 = update (star R1 R2) l a.
Proof.
  intros.
  erewrite <- star_update by eauto.
  f_equal.
  symmetry. eapply update_identity. assumption.
Qed.

Lemma star_update_identity_right:
  forall R1 R2 l a1 a2 a,
  maps R1 l a1 ->
  star a1 a2 = a ->
  star R1 (update R2 l a2) = update (star R1 R2) l a.
Proof.
  intros.
  erewrite <- star_update by eauto.
  f_equal.
  symmetry. eapply update_identity. assumption.
Qed.

(* As a particular case of [star_update_identity], if [R2] already maps [s] to
   the core of the new value [a], then [R2] does not change. *)

Lemma star_update_core_left:
  forall R1 R2 s a,
  maps R2 s (core a) ->
  star (update R1 s a) R2 = update (star R1 R2) s a.
Proof.
  eauto using star_update_identity_left, core_right_unit.
Qed.

Lemma star_update_core_right:
  forall R1 R2 s a,
  maps R1 s (core a) ->
  star R1 (update R2 s a) = update (star R1 R2) s a.
Proof.
  eauto using star_update_identity_right, core_left_unit.
Qed.

(* As a particular case of [star_update_core], if the old value [a1] is
   solitary, and if the new value [a2] has the same core as [a1], then
   [R2] does not change. *)

Lemma star_update_solitary_left:
  forall R1 R2 l a1 a2,
  maps R1 l a1 ->
  solitary a1 ->
  consistent (star R1 R2) ->
  core a1 = core a2 ->
  star (update R1 l a2) R2 =
  update (star R1 R2) l a2.
Proof.
  introv ? ? ? eq.
  (* By [star_update_core_left], it suffices to show that [R2] maps [l]
     to [core a2]. *)
  eapply star_update_core_left. simpl.
  (* By hypothesis, [core a2] is [core a1]. *)
  rewrite <- eq.
  (* And the fact that [R2] maps [l] to [core a1] follows from the fact
     that [R1] maps [l] to [a1] and [a1] is solitary. *)
  eauto using @maps_star_solitary_left.
Qed.

Lemma star_update_solitary_right:
  forall R1 R2 l a1 a2,
  maps R2 l a1 ->
  solitary a1 ->
  consistent (star R1 R2) ->
  core a1 = core a2 ->
  star R1 (update R2 l a2) =
  update (star R1 R2) l a2.
Proof.
  intros.
  forwards f: star_update_solitary_left.
    eauto.
    eauto.
    equates 1; eauto using @star_commutative.
    eauto.
  rewrite @star_commutative in f.
  rewrite f.
  f_equal. eapply star_commutative.
Qed.

(* The update of an existing address is permitted by [rely], provided
   the access right found at this address is updated in a manner that
   respects [rely], one level down. *)

Lemma rely_update:
  forall R s a1 a2,
  maps R s a1 ->
  rely a1 a2 ->
  rely R (update R s a2).
Proof.
  dependent_destruction. simpl. intros.
  eauto using growth_set_nth, rely_reflexive.
Qed.

(* Update at an existing address commutes with [rely]. *)

Lemma rely_update_update:
  forall R1 R2 s a a1 a2,
  maps R1 s a ->
  rely R1 R2 ->
  rely a1 a2 ->
  rely (update R1 s a1) (update R2 s a2).
Proof.
  dependent_destruction. msadam.
  eauto using growth_set_nth_set_nth.
Qed.

(* ------------------------------------------------------------------------- *)

(* Properties of [updatef]. *)

(* If [f] preserves consistency,then [updatef f] preserves consistency. *)

Lemma consistent_updatef:
  forall f,
  (forall a, consistent a -> consistent (f a)) ->
  forall R l,
  consistent R ->
  consistent (updatef R l f).
Proof.
  simpl. eauto using everywhere_updatef.
Qed.

(* If [f] preserves rely, then [updatef f] at an existing address preserves
   [rely]. *)

Lemma rely_updatef_updatef:
  forall f,
  (forall a1 a2, rely a1 a2 -> rely (f a1) (f a2)) ->
  forall R1 R2 s a,
  maps R1 s a ->
  rely R1 R2 ->
  rely (updatef R1 s f) (updatef R2 s f).
Proof.
  introv hf; dependent_destruction; msadam. eauto using growth_apply_nth_apply_nth.
Qed.

(* [star] commutes with [updatef]. *)

Lemma star_updatef:
  forall R1 R2 r f1 f2 f,
  (forall a1 a2, star (f1 a1) (f2 a2) = f (star a1 a2)) ->
  star (updatef R1 r f1) (updatef R2 r f2) = updatef (star R1 R2) r f.
Proof.
  msadam; do 2 rewrite length_apply_nth in *; redundancy;
  try solve [ simpl in *; false; omega ].
  (* zip_apply_nth *)
  generalize dependent r. induction_lists.
  eauto.
  destruct r; simpl; eauto with f_equal.
Qed.

(* If [f] commutes with [star], then [updatef f] commutes with [star]. *)

Lemma star_updatef_left_unconditional:
  forall f,
  (forall a1 a2, star (f a1) a2 = f (star a1 a2)) ->
  forall R1 R2 s,
  star (updatef R1 s f) R2 = updatef (star R1 R2) s f.
Proof.
  msadam; rewrite length_apply_nth in *; try congruence.
  redundancy. generalize dependent s.
  (* Let's prove this lemma about lists right here. *)
  induction_lists. 
  eauto.
  destruct s; simpl; eauto with f_equal.
Qed.

(* If [f] commutes with [star], then [updatef f] commutes with [star]. For
   greater applicability, we allow commutation to be subject to certain
   conditions. To begin with, we require consistency of the combination.
   Furthermore, we allow some knowledge [P] about [a1] to be exploited. *)

Lemma star_updatef_left:
  forall (P : A -> Prop) f, (
    forall a1 a2,
    consistent (star a1 a2) ->
    P a1 ->
    star (f a1) a2 = f (star a1 a2)
  ) ->
  forall R1 R2 s,
  consistent (star R1 R2) ->
  mapsp R1 s P ->
  star (updatef R1 s f) R2 = updatef (star R1 R2) s f.
Proof.
  msadam; rewrite length_apply_nth in *; try congruence.
  redundancy. mapsp_inversion. maps_inversion.
  generalize dependent s.
  (* Let's prove this lemma about lists right here. *)
  induction_lists. 
  eauto.
  is_nth_nonempty. eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If an evolution of elements is admissible, then allowing this evolution
   at a single address forms an admissible evolution of allocation maps. *)

Definition at_address l (evolution : A -> A -> Prop) (R1 R2 : aMap A) : Prop :=
  exists a1 a2,
  maps R1 l a1 /\
  R2 = update R1 l a2 /\
  evolution a1 a2.

Lemma admissible_at_address:
  forall evolution,
  admissible evolution ->
  forall l,
  admissible (at_address l evolution).
Proof.
  unfold admissible, at_address. introv hadm. intros. unpack. subst.
  forwards: maps_star; eauto. unpack.
  forwards: hadm. eauto. eauto. unpack.
  xplit; eauto using maps_star_construction, star_update, rely_update.
Qed.

(* ------------------------------------------------------------------------- *)

End Operations.

Hint Resolve @consistent_empty_map @consistent_allocate
@consistent_update @consistent_updatef @maps_consistent
@consistent_allocate : consistent.

Hint Resolve @mapsp_duplicablep : duplicable.

Hint Resolve @mapsp_maximalp : maximal.

Hint Resolve @mapsp_stablep : stable.

Hint Resolve admissible_at_address : admissible.

Section Agreement.

Context `{agrees : A -> B -> Prop}.

Lemma agrees_consistent:
  forall `{MSA B},
  (forall a b, agrees a b -> consistent b) ->
  forall R1 R2,
  aMap_agrees agrees R1 R2 ->
  consistent R2.
Proof.
  inversion 2; subst.
  unfold consistent; simpl.
  eauto using Forall2_Forall_modus_ponens.
Qed.

End Agreement.

Hint Resolve @agrees_consistent : consistent.

(* ------------------------------------------------------------------------- *)

(* A couple lemmas that help reasoning about equality of limits. *)

Lemma limit_limit_extend_right:
  forall A `{MSA B} (m1 : aMap A) (m2a m2b : aMap B),
  limit m1 = limit (star m2a m2b) ->
  consistent (star m2a m2b) ->
  limit m1 = limit m2a.
Proof.
  introv h ?. rewrite limit_split_left in h; eauto.
Qed.

Lemma limit_limit_extend_left:
  forall A `{MSA B} (m1 : aMap A) (m2a m2b : aMap B),
  limit m1 = limit (star m2b m2a) ->
  consistent (star m2a m2b) ->
  limit m1 = limit m2a.
Proof.
  introv h ?. rewrite limit_split_right in h. eauto. rewrite star_commutative. eauto.
Qed.

Lemma limit_limit_core:
  forall A `{MSA B} (m1 : aMap A) (m2 : aMap B),
  limit m1 = limit m2 ->
  limit m1 = limit (core m2).
Proof.
  intros. rewrite limit_core. assumption.
Qed.

Hint Resolve @agrees_limit @limit_limit_extend_left
@limit_limit_extend_right @limit_limit_core : limit_limit.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps use the [maps_maximal_*] lemmas in a situation
   where we have a hypothesis of the form [maps R _ a] and a goal of the form
   [maps (..... R .....) _ b]. The hints in [maximal] are supposed to prove
   that [a] is maximal. The hints in [maps_maximal_reflexivity_hints] are
   supposed to prove that [a] and [b] are equal. *)

Require Import MSATactics.

(* Create HintDb maximal. *) (* In MSA.v *)
Create HintDb maps_maximal_reflexivity_hints.

Ltac maps_maximal :=
  try eassumption; (* just in case *)
  instantiate;
  match goal with
  | h: maps ?R ?l ?a |- maps ?R' ?l ?b =>
      (* Force [a] and [b] to be the same element. If this fails, perhaps
         another hypothesis [h] will be tried; there might be several [maps]
         facts at hand. *)
      assert (a = b); [ solve [ eauto with maps_maximal_reflexivity_hints ] |
        first [
          (* First attempt. Try to prove that [R] and [R'] and equal. *)
          let eq := fresh in
          assert (eq: R' = R); [ aac_reflexivity | rewrite eq; eassumption ]
        | (* Second attempt. Try to prove that [R] appears within [R'].
             This is the same process as in the tactic [consistent_aac]. *)
          let haac := fresh in
          aac_match (fun x => star x R) R' haac;
          (* just [try], because [haac] could be a trivial equation *)
          try rewrite <- haac;
          eapply maps_maximal_right; [
            eexact h
          | try consistent_aac (* should work, but one never knows *)
          | eauto with maximal ]
        ]
      ]
  end.

Hint Extern 1 (maps (star _ _) _ _) => maps_maximal : maps_maximal.

(* Example. *)

Goal
  forall `{MSA A} (R : aMap A) l a,
  maps R l a ->
  maximal a ->
  exists b,
  forall R1 R2 R3,
  consistent (star (star R1 R) (star R2 R3)) ->
  maps (star (star R1 R) R2) l b /\ a = b.
Proof.
  eauto with maps_maximal.
Qed.

(* ------------------------------------------------------------------------- *)

(* This tactic offers another way of using the [maps_maximal] lemmas, in
   forward mode. The goal can be arbitrary. The tactic looks for a [maps]
   hypothesis that involves a resource [R1] and attempts to extend it to
   a one that involves a larger resource, [star R1 ?R2]. The resource [R2]
   is guessed by looking at the available consistency hypotheses. *)

Ltac hoist_maps_maximal :=
  repeat match goal with
  h: maps ?R1 _ _ |- _ =>
    first [
      let hh := fresh h in
      forwards hh: maps_maximal_left h; [
        (* This subgoal takes the form [consistent (star R1 ?R2)] where [R2]
           is unknown. We pick the first consistency hypothesis that we find
           (yes, this is crude), say [consistent R], and use [aac_guess] to
           try and prove that [R] is equal to [star R1 R2] for an appropriate
           choice of [R2]. *)
        equates 1; [ eassumption | aac_guess ]
      | solve [ eauto with maximal ]
      | (* Swap the two facts so that the newer one appears newer. *)
        revert hh; revert h
      ]
    | revert h
    ]
  end;
  intros.

(* ------------------------------------------------------------------------- *)

(* This tactic helps apply the lemma [maps_star_inversion_aatomic]. *)

Ltac maps_star_inversion_aatomic :=
  forwards: @maps_star_inversion_aatomic; [ eauto | eauto with aatomic | unpack ].

(* ------------------------------------------------------------------------- *)

(* This tactic pushes an [update] down into a [star]. *)

Ltac push_star_update_solitary_left :=
  erewrite <- @star_update_solitary_left; [
  | maps_maximal | eauto with solitary | eauto with consistent_split | eauto ].
Ltac push_star_update_solitary_right :=
  erewrite <- @star_update_solitary_right; [
  | maps_maximal | eauto with solitary | eauto with consistent_split | eauto ].
Ltac push_star_update_solitary :=
  repeat first [ push_star_update_solitary_left | push_star_update_solitary_right ].

(* This tactic pulls an [update] out of a [star]. *)

Ltac pull_star_update_solitary_left :=
  erewrite @star_update_solitary_left by eauto with solitary consistent_aac maps_maximal.
Ltac pull_star_update_solitary_right :=
  erewrite @star_update_solitary_right by eauto with solitary consistent_aac maps_maximal.
Ltac pull_star_update_solitary :=
  repeat first [ pull_star_update_solitary_left | pull_star_update_solitary_right ].

(* ------------------------------------------------------------------------- *)

(* The definition of the MSA of allocation maps should now be irrelevant. *)

Global Opaque msa_map.
