Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.

Section Definitions.

Variable A : Type.

(* ------------------------------------------------------------------------- *)

(* This a two-element monotonic separation algebra, where [BitZero] represents
   no access right, and [BitOne] represents a full access right. Accordingly,
   [BitZero] represents no knowledge, whereas [BitOne] represents full
   knowledge; this knowledge is represented by an element of type [A]. *)

Inductive bit :=
| BitZero: bit
| BitOne: A -> bit
| BitError: bit.

(* ------------------------------------------------------------------------- *)

(* A hammer. *)

Ltac destruct_bits :=
  intros;
  repeat match goal with R: bit |- _ => destruct R end;
  simpl in *;
  try solve [ tauto | congruence ].

Obligation Tactic := destruct_bits.

(* ------------------------------------------------------------------------- *)

(* The algebra. *)

Global Program Instance msa_bit : MSA bit := {

  star R1 R2 :=
    match R1, R2 return _ with
    | BitZero, R
    | R, BitZero =>
        (* [BitZero] is a unit. *)
        R
    | _, _ =>
        (* Every other combination is undefined. In particular, [star BitOne BitOne]
           is undefined. *)
        BitError
    end;

  consistent R :=
    R <> BitError;

  core R :=
    match R return _ with BitError => BitError | _ => BitZero end;

  (* The relation [rely] is equality: when a thread is inactive, the
     access rights that it holds do not change. *)

  rely R1 R2 :=
    R1 = R2

}.

(* Only one obligation is not automatically solved. *)

Next Obligation. (* rely_star *)
  do 2 eexists. splits; eauto. eauto.
  do 2 eexists. splits; eauto. eauto.
  do 2 eexists. splits; eauto. eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* [BitZero] is consistent and duplicable. *)

Lemma consistent_zero:
  consistent BitZero.
Proof.
  destruct_bits.
Qed.

Lemma duplicable_zero:
  duplicable BitZero.
Proof.
  destruct_bits.
Qed.

(* [BitOne _] is not duplicable. *)

Lemma not_duplicable_one:
  forall a,
  ~ duplicable (BitOne a).
Proof.
  destruct_bits.
Qed.

(* [BitOne a] is consistent and maximal. *)

Lemma consistent_one:
  forall a,
  consistent (BitOne a).
Proof.
  destruct_bits.
Qed.

Lemma maximal_one:
  forall a,
  maximal (BitOne a).
Proof.
  unfold maximal, maximalp. destruct_bits.
Qed.

(* Because [rely] is equality, every property and every element is stable. *)

Lemma stablep_bit:
  forall P,
  stablep P.
Proof.
  unfold stablep. destruct_bits.
Qed.

Lemma stable_bit:
  forall b,
  stable b.
Proof.
  unfold stable. eauto using stablep_bit.
Qed.

(* [BitZero] and [BitOne a] are atomic. *)

Lemma atomic_zero:
  atomic BitZero.
Proof.
  unfold atomic. destruct_bits.
Qed.

Lemma atomic_one:
  forall a,
  atomic (BitOne a).
Proof.
  unfold atomic. destruct_bits.
Qed.

(* [BitOne _] is solitary. *)

Lemma solitary_one:
  forall a,
  solitary (BitOne a).
Proof.
  eauto using atomic_one, maximal_one, @maximal_atomic_implies_solitary.
Qed.

(* ------------------------------------------------------------------------- *)

End Definitions.

Hint Resolve @duplicable_zero : duplicable aatomic.

Hint Resolve @consistent_zero @consistent_one : consistent.

Hint Resolve @maximal_one : maximal.

Hint Resolve @stablep_bit @stable_bit : stable.

Hint Resolve @atomic_zero @atomic_one : atomic aatomic.

Hint Resolve @solitary_one : solitary.

