Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.

(* ------------------------------------------------------------------------- *)

(* The product of two monotonic separation algebras forms a monotonic
   separation algebra. *)

Obligation Tactic := idtac.

Program Instance msa_pair `{MSA A, MSA B} : MSA (A * B) := {

  star R1 R2 :=
    let (RA1, RB1) := R1 in
    let (RA2, RB2) := R2 in
    (star RA1 RA2, star RB1 RB2);
  consistent R :=
    let (RA, RB) := R in
    consistent RA /\ consistent RB;
  core R :=
    let (RA, RB) := R in
    (core RA, core RB);
  rely R1 R2 :=
    let (RA1, RB1) := R1 in
    let (RA2, RB2) := R2 in
    rely RA1 RA2 /\ rely RB1 RB2

}.

Ltac decompose_pairs :=
  intros;
  repeat match goal with R: (_ * _)%type |- _ => destruct R end;
  repeat match goal with h: (_, _) = (_, _) |- _ => injection h; clear h; intros; try subst end;
  repeat match goal with h: _ /\ _ |- _ => destruct h end.

Next Obligation. (* star_commutative *)
  decompose_pairs. f_equal; eapply @star_commutative.
Qed.

Next Obligation. (* star_associative *)
  decompose_pairs. f_equal; eapply @star_associative.
Qed.

Next Obligation. (* consistent_split_left *)
  decompose_pairs. eauto using @consistent_split_left.
Qed.

Next Obligation. (* core_right_unit *)
  decompose_pairs. f_equal; eapply @core_right_unit.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  decompose_pairs. f_equal; eauto using @compatible_implies_common_core_left.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  decompose_pairs. f_equal; eauto using @duplicable_implies_self_core.
Qed.

Next Obligation. (* cores_are_duplicable *)
  decompose_pairs. f_equal; eauto using @cores_are_duplicable.
Qed.

Next Obligation. (* duplicable_split_left *)
  decompose_pairs. f_equal; eauto using @duplicable_split_left.
Qed.

Next Obligation. (* rely_reflexive *)
  decompose_pairs. eauto using @rely_reflexive.
Qed.

Next Obligation. (* rely_consistent *)
  decompose_pairs. eauto using @rely_consistent.
Qed.

Next Obligation. (* rely_star *)
  decompose_pairs.
  forwards [ RA2a [ RA2b ? ]]: @rely_star A; eauto.
  forwards [ RB2a [ RB2b ? ]]: @rely_star B; eauto.
  exists (RA2a, RB2a) (RA2b, RB2b).
  unpack. subst. tauto.
Qed.

Next Obligation. (* rely_core *)
  decompose_pairs. eauto using @rely_core.
Qed.

(* ------------------------------------------------------------------------- *)

(* Notations that help define predicates over pairs. *)

Section Predicates.

Variable A B : Type.

Definition respectively (P : A -> Prop) (Q : B -> Prop) (ab : A * B) : Prop :=
  let (a, b) := ab in
  P a /\ Q b.

Definition at_left (P : A -> Prop) (ab : A * B) : Prop :=
  let (a, b) := ab in
  P a.

Definition at_right (Q : B -> Prop) (ab : A * B) : Prop :=
  let (a, b) := ab in
  Q b.

End Predicates.

(* ------------------------------------------------------------------------- *)

(* Some properties. *)

Section Properties.

Context `{MSA A, MSA B}.

Lemma consistent_pair:
  forall (a : A) (b : B),
  consistent a ->
  consistent b ->
  consistent (a, b).
Proof.
  decompose_pairs. simpl in *. eauto with f_equal.
Qed.

Lemma duplicablep_respectively:
  forall (P : A -> Prop) (Q : B -> Prop),
  duplicablep P ->
  duplicablep Q ->
  duplicablep (respectively P Q).
Proof.
  unfold duplicablep, respectively. decompose_pairs. simpl in *. eauto.
Qed.

Lemma duplicablep_at_left:
  forall (P : A -> Prop),
  duplicablep P ->
  duplicablep (@at_left A B P).
Proof.
  unfold duplicablep, at_left. decompose_pairs. simpl in *. eauto.
Qed.

Lemma duplicablep_at_right:
  forall (Q : B -> Prop),
  duplicablep Q ->
  duplicablep (@at_right A B Q).
Proof.
  unfold duplicablep, at_right. decompose_pairs. simpl in *. eauto.
Qed.

Lemma duplicable_pair:
  forall (a : A) (b : B),
  duplicable a ->
  duplicable b ->
  duplicable (a, b).
Proof.
  decompose_pairs. simpl in *. eauto with f_equal.
Qed.

Lemma stablep_respectively:
  forall (P : A -> Prop) (Q : B -> Prop),
  stablep P ->
  stablep Q ->
  stablep (respectively P Q).
Proof.
  unfold stablep, respectively. decompose_pairs. simpl in *. unpack. eauto.
Qed.

Lemma stablep_at_left:
  forall (P : A -> Prop),
  stablep P ->
  stablep (@at_left A B P).
Proof.
  unfold stablep, at_left. decompose_pairs. simpl in *. unpack. eauto.
Qed.

Lemma stablep_at_right:
  forall (Q : B -> Prop),
  stablep Q ->
  stablep (@at_right A B Q).
Proof.
  unfold stablep, at_right. decompose_pairs. simpl in *. unpack. eauto.
Qed.

Lemma stable_pair:
  forall (a : A) (b : B),
  stable a ->
  stable b ->
  stable (a, b).
Proof.
  unfold stable, stablep. decompose_pairs. simpl in *. unpack. eauto with f_equal.
Qed.

Lemma maximalp_respectively:
  forall (P : A -> Prop) (Q : B -> Prop),
  maximalp P ->
  maximalp Q ->
  maximalp (respectively P Q).
Proof.
  unfold maximalp, respectively. decompose_pairs. simpl in *. unpack. injections. eauto.
Qed.

Lemma maximalp_at_left:
  forall (P : A -> Prop),
  maximalp P ->
  maximalp (@at_left A B P).
Proof.
  unfold maximalp, at_left. decompose_pairs. simpl in *. unpack. injections. eauto.
Qed.

Lemma maximalp_at_right:
  forall (Q : B -> Prop),
  maximalp Q ->
  maximalp (@at_right A B Q).
Proof.
  unfold maximalp, at_right. decompose_pairs. simpl in *. unpack. injections. eauto.
Qed.

Lemma maximal_pair:
  forall (a : A) (b : B),
  maximal a ->
  maximal b ->
  maximal (a, b).
Proof.
  unfold maximal, maximalp. decompose_pairs. simpl in *. unpack. injections. eauto with f_equal.
Qed.

Lemma aatomic_pair:
  forall (a : A) (b : B),
  aatomic a ->
  aatomic b ->
  aatomic (a, b).
Proof.
  introv ha hb. unfold aatomic. decompose_pairs. injections.
  forwards: ha; [ reflexivity | unpack ].
  forwards: hb; [ reflexivity | unpack ].
  split; congruence.
Qed.

Lemma solitary_pair:
  forall (a : A) (b : B),
  solitary a ->
  solitary b ->
  solitary (a, b).
Proof.
  unfold solitary. decompose_pairs. simpl in *. unpack. f_equal; eauto.
Qed.

Lemma narcissic_pair:
  forall (a : A) (b : B),
  narcissic a ->
  narcissic b ->
  narcissic (a, b).
Proof.
  unfold narcissic. decompose_pairs. simpl in *. unpack. f_equal; eauto.
Qed.

End Properties.

Hint Resolve @consistent_pair : consistent.

Hint Resolve @duplicablep_respectively @duplicablep_at_left @duplicablep_at_right @duplicable_pair : duplicable.

Hint Resolve @stablep_respectively @stablep_at_left @stablep_at_right @stable_pair : stable.

Hint Resolve @maximalp_respectively @maximalp_at_left @maximalp_at_right @maximal_pair : maximal.

Hint Resolve @aatomic_pair : aatomic.

Hint Resolve @solitary_pair : solitary.

Hint Resolve @narcissic_pair : narcissic.

(* ------------------------------------------------------------------------- *)

(* I would like to let the above instance declaration be transparent, but that
   would make it too fragile (we cannot always control exactly when [simpl] is
   invoked). So I make it opaque, but I define a set of rewriting rules that
   will allow simplification to be performed upon request. These rules can also
   be used to expand meta-variables. *)

Section Paraphrase.

Context `{MSA A, MSA B}.

Lemma msa_pair_star:
  forall (a1 a2 : A) (b1 b2 : B),
  star (a1, b1) (a2, b2) = (star a1 a2, star b1 b2).
Proof.
  reflexivity.
Qed.

Lemma msa_pair_core:
  forall (a : A) (b : B),
  core (a, b) = (core a, core b).
Proof.
  reflexivity.
Qed.

Lemma msa_pair_consistent:
  forall (a : A) (b : B),
  consistent (a, b) <-> consistent a /\ consistent b.
Proof.
  reflexivity. (* ?? *)
Qed.

Lemma msa_pair_rely:
  forall (a1 a2 : A) (b1 b2 : B),
  rely (a1, b1) (a2, b2) <-> rely a1 a2 /\ rely b1 b2.
Proof.
  reflexivity. (* ?? *)
Qed.

End Paraphrase.

(* Group these lemmas together as a rewrite database. *)

Hint Rewrite @msa_pair_star @msa_pair_core
@msa_pair_consistent @msa_pair_rely : msa_pair.

Ltac msa_pair :=
  autorewrite with msa_pair in *.
  (* Note: [repeat rewrite] can be more powerful than [autorewrite],
     as it will expand meta-variables. See also [explode] in file
     [MSAExplode.v]. *)

(* The definition can now be made opaque. *)

Global Opaque msa_pair.

(* ------------------------------------------------------------------------- *)

(* Examples. *)

Goal
  forall `{MSA A, MSA B} (a : A) (b : B),
  core (a, b) = (core a, core b).
Proof.
  intros.
  (* Simplification has no effect, because [msa_pair] is opaque. *)
  simpl.
  (* But expansion can be performed upon request. *)
  msa_pair.
  (* Note that, in this case, [reflexivity] would succeed directly. *)
  reflexivity.
Qed.

Goal
  forall `{MSA A, MSA B} (a : A) (b : B),
  consistent (a, b) -> consistent a.
Proof.
  intros.
  (* Simplification has no effect, because [msa_pair] is opaque. *)
  simpl in *.
  (* But expansion can be performed upon request. *)
  msa_pair.
  (* The goal can now be proved. *)
  tauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* Hints. *)

Hint Extern 1 (star _ _ = (_, _)) =>
  (rewrite @msa_pair_star; f_equal) : explode.

Hint Extern 1 (core _ = (_, _)) =>
  (rewrite @msa_pair_core; f_equal) : explode.

Hint Extern 1 (rely (_, _) _) =>
  (rewrite @msa_pair_rely; split) : explode.

Hint Extern 1 (consistent (_, _)) =>
  (rewrite @msa_pair_consistent; split) : explode.

(* ------------------------------------------------------------------------- *)

(* More hints. *)

(* Slightly random, but convenient. *)

Hint Unfold respectively at_left at_right : mapsp.

