Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.
Require Import AllocationMaps.
Require Import MSAAllocationMaps.

(* Let [A] be a monotonic separation algebra. Then, [aMap A], the type of the
   allocation maps over [A], is also a monotonic separation algebra. Let us
   think of them as heap fragments. *)

(* In general, a heap fragment is not ``round'': that is, it may contain
   pointers that point outside of itself, also known as dangling pointers. *)

(* Nevertheless, it is sometimes interesting to restrict one's attention to
   heap fragments that have no dangling edges, for a certain notion of
   ``edge''. Not every pointer needs be considered an edge; one is free to
   define a notion of ``edge'' that takes into account only certain pointers,
   such as the ``adopter'' pointers of Mezzo. *)
   
(* In this file, we prove that, under certain assumptions, ``round'' heap
   fragments form a monotonic separation algebra. *)

(* ---------------------------------------------------------------------------- *)

(* For the sake of modularity, we formulate the required assumptions in an
   abstract manner. We need two predicates, [present] and [pointsto], which
   respectively characterize which addresses are in the domain of a heap
   fragment and which edges leave these addresses. Each of these predicates
   must satisfy certain properties, which we group in a type class. *)

(* ---------------------------------------------------------------------------- *)

Class Present `{MSA A} := {

  (* The predicate [present] allows distinguishing a subset of ``present''
     elements of [A]. If [R] is an allocation map, we think of the addresses
     that are mapped to some ``present'' element as the ``domain'' of the heap
     fragment [R]. *)

  present : A -> Prop;

  (* If [x] is in the domain of [R1] or in the domain of [R2], then [x] must
     be in the domain of [R1 * R2]. *)

  (* Note that [x] could be in the domain of both [R1] and [R2], in which case
     the conjunction [R1 * R2] would be inconsistent. We do not wish to add
     the hypothesis [consistent (star a1 a2)] to the axiom below, because this
     would force us in turn to add a consistency hypothesis to the statement
     of the lemma [round_star], and [star] would no longer be a total function
     over round allocation maps, i.e. we would no longer be able to establish
     the property [star_compatible] of the type class [MSASig]. So, we omit
     the consistency hypothesis. In practice, this means that inconsistent
     elements must be considered ``present''. *)

  present_star_construction:
    forall a1 a2,
    (present a1 \/ present a2) ->
    present (star a1 a2);

  (* In order to prove that [rely] preserves roundness, we assume that no
     pointer can be created or removed in a thread's local state while this
     thread is passive. As far as [present] is concerned, this is expressed by
     the following requirement. *)

  (* An address cannot disappear while we are asleep. *)

  present_rely:
    forall a b,
    rely a b ->
    present a ->
    present b

}.

Class PresentMore `{M : MSA A, @Present A M} := {

  (* If two elements are present, then their [star] cannot be consistent.
     This means that two compatible heap fragments must have disjoint
     domains. *)

  present_exclusive:
    forall a b,
    present a ->
    present b ->
    consistent (star a b) ->
    False

}.

(* ---------------------------------------------------------------------------- *)

Class Pointsto `{MSA A} := {

  (* The predicate [pointsto] defines a relation between elements of [A] and
     addresses. We think of the assertion [pointsto a y] as the existence of a
     pointer from [a] to [y].  Thus, if the address [x] is mapped to some
     element [a] that points to the address [y], then there exists an edge in
     the heap from [x] to [y]. *)

  pointsto : A -> nat -> Prop;

  (* If there is an edge from [x] to [y] in [R1 * R2], then there must be such
     an edge either in [R1] or in [R2]. *)

  pointsto_star_inversion:
    forall a1 a2 y,
    pointsto (star a1 a2) y ->
    pointsto a1 y \/ pointsto a2 y;

  (* In order to prove that [rely] preserves roundness, we assume that no
     pointer can be created or removed in a thread's local state while this
     thread is passive. As far as [pointsto] is concerned, this is expressed
     by the following two requirements. *)

  (* An edge cannot appear at an existing address while we are asleep. *)

  pointsto_rely_reverse:
    forall a b y,
    rely a b ->
    pointsto b y ->
    pointsto a y;
    
  (* An edge cannot appear at a new address while we are asleep. *)

  (* Recall that our definition of [rely] for allocation maps requires that
     newly allocated addresses be mapped to duplicable elements (even though,
     as far as the theory of allocation maps is concerned, this requirement is
     not really useful). We are now requiring duplicable elements to carry no
     outgoing edges. The combined effect of these two points is that edges
     cannot appear at a new address while we are asleep, as desired. *)

  (* Furthermore, [pointsto_is_exclusive] is used also when proving that
     roundness is preserved by [core], and there, it seems to be pretty
     much a necessary condition. *)

  pointsto_is_exclusive:
    forall a y,
    duplicable a ->
    pointsto a y ->
    False

}.

Class PointstoMore `{M : MSA A, @Pointsto A M} := {

  pointsto_star:
    forall a1 a2 y,
    pointsto a1 y ->
    consistent (star a1 a2) ->
    pointsto (star a1 a2) y

}.

(* ---------------------------------------------------------------------------- *)

(* We now define roundness and prove that it is preserved by [star], [core],
   and [rely]. *)

Section Roundness.

  Context A `{M : MSA A, P : @Present A M, Q : @Pointsto A M,
              @PresentMore A M P, @PointstoMore A M Q}.

  (* In order to prove that [core] preserves roundness, it seems that we need
     to require either that [core] throws away all edges (in which case it may
     choose to keep or throw away the vertices, as it wishes) or that [core]
     preserves all vertices (in which case it may choose to keep or throw away
     the edges, as it wishes). We choose the first requirement, which
     corresponds to our intended application, where [core] throws away all
     vertices and all edges (the vertices are the exclusive objects and the
     edges are given by the adopter fields). We have already made a stronger
     requirement anyway: [pointsto_is_exclusive] implies that [core] throws
     away all edges. *)

  Lemma not_pointsto_core:
    forall a y,
    pointsto (core a) y ->
    False.
  Proof.
    intros.
    eapply pointsto_is_exclusive; [ | eassumption ].
    eapply cores_are_duplicable.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* An address [y] is in the domain of [R] if and only if [R] maps [y] to
     some present element. *)

  Definition domain R y :=
    exists b,
    maps R y b /\ present b.

  (* The limit is not part of the domain. *)

  Lemma domain_limit:
    forall R y,
    domain R y ->
    y <> limit R.
  Proof.
    unfold domain. intros. intro. subst. unpack.
    forwards: maps_limit. eauto. omega.
  Qed.

  (* If [y] is in the domain of [R1], then it is also in the domain of
     [R1 * R2], provided [R1] and [R2] agree on the allocation limit. *)

  Lemma domain_star_left:
    forall R1 R2 y,
    domain R1 y ->
    limit R1 = limit R2 ->
    domain (star R1 R2) y.
  Proof.
    unfold domain. intros. unpack.
    forwards: maps_star_extension; [ eauto | eauto | unpack ].
    eauto using present_star_construction.
  Qed.

  Lemma domain_star_right:
    forall R1 R2 y,
    domain R2 y ->
    limit R1 = limit R2 ->
    domain (star R1 R2) y.
  Proof.
    intros. rewrite star_commutative. eauto using domain_star_left.
  Qed.

  (* If [R1 * R2] has nonempty domain, then [R1] and [R2] have a common
     limit. *)

  Lemma domain_star_limit:
    forall R1 R2 y,
    domain (star R1 R2) y ->
    limit R1 = limit R2.
  Proof.
    unfold domain. intros. unpack. eauto using maps_star_limit.
  Qed.

  (* [domain] is preserved by [rely]. *)

  Lemma domain_rely:
    forall R1 R2 y,
    domain R1 y ->
    rely R1 R2 ->
    domain R2 y.
  Proof.
    unfold domain. intros. unpack.
    forwards: maps_rely; [ eauto | eauto | unpack ].
    solve [ eauto using present_rely ].
  Qed.

  (* Two compatible heap fragments have disjoint domains. This fact relies
     on the axiom [present_exclusive]. *)

  Lemma disjoint_domains:
    forall R1 R2 y,
    domain R1 y ->
    domain R2 y ->
    consistent (star R1 R2) ->
    False.
  Proof.
    unfold domain. intros. unpack.
    eauto using present_exclusive, maps_consistent, maps_star_construction.
  Qed.

  (* The domain of a heap fragment is enlarged by allocation. *)

  Lemma domain_allocate_elsewhere:
    forall R y a,
    domain R y ->
    domain (allocate R a) y.
  Proof.
    unfold domain. intros. unpack. eauto using maps_allocate_elsewhere.
  Qed.

  Lemma domain_allocate_there:
    forall R a,
    consistent R ->
    present a ->
    domain (allocate R a) (limit R).
  Proof.
    eexists; eauto using maps_allocate_there.
  Qed.

  (* The domain of a heap fragment is preserved (or increased) by an update,
     provided the new value is ``more present'' than the previous value. *)

  Lemma domain_updatef:
    forall f,
    (forall x, present x -> present (f x)) ->
    forall R l y,
    domain R y ->
    domain (updatef R l f) y.
  Proof.
    unfold domain. intros. unpack.
    destruct (eq_nat_dec y l); [ subst y | ];
    eauto using maps_updatef_there, maps_updatef_elsewhere.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* The predicate [into R P] means that every edge whose origin lies in [R]
     has some destination that satisfies the predicate [P]. *)

  Definition into R (P : nat -> Prop) :=
    forall x a y,
    maps R x a ->
    pointsto a y ->
    P y.

  (* [into] is contravariant in its first argument, in the following sense.
     This property relies on the axiom [pointsto_star]. *)

  Lemma into_contravariant:
    forall R1 R2 P,
    into (star R1 R2) P ->
    consistent (star R1 R2) ->
    into R1 P.
  Proof.
    unfold into. intros.
    forwards: maps_star; [ eauto | eauto | unpack ].
    eauto using maps_star_construction, pointsto_star.
  Qed.

  (* [into] is also contravariant in its first argument in the following
     sense. [into] is preserved by an update of its first argument, provided
     the update is restricted to removing edges and/or adding edges into [P].
     Thus, the proof obligation is that, if an edge exists after the update,
     then either it already existed before the update, or its destination
     satisfies [P]. *)

  Lemma into_updatef:
    forall f P,
    (forall x y, pointsto (f x) y -> (pointsto x y \/ P y)) ->
    forall R l,
    into R P ->
    into (updatef R l f) P.
  Proof.
    unfold into. introv hpointsto ? ? ?.
    forwards [ ? | ? ]: maps_updatef_inversion; eauto.
    { unpack. eauto. }
    { unfold mapsp in *. unpack. subst.
      forwards hdisj: hpointsto; eauto.
      destruct hdisj; eauto. }
  Qed.

  (* [into] is covariant in its second argument. *)

  Lemma into_covariant:
    forall R (P Q : nat -> Prop),
    into R P ->
    (forall y, P y -> Q y) ->
    into R Q.
  Proof.
    unfold into. eauto.
  Qed.

  (* [into] is preserved by [star] in the following sense. *)

  Lemma into_star:
    forall R1 R2 P,
    into R1 P ->
    into R2 P ->
    into (star R1 R2) P.
  Proof.
    (* Assume there is an edge from [x] to [y] in [R1 * R2]. *)
    unfold into; intros.
    (* Then, there must be such an edge in [R1] or in [R2]. *)
    forwards: maps_star_inversion; [ eauto | unpack; subst ].
    forwards [ ? | ? ]: pointsto_star_inversion;
    (* The two sub-cases are symmetric. *)
    eauto.
  Qed.

  (* [into] is trivially satisfied by any core. *)

  Lemma into_core:
    forall R P,
    into (core R) P.
  Proof.
    (* Assume there is an edge from [x] to [y] in [core R]. *)
    unfold into. intros.
    (* This is impossible! *)
    forwards: maps_core_reverse; [ eauto | unpack; subst ].
    false. eauto using not_pointsto_core.
  Qed.

  (* [into] is preserved by [rely]. *)

  Lemma into_rely:
    forall R1 R2 P,
    into R1 P ->
    rely R1 R2 ->
    into R2 P.
  Proof.
    (* Assume there is an edge from [x] to [y], where [x] lies in [R2]. *)
    unfold into; introv hi; intros.
    (* Either [x] was already present in [R1], or it is a newly allocated address. *)
    forwards h: maps_rely_reverse; [ eauto | eauto | ].
    destruct h; unpack.
    (* Sub-case: [x] was already present in [R1]. *)
    (* Then, the edge from [x] to [y] was already present in [R1]. *)
    forwards: pointsto_rely_reverse; [ eauto | eauto | ].
    (* There follows that [y] lies in [R']. *)
    solve [ eauto ].
    (* Sub_case: [x] is new. *)
    (* By definition of [rely], the element associated with [x] in [R2]
       must be duplicable. Thus, there cannot be an edge from [x] to [y]
       in [R2]. This case is impossible. *)
    false.
    eapply pointsto_is_exclusive; eauto.
  Qed.

  (* [into] is preserved by the allocation of a new vertex, provided this
     new vertex does not carry an edge, or carries an edge into [P]. *)

  Lemma into_allocate:
    forall R P a,
    into R P ->
    (forall y, pointsto a y -> P y) ->
    into (allocate R a) P.
  Proof.
    unfold into. intros.
    forwards [ ? | ? ]: maps_allocate_inversion; [ eauto | | ].
    eauto.
    subst. eauto.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* A heap fragment [R] to be round if and only if [into R (domain R)] holds,
     i.e., there are no dangling pointers. *)

  Definition round R :=
    into R (domain R).

  (* Roundness is preserved by conjunction. *)

  (* Intuitively speaking, this is obvious: if there are no dangling pointers
     in [R1] or in [R2], then there are no dangling pointers in [R1 * R2]. *)

  Lemma round_star:
    forall R1 R2,
    round R1 ->
    round R2 ->
    round (star R1 R2).
  Proof.
    (* Assume there is an edge from [x] to [y] in [R1 * R2]. *)
    unfold round, into; intros.
    (* Then, there must be such an edge in [R1] or in [R2]. *)
    forwards: maps_star_inversion; [ eauto | unpack; subst ].
    forwards [ ? | ? ]: pointsto_star_inversion;
    (* The two sub-cases are symmetric. There is a little miracle:
       the fact that [R1 * R2] maps [x] to something implies that
       [R1] and [R2] agree on the allocation limit. *)
    eauto using domain_star_left, domain_star_right, maps_star_limit.
  Qed.

  (* Roundness is preserved by [core]. In fact, we do not even need to
     assume that [R] is round in order to establish that [core R] is
     round. This is because [core R] has no outgoing edges at all. *)

  Lemma round_core:
    forall R,
    round (core R).
  Proof.
    unfold round. eauto using into_core.
  Qed.

  (* Roundness is preserved by [rely]. *)

  Lemma round_rely:
    forall R1 R2,
    round R1 ->
    rely R1 R2 ->
    round R2.
  Proof.
    unfold round. eauto using into_rely, into_covariant, domain_rely.
  Qed.

  (* Because roundness is compatible with [star], [core], and [rely], the type
     of round allocation maps forms a monotonic separation algebra. *)

  Require Import MSASig.

  Obligation Tactic := eauto using round_star, round_core, round_rely.
  Global Program Instance msasig_round: @MSASig (aMap A) _ round.

  (* An empty map is round. *)

  Lemma round_empty:
    round (empty_map _).
  Proof.
    unfold round, into. intros. false. eauto using maps_empty.
  Qed.

  (* Roundness is compatible with allocation. We prove a special case
     where the newly created vertex does not carry an edge, i.e., the
     the element [a] associated with the newly allocated adress does
     not point to anything. *)

  Lemma round_allocate:
    forall R a,
    round R ->
    (forall y, ~ pointsto a y) ->
    round (allocate R a).
  Proof.
    unfold not, round, into. intros.
    (* Assume there is an edge from [x] to [y] after allocation.
       Then, either [x] is an old address, or it is the new address. *)
    forwards [ | ]: maps_allocate_inversion; [ eauto | | ]; intros.
    (* In the first case, the result is immediate. *)
    eauto using domain_allocate_elsewhere.
    (* The second case is impossible. *)
    subst. false. eauto.
  Qed.

  (* Roundness is preserved by update, provided the new value is ``more
     present'' than the previous value and ``less pointsto'' than the
     previous value. These conditions are inherited from the lemmas
     [domain_updatef] and [into_updatef]. The new value can also be
     ``more pointsto'' than the previous value, provided the new edges
     that are installed are into [R]. *)

  Lemma round_updatef:
    forall f R,
    (forall x, present x -> present (f x)) ->
    (forall x y, pointsto (f x) y -> (pointsto x y \/ domain R y)) ->
    forall l,
    round R ->
    round (updatef R l f).
  Proof.
    unfold round. introv ? hpointsto ?.
    eapply into_updatef.
    { intros. forwards hdisj: hpointsto; eauto.
      destruct hdisj; eauto using domain_updatef. }
    { eauto using into_covariant, domain_updatef. }
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* A memory location [y] has no antecedents (with respect to the heap
     fragment [R]) if it is in the domain of [R] and no memory location
     in the domain of [R] points to it. *)

  Definition no_antecedents R y :=
    domain R y /\
    into R (fun z => z <> y).

  (* [no_antecedents] is preserved by [rely]. *)

  Lemma no_antecedents_rely:
    forall R1 y,
    no_antecedents R1 y ->
    forall R2,
    rely R1 R2 ->
    no_antecedents R2 y.
  Proof.
    unfold no_antecedents. intros. unpack.
    eauto using domain_rely, into_rely.
  Qed.

  (* [no_antecedents] is preserved by [star]. That is, if [y] has no
     antecedents in [R1], and if [R2] is compatible with [R1] and is
     round, then [y] has no antecedents in [R1 * R2]. *)

  Lemma no_antecedents_star:
    forall R1 y,
    no_antecedents R1 y ->
    forall R2,
    consistent (star R1 R2) ->
    round R2 ->
    no_antecedents (star R1 R2) y.
  Proof.
    unfold no_antecedents. intros. unpack. split.
    { eauto using domain_star_left, consistent_star_limit. }
    (* The goal is to show that no edge whose origin lies in [R1 * R2]
       has [y] as its destination. Thanks to the lemma [into_star], it
       suffices to check [R1] and [R2] separately, and by hypothesis,
       this is true of [R1], so there remains to check the edges whose
       origin lies in [R2]. *)
    eapply into_star; [ eauto | ].
    (* Now, because [R2] is round, every edge whose origin lies in [R2]
       has some vertex in [R2] as its destination. Thus, the goal boils
       down to checking that [y] is not in the domain of [R2]. *)
    eapply into_covariant; [ eassumption | intro z ].
    (* Assume, by way of contradiction, that [y] is in the domain of [R2].
       [y] is also in the domain of [R1] -- this comes from the definition
       of [no_antecedents R1 y] -- so it is in the domain of both, and this
       contradicts the fact that [R1 * R2] is consistent. *)
    do 2 intro; subst z.
    eauto using disjoint_domains.
  Qed.

  (* The reverse property holds as well. *)

  Lemma no_antecedents_star_reverse:
    forall R1 R2 y,
    no_antecedents (star R1 R2) y ->
    consistent (star R1 R2) ->
    domain R1 y ->
    no_antecedents R1 y.
  Proof.
    unfold no_antecedents. intros. unpack. eauto using into_contravariant.
  Qed.

  (* Provided we start off with a round heap fragment, and provided the
     new vertex does not point to itself, a newly allocated vertex has
     no antecedents. *)

  Lemma no_antecedents_allocate:
    forall R a,
    consistent R ->
    round R ->
    present a ->
    (forall y, pointsto a y -> y <> limit R) ->
    no_antecedents (allocate R a) (limit R).
  Proof.
    split.
    eauto using domain_allocate_there.
    eauto using into_allocate, into_covariant, domain_limit.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* The predicate [edge R x y] means that [R] contains (owns) an edge from [x]
     to [y]. *)

  Definition edge R x y :=
    mapsp R x (fun a => pointsto a y).

  (* If [star R1 R2] has an edge out of [x], then at least one of [R1] and [R2]
     has this edge. *)

  Lemma edge_star_inversion:
    forall R1 R2 x y,
    edge (star R1 R2) x y ->
    edge R1 x y \/ edge R2 x y.
  Proof.
    unfold edge, mapsp. intros. unpack.
    forwards: maps_star_inversion; [ eauto | unpack; subst ].
    forwards c: pointsto_star_inversion. eauto.
    destruct c; eauto.
  Qed.

  (* If [R] has an edge of [x] to [y] and is round, then [y] is in the domain
     of [R]. *)

  Lemma edge_round_domain:
    forall R x y,
    edge R x y ->
    round R ->
    domain R y.
  Proof.
    (* This is by definition of roundness. *)
    unfold edge, mapsp. intros. unpack. eauto.
  Qed.

  (* As a corollary, if [star R1 R2] has an edge out of [x] whose destination
     [y] lies in [R1], and if [R2] is round, then [x] cannot lie in the domain
     of [R2], hence must lie in the domain of [R1]. Thus, [R1] has this edge. *)

  Lemma edge_narrow_left:
    forall R1 R2 x y,
    edge (star R1 R2) x y ->
    domain R1 y ->
    round R2 ->
    consistent (star R1 R2) ->
    edge R1 x y.
  Proof.
    intros.
    forwards c: edge_star_inversion. eauto. destruct c.
    (* Case: [R1] contains an edge of [x] to [y]. *)
    { assumption. }
    (* Case: [R2] contains this edge. Contradiction. *)
    { false. eauto using disjoint_domains, edge_round_domain. }
  Qed.

End Roundness.

Hint Resolve @round_star @round_core @round_allocate : round.

