Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MyProofIrrelevance.
Require Import MSA.

(* ------------------------------------------------------------------------- *)

(* This auxiliary type class defines what it means for a property [P] to be
   compatible with a monotonic separation algebra. *)

Class MSASig `{MSA A} (P : A -> Prop) := {

  (* [P] must be preserved by the function [star]. *)

  star_compatible:
    forall a1 a2,
    P a1 ->
    P a2 ->
    P (star a1 a2);

  (* [P] must be preserved by the function [core]. *)

  core_compatible:
    forall a,
    P a ->
    P (core a);

  (* [P] must be preserved by the relation [rely]. *)

  rely_compatible:
    forall a1 a2,
    rely a1 a2 ->
    P a1 ->
    P a2

}.

(* ------------------------------------------------------------------------- *)

(* If the property [P] is compatible with the monotonic separation algebra
   structure of the type [A], then the subset type [sig P], which can also
   be written [{ a : A | P a }], forms a monotonic separation algebra. The
   behavior of this sub-algebra is the same as that of the mother algebra,
   but its carrier set is restricted to the elements that satisfy [P]. *)

Program Instance msa_sig
  {A : Type} {m : MSA A} {P : A -> Prop} {p : @MSASig A m P}
  : MSA (sig P) := {

  star a1 a2 :=
    let (a1, _) := a1 in
    let (a2, _) := a2 in
    star a1 a2;

  consistent a :=
    let (a, _) := a in
    consistent a;

  core a :=
    let (a, _) := a in
    core a;

  rely a1 a2 :=
    let (a1, _) := a1 in
    let (a2, _) := a2 in
    rely a1 a2

}.

Ltac decompose_sig :=
  repeat match goal with R: sig _ |- _ => destruct R end;
  exist_exist_hypotheses;
  exist_exist_goal.

Next Obligation. (* Proof obligation in the definition of [star] *)
  eapply star_compatible; eauto.
Qed.

Next Obligation. (* star_commutative *)
  decompose_sig. eauto using star_commutative.
Qed.

Next Obligation. (* star_associative *)
  decompose_sig. eauto using star_associative.
Qed.

Next Obligation. (* consistent_split_left *)
  decompose_sig. eauto using consistent_split_left.
Qed.

Next Obligation. (* Proof obligation in the definition of [core] *)
  eapply core_compatible; eauto.
Qed.

Next Obligation. (* core_right_unit *)
  decompose_sig. eauto using core_right_unit.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  decompose_sig. eauto using compatible_implies_common_core_left.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  decompose_sig. eauto using duplicable_implies_self_core.
Qed.

Next Obligation. (* cores_are_duplicable *)
  decompose_sig. eauto using cores_are_duplicable.
Qed.

Next Obligation. (* duplicable_split_left *)
  decompose_sig. eauto using duplicable_split_left.
Qed.

Next Obligation. (* rely_reflexive *)
  decompose_sig. eauto using rely_reflexive.
Qed.

Next Obligation. (* rely_consistent *)
  decompose_sig. eauto using rely_consistent.
Qed.

Next Obligation. (* rely_star *)
  decompose_sig.
  (* Apply the underlying [rely_star]. *)
  rely_star.
  (* Because the two resource fragments arise via [rely], they must
     satisfy the property [P]. This is where [rely_compatible] is
     needed, even though it was not needed in the definition of the
     new [rely] relation. *)
  match goal with h: rely _ (star ?R2a ?R2b) |- _ =>
    assert (PR2a: P R2a); [ eapply rely_compatible; eauto |
    assert (PR2b: P R2b); [ eapply rely_compatible; eauto |
  ]] end.
  exists (exist _ _ PR2a).
  exists (exist _ _ PR2b).
  splits; decompose_sig; eauto.
Qed.

Next Obligation. (* rely_core *)
  decompose_sig. eauto using rely_core.
Qed.

(* ------------------------------------------------------------------------- *)

(* See the example of use in [MSASigExample]. *)

(* ------------------------------------------------------------------------- *)

(* A set of rewriting rules that allows simplification to be performed upon
   request. These rules can also be used to expand meta-variables. *)

Section Paraphrase.

Context {A : Type} {m : MSA A} {P : A -> Prop} {p : @MSASig A m P}.

Lemma msa_sig_star:
  forall a1 a2 p1 p2,
  star (exist _ a1 p1) (exist _ a2 p2) =
  exist _ (star a1 a2) (star_compatible _ _ p1 p2).
Proof.
  intros. simpl. exist_exist_goal. reflexivity.
Qed.

Lemma msa_sig_core:
  forall a p,
  core (exist _ a p) =
  exist _ (core a) (core_compatible _ p).
Proof.
  intros. simpl. exist_exist_goal. reflexivity.
Qed.

Lemma msa_sig_consistent:
  forall a p,
  consistent (exist _ a p) <->
  consistent a.
Proof.
  reflexivity. (* ?? *)
Qed.

Lemma msa_sig_rely:
  forall a1 a2 p1 p2,
  rely (exist _ a1 p1) (exist _ a2 p2) <->
  rely a1 a2.
Proof.
  reflexivity. (* ?? *)
Qed.

End Paraphrase.

(* Group these lemmas together as a rewrite database. *)

Hint Rewrite @msa_sig_star @msa_sig_core
@msa_sig_consistent @msa_sig_rely : msa_sig.

Ltac msa_sig :=
  autorewrite with msa_sig in *.
  (* Note: [repeat rewrite] can be more powerful than [autorewrite],
     as it will expand meta-variables. See also [explode] in file
     [MSAExplode.v]. *)

(* ------------------------------------------------------------------------- *)

(* Hints. *)

Hint Extern 1 (star _ _ = exist _ _ _) =>
  (rewrite @msa_sig_star; exist_exist_goal) : explode.

Hint Extern 1 (core _ = exist _ _ _) =>
  (rewrite @msa_sig_core; exist_exist_goal) : explode.

Hint Extern 1 (rely (exist _ _ _) _) =>
  (rewrite @msa_sig_rely) : explode.

Hint Extern 1 (consistent (exist _ _ _)) =>
  (rewrite @msa_sig_consistent) : explode.

(* ------------------------------------------------------------------------- *)

Global Opaque msa_sig.

