Set Implicit Arguments.
Require Import Relation_Operators.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import GetSet.
Require Import Syntax.
Require Import WellKindedness.
Require Import AllocationMaps.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.

(* ---------------------------------------------------------------------------- *)

(* The small-step, call-by-value reduction semantics. *)

Inductive red : configuration -> configuration -> Prop :=

  (* The usual [beta-v]. Reduction is permitted only if the argument is a
     value. The distinction between values and terms is performed by the
     well-kindedness judgement. *)

  | RedBeta:
      forall s t v t',
      wk zilch v VALUE ->
      subst v 0 t = t' ->
      red
        (Cf s (TApp (TAbs t) v))
        (Cf s t')

  (* [fail] eats the evaluation context away. *)

  | RedContextFail:
      forall s E t,
      eval_context E ->
      fill E TFail = t ->
      red
        (Cf s t)
        (Cf s TFail)

  (* Reduction under a context. *)

  | RedContext:
      forall s s' E t t' u u',
      eval_context E ->
      red (Cf s t) (Cf s' t') ->
      fill E t = u ->
      fill E t' = u' ->
      red (Cf s u) (Cf s' u')

  (* Reduction in a thread pool. *)

  | RedThread:
      forall s s' t t',
      red (Cf s t) (Cf s' t') ->
      red (Cf s (TThread t)) (Cf s' (TThread t'))

  | RedParLeft:
      forall s s' t1 t1' t2,
      red (Cf s t1) (Cf s' t1') ->
      red (Cf s (TPar t1 t2)) (Cf s' (TPar t1' t2))

  | RedParRight:
      forall s s' t1 t2 t2',
      red (Cf s t2) (Cf s' t2') ->
      red (Cf s (TPar t1 t2)) (Cf s' (TPar t1 t2'))

  (* Spawning a new thread returns [TUnit] in the original thread
     and creates a new thread beside it. *)

  | RedSpawn:
      forall s D v1 v2,
      wk zilch v1 VALUE ->
      wk zilch v2 VALUE ->
      deep_eval_context D ->
      red
        (Cf s (TThread (fill D (TSpawn v1 v2))))
        (Cf s (TPar (TThread (fill D TUnit)) (TThread (TApp v1 v2))))

  (* A ghost instruction reduces in one step to a unit value. *)

  | RedGhost:
      forall s,
      red
        (Cf s TGhost)
        (Cf s TUnit)

  (* Allocation of a new memory block. The new block's adopter pointer is
     initially null. *)

  | RedNewRef:
      forall s s' h h' v,
      h = get s ->
      h' = allocate h (Block Null v) ->
      s' = set h' s ->
      red
        (Cf s (TNewRef v))
        (Cf s' (TLoc (limit h)))

  (* Reading a field in a memory block. *)

  | RedRead:
      forall s h l p v,
      h = get s ->
      maps h l (Block p v) ->
      red
        (Cf s (TRead (TLoc l)))
        (Cf s v)

  (* Writing a field in a memory block. This operation produces a unit result. *)

  | RedWrite:
      forall s s' h h' l p v v',
      h = get s ->
      maps h l (Block p v) ->
      h' = update h l (Block p v') ->
      s' = set h' s ->
      red
        (Cf s (TWrite (TLoc l) v'))
        (Cf s' TUnit)

  (* Allocation of a new lock. *)

  | RedNewLock:
      forall s s' lckh lckh',
      lckh = get s ->
      lckh' = allocate lckh LOCKED ->
      s' = set lckh' s ->
      red
        (Cf s TNewLock)
        (Cf s' (TLock (limit lckh)))

  (* Acquisition of a lock. *)

  | RedAcquire:
      forall s s' lckh lckh' lck,
      lckh = get s ->
      maps lckh lck UNLOCKED ->
      lckh' = update lckh lck LOCKED ->
      s' = set lckh' s ->
      red
        (Cf s (TAcquire (TLock lck)))
        (Cf s' TUnit)

  (* Release of a lock. *)

  | RedRelease:
      forall s s' lckh lckh' lck,
      lckh = get s ->
      maps lckh lck LOCKED ->
      lckh' = update lckh lck UNLOCKED ->
      s' = set lckh' s ->
      red
        (Cf s (TRelease (TLock lck)))
        (Cf s' TUnit)

  (* Adoption. *)

  | RedGive:
      forall s s' h h' x y p v,
      h = get s ->
      maps h x (Block p v) ->
      h' = update h x (Block (NonNull y) v) ->
      s' = set h' s ->
      red
        (Cf s (TGive (TLoc x) (TLoc y)))
        (Cf s' TUnit)

  (* Abandon is described by three reduction rules. *)

  (* Abandon 1. If the adopter pointer at address [x] contains the
     address [y], then the runtime test is considered successful,
     and one transitions to an intermediate state where this [take]
     operation is considered committed, but the adopter pointer has
     not been updated yet. *)

  | RedTakeSuccess:
      forall s h x y v,
      h = get s ->
      maps h x (Block (NonNull y) v) ->
      red
        (Cf s (TTake          (TLoc x) (TLoc y)))
        (Cf s (TTakeCommitted (TLoc x) (TLoc y)))

  (* Abandon 2. A [take] operation that has been committed is completed
     by writing [Null] into the adopter field at address [x]. *)

  | RedTakeCommitted:
      forall s s' h h' x y p v,
      h = get s ->
      maps h x (Block p v) ->
      h' = update h x (Block Null v) ->
      s' = set h' s ->
      red
        (Cf s (TTakeCommitted (TLoc x) (TLoc y)))
        (Cf s' TUnit)

  (* Abandon 3. If the runtime test fails, then one transitions to [TFail]. *)

  | RedTakeFailure:
      forall s h x y p v,
      h = get s ->
      maps h x (Block p v) ->
      p <> NonNull y ->
      red
        (Cf s (TTake (TLoc x) (TLoc y)))
        (Cf s TFail)

.

Hint Constructors red : red.

(* ---------------------------------------------------------------------------- *)

(* Basic properties of the reduction relation. *)

(* Reduction preserves well-kindedness. *)

(* For a long time, this property was not explicitly used; it was in part
   implicitly re-proved as part of the subject reduction statement. It is
   now explicitly exploited in the subject reduction proof, in the case of
   [JForallIntro]. Anyway, it is good to check up front that it holds. *)

Lemma red_wk:
  forall s1 t1 s2 t2,
  red (Cf s1 t1) (Cf s2 t2) ->
  forall kappa,
  wk zilch t1 kappa ->
  wks s1 ->
  wkc (Cf s2 t2) kappa.
Proof.
  introv hred; dependent induction hred; subst;
  try destruct_eval_context;
  introv hwk; dependent induction hwk; wk_inversion;
  try solve [
    (* Try exploiting the induction hypothesis, when there is one. *)
    forwards: IHhred; [ eauto | eauto | wkc_inversion; eauto with wk ]
  | (* Try a direct proof, in the easy base cases. *)
    eauto 14 using wk_substitution, wk_deep_eval_context with wk
  | (* In the cases that deal with the heap, do this. *)
    destruct_states; simpl; wks_inversion; wkh_inversion;
    eauto 10 using everywhere_allocate, everywhere_update with wk
  ].
Qed.

Lemma red_wkc:
  forall c1 c2,
  red c1 c2 ->
  forall kappa,
  wkc c1 kappa ->
  wkc c2 kappa.
Proof.
  destruct c1; destruct c2;
  introv hred hwkc; dependent destruction hwkc;
  eauto using red_wk with wk.
Qed.

(* Values are irreducible. *)

Lemma red_value:
  forall K s1 t1 c2,
  red (Cf s1 t1) c2 ->
  wk K t1 VALUE ->
  False.
Proof.
  introv hred; dependent induction hred; intros; wk_inversion;
  eauto using eval_context_value.
Qed.

(* [TFail] does not reduce. *)

Lemma red_fail:
  forall s1 c2,
  red (Cf s1 TFail) c2 ->
  False.
Proof.
  dependent_destruction; false; eauto using fill_fail.
Qed.

(* [TSpawn] does not reduce, not by itself. It reduces when nested
   (deep) within [TThread]; see the reduction rule [RedSpawn]. *)

Lemma red_spawn:
  forall s1 v1 v2 c2,
  red (Cf s1 (TSpawn v1 v2)) c2 ->
  False.
Proof.
  dependent_destruction; false; eauto using fill_spawn.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The reflexive-transitive closure of the reduction relation. *)

Notation redstar :=
  (clos_refl_trans_1n _ red).

Hint Constructors clos_refl_trans_1n : redstar.

(* ---------------------------------------------------------------------------- *)

(* This inversion tactic is useful when only one reduction rule is really
   applicable, but a little analysis is required to show that [RedSpawn]
   and [RedContext] are not. *)

Ltac red_inversion :=
  match goal with h: red _ _ |- _ =>
    revert h
  end;
  dependent_destruction;
  try (destruct_eval_context; discriminate).

(* ---------------------------------------------------------------------------- *)

(* The conjunction of [red] and [wkc] is sometimes used for conciseness. *)

Inductive redwkc : configuration -> configuration -> kind -> Prop :=
| RedWkc:
    forall c1 c2 kappa,
    wkc c1 kappa ->
    red c1 c2 ->
    redwkc c1 c2 kappa.

Ltac redwkc_inversion :=
  match goal with h: redwkc _ _ _ |- _ =>
    inversion h; clear h
  end;
  red_inversion;
  wkc_inversion;
  wk_inversion.

