Set Implicit Arguments.
Require Import MyTactics.
Require Export GetSet.
Require Export MSATactics.
Require Export MSAProduct.
Require Export MSASig.
Require Export MSAExplode.
Require Export HeapResources.
Require Export AdoptionResources.
Require Export LockResources.

(* ---------------------------------------------------------------------------- *)

(* A resource is a partial and/or instrumented machine state. *)

(* Just like the type of machine states, the type of resources should be
   thought of as open-ended, or partly unspecified. Some parts of the type
   system care care about only one component of the resource, while many parts
   do not care at all. *)

(* A resource is currently a triple of a heap resource (an instrumented heap),
   an adoption resource (another instrumented heap, whose allocation limit is
   the same as that of the heap resource), and a lock resource (an
   instrumented lock heap). *)

Notation resource :=
  (heap_resource * adoption_resource * lock_resource)%type.

(* ---------------------------------------------------------------------------- *)

(* Resources form a monotonic separation algebra. *)

Goal MSA resource.
Proof.
  eauto 10 with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [simplify_resources] performs simplification after resources have been
   exploded by one of the tactics below. *)

Ltac simplify_resources :=
  (* Because we have made [msa_pair] and [msa_sig] opaque, [simpl] cannot
     simplify away applications of the [MSAProduct] and [MSASig] operations;
     we must instead use [msa_pair]. We still use [simpl] to deal with other
     instances.  We add [injections] into the mix, because simplification may
     allow discovering equalities, which in turn (once substituted) may allow
     further simplification. *)
  repeat progress (msa_pair; msa_sig; simpl in *; injections);
  (* We might end up with some conjunctions in the context; decompose
     them. *)
  unpack.

(* ---------------------------------------------------------------------------- *)

(* [explode_resources] is supposed to explode a resource (which is a tuple)
   into its components, and perform simplification, so that all uses of [get]
   and [set] disappear. *)

Ltac explode_resources :=
  (* Destruct all resources. *)
  repeat match goal with R: resource |- _ =>
    let Rh := fresh R in
    let Ra := fresh R in
    let Rl := fresh R in
    destruct R as [ [ Rh [ Ra ? ] ] Rl ]
  end;
  (* Simplify. *)
  simplify_resources.

(* ---------------------------------------------------------------------------- *)

(* We also define less aggressive variants of [explode_resources], which
   perform just enough destruction to expose one component of interest, say,
   the block heap, or the lock heap. Avoiding unnecessary destruction reduces
   clutter and reduces the size of the proof terms. Avoiding the destruction
   and reconstruction of an object of type [sig P] is particularly beneficial,
   since reconstruction carries a proof obligation. *)

Ltac explode_resources_blocks :=
  repeat match goal with R: resource |- _ =>
    let Rh := fresh R in
    let Ra := fresh R in
    let Rl := fresh R in
    destruct R as [ [ Rh Ra ] Rl ]
  end;
  simplify_resources.

Ltac explode_resources_adoption_strong :=
  explode_resources. (* explodes the [sig] type *)

Ltac explode_resources_adoption_weak :=
  explode_resources_blocks. (* does not explode the [sig] type *)

Ltac explode_resources_locks :=
  (* One could choose not to destruct the pair [Rh, Ra], but this
     actually creates problems down the road, in [Agreement]. It
     seems preferable to always destruct the toplevel tuple. *)
  explode_resources_blocks.

(* ---------------------------------------------------------------------------- *)

(* With suitable hints, [eauto] can use [explode_resources] or its variants at
   appropriate times. *)

Hint Extern 1 (maps (get _) _ _) =>
  explode_resources_blocks : explode_resources.

Hint Extern 1 (rely (get _) _) =>
  explode_resources : explode_resources.

Hint Extern 1 (invariant (get _) _ _) =>
  explode_resources_locks : explode_resources.

Hint Extern 1 (taken (get _) _) =>
  explode_resources_locks : explode_resources.

Hint Extern 1 (adoptable (get _) _) =>
  explode_resources_adoption_strong : explode_resources.

Hint Extern 1 (unadopted (get _) _) =>
  explode_resources_adoption_strong : explode_resources.

Hint Extern 1 (adopts (get _) _) =>
  explode_resources_adoption_strong : explode_resources.

Hint Extern 1 (adopted (get _) _ _) =>
  explode_resources_adoption_strong : explode_resources.

(* ---------------------------------------------------------------------------- *)

(* We require an empty resource, which is used to type-check source programs.
   The empty resource must be consistent and duplicable. *)

Definition void : resource :=
  (heap_void, adoption_void, lock_void).

Lemma consistent_void:
  consistent void.
Proof.
  unfold void. explode; eauto with consistent.
Qed.

Lemma core_void:
  core void = void.
Proof.
  unfold void. explode.
Qed.

Lemma duplicable_void:
  duplicable void.
Proof.
  rewrite <- core_void. eapply cores_are_duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This lemma can be occasionally useful. *)

Lemma round_get:
  forall R : resource,
  round (get R).
Proof.
  intros. explode_resources_adoption_strong. assumption.
Qed.

Hint Resolve round_get : round.

