Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WellKindedness.

(* ---------------------------------------------------------------------------- *)

(* The predicate [row_length n Ts] asserts that the length of the row [Ts]
   is [n]. *)

Inductive row_length: nat -> term -> Prop :=
| RowLengthZero:
    row_length 0 RowNil
| RowLengthSucc:
    forall n T Ts,
    row_length n Ts ->
    row_length (S n) (RowCons T Ts).

Hint Constructors row_length : row_length.

Ltac select_row_length action :=
  match goal with h: row_length _ _ |- _ => action h end.
Ltac row_length_inversion :=
  gji select_row_length.

(* It is compatible with weakening and substitution. *)

Lemma row_length_weakening:
  forall n Ts,
  row_length n Ts ->
  forall x,
  row_length n (shift x Ts).
Proof.
  induction 1; intros; simpl_lift_goal; econstructor; eauto.
Qed.

Hint Resolve row_length_weakening : j_weakening.

Lemma row_length_substitution:
  forall n Ts,
  row_length n Ts ->
  forall U x,
  row_length n (subst U x Ts).
Proof.
  induction 1; intros; simpl_subst_goal; econstructor; eauto.
Qed.

Hint Resolve row_length_substitution : j_substitution.

(* It is really a partial function. *)

Lemma row_length_deterministic:
  forall n1 Ts,
  row_length n1 Ts ->
  forall n2,
  row_length n2 Ts ->
  n1 = n2.
Proof.
  induction 1; inversion 1; subst; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A notion of a row context with one hole helps read and write rows. *)

(* We represent a context with one hole as a type that exhibits exactly one
   occurrence of the variable 0. The operation of hole filling is then just
   substitution. *)

Definition fill_row vs v :=
  subst v 0 vs.

(* The index [f] represents the depth of the context, and can be thought of
   intuitively as a field number. *)

Inductive row_context : nat -> term -> Prop :=
| RowContextZero:
    forall Ts Ts',
    shift 0 Ts = Ts' ->
    row_context 0 (RowCons (TVar 0) Ts')
| RowContextSucc:
    forall f T T' Ts,
    shift 0 T = T' ->
    row_context f Ts ->
    row_context (S f) (RowCons T' Ts).

(* This notion is compatible with weakening and substitution. *)

Lemma row_context_weakening:
  forall f Ts,
  row_context f Ts ->
  forall x Ts',
  shift (1 + x) Ts = Ts' ->
  row_context f Ts'.
Proof.
  induction 1; intros; subst; simpl_lift_goal; econstructor; eauto with lift_lift.
Qed.

Hint Resolve row_context_weakening : j_weakening.

Lemma row_context_substitution:
  forall f Ts,
  row_context f Ts ->
  forall T T' x Ts',
  shift 0 T = T' ->
  subst T' (1 + x) Ts = Ts' ->
  row_context f Ts'.
Proof.
  induction 1; intros; subst; simpl_subst_goal; econstructor; eauto with lift_subst.
Qed.

Hint Resolve row_context_substitution : j_substitution.

(* If an application of a context [vs] to a term [v] is is well-kinded, then
   [v] and [vs] are well-kinded. *)

Lemma wk_row_context_inverse_hole:
  forall f Ts,
  row_context f Ts ->
  forall K kappa T,
  wk K (subst T 0 Ts) (ROW kappa) ->
  wk K T kappa.
Proof.
  induction 1; introv h; simpl_subst_all; dependent destruction h; eauto.
Qed.

Lemma wk_row_context_inverse_rest:
  forall f Ts,
  row_context f Ts ->
  forall K kappa T,
  wk K (subst T 0 Ts) (ROW kappa) ->
  wk (insert 0 kappa K) Ts (ROW kappa).
Proof.
  induction 1; introv h; subst; simpl_subst_all; dependent destruction h;
  match goal with h: wk _ _ _ |- _ => rewrite subst_lift in h end;
  eauto using wk_weakening with wk.
Qed.

Hint Resolve wk_row_context_inverse_hole : wk_inversion.

(* If [vs1[f:=v1]] and [vs2[f:=v2]] coincide, then [v1] and [v2] coincide
   and [vs1] and [vs2] coincide. *)

Lemma row_context_equality_inversion:
  forall f vs1,
  row_context f vs1 ->
  forall vs2 v1 v2,
  row_context f vs2 ->
  subst v1 0 vs1 = subst v2 0 vs2 ->
  v1 = v2 /\ vs1 = vs2.
Proof.
  induction 1; inversion 1; introv h; subst; simpl_subst_all;
  repeat rewrite subst_lift in h; injections.
  (* Zero *)
  eauto.
  (* Succ *)
  forwards: IHrow_context; [ eauto | eauto | unpack; subst ].
  eauto.
Qed.

Ltac row_context_equality_inversion :=
  match goal with h: subst _ 0 _ = subst _ 0 _ |- _ =>
    forwards: (>> row_context_equality_inversion h); [
      solve [ eauto ]
    | solve [ eauto ]
    | clear h; unpack; subst
    ]
  end.

(* ---------------------------------------------------------------------------- *)

(* If [vs] is a row of values and [Ts] is a row of types, then [zip vs Ts] is
   the iterated conjunction [vs @ Ts]. If [vs] and [Ts] have different lengths,
   then, by definition, this permission is unsatisfiable. *)

(* We could make [zip] a relation, as opposed to a function, with the benefit
   that we would not need to define what happens when [vs] and [Ts] are not
   closed-ended rows of matching length. However, a functional notation for
   [vs @ Ts] seems more natural than a relational one. *)

Fixpoint zip (vs : term) (Ts : term) : term :=
  match vs, Ts with
  | RowNil, RowNil =>
      TyEmpty
  | RowCons v vs, RowCons T Ts =>
      TyStar (TyAnchored v T) (zip vs Ts)
  | _, _ =>
      TyFalse
  end.

(* [zip] is preserved by lifting. *)

Lemma zip_weakening:
  forall vs Ts x,
  zip (shift x vs) (shift x Ts) = shift x (zip vs Ts).
Proof.
  induction vs; try solve [ reflexivity ];
  destruct Ts; try solve [ reflexivity ].
  intros. simpl. simpl_lift_goal. simpl. eauto with f_equal.
Qed.

Hint Resolve zip_weakening : j_weakening.

(* [zip] is preserved by substitution, except in the pathological case
   where one of the rows ends with a row variable. We avoid this case
   by requiring both rows to be closed-ended. We build this condition
   into the typing rule [JNew], where the lemma [zip_substitution] is
   used. *)

Lemma zip_substitution:
  forall n vs,
  row_length n vs ->
  forall Ts w x,
  row_length n Ts ->
  zip (subst w x vs) (subst w x Ts) = subst w x (zip vs Ts).
Proof.
  induction 1; inversion 1; subst; simpl; simpl_subst_goal; simpl; eauto with f_equal.
Qed.

Hint Resolve zip_substitution : j_substitution.

