Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WellKindedness.
Require Import Sugar.
Require Import Resources.
Require Import Dup.
Require Import Typing.
Require Import JBasic.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JWeakening.

(* We show how to translate the simply-typed lambda-calculus into Mezzo. *)

(* This is not really difficult. It is a sanity check: it shows that
   we did not forget some fundamental typing rule in the definition
   of Mezzo's type system. *)

(* ---------------------------------------------------------------------------- *)

(* The syntax of the lambda-calculus. *)

Inductive lambda :=
  | LVar: nat -> lambda
  | LAbs: lambda -> lambda
  | LApp: lambda -> lambda -> lambda.

(* ---------------------------------------------------------------------------- *)

(* The types of the simply-typed lambda-calculus. *)

(* These types do not contain any variables. They have no binding structure. *)

Inductive ty :=
  | STyIota: ty
  | STyArrow: ty -> ty -> ty.

(* ---------------------------------------------------------------------------- *)

(* The typing judgement of the simply-typed lambda-calculus. *)

Inductive st : env ty -> lambda -> ty -> Prop :=
  | STVar:
      forall E x T,
      lookup x E = Some T ->
      st E (LVar x) T
  | STAbs:
      forall E t T1 T2,
      st (insert 0 T1 E) t T2 ->
      st E (LAbs t) (STyArrow T1 T2)
  | STApp:
      forall E t1 t2 T1 T2,
      st E t1 (STyArrow T1 T2) ->
      st E t2 T1 ->
      st E (LApp t1 t2) T2.

(* ---------------------------------------------------------------------------- *)

(* Translating the terms, types, and judgements of the simply-typed
   lambda-calculus into Mezzo. *)

(* The translation of terms introduces a [let]-binding for the left-hand
   side of every application. *)

Fixpoint tr_term (t : lambda) : term :=
  match t with
  | LVar x =>
      TVar x
  | LAbs t =>
      TAbs (tr_term t)
  | LApp t u =>
      TLet (tr_term t) (TApp (TVar 0) (shift 0 (tr_term u)))
  end.

(* The translation of types is trivial. *)

Fixpoint tr_type (T : ty) : term :=
  match T with
  | STyIota =>
      TyTop (* for instance *)
  | STyArrow T1 T2 =>
      TyArrow (tr_type T1) (tr_type T2)
  end.

(* ---------------------------------------------------------------------------- *)

(* An entry in the typing environment, [x : T], is translated to an entry of the
   form [x : VALUE] together with a permission of the form [x @ T]. *)

Definition tr_env (E : env ty) : env kind :=
  (map (fun _ => VALUE) E).

Definition tr_perm (E : env ty) : term :=
  fold (fun (oT : option ty) P =>
    match oT with
    | None =>
        shift 0 P
    | Some T =>
        intro P (tr_type T)
    end
  ) E TyEmpty.

(* ---------------------------------------------------------------------------- *)

(* The following facts repeat the definitions. *)

Lemma tr_env_insert:
  forall E T,
  tr_env (insert 0 T E) = insert 0 VALUE (tr_env E).
Proof.
  intros. unfold tr_env. rewrite map_insert. reflexivity.
Qed.

Lemma tr_env_raw_insert:
  forall E,
  tr_env (raw_insert 0 None E) = raw_insert 0 None (tr_env E).
Proof.
  intros. unfold tr_env. rewrite map_raw_insert. reflexivity.
Qed.

Lemma tr_perm_insert:
  forall E T,
  tr_perm (insert 0 T E) = intro (tr_perm E) (tr_type T).
Proof.
  intros. unfold tr_perm. rewrite fold_insert. reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The translation of terms produces well-formed terms. *)

Lemma wk_tr_term:
  forall E t T,
  st E t T ->
  forall K,
  tr_env E = K ->
  wk K (tr_term t) TERM.
Proof.
  induction 1; intros; subst; simpl.
  (* STVar *)
  do 2 econstructor. eapply lookup_map_some with (f := fun _ => VALUE). eauto.
  (* STAbs *)
  eauto with wk.
  (* STApp *)
  eauto 8 using wk_weakening with wk.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The translation of types produces well-formed, closed, duplicable types. *)

Lemma wk_tr_type:
  forall T K,
  wk K (tr_type T) TYPE.
Proof.
  induction T; intros; simpl; eauto with wk.
Qed.

Lemma closed_tr_type:
  forall T,
  closed 0 (tr_type T).
Proof.
  eauto using wk_closed, wk_tr_type.
Qed.

Lemma duplicable_tr_type:
  forall T,
  dup (tr_type T).
Proof.
  induction T; simpl; eauto using dup_TyTop with dup.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The translation of environments produces well-formed, duplicable permissions. *)

Lemma wk_tr_perm:
  forall E,
  wk (tr_env E) (tr_perm E) PERM.
Proof.
  eapply fold_invariant with (P := fun E P =>
    wk (tr_env E) P PERM
  );
  unfold tr_env;
  intros;
  try match goal with o: option _ |- _ => destruct o end.
  (* Base. *)
  rewrite map_empty.
  solve [ eauto with wk ].
  (* Step/Some. *)
  rewrite map_insert.
  solve [ eauto 6 using wk_weakening, wk_tr_type with wk ].
  (* Step/None. *)
  rewrite map_raw_insert.
  solve [ eauto using wk_weakening ].
Qed.

Lemma duplicable_tr_perm:
  forall E,
  dup (tr_perm E).
Proof.
  eapply fold_invariant with (P := fun E P => dup P);
  intros;
  try match goal with o: option _ |- _ => destruct o end;
  unfold intro;
  eauto using dup_weakening, duplicable_tr_type with dup.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The following fact is used at variables: if the simple type environment [E]
   maps [x] to the simple type [T], then the translation of [E], a permission,
   entails the permission [x @ T]. *)

Lemma tr_perm_lookup:
  forall E x T,
  lookup x E = Some T ->
  exists n,
  jps n (tr_env E)
    (tr_perm E)
    (TyAnchored (TVar x) (tr_type T)).
Proof.
  intro E.
  eapply fold_invariant with (P := fun E P =>
    forall x T,
    lookup x E = Some T ->
    exists n,
    jps n (tr_env E) P (TyAnchored (TVar x) (tr_type T))
  ).
  (* Base. *)
  intros. false. eauto using lookup_empty_Some.
  (* Step/Some. *)
  intros [ | ]; introv ih h;
  (destruct x; [ | replace (S x) with (1 + x) in * by omega ]);
  lookup_insert_all;
  first [ rewrite tr_env_insert | rewrite tr_env_raw_insert ].
    (* Step/Some/[x = 0]. *)
    exists 0.
    unfold intro.
    erewrite closed_tr_type.
    solve [ eauto with j ].
    (* Step/Some/[x = 1 + ...]. *)
    forwards: ih. eauto. unpack. clear ih.
    eexists.
    jps_left JpsDisappearsRight.
    eapply JpsTransitive; [ | eapply JpsReflexiveEq ].
      eapply j_weakening; eauto.
      simpl_lift_goal. f_equal. eapply closed_tr_type.
    (* Step/None/[x=0]. *)
    congruence.
    (* Step/None[x = 1 + ...]. *)
    forwards: ih. eauto. unpack. clear ih.
    eexists.
    eapply j_weakening; eauto.
    simpl_lift_goal. f_equal. eapply closed_tr_type.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the simple typing judgement [st E t T] holds, then its encoding in Mezzo
   holds as well. *)

Lemma stlc_encoding:
  forall E t T,
  st E t T ->
  exists n,
  jt n void (tr_env E) (tr_perm E) (tr_term t) (tr_type T).
Proof.
  induction 1; simpl.
  (* STVar *)
  (* The key ingredients are the lemma [tr_perm_lookup] and the derived typing
     rule [JValue]. *)
  forwards: tr_perm_lookup. eassumption. unpack.
  eexists.
  eapply JSubLeft.
    eapply JValue.
      econstructor. eapply lookup_map_some with (f := fun _ => VALUE). eassumption.
      eauto using wk_tr_type.
    eassumption.
  (* STAbs *)
  (* The key ingredient is the lemma [duplicable_tr_perm], which allows the
     permission that encodes the typing environment to be accessed within the
     function body. *)
  unpack.
  eexists.
  eapply OldJFunction; [
    eauto using wk_tr_perm
  | eauto using wk_tr_type
  | eauto using duplicable_tr_perm
  | ].
  rewrite core_void.
  erewrite <- tr_env_insert.
  fold (intro (tr_perm E) (tr_type T1)). rewrite <- tr_perm_insert.
  erewrite closed_tr_type.
  eassumption.
  (* STApp *)
  (* The key ingredients are the derived typing rule [JNormalApp], which offers
     a simple way of type-checking applications, and the lemma [duplicable_tr_perm],
     which allows the permission that encodes the typing environment to be made
     available within each of the two branches. *)
  unpack.
  forwards: dup_complete_corollary; [ eapply wk_tr_perm | eapply duplicable_tr_perm | unpack ].
  eexists.
  eapply JSubLeft.
    eapply JNormalApp; eauto using wk_tr_perm, wk_tr_term.
    eassumption. (* from [dup_complete_corollary] *)
Qed.

