Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JTop.
Require Import JDecomposition.
Require Import SubsumptionSoundness.

(* ---------------------------------------------------------------------------- *)

(* This lemma is at the heart of subject reduction for [spawn]. It states that,
   if [D[spawn v1 v2]] is well-typed with respect to a resource [R], then, with
   respect to the same resource, separately, [D[()]] and [v1 v2] are well-typed. *)

(* This lemma is not really specific of [spawn], and could be turned into a more
   general statement that allows separating a term and its evaluation context as
   well as plugging a new term in that evaluation context. *)

Lemma decompose_fill_spawn_preliminary:
  forall n R1 K P D v1 v2 T,
  jt n R1 K P (fill D (TSpawn v1 v2)) T ->
  deep_eval_context D ->
  wk zilch v1 VALUE ->
  wk zilch v2 VALUE ->
  forall R2 R,
  jpi R2 K P ->
  star R1 R2 = R ->
  consistent R ->
  exists n1 n2 R'1 R'2 U,
  jt n1 R'1 K TyEmpty (fill D TUnit) T /\
  jt n2 R'2 K TyEmpty (TApp v1 v2) U /\
  star R'1 R'2 = R.
Proof.
  (* Well-founded induction over the height of the type derivation. *)
  intro n. induction_wf ih: lt_wf n.
  (* Introduce the remaining hypotheses. *)
  introv hjt; intros; subst.
  (* Reason by cases on the last typing rule that was used. *)
  dependent destruction hjt; subst;
  (* Remove some impossible cases. *)
  try solve [
    inconsistent_fill
  | forwards: deep_eval_context_value; [ eassumption | eassumption | wk_inversion ]
  ];
  decomposition.

  (* JForallIntro *)
  (* The terms [fill D (TSpawn v1 v2)] and [fill D TUnit] are closed. *)
  match goal with h: harmless (fill ?D (TSpawn ?v1 ?v2)) |- _ =>
    assert (wk zilch (fill D (TSpawn v1 v2)) TERM /\
            wk zilch (fill D TUnit) TERM); [
      split; eauto 6 using wk_fill_deep_eval_context with wk
    | unpack; closed ]
  end.
  (* Apply the induction hypothesis, and see what we have to prove. *)
  forwards: ih; eauto. unpack. xplit; [ | | eassumption ].
  (* We now have two separate threads, each of which expects a type
     variable [TVar 0] of kind [kappa] to exist. We wrap each of
     them independently in an instance of [JForallIntro]. *)
  eapply JForallIntroVariant.
    eauto using harmless_replace with harmless.
    simpl_lift_goal. closed. eassumption.
    eassumption.
  eapply JForallIntroVariant.
    eauto using harmless_fill_inverse, harmless_TSpawn_TApp.
    simpl_lift_goal. closed. eassumption.
    eassumption.

  (* JExistsElim. *)
  substitution. solve [ eauto ].

  (* JFrame *)
  (* Apply the induction hypothesis. *)
  forwards: ih; eauto with consistent_aac. unpack. xplit.
  (* Construct the goals. *)
  eapply JtTopCut; solve [ eauto with j jxtra ].
  eassumption.
  aac_congruence.

  (* JCut *)
  solve [ eauto with jpi aac ].

  (* JSubLeft *)
  solve [ eauto using jps_sound with consistent_aac ].

  (* JSubRight *)
  forwards: ih; eauto. unpack. xplit; solve [ eauto with j ].

  (* JApplication *)
  (* The evaluation context cannot be empty, so it must be [E[D[]]]. *)
  match goal with h: deep_eval_context _ |- _ => destruct h end.
  simpl_subst_all. discriminate.
  subst. rewrite fill_fill in * by eauto.
  (* And [E] must be an application context. *)
  destruct_eval_context. injections.
  (* We may now apply the induction hypothesis. *)
  forwards: ih; eauto with consistent_aac. unpack. xplit.
  (* And construct the goals. *)
  solve [ eauto 3 with j ].
  solve [ eauto ].
  aac_congruence.

  (* JSpawn *)
  (* The evaluation context must be empty. *)
  match goal with h: deep_eval_context _ |- _ => destruct h end; [
    simpl_subst_all; injections
  | subst; rewrite fill_fill in * by eauto; destruct_eval_context; discriminate ].
  (* Because [TSpawn v1 v2] is well-typed, [TApp v1 v2] is well-typed too. *)
  forwards: JcTopApp; eauto. unpack.
  (* Construct the goals. *)
  xplit.
  eapply JTop; eauto with wk.
  eapply JtTopCut; solve [ eauto with j ].
  eapply core_left_unit.

Qed.

(* ---------------------------------------------------------------------------- *)

(* A slightly simpler statement is obtained by taking [R2] to be [core R1] and
   taking [P] to be [TyEmpty]. *)

Lemma decompose_fill_spawn:
  forall n R K D v1 v2 T,
  jt n R K TyEmpty (fill D (TSpawn v1 v2)) T ->
  deep_eval_context D ->
  wk zilch v1 VALUE ->
  wk zilch v2 VALUE ->
  consistent R ->
  exists n1 n2 R'1 R'2 U,
  jt n1 R'1 K TyEmpty (fill D TUnit) T /\
  jt n2 R'2 K TyEmpty (TApp v1 v2) U /\
  star R'1 R'2 = R.
Proof.
  intros.
  forwards: decompose_fill_spawn_preliminary; eauto with jpi.
    rewrite @core_right_unit. eassumption.
  rewrite @core_right_unit in *.
  assumption.
Qed.

