Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JTop.
Require Import Agreement.

(* ---------------------------------------------------------------------------- *)

(* In preparation for the statement of subject reduction, we define a
   predicate [sr_target R1f K c2 T]. *)

(* This predicate appears as the conclusion of the preliminary subject
   reduction statement. *)

(* To a first approximation, this judgement means that the configuration [c2]
   has type [T] under environment [K]. *)

(* More precisely, this configuration is well-typed under some resource [R2]
   that corresponds exactly to the state [s2], but not all of [R2] is required
   to type-check the term [t2]. The resource [R2] can be split as [star R2t
   R2f]. The component [R2t] is sufficient to type-check the term [t2], while
   the component [R2f] is unused, but we guarantee that this component is an
   evolution of [R1f] along [rely]. *)

(* Thus, in short, the judgement [sr_target R1f K c2 T] means ``[c2] has type
   [T], and the evolution that is imposed to the outside world, whose initial
   state was [R1f], is acceptable. *)

(* The term [t2] is closed, but a non-empty environment [K] is needed because
   the type [T] is not necessarily closed. We do not have a value restriction,
   so reduction can take place under [JForallIntro]. *)

Inductive sr_target : resource -> env kind -> configuration -> term -> Prop :=
| SrTarget:
    forall R1f K s2 t2 T,
    forall n R2 R2f R2w,
    jsr s2 R2w ->
    star R2 R2f = R2w ->
    rely R1f R2f ->
    jt n R2 K TyEmpty t2 T ->
    sr_target R1f K (Cf s2 t2) T.

Hint Constructors sr_target : sr_target.

(* ---------------------------------------------------------------------------- *)

(* This variant of the constructor [SrTarget] uses a canonical derivation. *)

Lemma SrTargetCanonical:
  forall R1f K s2 t2 T,
  forall R2 R2f R2w,
  jsr s2 R2w ->
  star R2 R2f = R2w ->
  rely R1f R2f ->
  jctop R2 K t2 T ->
  sr_target R1f K (Cf s2 t2) T.
Proof.
  intros.
  forwards: jctop_jt. eauto. unpack.
  econstructor; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The constructor [SrTarget] is used as follows when there is no side effect,
   i.e., no evolution is imposed to the outside world. *)

Ltac sr_target_pure :=
  eapply SrTarget; eauto using @rely_reflexive.

(* ---------------------------------------------------------------------------- *)

(* This lemma helps establish an [sr_target] assertion by specifying an evolution
   of the resource that is admissible and is known to preserve [jsr]. *)

Lemma SrTargetAccordCanonical:
  forall change : state -> state -> Prop,
  forall evolution : resource -> resource -> Prop,
  admissible evolution ->
  accord jsr change evolution ->
  forall s1 s2 R1 R1f R2 K t2 T,
  jsr s1 (star R1 R1f) ->
  change s1 s2 ->
  evolution R1 R2 ->
  jctop R2 K t2 T ->
  sr_target R1f K (Cf s2 t2) T.
Proof.
  introv hadm. intros. subst. jsr_consistent.
  forwards: hadm; eauto. unpack.
  forwards: jctop_jt. eauto. unpack.
  econstructor; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This lemma justifies reduction under [JForallIntro]. *)

Lemma SrTargetForallIntro:
  forall R1f kappa K s2 t2 T,
  sr_target R1f (insert 0 kappa K) (Cf s2 t2) T ->
  wk zilch t2 TERM ->
  harmless t2 ->
  substitutable kappa ->
  sr_target R1f K (Cf s2 t2) (TyForall kappa T).
Proof.
  inversion 1; subst. econstructor; eauto.
  eapply JForallIntroVariant; eauto.
  simpl_lift_goal. closed. eassumption.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This lemma justifies reduction under [JFrame]. *)

Lemma SrTargetFrame:
  forall R1 R2 K Q c2 T,
  (* If [R1] justifies [Q], *)
  jpi R1 K Q ->
  (* and if the outside world includes [R1], *)
  sr_target (star R1 R2) K c2 T ->
  consistent (star R1 R2) ->
  (* then we are able to frame [Q] back in. *)
  sr_target R2 K c2 (TyMixStar T Q).
Proof.
  inversion 2; intros; subst.
  (* We are imposing an evolution along [rely] on the frame, which is a
     composite resource, [star R1 R2]. Thus, [rely_star] applies. *)
  rely_star.
  (* There remains to check that everything fits together. *)
  econstructor; [ eauto | | eauto | ].
    eauto using @star_associative.
    eapply JtTopCut.
      eauto with j.
      eapply j_rely; eauto with j jxtra consistent_split.
      eauto using @star_commutative.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This lemma justifies reduction under [JSubRight]. *)

Lemma SrTargetSubRight:
  forall R1f c n K T1 T2,
  sub n K T1 T2 ->
  sr_target R1f K c T1 ->
  sr_target R1f K c T2.
Proof.
  (* The proof is immediate: it suffices to apply the subtyping rule. *)
  inversion 2; intros; subst.
  econstructor; eauto.
  eapply JSubRight; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The following lemma explains why reduction under the evaluation context
   [TApp t _] is sound. *)

(* This lemma is extremely close to [SrTargetFrame], in its statement and
   in its proof. If the number of evaluation contexts grows, we will be
   tempted to share the common structure in these lemmas. For the moment,
   the duplication is tolerable (and simpler). *)

Lemma SrTargetApp:
  forall Rt R1f K s2 t u2 T U,
  (* If the framed-out resource includes [R], *)
  sr_target (star Rt R1f) K (Cf s2 u2) U ->
  consistent (star Rt R1f) ->
  (* and if [R] justifies that [t] is a function of suitable type, *)
  jctop Rt K t (TyArrow U T) ->
  (* then we are able to frame [t] back in and construct an application. *)
  sr_target R1f K (Cf s2 (TApp t u2)) T.
Proof.
  inversion 1; intros; subst.
  jpi_inversion.
  rely_star.
  econstructor; [ eauto | | eauto | ].
    eauto using @star_associative.
    eapply JtTopCut.
      eauto 2 with j.
      eapply j_rely; eauto with j jxtra consistent_split.
      eauto using @star_commutative.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [sr_target] is (contravariantly) affine in its first parameter.
   This means that if an evolution along [rely] is imposed to a large framed-out
   resource, then the same is true of a smaller framed-out resource. *)

Lemma sr_target_affine:
  forall R1 R2 R K c T,
  sr_target R K c T ->
  star R1 R2 = R ->
  consistent R ->
  sr_target R1 K c T.
Proof.
  inversion 1; intros; subst. jsr_consistent. rely_star.
  econstructor.
    eauto.
    2: eauto.
    aac_guess.
    eauto using j_drop_left with consistent_aac.
Qed.

(* The tactic [jsr_frame_away Ru] looks for a hypothesis of the form
   [jsr _ (star R Rf)] and re-associates the star so as to group [Rf]
   and [Ru], i.e., so as to obtain [jsr _ (star _ (star Rf Ru))]. In
   other words, [Ru] is being moved into the framed-out resource. *)

Ltac jsr_frame_away Ru :=
  match goal with
  h: jsr ?s1 (star ?R ?Rf) |- _ =>
    let heq := fresh in
    aac_match (fun Rk => star Rk (star Rf Ru)) (star R Rf) heq;
    rewrite <- heq in h;
    clear heq
  end.

(* The tactic [sr_forget] assumes that we have a [jsr] hypothesis
   of the above form and a goal of the form [sr_target Rf _ _ _].
   We first rewrite the [jsr] hypothesis as above; then, we change
   the goal to [sr_target (star Rf Ru) _ _ _]; then, we abstract
   away [star Rf Ru] and refer to it as just [Rf]. In the end,
   the useless resource [Ru] has vanished. *)

Ltac sr_forget Ru :=
  jsr_frame_away Ru;
  eapply sr_target_affine with (R2 := Ru); [
    (* nothing *) | eauto | jsr_consistent; eauto with consistent_aac
  ];
  match goal with |- sr_target (star ?Rf ?Ru) _ _ _ =>
    generalize dependent (star Rf Ru); clear Rf Ru; intro Rf; intros
  end.

