Set Implicit Arguments.
Require Import MyList.
        Import ListNotations.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ------------------------------------------------------------------------- *)

(* Freezing a memory block at address [l] means updating the heap resource,
   at this address, from [DX_EXCLUSIVE v] to [DX_DUPLICABLE v]. *)

Definition freeze l v (R1 R2 : resource) : Prop :=
  maps (get R1) l (DX_EXCLUSIVE v) /\
  R2 = set (update (get R1) l (DX_DUPLICABLE v)) R1.

(* ------------------------------------------------------------------------- *)

(* This is an admissible evolution. *)

Lemma admissible_freeze:
  forall l v,
  admissible (freeze l v).
Proof.
  intros l v.
  (* The basic reason why this is true is that we are performing a freeze
     at a single location [l] in the heap resource. *)
  assert (f: admissible (at_address l (dx_freeze v v))).
    { eauto with admissible. }
  unfold at_address, dx_freeze in f.
  (* There remains to lift this fact to the level of resources, ignoring
     the components other than the heap resource, as they are unaffected. *)
  unfold admissible, freeze. intros. unpack. subst.
  explode_resources_blocks.
  forwards: f; [ eauto 8 | eauto | unpack; subst ].
  xplit; explode; eauto using @rely_reflexive.
Qed.

(* ------------------------------------------------------------------------- *)

(* If the machine state does not change, and if the resource evolves along
   [freeze l v], then [agrees] is preserved. *)

Lemma accord_agrees_identity_freeze:
  forall l v,
  accord agrees (@identity _) (freeze l v).
Proof.
  unfold accord, identity, freeze. intros. unpack. subst.
  forwards: ag_maps_right_to_left Exclusive; eauto. unpack.
  agrees_inversion. explode_resources_blocks.
  eauto 6 using @agrees_update_identity with agrees.
Qed.

(* Hence, [jsr] is preserved as well. *)

Lemma accord_jsr_identity_freeze:
  forall l v,
  accord jsr (@identity _) (freeze l v).
Proof.
  intros. eapply lift_accord.
    eapply admissible_freeze.
    unfold freeze. prove_preserves_lock_resource.
    eapply accord_agrees_identity_freeze.
Qed.

Local Hint Unfold identity freeze.

(* ---------------------------------------------------------------------------- *)

(* We are now ready to establish subject reduction. *)

Lemma subject_reduction_JFreeze:
  forall s1 v s2 t2,
  redwkc (Cf s1 TGhost) (Cf s2 t2) TERM ->
  forall R K T,
  jctop R K v (TyRef Exclusive T) ->
  forall R1f,  
  jsr s1 (star R R1f) ->
  sr_target R1f K (Cf s2 t2) (TyMixStar TyTop (TyAnchored v (TyRef Duplicable T))).
Proof.
  intros. redwkc_inversion. jsr_consistent.
  classification. decomposition.

  (* A technical point. We have proved that [freeze l v] is admissible,
     so we want to apply this evolution to one specific heap fragment
     -- namely, the one that justifies the ownership of the address [l] --
     and work our way up from there. The following incantation identifies
     this heap fragment [R], as well as [l] and [v], and builds a heap
     fragment [R'] such that [freeze l v R R'] holds. *)
  match goal with h: maps (get ?R) ?l (mv2dx Exclusive ?v) |- _ =>
    with_evar resource (fun R' => assert (f: freeze l v R R'))
  end. { eauto. }

  (* Pull this evolution one step up, so as to encompass the resource
     that represents the ownership of the value [v]. *)
  forwards: @use_admissible_symmetric f;
    eauto 2 using admissible_freeze with consistent_split.
  unpack.

  (* We have already checked most of what we need. *)
  eapply SrTargetAccordCanonical;
    (instantiate; eauto using admissible_freeze, accord_jsr_identity_freeze).

  (* There only remains to check that the postcondition holds. *)
  eapply JcTopMix; [ eauto with wk | ].
  jctop JRef; eauto 4 using (last3 j_rely), @maps_update_there with consistent_split.

Qed.

