Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.
Require Import AdoptionAgreement.
Require Import BlockAgreement.
Require Import Zap.

(* ---------------------------------------------------------------------------- *)

(* At the level of machine states, the action of giving the adoptee at memory
   location [x] to the adopter at memory location [y] is represented by the
   following relation. *)

Definition state_give x y (s1 s2 : state) :=
  exists p v,
  maps (get s1) x (Block p v) /\
  s2 = set (update (get s1) x (Block (NonNull y) v)) s1.

(* At the level of resources, it is represented by the following relation. We
   must own [x] as an adoptee, as this is required for this relation to be
   admissible. We must also own [y] as an adopter, as this is required for
   the resource [R2] to be round. Note that we do not need to explicitly
   require [adopts (get R1) y], as this fact is implied by the proof [pf]. *)

Definition resource_give x y (R1 R2 : resource) :=
  exists pf,
  unadopted (get R1) x /\
  R2 = set (exist round (adopt (get R1) x y) pf) R1.

(* ---------------------------------------------------------------------------- *)

(* The desired roundness proof [pf] can in fact be established. *)

Lemma establish_resource_give:
  forall R1 x R2 xs y,
  unadopted (get R1) x ->
  all_adoptees (get R2) xs y ->
  consistent (star R1 R2) ->
  exists pf,
  resource_give x y (star R1 R2)
    (set (exist round (adopt (get (star R1 R2)) x y) pf) (star R1 R2)).
Proof.
  unfold resource_give. intros. xplit; eauto.
  eauto using unadopted_affine with explode_resources.
  Grab Existential Variables.
  eauto using adopts_affine, all_adoptees_adopts
         with round explode_resources aac_guess.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This relation is admissible. *)

Lemma admissible_resource_give:
  forall x y,
  admissible (resource_give x y).
Proof.
  unfold admissible, resource_give. intros. unpack. subst.
  explode_resources_adoption_strong.
  xplit; eauto using unadopted_affine, @rely_reflexive.
  explode. eauto using star_adopt_left.
  Grab Existential Variables.
  (* Here, one is tempted to use the lemma [round_adopt].  However, this lemma
     has a hypothesis of the form [adopts _ y], so we would have to prove that
     [round (adopt R2 x y)] implies [adopts R2 y]. The following proof is
     simpler. *)
  rewrite <- star_adopt_left by eauto.
  eauto with round.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Agreement is preserved by these relations. *)

Lemma accord_agrees_give:
  forall x y,
  accord agrees (state_give x y) (resource_give x y).
Proof.
  unfold accord, state_give, resource_give. intros. unpack. subst.
  explode_resources_adoption_strong. 
  agrees_inversion. adoption_resource_agrees_inversion.
  eauto with agrees.
Qed.

Lemma accord_jsr_give:
  forall x y,
  accord jsr (state_give x y) (resource_give x y).
Proof.
  intros. eapply lift_accord.
    eapply admissible_resource_give.
    unfold resource_give. prove_preserves_lock_resource.
    eapply accord_agrees_give.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TGive]. *)

Lemma subject_reduction_TGive:
  forall s1 v1 v2 s2 t2,
  redwkc (Cf s1 (TGive v1 v2)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K U,
  jpi R K (TyStar (TyStar
    (TyAnchored v1 U)
    (TyAnchored v1 TyUnadopted))
    (TyAnchored v2 (TyAdopts U))) ->
  sr_target R' K (Cf s2 t2)
    (TyMixStar TyTop (TyAnchored v2 (TyAdopts U))).
Proof.
  intros. redwkc_inversion. jsr_consistent.
  Opaque get. decomposition. Transparent get. (* TEMPORARY unpleasant *)

  (* Start with the heap fragment [R] that owns [x] and [y]. *)
  forwards: establish_resource_give; eauto with consistent_aac. unpack.

  (* Work our way up, so as to include the resources that own the new
     adoptee and the list of previous adoptees. *)
  forwards: @use_admissible_symmetric.
    eapply admissible_resource_give.
    eauto.
    match goal with hx: jctop ?Rx _ _ ?U, hxs: jpi ?Rxs _ (zap _ ?U) |- _ =>
      instantiate (1 := star Rx Rxs)
    end.
    eauto with consistent_aac.
  unpack.

  (* We have already checked most of what we need. *)
  eapply SrTargetAccordCanonical.
    eapply admissible_resource_give.
    eapply accord_jsr_give.
    eauto.
    unfold state_give. eauto with reflexivity.
    equates 2. eauto. aac_reflexivity.

  (* There remains to build a type derivation. *)
  eapply JcTopMix; [ eauto with wk | ].
  jctop JAdopts; [ | | eauto | eapply star_commutative ].
    (* The list of adoptees is extended with [x]. *)
    explode_resources_adoption_strong.
    eauto using all_adoptees_cons with consistent_aac.
    (* Every member of the extended list of adoptees has type [U]. *)
    simpl. eapply j_rely; eauto with j consistent_aac.

Qed.
