Set Implicit Arguments.
Require Import MyList.
        Import ListNotations.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import AdoptionAgreement.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ---------------------------------------------------------------------------- *)

(* At the level of machine states, the allocation of a new block is described
   by the following relation. *)

Definition machine_new v (s1 s1' : state) :=
  s1' = set (allocate (get s1) (Block Null v)) s1.

(* At the level of resources, it is described as follows. *)

Definition resource_new v (R1 R1' : resource) :=
  exists m Rh1' Ra1' pf,
  (* The new heap resource. *)
  Rh1' = allocate (get R1) (mv2dx m v) /\
  (* The new adoption resource. *)
  Ra1' = exist round (allocate (get R1) (AdopteeOwned Null, AdopterOwned)) pf /\
  (* The new resource. *)
  R1' = set Ra1' (set Rh1' R1).

(* ---------------------------------------------------------------------------- *)

(* This is an admissible evolution. *)

Lemma admissible_resource_new:
  forall v,
  admissible (resource_new v).
Proof.
  unfold admissible, resource_new. intros. unpack. subst.
  explode_resources_adoption_strong.
  xplit; explode; eauto using star_allocate_core_right_unit,
    @rely_reflexive, @rely_allocate with consistent duplicable.
  (* Roundness obligations. *)
  Grab Existential Variables.
  eauto with round.
  eauto with round.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This evolution preserves agreement. *)

Lemma accord_agrees_new:
  forall v,
  accord agrees (machine_new v) (resource_new v).
Proof.
  unfold accord, machine_new, resource_new. intros. unpack. subst.
  explode_resources_adoption_strong. agrees_inversion. adoption_resource_agrees_inversion. simpl.
  eauto 10 with agrees.
Qed.

Lemma accord_jsr_new:
  forall v,
  accord jsr (machine_new v) (resource_new v).
Proof.
  intros. eapply lift_accord.
    eapply admissible_resource_new.
    unfold resource_new. prove_preserves_lock_resource.
    eapply accord_agrees_new.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This auxiliary lemma helps send the adopter access right into the left-hand
   side and the adoptee access right into the right-hand side. *)

Lemma split_adoption_rights:
  forall R,
  star
    (allocate (core R) (AdopteeNotOwned,   AdopterOwned))
    (allocate (core R) (AdopteeOwned Null, AdopterNotOwned)) =
     allocate (core R) (AdopteeOwned Null, AdopterOwned).
Proof.
  intros.
  equates 1; [ eapply star_allocate | symmetry ].
  f_equal. eauto using @cores_are_duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A few trivial yet helpful facts. *)

Lemma adopts_allocate:
  forall R i l,
  consistent R ->
  l = limit R ->
  adopts (allocate (core R) (i, AdopterOwned)) l.
Proof.
  intros. subst. do 2 econstructor.
  eauto using @maps_allocate_there_core.
  econstructor.
Qed.

Lemma no_antecedents_allocate:
  forall R l,
  consistent R ->
  l = limit R ->
  no_antecedents (allocate (core R) (AdopteeNotOwned, AdopterOwned)) l.
Proof.
  intros. subst. equates 1.
    eapply no_antecedents_allocate.
      eauto using @consistent_core.
      eauto with round.
      simpl. tauto.
      simpl. tauto.
    symmetry. eapply limit_core.
Qed.

Lemma unadopted_allocate:
  forall R i l,
  consistent R ->
  l = limit R ->
  unadopted (allocate (core R) (AdopteeOwned Null, i)) l.
Proof.
  intros. subst. do 2 econstructor.
  eauto using @maps_allocate_there_core.
  econstructor.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TNewRef]. *)

Lemma subject_reduction_TNewRef:
  forall s1 v s2 t2,
  redwkc (Cf s1 (TNewRef v)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K m T,
  jctop R K v T ->
  sr_target R' K (Cf s2 t2) (intersection [
    TyRef m T;
    TyAdopts TyBottom;
    TyUnadopted
  ]).
Proof.
  intros. redwkc_inversion. jsr_consistent.
  eapply SrTargetAccordCanonical.
    eapply admissible_resource_new.
    eapply accord_jsr_new.
    eauto.
    unfold machine_new. explode.
    unfold resource_new. eauto 8.
  explode_resources_adoption_strong.

  (* The new memory location is well-typed. *)
  eapply JcTopIntersection. eauto with wk.
  (* Split the resource between the permissions that we must justify. The
     left-hand side is [TyRef m T], and the right-hand side are the adoption
     permissions. We wish to send everything into the left-hand side, except
     the adoption access rights for the new location. This is expressed as
     follows! *)
  eapply JpiStar.
  3: explode; first [ eapply (@star_allocate_core adoption_status) | eapply core_right_unit ].

  (* Sub-goal 1. The memory block. *)
  (* Split again. The left-hand side justifies the fields, while the right-hand
     side justifies the block itself. In the heap resource component, which is
     of the form [allocate R p], we wish to send [allocate R (core p)] into the
     left-hand side and [allocate (core R) p] into the right-hand side. In the
     other components, we wish to send everything into the left-hand side. This
     is expressed as follows! *)
  jctop JRef.
  explode; first [ eapply (@star_allocate_core heap_access_right) | eapply core_right_unit ].
  (* The value [v] is closed. *)
  eassumption.
  (* The value [v] was well-typed under some resource [R], and must now be
     well-typed under [allocate R (core p)]. Stability justifies this. *)
  eapply j_rely.
    eauto.
    explode; eauto 2 with consistent_split.
    explode; eauto using @rely_allocate_core, @rely_reflexive with consistent.
  (* And we own the new memory location. *)
  eapply maps_allocate_there_core.
    eauto with consistent_split.
    eauto with limit_limit reflexivity consistent_aac.

  (* Sub-goal 2. The adoption permissions. *)
  (* Split again, sending the adopter access right into the left-hand side
     and the adoptee access right into the right-hand side. We do not care
     about the rest. *)
  eapply JpiStar.
  3: explode; first [ eapply split_adoption_rights | eapply core_right_unit ].
  (* TyAdopts *)
  eapply JpiAtomicEmpty; [ eauto with wk | ].
  eapply (@JAdopts 0); try eapply core_right_unit.
    simpl. eapply AllAdopteesBase.
      { eapply adopts_allocate; eauto 3 with limit_limit reflexivity consistent_aac. }
      { eapply no_antecedents_allocate; eauto 3 with limit_limit reflexivity consistent_aac. }
    eapply JpiEmpty.
    eauto with wk substitutable logic.
  (* TyUnadopted *)
  eapply JpiStarEmpty.
  eapply JpiAtomicEmpty; [ eauto with wk | ].
  eapply (@JUnadopted 0).
    { eapply unadopted_allocate; eauto 3 with limit_limit reflexivity consistent_aac. }

  (* Roundness proof obligations. *)
  Grab Existential Variables.
  eauto with round.
  eauto with round.
  eauto with round.
  eauto with round.
  eauto with round.
  eauto with round.
  eauto with round.

Qed.

