Set Implicit Arguments.
Require Import MyList.
Require Import Permutation.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.
Require Import AdoptionAgreement.
Require Import BlockAgreement.
Require Import Zap.
Require Import ZapJpi.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TTake]. *)

Lemma subject_reduction_TTake:
  forall s1 v1 v2 s2 t2,
  redwkc (Cf s1 (TTake v1 v2)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K U,
  jpi R K (TyStar
    (TyAnchored v1 TyAdoptable)
    (TyAnchored v2 (TyAdopts U))) ->
  sr_target R' K (Cf s2 t2) (TyMixStar TyTop (TyStar (TyStar
    (TyAnchored v1 U)
    (TyAnchored v1 TyUnadopted))
    (TyAnchored v2 (TyAdopts U)))).
Proof.
  intros. jsr_consistent. redwkc_inversion.

  (* Case: RedTakeSuccess *)
  Opaque get. decomposition. Transparent get. (* TEMPORARY unpleasant *)
  sr_target_pure.
  eapply (JTakeCommitted 0); eauto.
    (* We have destructed this judgement. Reconstruct it. *)
    { jpi_drop_left. jctop JAdopts; eauto. }
    (* There remains to prove that [x] is adopted by [y] in our resource.
       This follows from 1- the fact that this is true in the heap, hence
       true in the global resource; and 2- we own [y], so we own [x] too;
       it cannot be owned by the framed-out resource. *)
    { forwards: jsr_adopted; 
      eauto using adopted_narrow_left, adopts_affine, all_adoptees_adopts
             with explode_resources aac_guess consistent_aac. }

  (* Case: RedTakeFail *)
  decomposition.
  sr_target_pure.
  eapply (JFail 0).
  eauto 10 with wk.

Qed.

(* ---------------------------------------------------------------------------- *)

(* At the level of machine states, the action of taking the adoptee at memory
   location [x] from the adopter at memory location [y] is represented by the
   following relation. *)

Definition state_take x (s1 s2 : state) :=
  exists p v,
  maps (get s1) x (Block p v) /\
  s2 = set (update (get s1) x (Block Null v)) s1.

(* At the level of resources, it is represented by the following relation.
   The condition [adopted (get R1) x y] is required for this relation to be
   admissible. *)

Definition resource_take x (R1 R2 : resource) :=
  exists y,
  adopted (get R1) x y /\
  R2 = set (exist _
    (abandon (get R1) x)
    (round_abandon x (round_get R1))
  ) R1.

(* ---------------------------------------------------------------------------- *)

(* This evolution is admissible. *)

Lemma admissible_resource_take:
  forall x,
  admissible (resource_take x).
Proof.
  unfold admissible, resource_take. intros. unpack. subst.
  explode_resources_adoption_strong.
  xplit; eauto using adopted_affine, @rely_reflexive.
  explode. eauto using star_abandon_left with consistent_split.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Agreement is preserved by these relations. *)

Lemma accord_agrees_take:
  forall x,
  accord agrees (state_take x) (resource_take x).
Proof.
  unfold accord, state_take, resource_take. intros. unpack. subst.
  explode_resources_adoption_strong. 
  agrees_inversion. adoption_resource_agrees_inversion.
  eauto with agrees.
Qed.

Lemma accord_jsr_take:
  forall x,
  accord jsr (state_take x) (resource_take x).
Proof.
  intros. eapply lift_accord.
    eapply admissible_resource_take.
    unfold resource_take. prove_preserves_lock_resource.
    eapply accord_agrees_take.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The following lemma states that the resource [abandon R x] separately
   justifies [unadopted _ x] and [all_adoptees _ xs y]. *)

Lemma split_after_take:
  forall R x xs y,
  round R ->
  adopted R x y ->
  all_adoptees (abandon R x) xs y ->
  exists R1 R2,
  round R1 /\
  round R2 /\
  unadopted R1 x /\
  all_adoptees R2 xs y /\
  star R1 R2 = abandon R x.
Proof.
  intros.
  (* The desired [R1] maps [x] to [AdopteeOwned Null], and owns nothing else,
     so it is obtained from [core R] by mapping [x] to [AdopteeOwned Null].
     This happens to be [abandon (core R) x], although the use of the word
     [abandon] here is somewhat counter-intuitive, as we are retaining the
     ownership of [x] as an (unadopted) adoptee. *)
  exists (abandon (core R) x).
  (* The desired [R2] maps [x] to [AdopteeNotOwned], and is otherwise equal
     to [R]. *)
  exists (updatef R x (set AdopteeNotOwned)).
  (* Now, let's see. *)
  xplit.
  { eauto with round. }
  { eapply round_updatef; eauto; intros; destruct_adoption_status; simpl in *; [ eauto | tauto ]. }
  { eauto using unadopted_abandon, adopted_adoptable, adoptable_core. }
  { equates 3.
    eapply all_adoptees_lose. eauto.
    eauto using all_adoptees_abandon_implies_not_in.
    unfold set_adopter_edge. erewrite compose_updatef_updatef. eauto.
    intros. destruct_adoption_status. eauto. }
  { unfold set_adopter_edge, set_owned.
    erewrite star_updatef.
    rewrite core_left_unit.
    eauto.
    intros. destruct_adoption_status. simpl. msa_pair. f_equal. }
Qed.

(* Same statement, this time at the level of resources. *)

Lemma split_after_take_fat:
  forall R x xs y,
  adopted (get R) x y ->
  all_adoptees (abandon (get R) x) xs y ->
  exists R1 R2,
  unadopted (get R1) x /\
  all_adoptees (get R2) xs y /\
  star R1 R2 = set (exist _
    (abandon (get R) x)
    (round_abandon x (round_get R))
  ) R.
Proof.
  (* It is a bit painful to have to state this, but this simplifies life
     down the way. Plus, it is not difficult to prove. *)
  intros.
  forwards: split_after_take; eauto with round. unpack.
  explode_resources_adoption_strong.
  xplit.
  3: explode; first [ eassumption | eapply core_left_unit (* arbitrary *) ].
  eauto.
  eauto.
  Grab Existential Variables.
  eauto.
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TTakeCommitted]. *)

Lemma subject_reduction_TTakeCommitted:
  forall s1 R R' K x y s2 t2 U,
  redwkc (Cf s1 (TTakeCommitted (TLoc x) (TLoc y))) (Cf s2 t2) TERM ->
  jsr s1 (star R R') ->
  adopted (get R) x y ->
  jctop R K (TLoc y) (TyAdopts U) ->
  sr_target R' K (Cf s2 t2) (TyMixStar TyTop (TyStar (TyStar
    (TyAnchored (TLoc x) U)
    (TyAnchored (TLoc x) TyUnadopted))
    (TyAnchored (TLoc y) (TyAdopts U)))).
Proof.
  intros. jsr_consistent. redwkc_inversion.
  Opaque get. decomposition. Transparent get. (* TEMPORARY unpleasant *)

  (* Start with the resource fragment that owns the edge of [x] to [y]. *)
  match goal with h: adopted (get (star ?R _)) ?x _ |- _ =>
    with_evar resource (fun R' => assert (f: resource_take x R R'))
  end.
  { unfold resource_take.
    eauto 7 using adopted_narrow_left, all_adoptees_adopts
             with explode_resources consistent_split. }

  (* Work our way up, so as to include the resource that owns the adoptees. *)
  forwards: admissible_resource_take f.
    { eauto 2 with consistent_split. }
  unpack.
  (* Take the adoptees through the rely. *)
  forwards: (last3 j_rely); eauto 3 with consistent_split.

  (* We have already checked most of what we need. *)
  eapply SrTargetAccordCanonical.
    { eapply admissible_resource_take. }
    { eapply accord_jsr_take. }
    { eauto. }
    { unfold state_take. eauto with reflexivity. }
    { eauto. }

  (* The address [x] is adopted by [y], and the list of all adoptees of [y]
     is [xs], so [x] must be a member of the list [xs]. *)
  forwards: no_lost_sheep; [ | eauto | ].
    { explode_resources_adoption_strong.
      eauto using all_adoptees_affine with consistent_split. }
  (* Bring [x] to the front of the list. *)
  forwards: the_last_will_be_the_first; eauto. unpack.
  forwards hzap: jpi_zap_permutation; eauto.
  dependent destruction hzap.

  (* There remains to build a type derivation. We must split into three parts
     that respectively justify [x @ U] and [x @ unadopted] and [y @ adopts U]. *)
  forwards: split_after_take_fat; eauto. unpack.
  eapply JcTopMix; [ eauto with wk | ].
  econstructor.
    econstructor; eauto.
      jctop JUnadopted; eauto.
    jctop JAdopts; eauto.
    aac_congruence.

Qed.

