Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ------------------------------------------------------------------------- *)

(* At the level of machine states, the action of writing the value [v2] at
   location [l] is represented by the following relation. *)

Definition state_write l v2 (s1 s2 : state) :=
  exists p v1,
  maps (get s1) l (Block p v1) /\
  s2 = set (update (get s1) l (Block p v2)) s1.

(* At the level of resources, it is represented by the following relation.
   We must own the memory location [l], and we keep this ownership. *)

Definition resource_write l v2 (R1 R2 : resource) :=
  exists v1,
  maps (get R1) l (mv2dx Exclusive v1) /\
  R2 = set (update (get R1) l (mv2dx Exclusive v2)) R1.

(* ------------------------------------------------------------------------- *)

(* This evolution of resources is admissible. *)

Lemma admissible_resource_write:
  forall l v2,
  admissible (resource_write l v2).
Proof.
  intros.
  (* The basic reason why this is true is that we are performing a freeze
     at a single location [l] in the heap resource. *)
  assert (f: admissible (at_address l (dx_write v2))).
    { eauto with admissible. }
  unfold at_address, dx_write in f.
  (* There remains to lift this fact to the level of resources, ignoring
     the components other than the heap resource, as they are unaffected. *)
  unfold admissible, resource_write. intros. unpack. subst.
  explode_resources_blocks.
  forwards: f; [ eauto 8 | eauto | unpack; subst ].
  xplit; explode; eauto using @rely_reflexive.
Qed.

(* ------------------------------------------------------------------------- *)

(* Updating the machine state and the resource at [l] with [v2] preserves
   agreement. *)

Lemma accord_agrees_write:
  forall l v2,
  accord agrees (state_write l v2) (resource_write l v2).
Proof.
  unfold accord, state_write, resource_write. intros. unpack. subst.
  forwards: ag_maps_maps. eauto. eauto. eauto. eapply core_right_unit.
  subst. agrees_inversion. explode_resources_blocks. eauto 7 with agrees.
Qed.

Lemma accord_jsr_write:
  forall l v2,
  accord jsr (state_write l v2) (resource_write l v2).
Proof.
  intros. eapply lift_accord.
    eapply admissible_resource_write.
    unfold resource_write. prove_preserves_lock_resource.
    eapply accord_agrees_write.
Qed.

Local Hint Unfold state_write resource_write.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TWrite]. *)

Lemma subject_reduction_TWrite:
  forall s1 v1 v2 s2 t2,
  redwkc (Cf s1 (TWrite v1 v2)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K T1 T2,
  jpi R K (TyStar (TyAnchored v1 (TyRef Exclusive T1)) (TyAnchored v2 T2)) ->
  sr_target R' K (Cf s2 t2) (TyMixStar TyTop (TyAnchored v1 (TyRef Exclusive T2))).
Proof.
  intros. redwkc_inversion. decomposition.
  
  (* We do not need the heap fragment that corresponds to the old value
     of the field, as this value is overwritten. Let's lose this heap
     fragment, so as to clear up the picture. *)
  match goal with hj: jctop ?Ru _ ?v1 T1, hw: wk _ ?v1 _ |- _ =>
    clear hj hw; sr_forget Ru
  end.
  jsr_consistent.

  (* Start with the heap fragment [R] that owns the location [l]. *)
  match goal with h: maps (get ?R) ?l (mv2dx Exclusive ?v) |- _ =>
    with_evar resource (fun R' => assert (f: resource_write l v2 R R'))
  end. { eauto. }

  (* Work our way up. *)
  forwards: @use_admissible_symmetric f;
    eauto 2 using admissible_resource_write with consistent_split.
  unpack.

  (* We have already checked most of what we need. *)
  eapply SrTargetAccordCanonical; (instantiate;
    eauto using admissible_resource_write, accord_jsr_write with reflexivity).

  (* There remains to build a type derivation. *)
  eapply JcTopMix; [ eauto with wk | ].
  jctop JRef; eauto 4 using (last3 j_rely),
                            @maps_update_there with consistent_split.

Qed.

