Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import Agreement.
Require Import JConf.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import SubjectReductionTerms.
Require Import Spawn.

(* ---------------------------------------------------------------------------- *)

(* The predicate [srs_target Rf c] serves as the conclusion of the main subject
   reduction lemma for soups. It means that the configuration [c] is well-typed
   and that the evolution imposed to the outside world, whose initial state was
   [R1f], is acceptable. *)

Inductive srs_target : resource -> configuration -> Prop :=
| SrsTarget:
    forall R1f s2 t2,
    forall R2t R2f R2,
    js R2t t2 ->
    jsr s2 R2 ->
    star R2t R2f = R2 ->
    rely R1f R2f ->
    srs_target R1f (Cf s2 t2).

Hint Constructors srs_target : srs_target.

(* ---------------------------------------------------------------------------- *)

(* These lemmas justify reduction under a [TPar] construct. *)

Lemma SrsTargetParLeft:
  forall R2 R1 h t1 t2,
  srs_target (star R2 R1) (Cf h t1) ->
  js R2 t2 ->
  consistent (star R2 R1) ->
  srs_target R1 (Cf h (TPar t1 t2)).
Proof.
  inversion 1; intros; subst.
  rely_star.
  eauto using js_rely, @star_associative with consistent_split js srs_target.
Qed.

Lemma SrsTargetParRight:
  forall R2 R1 h t1 t2,
  srs_target (star R1 R2) (Cf h t2) ->
  js R1 t1 ->
  consistent (star R1 R2) ->
  srs_target R2 (Cf h (TPar t1 t2)).
Proof.
  inversion 1; intros; subst.
  rely_star.
  eauto using js_rely with consistent_split js srs_target aac_guess.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The main subject reduction lemma for soups. *)

Lemma subject_reduction_SOUP:
  forall s1 t1 s2 t2,
  red (Cf s1 t1) (Cf s2 t2) ->
  wkc (Cf s1 t1) SOUP ->
  forall R1t R1f R1, 
  (* If the soup [t1] is well-typed under [R1t], *)
  js R1t t1 ->
  (* and if [R1t] is part of the resource [R1] that corresponds
     to the entire state [s1], *)
  jsr s1 R1 ->
  star R1t R1f = R1 ->
  (* then the new soup [t2] is well-typed under [R2t], part of a 
     a resource [R2] that corresponds to the entire state [s2],
     and the other part, namely [R2f], is an acceptable evolution
     of [R1f]. *)
  srs_target R1f (Cf s2 t2).
Proof.
  dependent_induction; inversion 1; introv hjs; intros; subst;
  try destruct_eval_context;
  dependent destruction hjs;
  wk_inversion;
  jsr_consistent.
  (* RedThread *)
  (* Just apply subject reduction for terms. *)
  forwards hsr: subject_reduction_TERM; eauto with wk.
  dependent destruction hsr.
  econstructor; solve [ eauto with js ].
  (* RedParLeft *)
  solve [ eauto using SrsTargetParLeft with wk consistent_aac aac_guess ].
  (* RedParRight *)
  solve [ eauto using SrsTargetParRight with wk consistent_aac aac_guess ].
  (* RedSpawn *)
  (* We have just the appropriate lemma for this case. *)
  forwards: decompose_fill_spawn; eauto with consistent_aac.
  unpack.
  econstructor; eauto using @rely_reflexive with js.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for configurations. A simplified statement. *)

Theorem subject_reduction:
  forall c1 c2,
  red c1 c2 ->
  jconf c1 ->
  jconf c2.
Proof.
  inversion 2; intros; subst. destruct_configurations.
  forwards hsr: subject_reduction_SOUP; eauto using @core_right_unit.
  dependent destruction hsr.
  jsr_consistent.
  econstructor; eauto using red_wkc, js_drop_right.
Qed.

