Set Implicit Arguments.
Require Import MyList.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WellKindedness.

(* ---------------------------------------------------------------------------- *)

(* A notation for introducing a new variable, together with a permission for
   it. Informally, [intro Q T] represents the permission [Q * (x @ T)], where
   [x] is ``a new variable''. *)

Definition intro Q T :=
  TyStar (shift 0 Q) (TyAnchored (TVar 0) (shift 0 T)).

(* ---------------------------------------------------------------------------- *)

(* [intro] preserves well-kindedness. *)

Lemma intro_wk:
  forall K Q T,
  wk K Q PERM ->
  wk K T TYPE ->
  wk (insert 0 VALUE K) (intro Q T) PERM.
Proof.
  unfold intro. eauto 7 using wk_weakening with wk.
Qed.

Hint Resolve intro_wk : wk.

(* [intro] commutes with [shift]. *)

Lemma shift_intro:
  forall sx Q T x,
  sx = 1 + x ->
  shift sx (intro Q T) = intro (shift x Q) (shift x T).
Proof.
  unfold intro. intros; subst; simpl_lift_goal; eauto with f_equal lift_lift.
Qed.

Hint Resolve shift_intro : j_weakening.

(* [intro] commutes with [subst]. *)

Lemma subst_intro:
  forall Q T U x U' sx,
  shift 0 U = U' ->
  1 + x = sx -> 
  subst U' sx (intro Q T) = intro (subst U x Q) (subst U x T).
Proof.
  unfold intro. intros; subst. simpl_subst_goal. eauto with f_equal lift_subst.
Qed.

Hint Resolve subst_intro : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* Notions of iterated star. *)

Notation TyIteratedStar Ps :=
  (fold_right TyStar TyEmpty Ps).

Notation TyIteratedStarReverse Ps :=
  (fold_left TyStar Ps TyEmpty).

(* ---------------------------------------------------------------------------- *)

(* [TyIteratedStar] maps a list of permissions to a permission. *)

Lemma wk_TyIteratedStar:
  forall K Ps,
  Forall (fun P => wk K P PERM) Ps ->
  wk K (TyIteratedStar Ps) PERM.
Proof.
  induction 1; simpl; intros; eauto with wk.
Qed.

Lemma wk_TyIteratedStar_inverse:
  forall K Ps,
  wk K (TyIteratedStar Ps) PERM ->
  Forall (fun P => wk K P PERM) Ps.
Proof.
  induction Ps; simpl; intros; wk_inversion; econstructor; eauto.
Qed.

(* [TyIteratedStar] commutes with [shift]. *)

Lemma shift_TyIteratedStar:
  forall x Ps Qs,
  List.map (shift x) Ps = Qs ->
  shift x (TyIteratedStar Ps) = TyIteratedStar Qs.
Proof.
  induction Ps; intros; subst; simpl; simpl_lift_goal; eauto with f_equal.
Qed.

(* [TyIteratedStar] commutes with [subst]. *)

Lemma subst_TyIteratedStar:
  forall U x Ps Qs,
  List.map (subst U x) Ps = Qs ->
  subst U x (TyIteratedStar Ps) = TyIteratedStar Qs.
Proof.
  induction Ps; intros; subst; simpl; simpl_subst_goal; eauto with f_equal.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A notation for an intersection type. The type [T /\ U] is encoded as
   {x: VALUE} (=x | x @ T * x @ U). *)

Definition intersection_body (Ts : list term) :=
  TyMixStar
    (TySingleton (TVar 0))
    (TyIteratedStar (List.map (fun T => TyAnchored (TVar 0) (shift 0 T)) Ts)).

Definition intersection (Ts : list term) :=
  TyExists VALUE (intersection_body Ts).

(* ---------------------------------------------------------------------------- *)

(* [intersection] maps a list of types to a type. *)

Lemma wk_intersection_body:
  forall K Ts,
  Forall (fun T => wk K T TYPE) Ts ->
  wk (insert 0 VALUE K) (intersection_body Ts) TYPE.
Proof.
  unfold intersection_body. intros.
  econstructor; eauto with wk.
  eapply wk_TyIteratedStar.
  eapply Forall_map; eauto.
  simpl. eauto using wk_weakening with wk.
Qed.

Hint Resolve wk_intersection_body : wk.

Lemma wk_intersection:
  forall K Ts,
  Forall (fun T => wk K T TYPE) Ts ->
  wk K (intersection Ts) TYPE.
Proof.
  unfold intersection. eauto with substitutable logic wk.
Qed.

Hint Resolve wk_intersection : wk.

(* [intersection] commutes with [shift]. *)

Lemma shift_intersection:
  forall x Ts Us,
  List.map (shift x) Ts = Us ->
  shift x (intersection Ts) = intersection Us.
Proof.
  unfold intersection, intersection_body.
  induction Ts; intros; subst; simpl; simpl_lift_goal; intros.
  (* Base case. *)
  eauto.
  (* Step case. *)
  do 3 f_equal.
    { eauto with f_equal lift_lift. }
    { eapply shift_TyIteratedStar.
      do 2 rewrite map_map.
      eapply map_ext.
      intros. simpl_lift_goal.
      eauto with f_equal lift_lift. }
Qed.

Hint Resolve shift_intersection : j_weakening.

(* [intersection] commutes with [subst]. *)

Lemma subst_intersection:
  forall U x Ts Us,
  List.map (subst U x) Ts = Us ->
  subst U x (intersection Ts) = intersection Us.
Proof.
  unfold intersection, intersection_body.
  induction Ts; intros; subst; simpl; simpl_subst_goal; intros.
  (* Base case. *)
  eauto.
  (* Step case. *)
  do 3 f_equal.
    { eauto with f_equal lift_subst. }
    { eapply subst_TyIteratedStar.
      do 2 rewrite map_map.
      eapply map_ext.
      intros. simpl_subst_goal.
      eauto with f_equal lift_subst. }
Qed.

Hint Resolve subst_intersection : j_substitution.

