Set Implicit Arguments.
Require Import EquivDec.
Require Import DeBruijn.
Require Import MSADXPermissions.

(* ---------------------------------------------------------------------------- *)

(* We wish to allow types to refer to terms: if [x] is a term variable, bound
   by an ordinary lambda-abstraction, then occurrences of [x] can appear both
   as part of a term and as part of a type. This dictates that term variables
   and type variables belong to a single namespace. Furthermore, since we wish
   to substitute terms for term variables, and types for type variables, there
   follows that terms and types must belong to a single syntactic category.
   Thus, in the following, we define a single Coq inductive type which
   represents the syntax of terms and types. Later, we use kinds and a
   well-kindedness judgement to distinguish between terms and types. This
   design decision might seem weird at first. Experience proves, in fact, that
   it is a good thing: it means that there is just one notion of lifting, one
   notion of substitution, and that fewer administrative lemmas (weakening,
   substitution, etc.) are needed. *)

(* ---------------------------------------------------------------------------- *)

(* Kinds. *)

(* Kinds play a double role in the type system and in its formalization.

   In the type system, kinds are used to distinguish between various ``kinds''
   of type variables. The following three kinds are used for this purpose:

     [TYPE]  classifies ordinary types.
     [PERM]  classifies permissions.
     [VALUE] classifies values (which exist at runtime); this kind is known
             as [term] in surface Mezzo

  In the formalization, kinds are also used to distinguish between terms and
  types (which are usually thought of as two disjoint syntactic categories)
  and to distinguish between terms and values (which are usually considered
  either two disjoint syntactic categories, or as two syntactic categories in
  a subset relationship). The following kinds are used for this purpose:

    [TERM]  classifies terms (also known as programs)
    [SOUP]  classifies concurrent programs made of one or more threads.

  Finally, we use the kind [ROW kappa] in order to represent lists of things
  of kind [kappa]. *)

Inductive kind :=
  | VALUE
  | TERM
  | SOUP
  | PERM
  | TYPE
  | ROW : kind -> kind
.

(* ---------------------------------------------------------------------------- *)

(* The syntax of terms and types. *)

Inductive term :=

(* The following case is common to terms and types. *)

(* A variable has an arbitrary kind. Thus, it can be a term variable (one that
   exists at runtime and will be bound to a value) or a type variable (one
   that does not exist at runtime) of arbitrary kind. *)

  | TVar: nat -> term

(* The following cases are terms. *)

(* If [t] is a term, then [TAbs t] is a term as well. It represents a
   lambda-abstraction. *)

  | TAbs: term -> term

(* If [t1] is a value and [t2] is a term, then [TApp t1 t2] is a term. It
   represents a function application. *)

  | TApp: term -> term -> term

(* [fail] is a term. It represents a runtime failure. It is well-typed in
   every context. *)

  | TFail: term

(* If [t] is a term, then [TThread t] is a soup. *)

  | TThread: term -> term

(* If [t] and [u] are two soups, then [TPar t u] is their parallel
   composition, which is still a soup. *)

  | TPar: term -> term -> term

(* If [t1] and [t2] are values, then [TSpawn t1 t2] is a term. Its
   semantics is to execute the function application [TApp t1 t2]
   in a new thread. *)

  | TSpawn: term -> term -> term

(* [TGhost] is a term. It reduces in one step to a unit value. It is used as a
   syntactic marker for a ghost instruction, e.g., freezing a memory block. *)

  | TGhost: term

(* If [l] is a memory location, then [TLoc l] is a value. *)

  | TLoc: nat -> term

(* If [t] is a value, then [TNewRef t] is a term. It is the operation of
   allocating a fresh memory cell, initialized with [t]. *)

  | TNewRef: term -> term

(* If [t] is a value, then [TRead t] is a term. It is the operation of reading
   the reference whose address is [t]. *)

  | TRead: term -> term

(* If [t] and [u] are values, then [TWrite t u] is a term. It is the operation
   of writing the value [u] to the memory block whose address is [t]. *)

  | TWrite: term -> term -> term

(* If [lck] is a lock address, then [TLock lck] is a value. *)

  | TLock: nat -> term

(* [TNewLock] is a lock allocation expression. *)

  | TNewLock: term

(* If [t] is a value, then [TAcquire t] is a term. *)

  | TAcquire: term -> term

(* If [t] is a value, then [TRelease t] is a term. *)

  | TRelease: term -> term

(* If [t1] and [t2] are values, then [TGive t1 t2] is a term. It is the
   operation of giving [t1] to [t2], or in other words, having [t2]
   adopt [t1]. *)

  | TGive: term -> term -> term

(* If [t1] and [t2] are values, then [TTake t1 t2] is a term. It is the act of
   taking [t1] from [t2], or in other words, having [t2] abandon [t1]. *)

  | TTake: term -> term -> term

(* The [take] operation that is available to the programmer is [TTake]. The
   construct [TTakeCommitted] represents the intermediate state that is
   reached after checking the adopter pointer and before updating this
   pointer. We introduce this intermediate state so as to prove that [take]
   need not be implemented in an atomic manner, even in a concurrent
   setting. *)

  | TTakeCommitted: term -> term -> term

(* The following cases are types. *)

(* If [t] has kind [VALUE], then [TySingleton t] is a type. The concrete
   syntax is [=t]. *)

  | TySingleton: term -> term

(* If [ty1] and [ty2] are types, then [TyArrow ty1 ty2] is a type. This
   is a standard function type. The concrete syntax is [t1 -> t2]. *)

  | TyArrow: term -> term -> term

(* Universal types are used (quite obviously) to encode the universal
   types of the surface language. They are also (less obviously) used
   to encode the ``dependent types'' of the source language. *)

  | TyForall: kind -> term -> term

(* Existential types. *)

  | TyExists: kind -> term -> term

(* If [ty] is a type and [P] is a permission, then [TyMixStar ty P] is a type.
   The concrete syntax is [ty | P]. This type is typically used to express
   function pre- and post-conditions. It also plays a role in the frame rule. *)

  | TyMixStar: term -> term -> term

(* If [t] has kind [VALUE] and [ty] is a type, then [TyAnchored t ty] is a
   permission, which we say is ``anchored'' at [t]. The concrete syntax is [t
   @ ty]. *)

  | TyAnchored: term -> term -> term

(* [TyEmpty] is the empty permission. It is a neutral element for [TyStar]. *)

  | TyEmpty: term

(* [TyStar] is the conjunction of permissions. *)

  | TyStar: term -> term -> term

(* If [T] is a type or permission, then [TyDuplicable T] is a permission,
   which asserts that [T] is duplicable. *)

  | TyDuplicable: term -> term

(* [RowNil] is the empty row. *)

  | RowNil: term

(* [RowCons] builds a non-empty row. *)

  | RowCons: term -> term -> term

(* [TyRef m T] is the type of a memory block whose field currently has type [T].
   The mode [m] indicates whether the block is mutable or immutable. If the
   former, then this type implies unique ownership of the block; if the latter,
   then the block may be shared. *)

  | TyRef: mode -> term -> term

(* If [P] is a permission, then [TyLock P] is a type. It is the type of
   a lock that protects the invariant [P]. *)

  | TyLock: term -> term

(* [TyLocked] is a type. It is the type of a lock that has been taken
   by the current thread. *)

  | TyLocked: term

(* [TyAdoptable] is a type. It is the type of a memory location that
   has an adopter field. This type used to be known as [TyDynamic].
   It is duplicable. *)

  | TyAdoptable: term

(* [TyUnadopted] is a type. It is the type of a memory location that has a
   [Null] adopter field. It is affine: it represents the ownership of this
   memory location in its role as an adoptee. *)

  | TyUnadopted: term

(* If [U] is a type, then [TyAdopts U] is a type. It is the type of a
   memory location that has adoptees of type [U]. It is affine: it
   represents the ownership of this memory location in its role as
   an adopter, and the ownership of the adoptees. *)

  | TyAdopts: term -> term

.

(* ---------------------------------------------------------------------------- *)

(* We don't need a sequencing construct, because we can use the following
   syntactic sugar instead. *)

Notation TLet u t :=
  (TApp (TAbs t) u).

(* If [v1] and [v2] have kind [VALUE], then [TyEquation v1 v2] is
   a permission. The concrete syntax is [v1 = v2]. It is a short-hand
   for the anchored permission [v1 @ =v2]. *)

Notation TyEquation v1 v2 :=
  (TyAnchored v1 (TySingleton v2)).

(* Some reduction rules, such as [RedSpawn] and [RedGhost], need to produce a
   dummy value as a result. We do not have a primitive unit value, so we use
   a fixed arbitrary value (here, a lambda-abstraction) for this purpose. *)

Notation TUnit :=
  (TAbs (TVar 0)).

(* The type [TyTop], defined as [{v : VALUE}.(=v)], also serves as the type
   of the unit value above. *)

Notation TyTop :=
  (TyExists VALUE (TySingleton (TVar 0))).

(* The type [TyBottom] is the least type. It is uninhabited. *)

Notation TyBottom :=
  (TyForall TYPE (TVar 0)).

(* The permission [TyFalse], defined as [forall p.p], is unsatisfiable. *)

Notation TyFalse :=
  (TyForall PERM (TVar 0)).

(* ---------------------------------------------------------------------------- *)

(* The binding structure of terms and types. *)

(* Because term variables and type variables form a single namespace, the
   definition of straightforward. Every binder counts. *)

Instance Var_term : Var term := {
  var := TVar (* avoid eta-expansion *)
}.

Fixpoint traverse_term (f : nat -> nat -> term) l t :=
  match t with
  | TVar x =>
      f l x
  | TAbs t =>
      TAbs (traverse_term f (1 + l) t)
  | TApp t1 t2 =>
      TApp (traverse_term f l t1) (traverse_term f l t2)
  | TFail =>
      TFail
  | TThread t =>
      TThread (traverse_term f l t)
  | TPar t1 t2 =>
      TPar (traverse_term f l t1) (traverse_term f l t2)
  | TSpawn v1 v2 =>
      TSpawn (traverse_term f l v1) (traverse_term f l v2)
  | TLoc l =>
      TLoc l
  | TNewRef t =>
      TNewRef (traverse_term f l t)
  | TRead t =>
      TRead (traverse_term f l t)
  | TWrite t u =>
      TWrite (traverse_term f l t) (traverse_term f l u)
  | TGhost =>
      TGhost
  | TLock lck =>
      TLock lck
  | TNewLock =>
      TNewLock
  | TAcquire t =>
      TAcquire (traverse_term f l t)
  | TRelease t =>
      TRelease (traverse_term f l t)
  | TGive t1 t2 =>
      TGive (traverse_term f l t1) (traverse_term f l t2)
  | TTake t1 t2 =>
      TTake (traverse_term f l t1) (traverse_term f l t2)
  | TTakeCommitted t1 t2 =>
      TTakeCommitted (traverse_term f l t1) (traverse_term f l t2)
  | TyForall kappa T =>
      TyForall kappa (traverse_term f (1 + l) T)
  | TyExists kappa T =>
      TyExists kappa (traverse_term f (1 + l) T)
  | TySingleton T =>
      TySingleton (traverse_term f l T)
  | TyArrow T1 T2 =>
      TyArrow (traverse_term f l T1) (traverse_term f l T2)
  | TyMixStar T1 T2 =>
      TyMixStar (traverse_term f l T1) (traverse_term f l T2)
  | TyAnchored T1 T2 =>
      TyAnchored (traverse_term f l T1) (traverse_term f l T2)
  | TyEmpty =>
      TyEmpty
  | TyStar T1 T2 =>
      TyStar (traverse_term f l T1) (traverse_term f l T2)
  | TyDuplicable T =>
      TyDuplicable (traverse_term f l T)
  | RowNil =>
      RowNil
  | RowCons T1 T2 =>
      RowCons (traverse_term f l T1) (traverse_term f l T2)
  | TyRef m T =>
      TyRef m (traverse_term f l T)
  | TyLock T =>
      TyLock (traverse_term f l T)
  | TyLocked =>
      TyLocked
  | TyAdoptable =>
      TyAdoptable
  | TyUnadopted =>
      TyUnadopted
  | TyAdopts U =>
      TyAdopts (traverse_term f l U)
  end.

Instance Traverse_term : Traverse term term := {
  traverse := traverse_term
}.

Instance TraverseVarInjective_term : @TraverseVarInjective term _ term _.
Proof.
  constructor. prove_traverse_var_injective.
Qed.

Instance TraverseFunctorial_term : @TraverseFunctorial term _ term _.
Proof.
  constructor. prove_traverse_functorial.
Qed.

Instance TraverseRelative_term : @TraverseRelative term term _.
Proof.
  constructor. prove_traverse_relative.
Qed.

Instance TraverseIdentifiesVar_term : @TraverseIdentifiesVar term _ _.
Proof.
  constructor. prove_traverse_identifies_var.
Qed.

Instance TraverseVarIsIdentity_term : @TraverseVarIsIdentity term _ term _.
Proof.
  constructor. prove_traverse_var_is_identity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Terms have decidable equality. *)

Lemma eq_term_dec:
  forall t1 t2 : term,
  { t1 = t2 } + { t1 <> t2 }.
Proof.
  repeat decide equality.
Qed.

Program Instance equiv_dec_term: EqDec term eq.
Next Obligation.
  eapply eq_term_dec.
Qed.

