Set Implicit Arguments.
Require Import MyList.
        Import ListNotations.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Zap.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.

(* ---------------------------------------------------------------------------- *)

(* In the following, there are two variants of the main typing judgement, [j].
   In the [Canonical] variant, the use of the permission subsumption rule
   [JSubLeft] and of the subtyping rule [JSubRight] is forbidden. In the
   [Unrestricted] variant, it is allowed. The canonical variant is used to
   define the semantic interpretation of permissions, which gives rise to a
   semantic notion of permission subsumption. The syntactic notion of
   permission subsumption, represented by the judgements [jps] and [sub], is
   then proved sound with respect to this semantic notion. The use of [jps]
   and [sub] is allowed, via [JSubLeft] and [JSubRight], in [Unrestricted]
   type derivations. *)

Inductive flag :=
  | Canonical (* subsumption disabled *)
  | Unrestricted (* subsumption enabled *)
.

(* ---------------------------------------------------------------------------- *)

(* Permission subsumption. *)

(* The subsumption judgements [jps], [sub], and [row_sub] are mutually
   recursive. The typing judgements [j], [jbs], and [jpi], are mutually
   recursive (and depend on the subsumption judgements). We make all of these
   judgements mutually inductive, because this allows us to define a combined
   induction scheme, which is convenient. *)

(* The judgements [jps] and [sub] carry an integer index, which measures the
   height of the subtyping derivation. The proof of soundness of subsumption
   involves a well-founded induction over this index. As far as the end user
   is concerned, this index has no meaning. *)

(* The judgement [jps] takes the form [K |- P <= Q].

   This judgement means that the permissions [P] can be re-arranged to yield
   the permissions [Q].

   The set of axioms and rules that defines [jps] is not minimal: some rules
   can be derived from others. This is not a problem: it is easier to prove
   these rules sound directly than to prove that they can be derived from a
   number of other rules. Indeed, we will prove that [jps] is sound with
   respect to a semantic notion of permission subsumption. *)

Inductive jps : nat -> env kind -> term -> term -> Prop :=

(* Permission subsumption is reflexive and transitive. *)

| JpsReflexive:
    forall n K P,
    jps n K P P

| JpsTransitive:
    forall n K P1 P2 P3,
    jps n K P1 P2 ->
    jps n K P2 P3 ->
    jps (S n) K P1 P3

(* Every permission is affine. This is indicated by the following
   rule, which states that every permission [P] can be replaced
   with [empty]. In other words, [empty] is the top element in
   the ordering of permissions. *)

| JpsEmptyTop:
    forall n K P,
    jps n K
      P
      TyEmpty

(* [TyEmpty] is a left and right unit for [TyStar]. This is indicated
   by the following four permission subsumption rules, which state
   that [P * empty] and [empty * P] are interconvertible with [P]. *)

| JpsEmptyAppearsLeft:
    forall n K P,
    jps n K
      P
      (TyStar TyEmpty P)

| JpsEmptyAppearsRight:
    forall n K P,
    jps n K
      P
      (TyStar P TyEmpty)

| JpsDisappearsLeft:
    forall n K P Q,
    jps n K
      (TyStar P Q)
      Q

| JpsDisappearsRight:
    forall n K P Q,
    jps n K
      (TyStar P Q)
      P

(* Conjunction is commutative and associative. *)

| JpsStarCommutative:
    forall n K P1 P2,
    jps n K
      (TyStar P1 P2)
      (TyStar P2 P1)

| JpsStarAssociative:
    forall n K P1 P2 P3,
    jps n K
      (TyStar P1 (TyStar P2 P3))
      (TyStar (TyStar P1 P2) P3)

(* Equality is reflexive. That is, if [v] is a value, a permission of the form
   [v = v] can spontaneously appear during a permission subsumption step. I
   use [TyEmpty] on the left-hand side, even though an arbitrary permission [P]
   could be used here, because this makes it clearer that the equation appears
   out of nowhere. *)

| JpsEqualityReflexive:
    forall n K v,
    wk K v VALUE ->
    jps n K TyEmpty (TyEquation v v)

(* Equality is symmetric and transitive. *)

(* I believe that these rules could be derived out of [JpsReflexive],
   [JpsEqualsForEquals], and the rules that govern [TyEmpty] and [TyStar].
   However, as explained above, it is less costly for us to give these rules
   than to derive them. *)

| JpsEqualitySymmetric:
    forall n K v1 v2,
    jps n K
      (TyEquation v1 v2)
      (TyEquation v2 v1)

| JpsEqualityTransitive:
    forall n K v1 v2 v3,
    jps n K
      (TyStar (TyEquation v1 v2) (TyEquation v2 v3))
      (TyEquation v1 v3)

(* Equals can be substituted for equals. *)

(* This is stated as follows: in the presence of the equation [v1 = v2],
   the permissions [P(v1)] and [P(v2)] are interconvertible. Here, [v1]
   and [v2] are logical values, and [P] is a permission context with a
   hole of kind [VALUE]. *)

(* For convenience, we give multiple variants of this rule, including variants
   where we go from [v2] to [v1] instead of from [v1] to [v2] and variants
   where the equation is dropped. *)

| JpsEqualsForEquals:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v1 v2) P1)
      (TyStar (TyEquation v1 v2) P2)

| JpsEqualsForEqualsReverse:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v2 v1) P1)
      (TyStar (TyEquation v2 v1) P2)

| JpsEqualsForEqualsDrop:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v1 v2) P1)
      P2

| JpsEqualsForEqualsReverseDrop:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v2 v1) P1)
      P2

(* A duplicable permission can be duplicated. *)

| JpsDuplicate:
    forall n K P,
    jps n K
      (TyStar (TyDuplicable P) P)
      (TyStar P P)

(* A function that requires a duplicable permission [P] can be turned into a
   function that does NOT require this permission, provided [P] is available
   now. *)

| JpsHideDuplicablePrecondition:
    forall n K v T U P,
    jps n K
      (TyStar (TyAnchored v (TyArrow (TyMixStar T P) U)) (TyStar (TyDuplicable P) P))
      (TyAnchored v (TyArrow T U))

(* A package of a type and a permission can be implicitly introduced
   and eliminated. That is, the permissions [x @ (T | P)] and
   [(x @ T) * P] are inter-convertible. *)

| JpsMixStarIntro:
    forall n K v T P,
    jps n K
      (TyStar (TyAnchored v T) P)
      (TyAnchored v (TyMixStar T P))

| JpsMixStarElim:
    forall n K v T P,
    jps n K
      (TyAnchored v (TyMixStar T P))
      (TyStar (TyAnchored v T) P)

(* An existential permission can be introduced via the standard rule. *)

| JpsExistsIntro:
    forall n K kappa T P Q,
    subst T 0 Q = P ->
    wk K T kappa ->
    wk (insert 0 kappa K) Q PERM ->
    substitutable kappa ->
    jps n K
      P
      (TyExists kappa Q)

(* An existential permission can be hoisted out of a conjunction. *)

| JpsExistsHoistLeft:
    forall n K kappa P Q Q',
    shift 0 Q = Q' ->
    jps n K
      (TyStar (TyExists kappa P) Q)
      (TyExists kappa (TyStar P Q'))

(* A permission at a universal type is equivalent to a universal permission. *)

| JpsForallType2Perm:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyAnchored v (TyForall kappa T))
      (TyForall kappa (TyAnchored v' T))

| JpsForallPerm2Type:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyForall kappa (TyAnchored v' T))
      (TyAnchored v (TyForall kappa T))

| JpsExistsType2Perm:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyAnchored v (TyExists kappa T))
      (TyExists kappa (TyAnchored v' T))

| JpsExistsPerm2Type:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyExists kappa (TyAnchored v' T))
      (TyAnchored v (TyExists kappa T))

(* A universal permission can be eliminated. *)

(* The analogous rule for universal types can be derived; see [DerivedSubtyping]. *)

(* TEMPORARY re-read these rules and think about redundancy/completeness *)

| JpsForallElimPerm:
    forall n K kappa P U P',
    subst U 0 P = P' ->
    wk K U kappa ->
    jps n K
      (TyForall kappa P)
      P'

(* A vacuous universal quantifier can be introduced. *)

| JpsVacuousForallIntro:
    forall n K kappa P P',
    substitutable kappa ->
    shift 0 P = P' ->
    jps n K
      P
      (TyForall kappa P')

(* The frame rule can also be formulated as a subtyping axiom, as follows. *)

| JpsFrame:
    forall n K v T1 T2 P,
    wk K P PERM ->
    jps n K
      (TyAnchored v (TyArrow T1 T2))
      (TyAnchored v (TyArrow (TyMixStar T1 P) (TyMixStar T2 P)))

(* Structural permissions can be decomposed and recomposed. *)

| JpsDecomposeRef:
    forall n K v v' m T T',
    shift 0 v = v' ->
    shift 0 T = T' ->
    jps n K
      (TyAnchored v (TyRef m T))
      (TyExists VALUE
        (TyStar
          (TyAnchored v' (TyRef m (TySingleton (TVar 0))))
          (TyAnchored (TVar 0) T')
        )
      )

| JpsRecomposeRef:
    forall n K v1 v2 m T,
    jps n K
      (TyStar
        (TyAnchored v1 (TyRef m (TySingleton v2)))
        (TyAnchored v2 T)
      )
      (TyAnchored v1 (TyRef m T))

(* If we have two names for the content of a single memory location, then
   these names denote the same value. *)

(* This rule is interesting in the case of two immutable blocks. If at least
   one of the blocks is mutable, then the rule is still sound, but too weak,
   as one can conclude in that case that we have a contradiction. *)

| JpsUnifyRef:
    forall n K v m1 m2 v1 v2 P,
    TyStar
        (TyAnchored v (TyRef m1 (TySingleton v1)))
        (TyAnchored v (TyRef m2 (TySingleton v2)))
    = P ->
    jps n K
      P
      (TyStar P (TyEquation v1 v2))

(* [TyUnadopted] gives rise to [TyAdoptable], and is not consumed in the
   process. *)

| JpsUnadoptedAdoptable:
    forall n K v,
    jps n K
      (TyAnchored v TyUnadopted)
      (TyStar
        (TyAnchored v TyUnadopted)
        (TyAnchored v TyAdoptable))

(* Functions are contravariant in their domain and covariant in their
   codomain. This rule (and others like it) make the judgements [jps]
   and [sub] mutually recursive. *)

(* The rule [JpsCoArrow] is a little bit special in that it is
   considered to have zero height, regardless of the height of
   its premises. This is not important -- we could use [S m]
   in the conclusion instead of [n] -- but it emphasizes the
   fact that the proof of the lemma [jps_sub_sound] does not
   exploit the induction hypothesis in the case of [JpsCoArrow]. *)

| JpsCoArrow:
    forall m n K v T1 T2 U1 U2,
    sub m K U1 T1 ->
    sub m K T2 U2 ->
    wk K U1 TYPE ->
    jps n K
      (TyAnchored v (TyArrow T1 T2))
      (TyAnchored v (TyArrow U1 U2))

(* The type constructor [TyAnchored] is covariant. This is in fact stated
   by the rule [Sub] below, which defines [sub] in terms of [jps]. See
   the lemma [JpsCoAnchored] in the file [DerivedSubtyping]. *)

(* Permission conjunction is covariant. *)

| JpsCoStar:
    forall n K P1 P2 Q1 Q2,
    jps n K P1 P2 ->
    jps n K Q1 Q2 ->
    jps (S n) K (TyStar P1 Q1) (TyStar P2 Q2)

(* Quantifiers are covariant. *)

| JpsCoForall:
    forall n K kappa P Q,
    jps n (insert 0 kappa K) P Q ->
    jps (S n) K (TyForall kappa P) (TyForall kappa Q)

| JpsCoExists:
    forall n K kappa P Q,
    jps n (insert 0 kappa K) P Q ->
    jps (S n) K (TyExists kappa P) (TyExists kappa Q)

(* References are covariant. *)

| JpsCoRef:
    forall n K v m T U,
    sub n K T U ->
    jps (S n) K
      (TyAnchored v (TyRef m T))
      (TyAnchored v (TyRef m U))

(* The type constructor [TyAdopts] is covariant, which means that it is
   permitted to forget information about the type of the adoptees. *)

| JpsCoAdopts:
    forall n K v T U,
    sub n K T U ->
    jps (S n) K
      (TyAnchored v (TyAdopts T))
      (TyAnchored v (TyAdopts U))

(* Each of the rules that define the judgement [dup] (see [Dup.v]) gives
   rise to a corresponding subtyping rule, which allows establishing an
   assertion of the form [TyDuplicable _]. *)

(* It would be nice if we could give just one generic subtyping rule,
   perhaps along the following lines: if [forall T U, dup T -> dup U],
   then [TyDuplicable T] subsumes [TyDuplicable U]. However, this rule
   exhibits some problems: 1- it is not general enough, as it has fixed
   arity 1; in general, we may need zero or more than one premise; 2-
   it is in fact unsound, because it is not stable by substitution; we
   would probably need to modify it along the following lines: if for
   all [T], [U] and for all substitutions [phi], [dup (phi T)] implies
   [dup (phi U)], then, etc. This sounds heavy, so I do not pursue this
   approach. *)

(* The following rules are construction rules for [TyDuplicable]. They are
   reversible: we could also provide matching destruction rules. However,
   these destruction rules would be less useful (the Mezzo type-checker does
   not currently implement them) so I do not bother. *)

| JpsDupForall:
    forall n K kappa P,
    jps n K
      (TyForall kappa (TyDuplicable P))
      (TyDuplicable (TyForall kappa P))

(* The rule for existential types (below) is sound but not fully satisfactory.
   It cannot prove that the type [exists a. (a | duplicable a)] is duplicable,
   even though it intuitively is. I suppose the fault lies in the definition
   of the predicate [dup], which would need be refined. *)

| JpsDupExists:
    forall n K kappa P,
    jps n K
      (TyForall kappa (TyDuplicable P))
      (TyDuplicable (TyExists kappa P))

| JpsDupSingleton:
    forall n K v,
    wk K v VALUE ->
    jps n K
      TyEmpty
      (TyDuplicable (TySingleton v))

| JpsDupArrow:
    forall n K T U,
    wk K T TYPE ->
    wk K U TYPE ->
    jps n K
      TyEmpty
      (TyDuplicable (TyArrow T U))

| JpsDupMixStar:
    forall n K T P,
    wk K T TYPE ->
    wk K P PERM ->
    jps n K
      (TyStar (TyDuplicable T) (TyDuplicable P))
      (TyDuplicable (TyMixStar T P))

| JpsDupAnchored:
    forall n K v T,
    wk K v VALUE ->
    wk K T TYPE ->
    jps n K
      (TyDuplicable T)
      (TyDuplicable (TyAnchored v T))

| JpsDupEmpty:
    forall n K,
    jps n K
      TyEmpty
      (TyDuplicable TyEmpty)

| JpsDupStar:
    forall n K P Q,
    wk K P PERM ->
    wk K Q PERM ->
    jps n K
      (TyStar (TyDuplicable P) (TyDuplicable Q))
      (TyDuplicable (TyStar P Q))

| JpsDupDup:
    forall n K T kappa,
    wk K T kappa ->
    logic kappa ->
    jps n K
      TyEmpty
      (TyDuplicable (TyDuplicable T))

| JpsDupNil:
    forall n K,
    jps n K
      TyEmpty
      (TyDuplicable RowNil)

| JpsDupCons:
    forall n K T Ts kappa,
    wk K T kappa ->
    wk K Ts (ROW kappa) ->
    logic kappa ->
    jps n K
      (TyStar (TyDuplicable T) (TyDuplicable Ts))
      (TyDuplicable (RowCons T Ts))

| JpsDupLock:
    forall n K P,
    wk K P PERM ->
    jps n K
      TyEmpty
      (TyDuplicable (TyLock P))

| JpsDupRef:
    forall n K T,
    wk K T TYPE ->
    jps n K
      (TyDuplicable T)
      (TyDuplicable (TyRef Duplicable T))

| JpsDupAdoptable:
    forall n K,
    jps n K
      TyEmpty
      (TyDuplicable TyAdoptable)

(* ---------------------------------------------------------------------------- *)

(* Subtyping is defined in terms of permission subsumption. *)

(* If, for a fresh variable [x], the permission [x @ T1] can be transformed
   to [x @ T2], then the type [T1] is considered a subtype of [T2]. *)

(* In other words, this rule temporarily invents a fresh name [x] for the
   root value, and this allows us to think in terms of permissions instead
   of in terms of types. This is handy because we already have many rules
   for reasoning about permission subsumption. *)

with sub : nat -> env kind -> term -> term -> Prop :=

(* K; x : VALUE |- (x @ T) <= (x @ U)
   ----------------------------------------
   K |- T <= U *)

| Sub:
    forall n K T U,
    jps n
      (insert 0 VALUE K)
      (TyAnchored (TVar 0) (shift 0 T))
      (TyAnchored (TVar 0) (shift 0 U)) ->
    sub (S n) K T U

(* ---------------------------------------------------------------------------- *)

(* Subtyping of rows is defined pointwise. *)

with row_sub : nat -> env kind -> term -> term -> Prop :=

| RowSubNil:
    forall n K,
    row_sub n K RowNil RowNil

| RowSubCons:
    forall n K T U Ts Us,
    sub n K T U ->
    row_sub n K Ts Us ->
    row_sub (S n) K (RowCons T Ts) (RowCons U Us)

(* ---------------------------------------------------------------------------- *)

(* The typing judgements. *)

(* There are two main judgements, only one of which is needed when
   type-checking source programs.

     [j] concerns values and terms.
       We write [jc] for the canonical variant, which is used in the
       interpretation of permissions, and [jt] for the unrestricted
       variant, which is used to type-check source values and terms.
     [jpi] defines the semantic interpretation of permissions.

   [j] depends on [jpi] because a semantic pre-condition (a resource) can be
   turned into a syntactic pre-condition (a permission): [JCut].

   [jpi] depends on [jc] because the permission [v @ T] is considered true
   when the value [v] has type [T]: [JpiAtomic]. *)

(* The judgement [j] takes the form [R, K |- { P } t : T], where [R] is a
   resource, [K] is a kind environment (which maps variables to kinds), [t] is
   a term, [P] is a permission, and [T] is a type.

   The components [R] and [P] play the role of pre-conditions. [R] can be viewed
   as a semantic pre-condition, a resource that must be held for the term to be
   well-typed; and [P] can be viewed as a syntactic pre-condition, a permission
   that must be held for the term to be well-typed. Permissions are ultimately
   interpreted in terms of resources, so [R] and [P] serve analogous purposes,
   and indeed we provide a rule, [JCut], which allows turning a syntactic
   pre-condition into a semantic one (or vice-versa, depending on whether the rule
   is read bottom up or top down).

   The type [T] is a post-condition: it describes the value produced by
   executing [t] as well as the permissions that exist at the end of the
   execution.

   Instead of a type [T] on the right-hand side, one could prefer to use a
   permission [Q], which would serve as a post-condition, and where one more
   variable would be bound, allowing the post-condition to refer to the result
   of evaluating [t]. The two options should ultimately be equivalent, and
   this one seems simpler. *)

(* ---------------------------------------------------------------------------- *)

(* Typing rules for values and terms. *)

(* The judgement [j] carries an integer index [n], which measures the height
   of the typing derivation. The proof of canonicalization involves a
   well-founded induction over this index. As far as the end user is
   concerned, this index has no meaning. *)

with j : flag -> nat -> resource -> env kind -> term -> term -> term -> Prop :=

(* An introduction rule for [TySingleton]. *)

(* If [v] is a value, then [v] admits the singleton type [=v]. *)

  | JSingleton:
      forall n f R K P v1 v2,
      v1 = v2 ->
      wk K v2 VALUE ->
      j f n R K P v1 (TySingleton v2)

(* An introduction rule for [TyArrow]. *)

(* A lambda-bound variable has kind [VALUE]. This indicates that it is
   a value and that it exists at runtime. *)

(* A permission [x @ T1] for the formal parameter [x] is available in the
   function body [t]. If a duplicable permission [P] is available outside the
   function body, it is also available inside. The resource [R] that is
   available outside is replaced with [core R] inside. A closure can capture
   only duplicable resources and permissions. *)

(* Inside the function body, the use of permission subsumption is allowed: the
   inside flag is [Unrestricted], regardless of the outside flag [f]. *)

  | JFunction:
      forall n f R K P T t U,
      wk K T TYPE ->
      j
        Unrestricted
        n
        (core R)
        (insert 0 VALUE K)
        (intro P T)
        t
        (shift 0 U)
      ->
      j f (S n) R K (TyStar (TyDuplicable P) P) (TAbs t) (TyArrow T U)

(* An introduction rule for [TyForall]. *)

(* The introduction of a universal quantifier is restricted to harmless
   terms. This is a class of expressions whose execution definitely
   cannot allocate a new lock. *)

  | JForallIntro:
      forall f n R K P t kappa T,
      harmless t ->
      j f n R (insert 0 kappa K) P (shift 0 t) T ->
      j f (S n) R K (TyForall kappa P) t (TyForall kappa T)

(* The elimination of [TyForall] is performed by subsumption. *)

(* An introduction rule for [TyExists]. *)

(* We restrict this introduction rule to values. This is not essential,
   but simplifies our life by removing one case in the proof of subject
   reduction. There is no loss of expressiveness, since [TyExists] can
   be introduced via subtyping: [JpsExistsIntro] and [SubExistsIntro]. *)

(* This rule, restricted to values, must be present precisely because it
   serves as a justification for the subtyping rule [JpsExistsIntro].
   Without it, we would not be able to prove [sub_sound] for this rule. *)

  | JExistsIntro:
      forall f n R K kappa P U v T,
      j f n R K P v (subst U 0 T) ->
      wk K U kappa ->
      wk (insert 0 kappa K) T TYPE ->
      substitutable kappa ->
      wk K v VALUE ->
      j f (S n) R K P v (TyExists kappa T)

(* An elimination rule for [TyExists]. *)

(* Eliminating an existential permission introduces a new name. For this
   reason, it cannot be considered a special case of [JSubLeft]. *)

  | JExistsElim:
      forall n R K kappa P t T,
      j Unrestricted n R (insert 0 kappa K) P (shift 0 t) (shift 0 T) ->
      j Unrestricted (S n) R K (TyExists kappa P) t T

(* An introduction rule for [TyMixStar]. *)

(* [TyMixStar] can be introduced and eliminated via permission subsumption:
   see [JpsMixStarIntro] and [JpsMixStarElim]. Nevertheless, we do need an
   introduction rule for [TyMixStar] as part of the judgement [j]. Indeed, we
   must prove that permission subsumption is sound with respect to [jpi] (see
   the lemma [jps_sound]), and because the definition of [jpi] relies on [j],
   we need [j] to justify that a value [v] can have a type of the form
   [TyMixStar T P]. *)

(* Furthermore, this rule is useful when type-checking source programs: it is
   the frame rule. *)

  | JFrame:
      forall n f R K P Q t T,
      j f n R K P t T ->
      j f (S n) R K (TyStar P Q) t (TyMixStar T Q)

(* The judgement [j] is parameterized with a resource [R] and a permission
   [P], which both serve as pre-conditions. There is a communication between
   the two: (part of) the semantic pre-condition [R] can be used to discharge
   (part of) the syntactic pre-condition [P]. *)

(* This rule is required in order to type-check source programs. It is in fact
   the only branching rule, that is, the only rule that has two typing
   judgements among its premises. *)

  | JCut:
      forall n f R1 R2 R K P1 P2 t T,
      j f n R2 K (TyStar P1 P2) t T ->
      jpi R1 K P1 ->
      star R1 R2 = R ->
      j f (S n) R K P2 t T

(* At any point in a program, the currently available permissions
   can be re-organized, as per the permission subsumption rules. *)

(* This rule is enabled only when the flag is [Unrestricted]. *)

  | JSubLeft:
      forall m n R K P1 P2 t T,
      j Unrestricted n R K P2 t T ->
      jps m K P1 P2 ->
      j Unrestricted (S n) R K P1 t T

(* Subtyping on the right-hand side. *)

(* This rule is enabled only when the flag is [Unrestricted]. *)

(* If one expands the definition of [sub] in terms of [jps], one finds that
   this rule allows inventing a temporary name [x] for the return value,
   re-arranging the return permissions using [jps], and finally forgetting the
   name [x] again. *)

  | JSubRight:
      forall m n R K P t T1 T2,
      j Unrestricted n R K P t T1 ->
      sub m K T1 T2 ->
      j Unrestricted (S n) R K P t T2

(* The function application rule requires the left-hand term, [v], to be a
   value.

   There is no loss of generality in doing so; we can convert terms to
   administrative normal form before type-checking them.

   Because [v] is required to be a value, we are allowed to mention [v] within
   a type. More precisely, [v @ T -> U] is a well-formed permission, which
   states that [v] is a function. Thus, instead of using a premise that
   explicitly requires [v] to have type [T -> U], we simply make the
   permission [v @ T -> U] part of the pre-condition that appears in the
   conclusion.

   Why adopt this formulation? By placing the requirement [v @ T -> U] in the
   (syntactic) pre-condition, we avoid the need to split up the resource [R]
   between [v] and [t]. All of [R] goes to [t]. Thus, the only typing rule
   where [R] is split is [JCut]. *)

  | JApplication:
      forall n R K Q v t T U,
      wk K v VALUE ->
      j Unrestricted n R K Q t T ->
      j Unrestricted (S n) R K (TyStar (TyAnchored v (TyArrow T U)) Q) (TApp v t) U

(* The [fail] construct is well-typed in every context. It represents a
   runtime failure that the type system cannot detect. The [take] operation
   reduces to [fail] when it fails. *)

  | JFail:
      forall n R K P T,
      wk K T TYPE ->
      j Unrestricted n R K P TFail T

(* The construct [TSpawn v1 v2] requires [v1] and [v2] to be values. Like
   a function application, it requires [v1 @ T -> U] and [v2 @ T]. The
   function application is executed in a new thread. In the parent thread,
   a value of type [TyTop] is returned. *)

  | JSpawn:
      forall n R K v1 v2 T U,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      j Unrestricted n R K
        (TyStar (TyAnchored v1 (TyArrow T U)) (TyAnchored v2 T))
        (TSpawn v1 v2)
        TyTop

(* The typing rule [JRef] assigns a structural type to a memory location. *)

  | JRef:
      forall n f R1 R2 R K P l m v T,
      (* We check that [R] can be split in two fragments. *)
      star R1 R2 = R ->
      (* The value [v] must be closed, but its type [T] need not be closed. *)
      wk zilch v VALUE ->
      (* The first fragment, [R1], justifies that [v] has type [T]. *)
      jpi R1 K (TyAnchored v T) ->
      (* The second fragment, [R2], justifies the ownership of the location [l]. *)
      maps (get R2) l (mv2dx m v) ->
      (* The conclusion! *)
      j f n R K P (TLoc l) (TyRef m T)

(* A memory allocation expression receives the structural type [TyRef T] if
   the new block is initialized with a value of type [T]. *)

(* Furthermore, after the introduction of adoption & abandon, we obtain the
   ownership of the new memory location, both as an adopter and as an adoptee.
   Note that these permissions are affine, regardless of whether the block is
   mutable or immutable. *)

  | JNew:
      forall n R K P m v T I,
      wk K v VALUE ->
      wk K T TYPE ->
      TyAnchored v T = P ->
      I = intersection [
        TyRef m T;
        TyAdopts TyBottom;
        TyUnadopted
      ] ->
      j Unrestricted n R K P (TNewRef v) I

(* Reading a memory block of type [TyRef T] produces a value of type [T].
   The permission for the memory block is not consumed. The type [T] must
   be duplicable. *)

  | JRead:
      forall n R K P m v T,
      wk K v VALUE ->
      P = TyAnchored v (TyRef m T) ->
      j Unrestricted n R K
        (TyStar (TyDuplicable T) P)
        (TRead v)
        (TyMixStar T P)

(* Writing a value of type [T2] to a memory block of type [TyRef T1] changes
   the type of the memory block to [TyRef T2] and produces a unit result. *)

  | JWrite:
      forall n R K v1 v2 T1 T2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      j Unrestricted n R K
        (TyStar (TyAnchored v1 (TyRef Exclusive T1)) (TyAnchored v2 T2))
        (TWrite v1 v2)
        (TyMixStar TyTop (TyAnchored v1 (TyRef Exclusive T2)))

(* Freezing a memory block is a ghost instruction. It changes the block's mode
   from [Exclusive] to [Duplicable]. *)

  | JFreeze:
      forall n R K v T,
      j Unrestricted n R K
        (TyAnchored v (TyRef Exclusive T))
        TGhost
        (TyMixStar TyTop (TyAnchored v (TyRef Duplicable T)))

(* The value [TLock lck] has type [TyLock P] iff (according to [R]) the lock
   [lck] exists and its invariant is [P]. *)

  | JLock:
      forall n f R K Q lck P,
      wk zilch P PERM ->
      invariant (get R) lck P ->
      j f n R K Q (TLock lck) (TyLock P)

(* The value [TLock lck] has type [TyLocked] iff (according to [R]) the lock
   [lck] exists and is taken. *)

  | JLocked:
      forall n f R K Q lck,
      taken (get R) lck ->
      j f n R K Q (TLock lck) TyLocked

(* A new lock is created locked. So, one does not need to consume [P] at
   creation time. In surface syntax, the result type of this expression
   is [{ l : term } (=l | l @ lock P * l @ locked)]. *)

  | JNewLock:
      forall n R K P Q I,
      wk K P PERM ->
      I = intersection [
        TyLock P;
        TyLocked
      ] ->
      j Unrestricted n R K Q TNewLock I

(* Acquiring a lock causes the permission that it protects to appear,
   seemingly out of thin air. The precondition is [v @ lock P]. The
   postcondition is [P * v @ locked]. *)

  | JAcquire:
      forall n R K P v,
      wk K v VALUE ->
      j Unrestricted n R K
        (TyAnchored v (TyLock P))
        (TAcquire v)
        (TyMixStar TyTop (TyStar P (TyAnchored v TyLocked)))

(* Releasing a lock requires reliquinshing the lock invariant as well
   as the token that proves that the lock is held. The precondition is
   thus [P * v @ lock P * v @ locked]. The postcondition is empty. *)

  | JRelease:
      forall n R K P v,
      wk K v VALUE ->
      j Unrestricted n R K
        (TyStar
          P 
          (TyStar
            (TyAnchored v TyLocked)
            (TyAnchored v (TyLock P))))
        (TRelease v)
        TyTop

(* A memory location has type [TyAdoptable] if and only if it is adoptable,
   i.e., there is an adopter field at this address. *)

  | JAdoptable:
      forall n f R K P l,
      adoptable (get R) l ->
      j f n R K P (TLoc l) TyAdoptable

(* A memory location has type [TyUnadopted] if and only if it is unadopted,
   i.e., there is an adopter field at this address, this field is currently
   [Null], and we own this field (i.e. only we can assign it). *)

  | JUnadopted:
      forall n f R K P l,
      unadopted (get R) l ->
      j f n R K P (TLoc l) TyUnadopted

(* A memory location [y] has type [TyAdopts U] if and only if: 1- [xs]
   is a list of all adoptees of [y], and we own [xs] as adoptees and
   [y] as adopter; and (separately) 2- every member of the list [xs]
   has type [U]. *)

  | JAdopts:
      forall n f R1 R2 R K P xs y U,
      all_adoptees (get R1) xs y ->
      jpi R2 K (zap xs U) ->
      wk K U TYPE ->
      star R1 R2 = R ->
      j f n R K P (TLoc y) (TyAdopts U)

(* For [give v1 v2] to be well-typed, it must be the case that:
   1- [v1] admits the types [TyUnadopted] and [U]; and (separately)
   2- [v2] adopts objects of type [U]. *)

  | JGive:
      forall n R K v1 v2 U,
      wk K v1 VALUE ->
      wk K v2 VALUE ->      
      j Unrestricted n R K
        (TyStar (TyStar
          (TyAnchored v1 U)
          (TyAnchored v1 TyUnadopted))
          (TyAnchored v2 (TyAdopts U)))
        (TGive v1 v2)
        (TyMixStar TyTop
           (TyAnchored v2 (TyAdopts U)))

(* The instruction [take x from y] requires [x @ adoptable] and [y @ adopts U].
   The former permission is replaced with [x @ U] and [x @ unadopted], while the
   latter permission is preserved. *)

  | JTake:
      forall n R K v1 v2 U,
      wk K v1 VALUE ->
      wk K v2 VALUE ->      
      j Unrestricted n R K
        (TyStar
          (TyAnchored v1 TyAdoptable)
          (TyAnchored v2 (TyAdopts U)))
        (TTake v1 v2)
        (TyMixStar TyTop (TyStar (TyStar
          (TyAnchored v1 U)
          (TyAnchored v1 TyUnadopted))
          (TyAnchored v2 (TyAdopts U))))

(* The constructed [TTakeCommitted (TLoc x) (TLoc y)] is not available to the
   programmer. For this reason, it does not have a syntactic precondition.
   Instead, it requires [R] to justify [y @ adopts U] and, in addition,
   requires [x] to be currently adopted by [y] in [R]. This last condition
   records the fact that the runtime test performed by [TTake] has just
   succeeded. The postcondition is the same as that of a [TTake] construct. *)

  | JTakeCommitted:
      forall n R K P v1 v2 x y U,
      v1 = TLoc x ->
      v2 = TLoc y ->
      jpi R K (TyAnchored v2 (TyAdopts U)) ->
      adopted (get R) x y ->
      j Unrestricted n R K P
        (TTakeCommitted v1 v2)
        (TyMixStar TyTop (TyStar (TyStar
          (TyAnchored v1 U)
          (TyAnchored v1 TyUnadopted))
          (TyAnchored v2 (TyAdopts U))))

(* ---------------------------------------------------------------------------- *)

(* The auxiliary judgement [jbs f n R K P tag Tss ts T] is used to check a row
   of terms [ts], which represent the branches of a [TSwitch] construct,
   against a list of branches [Tss], which represent the definition of an
   algebraic data type. The pre-condition [P] is a *function* of the tag [tag]
   and fields [Ts] in each branch. *)

(* One could give a more readable definition by defining a higher-order predicate
   [in_each_branch] independently of the judgement [j]. Unfortunately, that would
   make it difficult for us to construct a mutual induction principle. *)

with jbs : flag -> nat -> resource -> env kind -> (* f, n, R, K *)
           (nat -> term -> term) ->               (* the parameterized pre-condition *)
           nat ->                                 (* the tag, which counts up *)
           list term ->                           (* the branches [Tss] *)
           term ->                                (* the branches [ts] *)
           term ->                                (* the return type [T] *)
           Prop :=

  | JbsZero:
      forall f n R K P tag T,
      jbs f n R K P tag nil RowNil T

  | JbsSucc:
      forall f n R K P tag Ts Tss t ts T,
      jbs f n R K P          (S tag) Tss         ts             T ->
      j   f n R K (P tag Ts)                     t              T ->
      jbs f n R K P             tag  (Ts :: Tss) (RowCons t ts) T

(* ---------------------------------------------------------------------------- *)

(* A semantic interpretation of permissions. *)

(* A permission is analogous to a separation logic formula: it is an assertion
   about the variables and about the current machine state, or more precisely,
   about the current resource. *)

(* One might think that we need only give meaning to closed permissions, and
   the meaning of open permissions will follow. In fact, in an early version
   of this proof, the judgement [jpi] was only about closed permissions.
   However, now that the judgement [j] depends on [jpi], we need to interpret
   open permissions as well. This seems to be necessary in order to allow a
   smooth transition (by substitution) between type derivations for source
   programs and type derivations for programs under execution.

   Because we give meaning to open permissions, the judgement [jpi] is
   parameterized with an environment [K]. This environment is effectively used
   only by [JpiAtomic], the rule that states that the permission [v @ T] if
   [v] has type [T]. There, [v] can have free variables. *)

(* A permission is an assertion about the current resource. A permission [P]
   is either true or false of a resource [R]. This is the interpretation of
   permissions. It is defined by the judgement [jpi R K P]. *)

(* This predicate is not used to type-check source programs. It plays a role
   in the type soundness proof only. *)

(* Technically, we could either build the consistency of [R] into the
   definition of [jpi], so that [jpi R K P] implies [consistent R], or not
   build it in, in which case we have to explicitly require both [jpi R K P]
   and [consistent R] in later lemmas and definitions, such as [jps_sound] and
   [jconf]. It seems that the proof of [jps_sound] is easier if we do not
   build consistency in -- even though this lemma needs a consistency
   hypothesis. *)

with jpi : resource -> env kind -> term -> Prop :=

(* The empty permission is true of an arbitrary resource. *)

| JpiEmpty:
    forall R K,
    jpi R K TyEmpty

(* A conjunction of two permissions is true of a resource [R] if and
   only [R] can be split as [star R1 R2] and each of the conjuncts is
   true of one fragment. *)

| JpiStar:
    forall R1 R2 R K P1 P2,
    jpi R1 K P1 ->
    jpi R2 K P2 ->
    star R1 R2 = R ->
    jpi R K (TyStar P1 P2)

(* Universal permissions are interpreted in the obvious manner. *)

| JpiForall:
    forall R kappa K P,
    substitutable kappa ->
    jpi R (insert 0 kappa K) P ->
    jpi R K (TyForall kappa P)

(* Existential permissions are interpreted in the obvious manner. *)

| JpiExists:
    forall R K kappa U P,
    jpi R K (subst U 0 P) ->
    wk K U kappa ->
    wk (insert 0 kappa K) P PERM ->
    substitutable kappa ->
    jpi R K (TyExists kappa P)

(* A permission of the form [v @ T], where [v] is a value, is true if
   [v] has type [T] under a permission [P] which itself is true. *)

(* We cannot require [P] to be [TyEmpty]. This would be too demanding. The
   derivation of the premise [jc _ _ P v T] must be canonical, so it cannot
   exploit [JSubLeft], and as a result, we cannot control the syntactic
   shape of [P]. I conjecture that we could add a semantic variant of
   [JSubLeft], which uses semantic subtyping, and can be used in a canonical
   derivation; that would allow us to require [P] to be [TyEmpty] here,
   and get rid of the second premise. This would simplify [JpiAtomic],
   but would introduce some noise into the system by adding one rule. *)

| JpiAtomic:
    forall n R1 R2 R K P v T,
    wk K v VALUE ->
    j Canonical n R1 K P v T ->
    jpi R2 K P ->
    star R1 R2 = R ->
    jpi R K (TyAnchored v T)

(* A permission of the form [TyDuplicable T] is true if [T] is duplicable,
   according to the (meta-)predicate [dup]. This is independent of the
   resource [R]. *)

| JpiDuplicable:
    forall R K T kappa,
    wk K T kappa ->
    logic kappa ->
    dup T ->
    jpi R K (TyDuplicable T)

.

Hint Constructors jps sub row_sub j jbs jpi : j.

Hint Constructors jpi : jpi.

Notation jt := (j Unrestricted).

Notation jc := (j Canonical).

