Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.

(* ---------------------------------------------------------------------------- *)

(* Some kinds classify "logical" things -- things that do not exist at runtime
   -- and some kinds classify "physical" things. The following function
   performs a distinction between these kinds. *)

Fixpoint logic (kappa : kind) :=
  match kappa with
  | VALUE
  | TERM
  | SOUP =>
      False
  | PERM
  | TYPE =>
      True
  | ROW kappa =>
      logic kappa
  end.

Hint Unfold logic : logic.

Lemma logic_PERM:
  logic PERM.
Proof.
  simpl. tauto.
Qed.

Hint Resolve logic_PERM : logic.
  (* This can prove a goal of the form [logic ?kappa] where [?kappa]
     is a meta-variable, so it is stronger than Hint Unfold above. *)

(* ---------------------------------------------------------------------------- *)

(* A substitutable kind is one at which we may perform substitutions. This
   includes all logic kinds, as well as the kind [VALUE]. *)

Definition substitutable kappa :=
  logic kappa \/ kappa = VALUE.

Hint Unfold substitutable : substitutable.

(* ---------------------------------------------------------------------------- *)

(* The empty environment. *)

Notation zilch :=
  (@empty kind).

(* ---------------------------------------------------------------------------- *)

(* Well-kindedness. *)

(* In several places in the syntax of terms, we require a certain sub-term to
   be a value, as opposed to an arbitrary term. This is administrative normal
   form (ANF). This restriction can be worked around by adding extra [TLet]
   bindings. It leads to a couple of simplifications: 1- the typing rules
   become simpler; 2- there are fewer evaluation contexts. *)

Inductive wk : env kind -> term -> kind -> Prop :=

(* A value is also a term. In other words, values are a subset of terms. This
   allows a value to appear directly within a term, without an explicit
   [return] construct. *)

  | WkValuesAreTerms:
      forall K v,
      wk K v VALUE ->
      wk K v TERM

(* The kind of a variable is dictated by the kind environment. *)

  | WkTVar:
      forall K x kappa,
      lookup x K = Some kappa ->
      wk K (TVar x) kappa

(* A lambda-abstraction is a value. The formal parameter [x] is considered a
   value, so it can appear within terms and within types. The body [t] must be
   a term. *)

  | WkTAbs:
      forall K t,
      wk (insert 0 VALUE K) t TERM ->
      wk K (TAbs t) VALUE

(* A function application is a term if its left-hand side is a value and its
   right-hand side is a term (ANF). *)

  | WkTApp:
      forall K v t,
      wk K v VALUE ->
      wk K t TERM ->
      wk K (TApp v t) TERM

(* [fail] is a term. *)

  | WkTFail:
      forall K,
      wk K TFail TERM

(* [TThread] encapsulates a term as a soup (composed of a single thread). *)

  | WkTThread:
      forall K t,
      wk K t TERM ->
      wk K (TThread t) SOUP

(* The parallel composition of two soups is a soup. *)

  | WkTPar:
      forall K t1 t2,
      wk K t1 SOUP ->
      wk K t2 SOUP ->
      wk K (TPar t1 t2) SOUP

(* The arguments of a [spawn] operation must be values. *)

  | WkTSpawn:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TSpawn v1 v2) TERM

(* A ghost instruction is a term. *)

  | WkTGhost:
      forall K,
      wk K TGhost TERM

(* A memory location is a value. *)

  | WkTLoc:
      forall K l,
      wk K (TLoc l) VALUE

(* The argument of a memory allocation operation is a row of values (ANF). *)

  | WkTNew:
      forall K v,
      wk K v VALUE ->
      wk K (TNewRef v) TERM

(* The argument of a read operation is a value (ANF). *)

  | WkTRead:
      forall K v,
      wk K v VALUE ->
      wk K (TRead v) TERM

(* The arguments of a write operation are values (ANF). *)

  | WkTWrite:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TWrite v1 v2) TERM

(* A lock address is a value. *)

  | WkTLock:
      forall K lck,
      wk K (TLock lck) VALUE

(* A lock allocation expression is a term, not a value. *)

  | WkTNewLock:
      forall K,
      wk K TNewLock TERM

(* An [acquire] construct is a term. *)

  | WkTAcquire:
      forall K v,
      wk K v VALUE ->
      wk K (TAcquire v) TERM

(* A [release] construct is a term. *)

  | WkTRelease:
      forall K v,
      wk K v VALUE ->
      wk K (TRelease v) TERM

(* A [give] construct is a term. *)

  | WkTGive:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TGive v1 v2) TERM

(* A [take] construct is a term. *)

  | WkTTake:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TTake v1 v2) TERM

  | WkTTakeCommitted:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TTakeCommitted v1 v2) TERM

(* The quantifiers bind a variable of substitutable kind within a thing
   of logic kind. *)

  | WkTyForall:
      forall K kappa1 kappa2 T,
      substitutable kappa1 ->
      logic kappa2 ->
      wk (insert 0 kappa1 K) T kappa2 ->
      wk K (TyForall kappa1 T) kappa2

  | WkTyExists:
      forall K kappa1 kappa2 T,
      substitutable kappa1 ->
      logic kappa2 ->
      wk (insert 0 kappa1 K) T kappa2 ->
      wk K (TyExists kappa1 T) kappa2

(* The singleton type constructor transforms a value into a type. *)

  | WkTySingleton:
      forall K v,
      wk K v VALUE ->
      wk K (TySingleton v) TYPE

(* A function type is a type if its domain and codomain are types. *)

  | WkTyArrow:
      forall K T U,
      wk K T TYPE ->
      wk K U TYPE ->
      wk K (TyArrow T U) TYPE

(* [TyMixStar] builds a conjunction of a type and a permission. The result
   is a type. *)

  | WkTyMixStar:
      forall K T P,
      wk K T TYPE ->
      wk K P PERM ->
      wk K (TyMixStar T P) TYPE

(* An anchored permission [t @ ty] is a permission if [t] is a value and [ty]
   is a type. *)

  | WkTyAnchored:
      forall K v T,
      wk K v VALUE ->
      wk K T TYPE ->
      wk K (TyAnchored v T) PERM

(* The empty permission is a permission. *)

  | WkTyEmpty:
      forall K,
      wk K TyEmpty PERM

(* The conjunction of two permissions is a permission. *)

  | WkTyStar:
      forall K P Q,
      wk K P PERM ->
      wk K Q PERM ->
      wk K (TyStar P Q) PERM

(* If [T] is a type or permission, then [TyDuplicable T] is a permission. *)

  | WkTyDuplicable:
      forall K T kappa,
      wk K T kappa ->
      logic kappa ->
      wk K (TyDuplicable T) PERM

(* The empty row is a valid row. *)

  | WkRowNil:
      forall kappa K,
      wk K RowNil (ROW kappa)

(* A nonempty row is a valid row if its head is a valid element
   and its tail is a valid row. *)

  | WkRowCons:
      forall K kappa t ts,
      wk K t kappa ->
      wk K ts (ROW kappa) ->
      wk K (RowCons t ts) (ROW kappa)

(* If [T] is a type, then [TyRef m T] is a type. *)

  | WkTyRef:
      forall K m T,
      wk K T TYPE ->
      wk K (TyRef m T) TYPE

(* If [P] is a permission, then [TyLock P] is a type. *)

  | WkTyLock:
      forall K P,
      wk K P PERM ->
      wk K (TyLock P) TYPE

(* [TyLock] is a type. *)

  | WkTyLocked:
      forall K,
      wk K TyLocked TYPE

(* [TyAdoptable] is a type. *)

  | WkTyAdoptable:
      forall K,
      wk K TyAdoptable TYPE

(* [TyUnadopted] is a type. *)

  | WkTyUnadopted:
      forall K,
      wk K TyUnadopted TYPE

(* If [U] is a type, then [TyAdopts U] is a type. *)

  | WkTyAdopts:
      forall K U,
      wk K U TYPE ->
      wk K (TyAdopts U) TYPE

.

Hint Constructors wk : wk.

(* ---------------------------------------------------------------------------- *)

(* [TyTop] is well-kinded. *)

Lemma wk_TyTop:
  forall K,
  wk K TyTop TYPE.
Proof.
  eauto 8 with wk logic substitutable.
Qed.

(* [TyFalse] is well-kinded. *)

Lemma wk_TyFalse:
  forall K,
  wk K TyFalse PERM.
Proof.
  eauto 8 with wk logic substitutable.
Qed.

Hint Resolve wk_TyTop : wk.

(* ---------------------------------------------------------------------------- *)

(* Basic properties of the well-kindedness judgement. *)

(* If the environment [K2] extends the environment [K1], then a term
   that is well-kinded under [K1] is also well-kinded under [K2]. *)

Lemma wk_agree:
  forall K1 t kappa,
  wk K1 t kappa ->
  forall K2 k,
  agree K1 K2 k ->
  length K1 <= k ->
  wk K2 t kappa.
Proof.
  induction 1; intros; eauto with wk length agree omega.
Qed.

(* A corollary of the previous lemma: a term that is well-kinded
   under the empty environment is well-kinded under an arbitrary
   environment. *)

Lemma wk_empty:
  forall K t kappa,
  wk (@empty _) t kappa ->
  wk K t kappa.
Proof.
  eauto using wk_agree with length agree.
Qed.

(* If [t] has kind [kappa2] under [K2] and [t] is also well-kinded
   under a shorter environment [K1], then [t] has kind [kappa2]
   under [K1]. *)

Lemma wk_agree_reverse:
  forall K2 t kappa2,
  wk K2 t kappa2 ->
  forall K1 k,
  agree K2 K1 k ->
  length K1 <= k ->
  forall kappa1,
  wk K1 t kappa1 ->
  wk K1 t kappa2.
Proof.
  induction 1; introv ha hl h; dependent induction h;
  eauto using (WkRowNil VALUE) with wk length agree omega.
Qed.

(* A corollary of the previous lemma. *)

Lemma wk_empty_reverse:
  forall K t kappa,
  wk K t kappa ->
  (exists kappa, wk zilch t kappa) ->
  wk zilch t kappa.
Proof.
  intros. unpack. eauto using wk_agree_reverse with length agree.
Qed.

(* A term that is well-kinded under the empty environment is closed. *)

Lemma wk_closed:
  forall t kappa,
  wk zilch t kappa ->
  closed 0 t.
Proof.
  cut (
    forall K t kappa,
    wk K t kappa ->
    forall k,
    length K <= k ->
    closed k t
  ); [ eauto | ].
  induction 1; intros; construction_closed.
Qed.

(* Well-kindedness is preserved by the insertion of a new variable. *)

Lemma wk_weakening:
  forall K T2 kappa2,
  wk K T2 kappa2 ->
  forall x1 okappa1 K',
  raw_insert x1 okappa1 K = K' ->
  wk K' (shift x1 T2) kappa2.
Proof.
  induction 1; intros; subst; simpl_lift_goal;
  try econstructor (solve [ eauto with lookup_insert insert_insert insert_concat ]).
Qed.

Hint Resolve wk_weakening : j_weakening j_substitution.

(* Well-kindedness is preserved by the removal of an unused variable. *)

Lemma wk_weakening_inverse:
  forall T2 K' x1 kappa2,
  wk K' (shift x1 T2) kappa2 ->
  forall kappa1 K,
  insert x1 kappa1 K = K' ->
  wk K T2 kappa2.
Proof.
  intro T2; introv h; dependent induction h; intros; subst;
  (* This deals with the non-syntax-directed rules -- that is,
     the sub-kinding rules. *)
  try solve [ eauto with wk ];
  (* Every other case. *)
  (* Perform case analysis on [T2]; only one case is possible, because
     the well-kindedness rule exhibits an explicit head constructor;
     exploit this to simplify [shift x1 T2]. *)
  destruct T2; try discriminate; simpl_lift_all;
  (* Decompose the equation (if possible). *)
  injections;
  (* Conclude. *)
  econstructor (solve [ eauto with lookup_insert insert_insert insert_concat ]).
  (* Because removing an unused variable amounts to replacing it with an
     arbitrary type of the appropriate kind, this lemma would be a consequence
     of [wk_substitution] below, if we were willing to exploit the fact that
     every kind is inhabited. *)
Qed.

(* Well-kindedness is preserved by substitution. *)

Lemma wk_substitution:
  forall K' T2 kappa2,
  wk K' T2 kappa2 ->
  forall x kappa1 K,
  insert x kappa1 K = K' ->
  forall T1,
  wk K T1 kappa1 ->
  wk K (subst T1 x T2) kappa2.
Proof.
  induction 1; intros; subst; simpl_subst_goal;
  (* General rule. *)
  try econstructor (solve [
    eauto using wk_weakening with insert_insert insert_concat
  ]).
  (* Case TVar. *)
  unfold subst_idx; dblib_by_cases; lookup_insert_all; eauto with wk.
Qed.

Hint Resolve wk_substitution : j_substitution.

(* Substituting an object [t1] of logical kind into an object [t2] of
   physical kind has no effect -- the result is [t2], up to a shift. *)

Lemma wk_substitution_logic:
  forall K' t2 kappa2,
  wk K' t2 kappa2 ->
  ~ logic kappa2 ->
  forall t1 x K kappa1,
  insert x kappa1 K = K' ->
  wk K t1 kappa1 ->
  logic kappa1 ->
  shift x (subst t1 x t2) = t2.
Proof.
  induction 1; intros; subst; simpl_subst_goal; simpl_lift_goal;
  try solve [
    (* If we have a hypothesis of the form [~ logic PERM] or [~ logic TYPE],
       this will contradict it. *)
    false; unfold not in *; eauto with logic
    (* And this is the general method. *)
  | eauto using wk_weakening with f_equal insert_insert
  ].
  (* TVar *)
  unfold subst_idx. dblib_by_cases; try solve [
    do 2 simpl_lift_goal; eauto with f_equal omega
  ].
  subst. lookup_insert_all. tauto. (* Contradiction. *)
Qed.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps destruct the well-kindedness judgement. *)

Ltac select_wk action :=
  (* We apply a number of lemmas at this particular point, inside
     the loop executed by [gji], because the use of these lemmas and
     the destruction of [wk] facts are mutually beneficial: each of
     them enables, in some situations, the other. *)
  match goal with
  h: wk ?K ?t ?kappa |- _ =>
    (* If [t] is a variable, then inverting this hypothesis will not
       help. We test for this situation in order to save time. *)
    first [
    is_var t; revert h |
    (* If [t] is not a variable, then look for the following situations. *)
    match type of h with
      (* Apply the lemma [wk_weakening_inverse] if possible. *)
    | wk (insert ?x _ _) (shift ?x _) _ =>
        forwards: wk_weakening_inverse; [ eexact h | reflexivity | clear h ]
      (* Some well-kindedness judgements produce two sub-goals when destructed,
         but the first one corresponds to [WkValuesAreTerms] and can be immediately
         contradicted if the term at hand is clearly not a value. *)
    | wk _ ?t TERM =>
        inversion h; [
          solve [ match goal with h: wk _ t VALUE |- _ => inversion h end ]
        | clear h; try subst
        ]
      (* Destruct the well-kindedness judgement. *)
    | _ =>
        action h
    end
    ]
  end.

Ltac wk_inversion :=
  gji select_wk.

(* ---------------------------------------------------------------------------- *)

(* This tactic helps exploit the fact that some term or type is closed. It
   rewrites [shift k T] and [subst U k T] to just [T], in the context or in
   the goal. It uses a well-kindedness fact in order to prove that [T] is
   closed. *)

(* This auxiliary tactic has a specific target, the term [t]. We assume
   that we already have a hypothesis of the form [wk zilch t _]. *)

Ltac internal_closed_target t :=
  let heq := fresh in
  assert (heq: forall x, shift x t = t); [
    solve [ eauto 6 using @closed_lift_invariant, wk_closed
                     with typeclass_instances omega ]
  | repeat rewrite heq in *; clear heq
  ];
  assert (heq: forall u x, subst u x t = t); [
    solve [ eauto 6 using @closed_subst_invariant, wk_closed
                     with typeclass_instances omega ]
  | repeat rewrite heq in *; clear heq
  ].

(* The main entry point. This one repeatedly looks for a term that is known to
   be closed, and chooses it for the target. Having to look for an appropriate
   hypothesis is painful, but preferable to blindly rewriting with the lemma
   [closed_lift_invariant], as an incorrect choice of target could then be
   made. *)

Ltac closed :=
  repeat match goal with
  | h: wk zilch ?T _ |- _ =>
      internal_closed_target T;
      revert h
  end;
  intros.

(* ---------------------------------------------------------------------------- *)

(* With each algebraic data type constructor, we associate a list of branches,
   where a branch is just a row of closed types. *)

Definition wkbs (Tss : list term) :=
  List.Forall (fun Ts => wk zilch Ts (ROW TYPE)) Tss.

Lemma wkbs_inversion_head:
  forall Ts Tss,
  wkbs (Ts :: Tss) ->
  wk zilch Ts (ROW TYPE).
Proof.
  inversion 1; subst. assumption.
Qed.

Lemma wkbs_inversion_tail:
  forall Ts Tss,
  wkbs (Ts :: Tss) ->
  wkbs Tss.
Proof.
  inversion 1; subst. assumption.
Qed.

Hint Resolve wkbs_inversion_head wkbs_inversion_tail : j_weakening j_substitution.

