Set Implicit Arguments.
Generalizable All Variables.
Require Import EquivDec.
Require Import MyTactics.
Require Import MSA.

Section Definitions.

Context `{EqDec A eq, B}.

(* ------------------------------------------------------------------------- *)

(* Duplicable-or-exclusive (DX) access rights. *)

(* A duplicable-or-exclusive access right on an object is one of: [DX_NONE],
   which means we have no access right for the object, and know nothing about
   its current state; [DX_DUPLICABLE a], which means we know that the object
   is in state [a], others may know it as well, and this state is fixed
   forever; [DX_EXCLUSIVE b], which means we have an exclusive access right
   over the object, and the object is currently in state [b]; and [DX_ERROR],
   which represents an inconsistent access right. *)

(* The descriptions associated with duplicable objects have type [A], while
   the descriptions associated with exclusive objects have type [B]. There
   is no requirement that the types [A] and [B] coincide. *)

(* Note: I initially allowed [DX_NONE] and [DX_DUPLICABLE] to co-exist. But
   this breaks [compatible_implies_common_core_left], which states that two
   compatible resources have the same core. Thus, I do not allow them to
   co-exist. If an active thread changes its access right from [DX_EXCLUSIVE]
   to [DX_DUPLICABLE], then every inactive thread will have to adapt by
   changing their access right from [DX_NONE] to [DX_DUPLICABLE]. *)

Inductive dx :=
| DX_NONE: dx
| DX_DUPLICABLE: A -> dx
| DX_EXCLUSIVE: B -> dx
| DX_ERROR: dx.

(* DX access rights form a monotonic separation algebra. *)

(* Equality over [A] must be decidable. This is exploited in the definition of
   the function [star], where one needs to determine whether the two sides
   agree on a single value [a]. *)

Obligation Tactic := idtac.

Global Program Instance msa_dx : MSA dx := {

  star p1 p2 :=
    match p1, p2 return _ with
    | DX_ERROR, _
    | _, DX_ERROR
    | DX_EXCLUSIVE _, DX_EXCLUSIVE _
    | DX_EXCLUSIVE _, DX_DUPLICABLE _
    | DX_DUPLICABLE _, DX_EXCLUSIVE _
    | DX_DUPLICABLE _, DX_NONE
    | DX_NONE, DX_DUPLICABLE _ =>
        DX_ERROR
    | DX_DUPLICABLE a1, DX_DUPLICABLE a2 =>
        (* [DX_DUPLICABLE a] is compatible with itself, and is idempotent:
           it is duplicable. *)
        if equiv_dec a1 a2 then DX_DUPLICABLE a1 else DX_ERROR
    | DX_NONE, DX_NONE =>
        (* [DX_NONE] is also idempotent. *)
        DX_NONE
    | DX_NONE, DX_EXCLUSIVE b
    | DX_EXCLUSIVE b, DX_NONE =>
        (* [DX_NONE] is compatible with, and a neutral element for,
           [DX_EXCLUSIVE b]. *)
        DX_EXCLUSIVE b
    end;

  consistent p :=
    match p return _ with
    | DX_ERROR =>
        False
    | _ =>
        True
    end;

  core p :=
    match p return _ with
    | DX_ERROR =>
        DX_ERROR
    | DX_DUPLICABLE a =>
        (* [DX_DUPLICABLE a] is preserved by [core], because it represents
           duplicable information. *)
        DX_DUPLICABLE a
    | DX_NONE
    | DX_EXCLUSIVE _ =>
        DX_NONE
    end;

  (* To a first approximation, the relation [rely] over access rights is
     equality: when a thread is inactive, the access rights that it holds do not
     change. There is one exception, though: [DX_NONE] can change to
     [DX_DUPLICABLE], which means that the active thread has changed its
     access right from exclusive to duplicable, and the passive thread must
     adapt. *)

  rely p1 p2 :=
    match p1, p2 return _ with
    | DX_NONE, DX_DUPLICABLE _ =>
        True
    | _, _ =>
        p1 = p2
    end

}.

Ltac dx :=
  intros;
  repeat match goal with R: dx |- _ => destruct R end;
  simpl in *;
  equiv_dec_by_cases;
  try congruence.

Next Obligation. (* star_commutative *)
  dx.
Qed.

Next Obligation. (* star_associative *)
  dx.
Qed.

Next Obligation. (* consistent_split_left *)
  dx.
Qed.

Next Obligation. (* core_right_unit *)
  dx.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  dx.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  dx.
Qed.

Next Obligation. (* cores_are_duplicable *)
  dx.
Qed.

Next Obligation. (* duplicable_split_left *)
  dx.
Qed.

Next Obligation. (* rely_reflexive *)
  dx.
Qed.

Next Obligation. (* rely_consistent *)
  dx.
Qed.

Next Obligation. (* rely_star *)
  intros.
  case_eq R1; case_eq R2; case_eq R'; intros;
  try solve [
    (* In most cases, [rely] implies that [R1] and [R2] are equal,
       and the result is immediate: just take [R2a] and [R2b] to
       be [R1a] and [R1b]. This solves 63 out of 64 cases. *)
    exists R1 R2; subst; equiv_dec_by_cases; eauto
  ].
  (* In the remaining case, [R1] is [DX_NONE] and [R2] is [DX_DUPLICABLE _]. *)
  exists R' R'. subst. equiv_dec_by_cases.
Qed.

Next Obligation. (* rely_core *)
  dx.
Qed.

(* ------------------------------------------------------------------------- *)

(* The access right [DX_DUPLICABLE _] is duplicable. *)

Lemma duplicable_duplicable:
  forall a,
  duplicable (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

(* An evolution from [DX_NONE] to [DX_DUPLICABLE] is permitted by [rely]. *)

Lemma rely_none_duplicable:
  forall a,
  rely DX_NONE (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

Lemma rely_core_duplicable:
  forall a b,
  rely (core (DX_EXCLUSIVE a)) (DX_DUPLICABLE b).
Proof.
  dx.
Qed.

(* If [p] evolves to [DX_EXCLUSIVE b] along [rely], then [p] must be equal
   to [DX_EXCLUSIVE b]. *)

Lemma rely_exclusive_reverse:
  forall { p b },
  rely p (DX_EXCLUSIVE b) ->
  p = DX_EXCLUSIVE b.
Proof.
  dx.
Qed.

(* ------------------------------------------------------------------------- *)

(* In the algebra of duplicable-or-exclusive access rights, [DX_DUPLICABLE a]
   and [DX_EXCLUSIVE a] are consistent, maximal, stable, atomic elements. *)

Lemma consistent_duplicable:
  forall a,
  consistent (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

Lemma consistent_exclusive:
  forall b,
  consistent (DX_EXCLUSIVE b).
Proof.
  dx.
Qed.

Lemma maximal_duplicable:
  forall a,
  maximal (DX_DUPLICABLE a).
Proof.
  unfold maximal, maximalp. intros. dx.
Qed.

Lemma maximal_exclusive:
  forall b,
  maximal (DX_EXCLUSIVE b).
Proof.
  unfold maximal, maximalp. intros. dx.
Qed.

Lemma stable_duplicable:
  forall a,
  stable (DX_DUPLICABLE a).
Proof.
  unfold stable, stablep. intros. dx.
Qed.

Lemma stable_exclusive:
  forall b,
  stable (DX_EXCLUSIVE b).
Proof.
  unfold stable, stablep. intros. dx.
Qed.

Lemma atomic_duplicable:
  forall a,
  atomic (DX_DUPLICABLE a).
Proof.
  unfold atomic. dx.
Qed.

Lemma atomic_exclusive:
  forall a,
  atomic (DX_EXCLUSIVE a).
Proof.
  unfold atomic. dx.
Qed.

Lemma solitary_duplicable:
  forall a,
  solitary (DX_DUPLICABLE a).
Proof.
  eauto using atomic_duplicable, maximal_duplicable, @maximal_atomic_implies_solitary.
Qed.

Lemma solitary_exclusive:
  forall a,
  solitary (DX_EXCLUSIVE a).
Proof.
  eauto using atomic_exclusive, maximal_exclusive, @maximal_atomic_implies_solitary.
Qed.

Lemma narcissic_duplicable:
  forall a,
  narcissic (DX_DUPLICABLE a).
Proof.
  unfold narcissic. dx.
Qed.

(* ------------------------------------------------------------------------- *)

(* The act of updating the information associated with an object that we own
   is an admissible evolution. *)

Definition dx_write a2 r1 r2 :=
  exists a1,
  r1 = DX_EXCLUSIVE a1 /\
  r2 = DX_EXCLUSIVE a2.

Lemma admissible_dx_write:
  forall a,
  admissible (dx_write a).
Proof.
  unfold admissible, dx_write. intros. unpack. subst.
  (* If the active thread has an exclusive right, then the passive
     thread has nothing, i.e., [DX_NONE]. *)
  forwards: use_solitary. eauto. eapply solitary_exclusive. subst.
  (* Thus, the passive thread sees no evolution. And the two
     threads together evolve along [dx_write]. *)
  eauto using core_right_unit, rely_reflexive.
Qed.

(* ------------------------------------------------------------------------- *)

(* The act of freezing (moving from exclusive to duplicable) is an admissible
   evolution. *)

Definition dx_freeze a b r1 r2 :=
  r1 = DX_EXCLUSIVE a /\
  r2 = DX_DUPLICABLE b.

Lemma admissible_dx_freeze:
  forall a b,
  admissible (dx_freeze a b).
Proof.
  unfold admissible, dx_freeze. intros. unpack. subst.
  (* If the active thread has an exclusive right, then the passive
     thread has nothing, i.e., [DX_NONE]. *)
  forwards: use_solitary. eauto. eapply solitary_exclusive. subst.
  (* Thus, the passive thread sees an evolution from [DX_NONE] to
     [DX_DUPLICABLE _], which is permitted by [rely]. And the two
     threads together evolve from [DX_EXCLUSIVE] to [DX_DUPLICABLE],
     which is, again, an evolution along [dx_freeze]. *)
  exists (DX_DUPLICABLE b).
  eauto using core_right_unit, duplicable_duplicable, rely_core_duplicable.
Qed.

(* ------------------------------------------------------------------------- *)

End Definitions.

Hint Resolve @consistent_duplicable @consistent_exclusive : consistent.

Hint Resolve @duplicable_duplicable : duplicable aatomic.

Hint Resolve @maximal_duplicable @maximal_exclusive : maximal.

Hint Resolve @stable_duplicable @stable_exclusive : stable.

Hint Resolve @atomic_duplicable @atomic_exclusive : atomic aatomic.

Hint Resolve @solitary_duplicable @solitary_exclusive : solitary.

Hint Resolve @narcissic_duplicable : narcissic.

Hint Resolve @admissible_dx_write @admissible_dx_freeze : admissible.

Implicit Arguments dx [ ].

Section Agreement.

Context `{A : Type, B : Type, C : Type}.
Context `{d_agrees : C -> A -> Prop, x_agrees : C -> B -> Prop}.
Context `{H : EqDec A eq}.

(* ------------------------------------------------------------------------- *)

(* The relation [dx_agrees c p] indicates that the access right [p]
   is compatible with the underlying value [c]. *)

Inductive dx_agrees : C -> dx A B -> Prop :=
| DxAgreesDuplicable:
    forall c a,
    d_agrees c a ->
    dx_agrees c (DX_DUPLICABLE a)
| DxAgreesExclusive:
    forall c b,
    x_agrees c b ->
    dx_agrees c (DX_EXCLUSIVE b).

Lemma dx_agrees_consistent:
  forall c p,
  dx_agrees c p ->
  consistent p.
Proof.
  inversion 1; subst; simpl; tauto.
Qed.

(* ------------------------------------------------------------------------- *)

End Agreement.

Implicit Arguments dx_agrees [ A B C ].

Hint Constructors dx_agrees : agrees.

Ltac select_dx_agrees action :=
  match goal with h: dx_agrees _ _ _ _ |- _ => action h end.
Ltac dx_agrees_inversion :=
  gji select_dx_agrees.

Hint Resolve @dx_agrees_consistent : consistent.

(* ---------------------------------------------------------------------------- *)

(* When the types [A] and [B] coincide, an element [a] of type [A]/[B] can be
   turned into [DX_DUPLICABLE a] and [DX_EXCLUSIVE a], as desired. We define a
   notion of ``mode'' that indicates which way we wish to go, and a function
   [mv2dx] that performs this injection. *)

Section Mode.

Context `{EqDec A eq}.

Inductive mode :=
  | Duplicable
  | Exclusive
.

Fixpoint mv2dx (m : mode) (a : A) : dx A A :=
  match m with
  | Duplicable =>
      DX_DUPLICABLE a
  | Exclusive =>
      DX_EXCLUSIVE a
  end.

Lemma maximal_mv2dx:
  forall m a,
  maximal (mv2dx m a).
Proof.
  destruct m; simpl; eauto with maximal.
Qed.

Lemma stable_mv2dx:
  forall m a,
  stable (mv2dx m a).
Proof.
  destruct m; simpl; eauto with stable.
Qed.

Lemma agrees_mv2dx:
  forall m a,
  dx_agrees eq eq a (mv2dx m a).
Proof.
  destruct m; simpl; eauto with agrees.
Qed.

Lemma consistent_mv2dx:
  forall m a,
  consistent (mv2dx m a).
Proof.
  destruct m; simpl; eauto with consistent.
Qed.

Lemma solitary_mv2dx:
  forall m a,
  solitary (mv2dx m a).
Proof.
  destruct m; simpl; eauto with solitary.
Qed.

Lemma unify_mv2dx:
  forall m1 m2 a1 a2,
  consistent (star (mv2dx m1 a1) (mv2dx m2 a2)) ->
  m1 = Duplicable /\
  m2 = Duplicable /\
  a1 = a2.
Proof.
  destruct m1; destruct m2; simpl; intros; solve [ tauto | equiv_dec_by_cases ].
Qed.

End Mode.

Hint Resolve @maximal_mv2dx : maximal.

Hint Resolve @stable_mv2dx : stable.

Hint Resolve @agrees_mv2dx : agrees.

Hint Resolve @consistent_mv2dx : consistent.

Hint Resolve @solitary_mv2dx : solitary.

Ltac destruct_modes :=
  repeat match goal with m: mode |- _ => destruct m end.

