Set Implicit Arguments.
Generalizable All Variables.
Require Import EquivDec.
Require Import MyTactics.
Require Import MSA.

Section Definitions.

Context `{EqDec A eq}.

(* ------------------------------------------------------------------------- *)

(* If the type [A] is equipped with a decidable notion of equality, then the
   type [discrete A] forms a trivial monotonic separation algebra.  The
   decidability of equality is exploited in the definition of the function
   [star], where one needs to determine whether the two sides agree on a
   single value [a]. The element [DiscreteError] is considered inconsistent
   and serves as the result of attempting to combine two distinct elements.
   The element [DiscreteValue a] is considered duplicable. *)

Inductive discrete :=
| DiscreteError : discrete
| DiscreteValue : A -> discrete.

Ltac destruct_discrete :=
  intros;
  repeat match goal with R: discrete |- _ => destruct R end;
  simpl in *;
  equiv_dec_by_cases;
  try congruence.

Obligation Tactic := try solve [ destruct_discrete ].

Global Program Instance msa_discrete : MSA discrete := {

  star d1 d2 :=
    match d1, d2 return _ with
    | DiscreteError, _
    | _, DiscreteError =>
        DiscreteError
    | DiscreteValue a1, DiscreteValue a2 =>
        (* [DiscreteValue a] is compatible with itself, and is idempotent:
           it is duplicable. *)
        if equiv_dec a1 a2 then DiscreteValue a1 else DiscreteError
    end;

  consistent d :=
    match d return _ with
    | DiscreteError =>
        False
    | _ =>
        True
    end;

  core d :=
    match d return _ with
    | DiscreteError =>
        DiscreteError
    | DiscreteValue a =>
        (* [DiscreteValue a] is preserved by [core], because it represents
           duplicable information. *)
        DiscreteValue a
    end;

  (* The relation [rely] is equality: duplicable knowledge never changes. *)

  rely d1 d2 :=
    d1 = d2

}.

Next Obligation. (* rely_star *)
  intros.
  case_eq R1; case_eq R2; intros;
  try solve [
    (* [rely] implies that [R1] and [R2] are equal, and the result is
       immediate: just take [R2a] and [R2b] to be [R1a] and [R1b]. *)
    exists R1 R2; subst; equiv_dec_by_cases; eauto
  ].
Qed.

(* ------------------------------------------------------------------------- *)

(* In a discrete monotonic separation algebra, any two (consistent)
   compatible elements are equal. In other words, every element is
   narcissic. *)

Lemma narcissic_discrete:
  forall d,
  narcissic d.
Proof.
  unfold narcissic. destruct_discrete.
Qed.

(* This implies that every element is duplicable. *)

Lemma duplicable_discrete:
  forall d,
  duplicable d.
Proof.
  (* We could make this a corollary of the above lemma, but that
     would add a consistency hypothesis. A direct proof is used. *)
  destruct_discrete.
Qed.

(* ------------------------------------------------------------------------- *)

(* [DiscreteValue a] is consistent, maximal, stable, atomic, solitary. *)

(* Every property is maximal and stable. *)

Lemma consistent_discrete:
  forall a,
  consistent (DiscreteValue a).
Proof.
  destruct_discrete.
Qed.

Lemma maximalp_discrete:
  forall P,
  maximalp P.
Proof.
  unfold maximalp. destruct_discrete.
Qed.

Lemma maximal_discrete:
  forall R,
  maximal R.
Proof.
  unfold maximal. eauto using maximalp_discrete.
Qed.

Lemma stablep_discrete:
  forall P,
  stablep P.
Proof.
  unfold stablep. destruct_discrete.
Qed.

Lemma stable_discrete:
  forall R,
  stable R.
Proof.
  unfold stable. eauto using stablep_discrete.
Qed.

Lemma atomic_discrete:
  forall a,
  atomic (DiscreteValue a).
Proof.
  unfold atomic. destruct_discrete.
Qed.

Lemma solitary_discrete:
  forall d,
  solitary d.
Proof.
  (* Note that [maximal] and [atomic] imply [solitary]. Nevertheless,
     we give a direct proof. *)
  unfold solitary. destruct_discrete.
Qed.

(* ------------------------------------------------------------------------- *)

End Definitions.

Hint Resolve @consistent_discrete : consistent.

Hint Resolve @duplicable_discrete : duplicable aatomic.

Hint Resolve @maximalp_discrete @maximal_discrete : maximal.

Hint Resolve @stablep_discrete @stable_discrete : stable.

Hint Resolve @atomic_discrete : atomic aatomic.

Hint Resolve @solitary_discrete : solitary.

Hint Resolve @narcissic_discrete : narcissic.

Implicit Arguments discrete [ ].

Section Agreement.

Context `{A : Type, B : Type}.
Context `{agrees : A -> B -> Prop}.
Context `{EqDec B eq}.

(* ------------------------------------------------------------------------- *)

(* The relation [discrete_agrees a d] indicates that [d] is compatible with
   the underlying value [a]. *)

Inductive discrete_agrees : A -> discrete B -> Prop :=
| DiscreteAgrees:
    forall a b,
    agrees a b ->
    discrete_agrees a (DiscreteValue b).

Lemma discrete_agrees_consistent:
  forall a d,
  discrete_agrees a d ->
  consistent d.
Proof.
  inversion 1; subst; simpl; tauto.
Qed.

(* ------------------------------------------------------------------------- *)

End Agreement.

Implicit Arguments discrete_agrees [ A B ].

