Set Implicit Arguments.
Generalizable All Variables.
Require Import AAC.
Require Import MyTactics.
Require Import MSA.

(* ---------------------------------------------------------------------------- *)

(* Set up the [AAC] library and plugin to help reason about resources. *)

Instance sc : forall `{MSA A}, Commutative (@eq A) star.
Proof. unfold Commutative. eauto using @star_commutative. Qed.
Instance sa: forall `{MSA A}, Associative (@eq A) star.
Proof. unfold Associative. eauto using @star_associative. Qed.

(* Examples. *)

Goal
  forall `{MSA A} (R1 R2 : A),
  star R1 R2 = star R2 R1.
Proof.
  intros. aac_reflexivity.
Qed.

Goal
  forall `{MSA A} (R1 R2 R3 : A),
  star R1 (star R2 R3) = star (star R2 R3) R1.
Proof.
  intros. aac_reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This hint causes [eauto] to invoke the [AAC] plugin when trying to prove
   equations between elements of a monotonic separation algebra. Any
   meta-variables whose value has been determined by other branches of the
   search are instantiated first, because [aac_reflexivity] does not like
   meta-variables. *)

Hint Extern 1 (_ = _) => (instantiate; aac_reflexivity) : aac.

(* Example. *)

Goal
  forall `{MSA A} (R1 R2 R3 : A),
  star R1 R2 = star R2 R1 /\
  star R1 (star R2 R3) = star (star R2 R3) R1.
Proof.
  eauto with aac.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This tactic can sometimes prove an equation between resources when there is
   one Coq meta-variable in the goal. *)

(* TEMPORARY it would be great if Coq could find for us where the
   single meta-variable in the goal is found, and abstract it. *)

Ltac aac_guess :=
  match goal with
  | |- star ?R1 ?R = ?R2 =>
      (* We are assuming [R1] and [R2] are rigid and [R] is a meta-variable. *)
      is_evar R; check_noevar R1; check_noevar R2;
      (* [aac_match] finds the appropriate value for this variable, and we are happy. *)
      let haac := fresh in
      aac_match (fun R => star R1 R) R2 haac;
      eexact haac
  | |- star ?R ?R1 = ?R2 =>
      (* We are assuming [R1] and [R2] are rigid and [R] is a meta-variable. *)
      is_evar R; check_noevar R1; check_noevar R2;
      (* [aac_match] finds the appropriate value for this variable, and we are happy. *)
      let haac := fresh in
      aac_match (fun R => star R R1) R2 haac;
      eexact haac
  | |- star ?R1a (star ?R ?R1b) = ?R2 =>
      (* We are assuming [R1] and [R1a] and [R2] are rigid and [R] is a meta-variable. *)
      is_evar R; check_noevar R1a; check_noevar R1a; check_noevar R2;
      (* [aac_match] finds the appropriate value for this variable, and we are happy. *)
      let haac := fresh in
      aac_match (fun R => star R1a (star R R1b)) R2 haac;
      eexact haac
  | |- star (star ?R1a ?R) ?R1b = ?R2 =>
      (* We are assuming [R1] and [R1a] and [R2] are rigid and [R] is a meta-variable. *)
      is_evar R; check_noevar R1a; check_noevar R1a; check_noevar R2;
      (* [aac_match] finds the appropriate value for this variable, and we are happy. *)
      let haac := fresh in
      aac_match (fun R => star (star R1a R) R1b) R2 haac;
      eexact haac
  end.

Hint Extern 1 (_ = _) => (instantiate; aac_guess) : aac_guess.

(* Examples. *)

Goal
  forall `{MSA A} (R1 R2 : A),
  exists R,
  star R1 R = star R2 R1.
Proof.
  intros. eexists. aac_guess.
Qed.

Goal
  forall `{MSA A} (R1 R2 R3 : A),
  exists R,
  star R1 (star R R2) = star (star R2 R1) R3.
Proof.
  intros. eexists. aac_guess.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This tactic proves a goal of the form [consistent R] by exploiting
   a hypothesis of the form [consistent R']. If we can prove that [R]
   is a conjunct of [R'], then we are done. The tactic [aac_match] is
   very useful for this purpose. *)

Ltac consistent_aac :=
  match goal with
    | hc: consistent ?R' |- consistent ?R =>
        first [
          (* First attempt: try to prove [R] and [R'] equal. *)
          assert (R = R'); [ aac_reflexivity | congruence ]
        | (* Second attempt: try to prove that [R] is a fragment of [R'],
             i.e., [R'] is the form [star x R], for some resource [x]. *)
          let haac := fresh in
          aac_match (fun x => star x R) R' haac;
          (* just [try], because [haac] could be a trivial equation *)
          try rewrite <- haac in hc;
          eauto using @consistent_split_right
        ]
  end.

Hint Extern 1 (consistent _) => consistent_aac : consistent_aac rely_star_hints.

(* Examples. *)

Goal
  forall `{MSA A} (R1 R2 : A),
  consistent (star R1 R2) ->
  consistent (star R1 R2).
Proof.
  intros. consistent_aac.
Qed.

Goal
  forall `{MSA A} (R1 R2 : A),
  consistent (star R1 R2) ->
  consistent (star R2 R1).
Proof.
  intros. consistent_aac.
Qed.

Goal
  forall `{MSA A} (R1 R2 R3 : A),
  consistent (star R1 (star R2 R3)) ->
  consistent (star R3 R1).
Proof.
  intros. consistent_aac.
Qed.

Goal
  forall `{MSA A} (R1 R2 R3 R4 : A),
  consistent (star (star R1 (star R2 R3)) R4) ->
  consistent (star R3 R1) /\ consistent (star R2 R4).
Proof.
  intros. eauto with consistent_aac.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This ad hoc tactic can sometimes prove an equation between resources in the
   presence of another such equation as a hypothesis. *)

Ltac aac_congruence :=
  match goal with
  h: star _ _ = _ |- _ = _ =>
    aac_rewrite h; aac_reflexivity
  end.

(* Example. *)

Goal
  forall `{MSA A} (R1 R2 R3 R4 R5 : A),
  star R1 R2 = star R3 R4 ->
  star (star R5 R1) R2 = star (star R5 R4) R3.
Proof.
  intros. aac_congruence.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Miscellaneous consistency lemmas. *)

(* These lemmas are ugly, but can be used by [eauto] in order to instantiate
   meta-variables, which makes them necessary in some situations. *)

Lemma consistent_12p3_23:
  forall `{MSA A} (R1 R2 R3 : A),
  consistent (star (star R1 R2) R3) ->
  consistent (star R2 R3).
Proof.
  intros. consistent_aac.
Qed.

