Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JSubstitution.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import Agreement.

(* ---------------------------------------------------------------------------- *)

(* A term is an answer iff it is a value or [fail]. *)

Inductive answer : term -> Prop :=
| AnswerValue:
    forall v,
    wk zilch v VALUE ->
    answer v
| AnswerFail:
    answer TFail.

(* ---------------------------------------------------------------------------- *)

(* A term spawns iff it is a [spawn] operation in a deep evaluation
   context. *)

Inductive spawns : term -> Prop :=
| Spawns:
    forall D v1 v2 t,
    wk zilch v1 VALUE ->
    wk zilch v2 VALUE ->
    deep_eval_context D ->
    fill D (TSpawn v1 v2) = t ->
    spawns t.

Hint Constructors spawns : spawns.

(* ---------------------------------------------------------------------------- *)

(* If [t] spawns, then [E[t]] spawns. *)

Lemma spawns_fill:
  forall E t u,
  spawns t ->
  eval_context E ->
  fill E t = u ->
  spawns u.
Proof.
  inversion 1; intros; subst.
  rewrite <- fill_fill by eauto.
  eauto with spawns deep_eval_context.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A thread [t] is waiting in the machine state [s] if it is trying to acquire 
   a lock that is in the [LOCKED] state. *)

Inductive waiting : state -> term -> Prop :=
| Waiting:
    forall s lck,
    maps (get s) lck LOCKED ->
    waiting s (TAcquire (TLock lck))
| WaitingContext:
    forall s t u,
    waiting s t ->
    waiting s (TLet t u).

Hint Constructors waiting : waiting.

(* ---------------------------------------------------------------------------- *)

(* The main progress lemma for configurations of kind [TERM]. *)

(* The lemma states that if the configuration [Cf h t] is well-typed, then
   either [t] is an answer, or [t] is about to spawn a new thread, or the
   configuration reduces. *)

(* As usual, because of the presence of [JExistsElim], the proof must be
   by induction over the height of the type derivation. *)

(* Because of the presence of [JForallIntro], which is not restricted
   to values, the statement must allow for a nonempty environment [K].
   Nevertheless, the term [t] itself is closed; this is stated
   separately. An alternative approach would be to keep a statement of
   progress under an empty environment, and in [JForallIntro], to use
   substitution in order to replace the type variable with an arbitrary
   type. This would require exploiting the fact that every kind is
   inhabited. I avoid this route because I do not wish to rely on this
   fact. *)

Lemma progress_TERM_preliminary:
  forall n R1t1 K P t T,
  jt n R1t1 K P t T -> (* [t] is well-typed *)
  wk zilch t TERM ->   (* [t] is closed *)
  forall R1t2 R1t,
  jpi R1t2 K P ->
  star R1t1 R1t2 = R1t ->
  forall R1f R1 s,
  jsr s R1 ->
  star R1t R1f = R1 ->
  answer t \/
  spawns t \/
  waiting s t \/
  exists c, red (Cf s t) c.

(* This tactic means that the term reduces. *)
Local Ltac reduces :=
  right; right; right; eexists.

Proof.
  intro n. induction_wf ih: lt_wf n.
  introv h; dependent destruction h; intros; try subst; decomposition;
  try solve [
    (* If the term at hand is obviously a value, we are done. *)
    left; eapply AnswerValue; eauto using wk_empty_reverse with wk
  ];
  jsr_consistent;
  wk_inversion.

  (* JForallIntro *)
  (* The term [t] is closed, so the result follows by the induction
     hypothesis. *)
  closed. solve [ eauto ].

  (* Case: JExistsElim *)
  substitution. solve [ eauto ].

  (* JFrame *)
  solve [ eauto with aac_guess ].

  (* JCut *)
  solve [ eauto with j aac_guess ].

  (* JSubLeft *)
  solve [ eauto using jps_sound with consistent_aac ].

  (* JSubRight *)
  (* Just discard this instance of [JSubRight] and work with what's under it. *)
  solve [ eauto ].

  (* JApplication *)
  (* We have a canonical type derivation for the term [t] on the left-hand
     side of the application. Thus, [t] must be a lambda-abstraction. *)
  classification.
  (* Now, either the term [u] on the right-hand side of the application is a
     value, in which case the whole application is a beta-redex; or [u] is
     [fail], in which case the whole application again reduces; or [u] spawns,
     in which the whole application spawns; or [u] reduces, in which case the
     whole application reduces. *)
  forwards f: ih; eauto with aac_guess.
  destruct f as [ [ | ] | [ ? | [ ? | ? ]]].
  (* Sub-case: [u] is a value. *)
  reduces. solve [ eauto with red ].
  (* Sub-case: [u] is [fail]. *)
  reduces. solve [ eauto 6 using fill_eval_context with red eval_context wk ].
  (* Sub-case: [u] spawns. *)
  solve [ eauto 8 using spawns_fill, fill_eval_context with eval_context wk ].
  (* Sub-case: [u] is waiting. *)
  solve [ eauto with waiting ].
  (* Sub-case: [u] reduces. *)
  unpack. destruct_configurations. reduces.
  solve [ eauto 6 using fill_eval_context with eval_context wk red ].

  (* JFail *)
  solve [ eauto using AnswerFail ].

  (* JSpawn *)
  solve [ eauto with spawns deep_eval_context simpl_subst_goal ].

  (* JNew *)
  reduces. solve [ eauto with red ].

  (* JRead *)
  classification. decomposition.
  jsr_maps_right_to_left.
  reduces. solve [ eauto with red ].

  (* JWrite *)
  classification. decomposition.
  jsr_maps_right_to_left.
  reduces. solve [ eauto with red ].

  (* JFreeze *)
  reduces. solve [ eauto with red ].

  (* JNewLock *)
  reduces. solve [ eauto with red ].

  (* JAcquire *)
  classification. decomposition.
  jsr_invariant.
  (* Sub-case: the lock is available. The term reduces. *)
  reduces. solve [ eauto with red ].
  (* Sub-case: the lock is taken. The term waits. *)
  solve [ eauto with waiting ].

  (* JRelease *)
  classification. decomposition. jsr_taken.
  reduces. solve [ eauto with red ].

  (* JGive *)
  classification. decomposition. jsr_unadopted.
  reduces. solve [ eauto with red ].

  (* JTake *)
  classification. decomposition. jsr_adoptable.
  (* The adopter pointer is either [NonNull y] or something else. *)
  match goal with p: pointer |- context[TTake _ (TLoc ?y)] =>
    destruct (eq_pointer_dec p (NonNull y))
  end;
  (* In either case, one of the reduction rules is applicable. *)
  subst; reduces; solve [ eauto with red ].

  (* JTakeCommitted *)
  forwards: adopted_adoptable; [ eauto | ]. jsr_adoptable.
  reduces. solve [ eauto with red ].
    
Qed.

(* ---------------------------------------------------------------------------- *)

(* A slightly simplified statement. *)

Lemma progress_TERM:
  forall n R1t K t T,
  jt n R1t K TyEmpty t T ->
  wk zilch t TERM ->
  forall R1f R1 s,
  jsr s R1 ->
  star R1t R1f = R1 ->
  answer t \/
  spawns t \/
  waiting s t \/
  exists c, red (Cf s t) c.
Proof.
  eauto using progress_TERM_preliminary, @core_right_unit with j.
Qed.

