Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.
Require Import SubjectReductionCaseNew.
Require Import SubjectReductionCaseRead.
Require Import SubjectReductionCaseWrite.
Require Import SubjectReductionCaseFreeze.
Require Import SubjectReductionCaseNewLock.
Require Import SubjectReductionCaseAcquire.
Require Import SubjectReductionCaseRelease.
Require Import SubjectReductionCaseGive.
Require Import SubjectReductionCaseTake.

(* ---------------------------------------------------------------------------- *)

(* Some of the trickier proof cases have been isolated as separate lemmas. The
   following hints help apply these lemmas. *)

Local Hint Constructors redwkc : srcase.

Local Hint Resolve jpi_drop_right jpi_drop_left : srcase.

(* ---------------------------------------------------------------------------- *)

(* The main subject reduction lemma, for terms of kind [TERM]. *)

Lemma subject_reduction_TERM_preliminary:
  (* The measure. *)
  forall n,
  (* The reduction hypothesis. *)
  forall s1 t1 s2 t2,
  red (Cf s1 t1) (Cf s2 t2) ->
  (* The typing hypothesis. The following three hypotheses are equivalent
     to [jt _ R1t K TyEmpty t1 T]. *)
  forall R1t1 K P T,
  jt n R1t1 K P t1 T ->
  forall R1t2 R1t,
  jpi R1t2 K P ->
  star R1t1 R1t2 = R1t ->
  (* Even though [K] is non-empty, the program [t1] must be closed. (The
     environment [K] introduces only ghost variables, so to speak.) We
     require the configuration [Cf s1 t1] to be well-sorted under an
     empty environment. This implies that [t1] is closed, and is preserved
     by reduction (see the lemma [red_wk]). *)
  wkc (Cf s1 t1) TERM ->
  (* Because [JForallIntro] allows harmless terms only, we know that if
     [K] is non-empty, then [t1] must be harmless. This information is
     exploited in the case of [JNewLock], where we need to know that the
     invariant associated with the newly-created lock is closed. *)
  (K <> zilch -> harmless t1) ->
  (* We keep track of the fact that [K] contains substitutable kinds
     only. This is true because [JForallIntro] is the only rule that
     extends [K]. This hypothesis allows us to exploit the strengthening
     lemma where needed (e.g., in [JRelease]). *)
  supports_strengthening K ->
  (* We have the entire state [s1] at hand, of which the resource [R1t]
     describes only a part. The rest, which has been framed out already,
     is [R1f]. The combination of [R1t] and [R1f] corresponds to the
     state [s1] (and must therefore be consistent). *)
  forall R1f R1,
  jsr s1 R1 ->
  star R1t R1f = R1 ->
  (* If all this holds, then the term obtained after this reduction step
     is well-typed under some resource [R2t]. The new state [s2] corresponds
     to the combination of [R2t] and [R2f], and the evolution that has been
     imposed to the outside world, from [R1f] to [R2f], respects [rely]. *)
  sr_target R1f K (Cf s2 t2) T.

Proof.
  (* Well-founded induction over the height of the type derivation. *)
  intro n. induction_wf ih: lt_wf n.
  (* Introduce the remaining hypotheses. *)
  introv hred hjt. intros. subst.
  jsr_consistent.
  (* Reason by cases on the last typing rule that was used. *)
  dependent destruction hjt; subst;
  (* If this typing rule implies that [t1] must be a value, then
     we obtain a contradiction, because values are irreducible.
     Same thing if [t1] is a term that does not reduce, such as
     [TFail] or [TSpawn]. *)
  try solve [ false; eauto using (@red_value zilch),
              red_value, red_fail, red_spawn with wk ].

  (* Case: JForallIntro *)
  (* This case is present because we do not have a value restriction. We
     do have a restriction to harmless terms, so we need to argue that
     harmlessness is preserved by reduction. *)
  decomposition.
  forwards: red_wkc; [ eauto | eauto | ].
  wkc_inversion. closed.
  eapply SrTargetForallIntro;
  solve [ eauto using harmless_red with wk supports_strengthening ].

  (* Case: JExistsElim *)
  decomposition.
  (* We have a type derivation for the term [t] under the precondition [P],
     which refers to the name [x]. We know that [[U/x]P] holds. Hence, we
     substitute [U] for [x] throughout the type derivation. Note that this
     does not affect the term [t] or its type [T], because they do not refer
     to [x]. We are then able to apply the induction hypothesis. *)
  substitution.
  solve [ eauto ].

  (* Case: JFrame *)
  decomposition.
  (* A permission [Q] is framed out. Apply the induction hypothesis, which
     allows us to take a type-preserving step in the absence of [Q], then
     apply the lemma [SrTargetFrame] to recover [Q]. *)
  solve [ eauto using SrTargetFrame with consistent_aac aac ].

  (* Case: JCut *)
  (* Just apply the induction hypothesis. *)
  solve [ eauto with j aac ].

  (* Case: JSubLeft *)
  (* The lemma [jps_sound] allows us to transport the [jpi] hypothesis
     up to the level of the premise. There, we apply the induction
     hypothesis, and we are done. We do not even need to reconstruct
     an instance of [JSubLeft]. *)
  solve [ eauto 7 using jps_sound with consistent_aac ].

  (* Case: JSubRight *)
  (* Just work under this instance of [JSubRight]. *)
  solve [ eauto using SrTargetSubRight ].

  (* Case: JApplication *)
  (* Reason by cases on the last reduction rule that was used. In this
     rule, the redex must exhibit an application at the root. *)
  dependent destruction hred;
  decomposition.

    (* Sub-case: RedBeta *)
    clear ih.
    (* We have an application of the form [TApp (TAbs t) u]. Because of the form
       of the pre-condition in the conclusion of [JApplication], a permission of
       the form [TAbs t @ U -> T] is known to hold. This implies that the term [t]
       has type [T] under the hypothesis [0 @ U]. *)
    (* Because [t] is well-typed under [core R], it is also well-typed under [R].
       Here, the function [TAbs t] is applied just once, so there is no need to
       argue that it is well-typed under a duplicable resource. *)
    forwards: j_core; [ eassumption | eauto with consistent_split | ].
    (* Because [u] is a value, and is well-typed according to [jt], it must be
       well-typed according to [jctop] as well. *)
    forwards: (canonicalization hjt); eauto using wk_empty with consistent_aac.
    (* By substituting [u] for the variable [0] throughout the type derivation
       for [t], we find that [subst u 0 t] is well-typed. *)
    forwards: jctop_substitution; eauto using wk_empty.
    (* The remains to package things up. *)
    sr_target_pure.
    solve [ eauto using JSameResource with aac ].

    (* Sub-case: RedContextFail *)
    (* This case is easy, since [TFail] is always well-typed. *)
    forwards: jctop_wk; [ eauto | wk_inversion ].
    sr_target_pure. solve [ eauto using (JFail 0) ].

    (* Sub-case: RedContext *)
    (* We are looking at the reduction of [ E[u1] ] to [ E[u2] ]. First, decompose
       the context [E]. There is just one case, so we find that we are looking at
       a reduction of [TApp t u1] to [TApp t u2]. Then apply the induction hypothesis,
       which proves that the reduction of [u1] to [u2] is type-preserving. Finally,
       use [JConfApp] to reconstruct a type derivation for [TApp t u2]. *)
    destruct_eval_context. injections. wkc_inversion; wk_inversion.
    solve [ eauto 7 using SrTargetApp, harmless_TApp_inversion with consistent_aac wk aac_guess ].

  (* Case: JNew *)
  { eauto using subject_reduction_TNewRef with srcase. }

  (* Case: JRead *)
  { eauto using subject_reduction_TRead with srcase. }

  (* Case: JWrite *)
  { eauto using subject_reduction_TWrite with srcase. }

  (* JFreeze *)
  { eauto using subject_reduction_JFreeze with srcase. }

  (* Case: JNewLock *)
  { eauto using subject_reduction_TNewLock with srcase. }

  (* Case: JAcquire *)
  { eauto using subject_reduction_TAcquire with srcase. }

  (* Case: JRelease *)
  { eauto using subject_reduction_TRelease with srcase. }

  (* Case: JGive *)
  { eauto using subject_reduction_TGive with srcase. }

  (* Case: JTake *)
  { eauto using subject_reduction_TTake with srcase. }

  (* Case: JTakeCommitted *)
  { eauto 7 using subject_reduction_TTakeCommitted, adopted_affine
             with srcase explode_resources consistent_split. }

Qed.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for terms of kind [TERM]. A slightly simplified statement. *)

Lemma subject_reduction_TERM:
  forall s1 t1 s2 t2,
  red (Cf s1 t1) (Cf s2 t2) ->
  forall n R1t T,
  jt n R1t zilch TyEmpty t1 T ->
  wkc (Cf s1 t1) TERM ->
  forall R1f R1,
  jsr s1 R1 ->
  star R1t R1f = R1 ->
  sr_target R1f zilch (Cf s2 t2) T.
Proof.
  eauto using subject_reduction_TERM_preliminary, @core_right_unit
         with j congruence supports_strengthening.
Qed.

