Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Zap.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JDuplication.
Require Import JTop.
Require Import JClassification.
Require Import JDecomposition.

(* ------------------------------------------------------------------------- *)

(* This preliminary lemma states that if the subtyping fact [T <= U] is
   semantically sound, then the permission subsumption fact [ys @ T <= ys @ U]
   is semantically sound as well. This lemma is later used to apply the fact
   [T <= U] to a list of adoptees, so as to prove that it is sound for
   [adopts] clauses to be covariant. *)

Lemma sub_zap:
  forall K T U,
  (forall R v , jctop R K       v T  -> consistent R -> jctop R K       v U) ->
  (forall xs R, jpi   R K (zap xs T) -> consistent R -> jpi   R K (zap xs U)).
Proof.
  induction xs; simpl; intros; decomposition; econstructor; eauto with consistent_split.
Qed.

(* ------------------------------------------------------------------------- *)

(* Subsumption is sound with respect to the natural notion of implication that
   arises out of the interpretation of permissions. *)

(* This key lemma is the reason why we do not care how complicated the
   definition of [jps] might be. *)

(* The proof of this lemma is carried out by well-founded induction over the
   index [n], and is the reason why we introduce this index in the first
   place. Ordinary structural induction would not work here because there is
   at least one case where a substitution must be applied to a sub-term before
   exploiting the induction hypothesis that comes with it. *)

Lemma jps_sub_sound:
(
  forall n K P Q,
  jps n K P Q ->
  forall R,
  jpi R K P -> 
  consistent R ->
  jpi R K Q
) /\ (
  forall n K T U,
  sub n K T U ->
  forall R v,
  jctop R K v T ->
  consistent R ->
  jctop R K v U
) /\ (
  forall n K Ts Us,
  row_sub n K Ts Us ->
  forall R vs,
  jpi R K (zip vs Ts) ->
  consistent R ->
  jpi R K (zip vs Us)
).
Local Hint Resolve JcTopSingleton : singleton.
Proof.
  eapply simultaneous_well_founded_induction_3; intros n ih_jps ih_sub ih_row_sub;
  introv h; dependent destruction h; intros; subst;
  (* A general recipe for the subtyping rules that prove [TyDuplicable _]. *)
  try match goal with |- jpi _ _ (TyDuplicable _) =>
    decomposition; eauto with jpi wk logic dup
  end.

  (* JpsReflexive *)
  { eauto 1. }

  (* JpsTransitive *)
  { eauto. }

  (* JpsEmptyTop *)
  { eauto 1 with jpi. }

  (* JpsEmptyAppearsLeft *)
  { eauto 2 with jpi core_unit. }

  (* JpsEmptyAppearsRight *)
  { eauto 2 with jpi core_unit. }

  (* JpsDisappearsLeft *)
  { decomposition. eauto 2 with jpi_drop. }

  (* JpsDisappearsRight *)
  { decomposition. eauto 2 with jpi_drop. }

  (* JpsStarCommutative *)
  { decomposition. eauto 2 using @star_commutative with jpi. }

  (* JpsStarAssociative *)
  { decomposition. eauto 3 using @star_associative with jpi. }

  (* JpsEqualityReflexive *)
  { eauto 2 using JcTopSingleton. }

  (* JpsEqualitySymmetric *)
  { decomposition. eauto 2 with singleton. }

  (* JpsEqualityTransitive *)
  { decomposition. eauto 2 with singleton. }

  (* JpsEqualsForEquals *)
  { decomposition. eauto 3 with jpi singleton. }

  (* JpsEqualsForEqualsReverse *)
  { decomposition. eauto 3 with jpi singleton. }

  (* JpsEqualsForEqualsDrop *)
  { decomposition. eauto 2 with jpi jpi_drop. }

  (* JpsEqualsForEqualsReverseDrop *)
  { decomposition. eauto 2 with jpi jpi_drop. }

  (* JpsDuplicate *)
  { forwards: decomposition_TyStar_TyDuplicable_left; [ eauto | unpack ].
    eauto 3 using j_duplicable with jpi core_unit. }

  (* JpsHideDuplicablePrecondition *)
  (* A value that admits a function type must be a function. *)
  { decomposition. classification. decomposition.
    (* Build a new proof. *)
    econstructor; [ eauto with wk | | | eauto ].
      (* Premise 1. Type-check the function. *)
      eapply JFunction; [ eauto | ].
        (* Upon entry to the function body, move the permission [P], which is
           ambient, to the argument, where the function expects it. *)
        eapply (@JSubLeft 1); [ eauto | ].
        jps_left JpsStarCommutative.
        simpl_lift_goal. eapply JpsMixStarIntro.
      (* Premise 2. Justify [TyDuplicable P] and [P]. *)
      eauto with jpi. }

  (* JpsMixStarIntro *)
  { decomposition. eauto 2 using JcTopFrame. }

  (* JpsMixStarElim *)
  { decomposition. eauto 2 with jpi. }

  (* JpsExistsIntro *)
  { eauto 2 with jpi. }

  (* JpsExistsHoistLeft *)
  { decomposition.
    econstructor; eauto using wk_weakening with wk.
    simpl_subst_goal. rewrite subst_lift.
    eauto 2 with jpi. }

  (* JpsForallType2Perm *)
  (* The goal boils down to proving that [v] has type [T] in the presence
     of a fresh variable [x]. *)
  { decomposition. eauto 2 with jpi. }

  (* JpsForallPerm2Type *)
  { decomposition. eauto 2 using JcTopForallIntro. }

  (* JpsExistsType2Perm *)
  { decomposition.
    econstructor; [ | eauto | eauto using wk_weakening with wk | eauto ].
    simpl_subst_goal. rewrite subst_lift. eassumption. }

  (* JpsExistsPerm2Type *)
  { decomposition.
    wk_inversion.
    match goal with h: jpi _ _ _ |- _ =>
      simpl_subst_in h;
      rewrite subst_lift in h
    end.
    eauto 2 using JcTopExistsIntro. }

  (* JpsForallElimPerm *)
  { decomposition. eapply j_substitution; eauto 1. }

  (* JpsVacuousForallIntro *)
  { econstructor. eauto.
    eapply j_weakening; eauto 1. }

  (* JpsFrame *)
  { (* A value that admits a function type must be a function. *)
    classification. decomposition.
    (* Build a new proof. *)
    econstructor; [ eauto with wk | | | eapply core_right_unit ].
      (* Premise 1. Type-check the function. *)
      eapply JFunction with (P := TyEmpty); [ eauto with wk | ].
      unfold intro; simpl_lift_goal.
        (* Upon entry to the function body, move the permission [P], which is
           attached to the argument, and make it ambient. Then, apply the
           frame rule, so as to get rid of [P]. *)
        eapply (@JSubLeft 1); [ eapply JFrame; eassumption | ].
        jps_left JpsDisappearsLeft.
        eapply JpsMixStarElim.
      (* Premise 2. Justify [TyDuplicable TyEmpty] and [TyEmpty]. *)
      eauto 3 with wk logic dup jpi core_unit. }

  (* JpsDecomposeRef *)
  { classification. decomposition.
    eapply JpiExists.
    simpl_subst_goal. repeat rewrite subst_lift.
    (* The focused value. *)
    eapply JpiStar; [ | eauto | aac_guess ].
    (* The block. *)
    jctop JRef; eauto using JcTopSingleton, wk_empty, @core_left_unit.
    (* The administrative premises of [JpiExists]. *)
    eauto using wk_empty with wk.
    do 3 (econstructor; eauto using wk_weakening, wk_substitution with wk insert_insert).
    eauto with substitutable. }

  (* JpsRecomposeRef *)
  { decomposition. classification. decomposition.
    jctop JRef.
      3: eauto 1.
      aac_guess.
      eauto 1.
      eauto 2 with explode_resources maps_maximal. }

  (* JpsUnifyRef *)
  { decomposition.
    (* The goal rapidly boils down to proving [v1 = v2]. *)
    econstructor; eauto with jpi core_unit.
    classification. decomposition.
    eapply JcTopSingletonEq; [ | eauto ].
    (* This equation follows from the fact that the access rights [mv2dx m1 v1]
       and [mv2dx m2 v2] are compatible. *)
    explode_resources_blocks.
    eapply unify_mv2dx.
    eapply maps_consistent.
      eapply maps_star_construction; [ | eauto | eauto ].
        eauto with consistent_aac.
      eauto with consistent_aac. }

  (* JpsUnadoptedAdoptable *)
  { classification. decomposition.
    econstructor; [ | | eapply core_right_unit ].
      jctop JUnadopted. eauto.
      jctop JAdoptable. eauto using adoptable_core, unadopted_implies_adoptable
                               with explode_resources. }

  (* JpsCoArrow *)
  { (* A value that admits a function type must be a lambda-abstraction. *)
    classification.
    (* The function body must be well-typed. *)
    decomposition.
    (* The subtyping judgement for the domain is expressed using [sub].
       Unfold the definition of [sub] so as to express it in terms of
       [jps]. *)
    sub_inversion.
    (* We now build a new type derivation by pushing the subtyping rule
       under the lambda-abstraction. We use [JSubLeft] for the domain 
       and [JSubRight] for the codomain. The resulting type derivation
       is canonical because the use of subtyping under a lambda-abstraction
       is allowed: it takes place one step ``later''. Thus, the induction
       hypotheses are not needed. *)
    eapply JFunctionEmpty; [ eauto | | eauto ].
    eapply JSubLeft; [ | eassumption ].
    eapply JSubRight; [ | eapply j_weakening; eauto ].
    eassumption. }

  (* JpsCoStar *)
  { decomposition. eauto 8 with jpi consistent_split. }

  (* JpsCoForall *)
  { decomposition. eauto 4 with jpi jpi_drop core_unit. }

  (* JpsCoExists *)
  { decomposition. eauto 6 using (get1 j_substitution) with wk jpi. }

  (* JpsCoRef *)
  { classification. decomposition. jctop JRef; eauto 4 with consistent_split. }

  (* JpsCoAdopts *)
  { classification. Opaque get. decomposition. Transparent get. (* TEMPORARY unpleasant *)
    jctop JAdopts; eauto using sub_zap with wk consistent_split. }

  (* Sub *)
  { (* Because [jctop] is defined in terms of [jpi], we are able to
       directly apply the induction hypothesis [ih_jps]. *)
    eapply ih_jps; eauto.
    (* The judgement that justifies this use of the induction hypothesis
       is obtained by substituting [v] for the name [x] that is
       universally quantified in the definition of [sub]. *)
    eapply j_substitution; eauto using jctop_value with substitutable;
    simpl_subst_goal; f_equal; eapply subst_lift. }

  (* RowSubNil *)
  { eauto 1. }

  (* RowSubCons *)
  { (* If the row of types has at least one component, then the row of values
       has at least one component. *)
    match goal with h: jpi _ _ (zip ?vs (RowCons _ _)) |- _ =>
      destruct vs; simpl in h;
      try solve [ false; eauto using jpi_TyFalse ];
      revert h
    end; intro hjpi; dependent destruction hjpi; subst; simpl.
    (* Build a new type derivation. *)
    econstructor; eauto 4 with consistent_split. }

Qed.

Definition jps_sound := get1 jps_sub_sound.
Definition sub_sound := get2 jps_sub_sound.

